import { EncryptionAuthStatus, MailFolderType, SYSTEM_GROUP_MAIL_ADDRESS, TUTANOTA_MAIL_ADDRESS_DOMAINS } from "../TutanotaConstants.js";
import { assertNotNull, endsWith } from "@tutao/tutanota-utils";
/**
 * Returns true if given folder is the {@link MailFolderType.SPAM} or {@link MailFolderType.TRASH} folder, or a descendant of those folders.
 */
export function isSpamOrTrashFolder(system, folder) {
    // not using isOfTypeOrSubfolderOf because checking the type first is cheaper
    return (folder.folderType === MailFolderType.TRASH ||
        folder.folderType === MailFolderType.SPAM ||
        isSubfolderOfType(system, folder, MailFolderType.TRASH) ||
        isSubfolderOfType(system, folder, MailFolderType.SPAM));
}
export function isOfTypeOrSubfolderOf(system, folder, type) {
    return folder.folderType === type || isSubfolderOfType(system, folder, type);
}
export function isSubfolderOfType(system, folder, type) {
    const systemFolder = system.getSystemFolderByType(type);
    return systemFolder != null && system.checkFolderForAncestor(folder, systemFolder._id);
}
/**
 * Gets a system folder of the specified type and unwraps it.
 * Some system folders don't exist in some cases, e.g. spam or archive for external mailboxes!
 *
 * Use with caution.
 */
export function assertSystemFolderOfType(system, type) {
    return assertNotNull(system.getSystemFolderByType(type), "System folder of type does not exist!");
}
/**
 * NOTE: DOES NOT VERIFY IF THE MESSAGE IS AUTHENTIC - DO NOT USE THIS OUTSIDE OF THIS FILE OR FOR TESTING
 * @VisibleForTesting
 */
export function isTutanotaTeamAddress(address) {
    return endsWith(address, "@tutao.de") || address === "no-reply@tutanota.de";
}
function hasValidEncryptionAuthForTeamOrSystemMail({ encryptionAuthStatus }) {
    switch (encryptionAuthStatus) {
        // emails before tuta-crypt had no encryptionAuthStatus
        case null:
        case undefined:
        case EncryptionAuthStatus.RSA_NO_AUTHENTICATION:
        case EncryptionAuthStatus.PQ_AUTHENTICATION_SUCCEEDED:
            return true;
        case EncryptionAuthStatus.AES_NO_AUTHENTICATION:
        case EncryptionAuthStatus.PQ_AUTHENTICATION_FAILED:
        // we have to be able to handle future cases, to be safe we say that they are not valid encryptionAuth
        default:
            return false;
    }
}
/**
 * Is this a tutao team member email or a system notification
 */
export function isTutanotaTeamMail(mail) {
    const { confidential, sender, state } = mail;
    return (confidential &&
        state === "2" /* MailState.RECEIVED */ &&
        hasValidEncryptionAuthForTeamOrSystemMail(mail) &&
        (sender.address === SYSTEM_GROUP_MAIL_ADDRESS || isTutanotaTeamAddress(sender.address)));
}
/**
 * Is this a system notification?
 */
export function isSystemNotification(mail) {
    const { confidential, sender, state } = mail;
    return (state === "2" /* MailState.RECEIVED */ &&
        confidential &&
        hasValidEncryptionAuthForTeamOrSystemMail(mail) &&
        (sender.address === SYSTEM_GROUP_MAIL_ADDRESS ||
            // New emails will have sender set to system and will only have replyTo set to no-reply
            // but we should keep displaying old emails correctly.
            isNoReplyTeamAddress(sender.address)));
}
export function isNoReplyTeamAddress(address) {
    return address === "no-reply@tutao.de" || address === "no-reply@tutanota.de";
}
export function getDisplayedSender(mail) {
    const realSender = mail.sender;
    return { address: realSender.address, name: realSender.name };
}
export function isTutanotaMailAddress(mailAddress) {
    return TUTANOTA_MAIL_ADDRESS_DOMAINS.some((tutaDomain) => mailAddress.endsWith("@" + tutaDomain));
}
//# sourceMappingURL=CommonMailUtils.js.map