import { bootstrapWorker } from "./WorkerClient";
import { EventController } from "./EventController";
import { EntropyCollector } from "./EntropyCollector";
import { SearchModel } from "../../search/model/SearchModel";
import { MailModel } from "../../mail/model/MailModel";
import { assertMainOrNode, isAndroidApp, isApp, isBrowser, isDesktop, isElectronClient, isIOSApp } from "../common/Env";
import { notifications } from "../../gui/Notifications";
import { LoginController } from "./LoginController";
import { EntityClient } from "../common/EntityClient";
import { assert, defer, lazyMemoized, noOp, ofClass } from "@tutao/tutanota-utils";
import { ProgressTracker } from "./ProgressTracker";
import { MinimizedMailEditorViewModel } from "../../mail/model/MinimizedMailEditorViewModel";
import { SchedulerImpl } from "../common/utils/Scheduler.js";
import { createCredentialsProvider } from "../../misc/credentials/CredentialsProviderFactory";
import { guiDownload } from "../../file/FileController";
import { ProgrammingError } from "../common/error/ProgrammingError";
import { SecondFactorHandler } from "../../misc/2fa/SecondFactorHandler";
import { WebauthnClient } from "../../misc/2fa/webauthn/WebauthnClient";
import { BrowserWebauthn } from "../../misc/2fa/webauthn/BrowserWebauthn.js";
import { UsageTestController } from "@tutao/tutanota-usagetests";
import { EphemeralUsageTestStorage, UsageTestModel } from "../../misc/UsageTestModel";
import { deviceConfig } from "../../misc/DeviceConfig";
import { PageContextLoginListener } from "./PageContextLoginListener.js";
import { FileControllerBrowser } from "../../file/FileControllerBrowser.js";
import { FileControllerNative } from "../../file/FileControllerNative.js";
import { windowFacade } from "../../misc/WindowFacade.js";
import { NewsModel } from "../../misc/news/NewsModel.js";
import { NoZoneDateProvider } from "../common/utils/NoZoneDateProvider.js";
import { WebsocketConnectivityModel } from "../../misc/WebsocketConnectivityModel.js";
import { OperationProgressTracker } from "./OperationProgressTracker.js";
import { InfoMessageHandler } from "../../gui/InfoMessageHandler.js";
import { OfflineIndicatorViewModel } from "../../gui/base/OfflineIndicatorViewModel.js";
import { Const, FeatureType, GroupType, KdfType } from "../common/TutanotaConstants.js";
import { showProgressDialog } from "../../gui/dialogs/ProgressDialog.js";
import { InboxRuleHandler } from "../../mail/model/InboxRuleHandler.js";
import { ScopedRouter, ThrottledRouter } from "../../gui/ScopedRouter.js";
import { DomainConfigProvider } from "../common/DomainConfigProvider.js";
import { getEnabledMailAddressesWithUser } from "../../mail/model/MailUtils.js";
import { getDisplayedSender } from "../common/mail/CommonMailUtils.js";
import { isCustomizationEnabledForCustomer } from "../common/utils/CustomerUtils.js";
import { NativeContactsSyncManager } from "../../contacts/model/NativeContactsSyncManager.js";
import { PermissionError } from "../common/error/PermissionError.js";
assertMainOrNode();
class MainLocator {
    async noZoneDateProvider() {
        return new NoZoneDateProvider();
    }
    async sendMailModel(mailboxDetails, mailboxProperties) {
        const factory = await this.sendMailModelSyncFactory(mailboxDetails, mailboxProperties);
        return factory();
    }
    async appHeaderAttrs() {
        return {
            offlineIndicatorModel: await this.offlineIndicatorViewModel(),
            newsModel: this.newsModel,
        };
    }
    inboxRuleHanlder() {
        return new InboxRuleHandler(this.mailFacade, this.entityClient, this.logins);
    }
    async searchViewModelFactory() {
        const { SearchViewModel } = await import("../../search/view/SearchViewModel.js");
        const conversationViewModelFactory = await this.conversationViewModelFactory();
        const redraw = await this.redraw();
        const searchRouter = await this.scopedSearchRouter();
        return () => {
            return new SearchViewModel(searchRouter, this.search, this.searchFacade, this.mailModel, this.logins, this.indexerFacade, this.entityClient, this.eventController, this.mailOpenedListener, this.calendarFacade, this.progressTracker, conversationViewModelFactory, redraw, deviceConfig.getMailAutoSelectBehavior());
        };
    }
    async receivedGroupInvitationsModel(groupType) {
        const { ReceivedGroupInvitationsModel } = await import("../../sharing/model/ReceivedGroupInvitationsModel.js");
        return new ReceivedGroupInvitationsModel(groupType, this.eventController, this.entityClient, this.logins);
    }
    /** This ugly bit exists because CalendarEventWhoModel wants a sync factory. */
    async sendMailModelSyncFactory(mailboxDetails, mailboxProperties) {
        const { SendMailModel } = await import("../../mail/editor/SendMailModel");
        const recipientsModel = await this.recipientsModel();
        const dateProvider = await this.noZoneDateProvider();
        return () => new SendMailModel(this.mailFacade, this.entityClient, this.logins, this.mailModel, this.contactModel, this.eventController, mailboxDetails, recipientsModel, dateProvider, mailboxProperties);
    }
    async calendarEventModel(editMode, event, mailboxDetail, mailboxProperties, responseTo) {
        const [{ makeCalendarEventModel }, { getTimeZone }, { calendarNotificationSender }] = await Promise.all([
            import("../../calendar/gui/eventeditor-model/CalendarEventModel.js"),
            import("../../calendar/date/CalendarUtils.js"),
            import("../../calendar/view/CalendarNotificationSender.js"),
        ]);
        const sendMailModelFactory = await this.sendMailModelSyncFactory(mailboxDetail, mailboxProperties);
        const showProgress = (p) => showProgressDialog("pleaseWait_msg", p);
        return await makeCalendarEventModel(editMode, event, await this.recipientsModel(), await this.calendarModel(), this.logins, mailboxDetail, mailboxProperties, sendMailModelFactory, calendarNotificationSender, this.entityClient, responseTo, getTimeZone(), showProgress);
    }
    async recipientsSearchModel() {
        const { RecipientsSearchModel } = await import("../../misc/RecipientsSearchModel.js");
        const suggestionsProvider = isApp()
            ? (query) => this.mobileContactsFacade.findSuggestions(query).catch(ofClass(PermissionError, () => []))
            : null;
        return new RecipientsSearchModel(await this.recipientsModel(), this.contactModel, suggestionsProvider, this.entityClient);
    }
    async conversationViewModel(options) {
        const factory = await this.conversationViewModelFactory();
        return factory(options);
    }
    async mailViewerViewModelFactory() {
        const { MailViewerViewModel } = await import("../../mail/view/MailViewerViewModel.js");
        return ({ mail, showFolder }) => new MailViewerViewModel(mail, showFolder, this.entityClient, this.mailModel, this.contactModel, this.configFacade, this.fileController, this.logins, async (mailboxDetails) => {
            const mailboxProperties = await this.mailModel.getMailboxProperties(mailboxDetails.mailboxGroupRoot);
            return this.sendMailModel(mailboxDetails, mailboxProperties);
        }, this.eventController, this.workerFacade, this.search, this.mailFacade, this.cryptoFacade, () => this.contactImporter());
    }
    async externalLoginViewModelFactory() {
        const { ExternalLoginViewModel } = await import("../../login/ExternalLoginView.js");
        return () => new ExternalLoginViewModel(this.credentialsProvider);
    }
    get native() {
        return this.getNativeInterface("native");
    }
    get fileApp() {
        return this.getNativeInterface("fileApp");
    }
    get pushService() {
        return this.getNativeInterface("pushService");
    }
    get commonSystemFacade() {
        return this.getNativeInterface("commonSystemFacade");
    }
    get themeFacade() {
        return this.getNativeInterface("themeFacade");
    }
    get systemFacade() {
        return this.getNativeInterface("mobileSystemFacade");
    }
    get mobileContactsFacade() {
        return this.getNativeInterface("mobileContactsFacade");
    }
    async mailAddressTableModelForOwnMailbox() {
        const { MailAddressTableModel } = await import("../../settings/mailaddress/MailAddressTableModel.js");
        const nameChanger = await this.ownMailAddressNameChanger();
        return new MailAddressTableModel(this.entityClient, this.serviceExecutor, this.mailAddressFacade, this.logins, this.eventController, this.logins.getUserController().userGroupInfo, nameChanger, await this.redraw());
    }
    async mailAddressTableModelForAdmin(mailGroupId, userId, userGroupInfo) {
        const { MailAddressTableModel } = await import("../../settings/mailaddress/MailAddressTableModel.js");
        const nameChanger = await this.adminNameChanger(mailGroupId, userId);
        return new MailAddressTableModel(this.entityClient, this.serviceExecutor, this.mailAddressFacade, this.logins, this.eventController, userGroupInfo, nameChanger, await this.redraw());
    }
    async ownMailAddressNameChanger() {
        const { OwnMailAddressNameChanger } = await import("../../settings/mailaddress/OwnMailAddressNameChanger.js");
        return new OwnMailAddressNameChanger(this.mailModel, this.entityClient);
    }
    async adminNameChanger(mailGroupId, userId) {
        const { AnotherUserMailAddressNameChanger } = await import("../../settings/mailaddress/AnotherUserMailAddressNameChanger.js");
        return new AnotherUserMailAddressNameChanger(this.mailAddressFacade, mailGroupId, userId);
    }
    async drawerAttrsFactory() {
        return () => ({
            logins: this.logins,
            newsModel: this.newsModel,
            desktopSystemFacade: this.desktopSystemFacade,
        });
    }
    domainConfigProvider() {
        return new DomainConfigProvider();
    }
    async credentialsRemovalHandler() {
        const { NoopCredentialRemovalHandler, AppsCredentialRemovalHandler } = await import("../../login/CredentialRemovalHandler.js");
        return isBrowser()
            ? new NoopCredentialRemovalHandler()
            : new AppsCredentialRemovalHandler(this.indexerFacade, this.pushService, this.configFacade, isApp() ? this.mobileContactsFacade : null);
    }
    async loginViewModelFactory() {
        const { LoginViewModel } = await import("../../login/LoginViewModel.js");
        const credentialsRemovalHandler = await locator.credentialsRemovalHandler();
        return () => {
            const domainConfig = isBrowser()
                ? locator.domainConfigProvider().getDomainConfigForHostname(location.hostname, location.protocol, location.port)
                : // in this case, we know that we have a staticUrl set that we need to use
                    locator.domainConfigProvider().getCurrentDomainConfig();
            return new LoginViewModel(locator.logins, locator.credentialsProvider, locator.secondFactorHandler, deviceConfig, domainConfig, credentialsRemovalHandler, isBrowser() ? null : this.pushService);
        };
    }
    getNativeInterface(name) {
        if (!this.nativeInterfaces) {
            throw new ProgrammingError(`Tried to use ${name} in web`);
        }
        return this.nativeInterfaces[name];
    }
    get initialized() {
        return this._deferredInitialized.promise;
    }
    constructor() {
        this.nativeInterfaces = null;
        this.exposedNativeInterfaces = null;
        this.recipientsModel = lazyMemoized(async () => {
            const { RecipientsModel } = await import("./RecipientsModel.js");
            return new RecipientsModel(this.contactModel, this.logins, this.mailFacade, this.entityClient);
        });
        this.redraw = lazyMemoized(async () => {
            const m = await import("mithril");
            return m.redraw;
        });
        this.offlineIndicatorViewModel = lazyMemoized(async () => {
            return new OfflineIndicatorViewModel(this.cacheStorage, this.loginListener, this.connectivityModel, this.logins, this.progressTracker, await this.redraw());
        });
        this.mailViewModel = lazyMemoized(async () => {
            const { MailViewModel } = await import("../../mail/view/MailViewModel.js");
            const conversationViewModelFactory = await this.conversationViewModelFactory();
            const router = new ScopedRouter(this.throttledRouter(), "/mail");
            return new MailViewModel(this.mailModel, this.entityClient, this.eventController, this.connectivityModel, this.cacheStorage, conversationViewModelFactory, this.mailOpenedListener, deviceConfig, this.inboxRuleHanlder(), router, await this.redraw());
        });
        this.throttledRouter = lazyMemoized(() => new ThrottledRouter());
        this.scopedSearchRouter = lazyMemoized(async () => {
            const { SearchRouter } = await import("../../search/view/SearchRouter.js");
            return new SearchRouter(new ScopedRouter(this.throttledRouter(), "/search"));
        });
        this.unscopedSearchRouter = lazyMemoized(async () => {
            const { SearchRouter } = await import("../../search/view/SearchRouter.js");
            return new SearchRouter(this.throttledRouter());
        });
        this.mailOpenedListener = {
            onEmailOpened: isDesktop()
                ? (mail) => {
                    this.desktopSystemFacade.sendSocketMessage(getDisplayedSender(mail).address);
                }
                : noOp,
        };
        this.contactViewModel = lazyMemoized(async () => {
            const { ContactViewModel } = await import("../../contacts/view/ContactViewModel.js");
            const router = new ScopedRouter(this.throttledRouter(), "/contact");
            return new ContactViewModel(this.contactModel, this.entityClient, this.eventController, router, await this.redraw());
        });
        this.contactListViewModel = lazyMemoized(async () => {
            const { ContactListViewModel } = await import("../../contacts/view/ContactListViewModel.js");
            const router = new ScopedRouter(this.throttledRouter(), "/contactlist");
            return new ContactListViewModel(this.entityClient, this.groupManagementFacade, this.logins, this.eventController, this.contactModel, await this.receivedGroupInvitationsModel(GroupType.ContactList), router, await this.redraw());
        });
        this.calendarViewModel = lazyMemoized(async () => {
            const { CalendarViewModel } = await import("../../calendar/view/CalendarViewModel.js");
            const { DefaultDateProvider } = await import("../../calendar/date/CalendarUtils");
            const timeZone = new DefaultDateProvider().timeZone();
            return new CalendarViewModel(this.logins, async (mode, event) => {
                const mailboxDetail = await this.mailModel.getUserMailboxDetails();
                const mailboxProperties = await this.mailModel.getMailboxProperties(mailboxDetail.mailboxGroupRoot);
                return await this.calendarEventModel(mode, event, mailboxDetail, mailboxProperties, null);
            }, (...args) => this.calendarEventPreviewModel(...args), await this.calendarModel(), await this.calendarEventsRepository(), this.entityClient, this.eventController, this.progressTracker, deviceConfig, await this.receivedGroupInvitationsModel(GroupType.Calendar), timeZone, this.mailModel);
        });
        this.calendarEventsRepository = lazyMemoized(async () => {
            const { CalendarEventsRepository } = await import("../../calendar/date/CalendarEventsRepository.js");
            const { DefaultDateProvider } = await import("../../calendar/date/CalendarUtils");
            const timeZone = new DefaultDateProvider().timeZone();
            return new CalendarEventsRepository(await this.calendarModel(), this.calendarFacade, timeZone, this.entityClient, this.eventController);
        });
        this.conversationViewModelFactory = async () => {
            const { ConversationViewModel } = await import("../../mail/view/ConversationViewModel.js");
            const factory = await this.mailViewerViewModelFactory();
            const m = await import("mithril");
            return (options) => {
                return new ConversationViewModel(options, (options) => factory(options), this.entityClient, this.eventController, deviceConfig, this.mailModel, m.redraw);
            };
        };
        this.contactImporter = async () => {
            const { ContactImporter } = await import("../../contacts/ContactImporter.js");
            return new ContactImporter(this.contactFacade, this.systemPermissionHandler);
        };
        this._deferredInitialized = defer();
        this.calendarModel = lazyMemoized(async () => {
            const { DefaultDateProvider } = await import("../../calendar/date/CalendarUtils");
            const { CalendarModel } = await import("../../calendar/model/CalendarModel");
            const timeZone = new DefaultDateProvider().timeZone();
            return new CalendarModel(notifications, this.alarmScheduler, this.eventController, this.serviceExecutor, this.logins, this.progressTracker, this.entityClient, this.mailModel, this.calendarFacade, this.fileController, timeZone);
        });
        this.calendarInviteHandler = lazyMemoized(async () => {
            const { CalendarInviteHandler } = await import("../../calendar/view/CalendarInvites.js");
            const { calendarNotificationSender } = await import("../../calendar/view/CalendarNotificationSender.js");
            return new CalendarInviteHandler(this.mailModel, await this.calendarModel(), this.logins, calendarNotificationSender, (...arg) => this.sendMailModel(...arg));
        });
        this.alarmScheduler = lazyMemoized(async () => {
            const { AlarmScheduler } = await import("../../calendar/date/AlarmScheduler");
            const { DefaultDateProvider } = await import("../../calendar/date/CalendarUtils");
            const dateProvider = new DefaultDateProvider();
            return new AlarmScheduler(dateProvider, await this.scheduler());
        });
        this.nativeContactsSyncManager = lazyMemoized(() => {
            assert(isApp(), "isApp");
            return new NativeContactsSyncManager(this.logins, this.mobileContactsFacade, this.entityClient, this.eventController, this.contactModel, deviceConfig);
        });
        this.postLoginActions = lazyMemoized(async () => {
            const { PostLoginActions } = await import("../../login/PostLoginActions");
            return new PostLoginActions(this.credentialsProvider, this.secondFactorHandler, this.connectivityModel, this.logins, await this.noZoneDateProvider(), this.entityClient, this.userManagementFacade, this.customerFacade, () => this.showSetupWizard(), this.loginFacade);
        });
        this.showSetupWizard = async () => {
            if (isApp()) {
                const { showSetupWizard } = await import("../../native/main/wizard/SetupWizard.js");
                return showSetupWizard(this.systemPermissionHandler, this.webMobileFacade, await this.contactImporter(), this.systemFacade, this.credentialsProvider, await this.nativeContactsSyncManager(), deviceConfig);
            }
        };
        this._workerDeferred = defer();
    }
    async init() {
        // Split init in two separate parts: creating modules and causing side effects.
        // We would like to do both on normal init but on HMR we just want to replace modules without a new worker. If we create a new
        // worker we end up losing state on the worker side (including our session).
        this.worker = bootstrapWorker(this);
        await this._createInstances();
        this._entropyCollector = new EntropyCollector(this.entropyFacade, await this.scheduler(), window);
        this._entropyCollector.start();
        this._deferredInitialized.resolve();
    }
    async _createInstances() {
        const { loginFacade, customerFacade, giftCardFacade, groupManagementFacade, configFacade, calendarFacade, mailFacade, shareFacade, counterFacade, indexerFacade, searchFacade, bookingFacade, mailAddressFacade, blobFacade, userManagementFacade, deviceEncryptionFacade, restInterface, serviceExecutor, cryptoFacade, cacheStorage, random, eventBus, entropyFacade, workerFacade, sqlCipherFacade, contactFacade, } = this.worker.getWorkerInterface();
        this.loginFacade = loginFacade;
        this.customerFacade = customerFacade;
        this.giftCardFacade = giftCardFacade;
        this.groupManagementFacade = groupManagementFacade;
        this.configFacade = configFacade;
        this.calendarFacade = calendarFacade;
        this.mailFacade = mailFacade;
        this.shareFacade = shareFacade;
        this.counterFacade = counterFacade;
        this.indexerFacade = indexerFacade;
        this.searchFacade = searchFacade;
        this.bookingFacade = bookingFacade;
        this.mailAddressFacade = mailAddressFacade;
        this.blobFacade = blobFacade;
        this.userManagementFacade = userManagementFacade;
        this.deviceEncryptionFacade = deviceEncryptionFacade;
        this.contactFacade = contactFacade;
        this.serviceExecutor = serviceExecutor;
        this.logins = new LoginController();
        // Should be called elsewhere later e.g. in mainLocator
        this.logins.init();
        this.eventController = new EventController(locator.logins);
        this.progressTracker = new ProgressTracker();
        this.search = new SearchModel(this.searchFacade, () => this.calendarEventsRepository());
        this.entityClient = new EntityClient(restInterface);
        this.cryptoFacade = cryptoFacade;
        this.cacheStorage = cacheStorage;
        this.entropyFacade = entropyFacade;
        this.workerFacade = workerFacade;
        this.connectivityModel = new WebsocketConnectivityModel(eventBus);
        this.mailModel = new MailModel(notifications, this.eventController, this.connectivityModel, this.mailFacade, this.entityClient, this.logins, this.inboxRuleHanlder());
        this.operationProgressTracker = new OperationProgressTracker();
        this.infoMessageHandler = new InfoMessageHandler(this.search);
        this.Const = Const;
        if (!isBrowser()) {
            const { WebDesktopFacade } = await import("../../native/main/WebDesktopFacade");
            const { WebMobileFacade } = await import("../../native/main/WebMobileFacade.js");
            const { WebCommonNativeFacade } = await import("../../native/main/WebCommonNativeFacade.js");
            const { WebInterWindowEventFacade } = await import("../../native/main/WebInterWindowEventFacade.js");
            const { WebAuthnFacadeSendDispatcher } = await import("../../native/common/generatedipc/WebAuthnFacadeSendDispatcher.js");
            const { createNativeInterfaces, createDesktopInterfaces } = await import("../../native/main/NativeInterfaceFactory.js");
            this.webMobileFacade = new WebMobileFacade(this.connectivityModel, this.mailModel);
            this.nativeInterfaces = createNativeInterfaces(this.webMobileFacade, new WebDesktopFacade(), new WebInterWindowEventFacade(this.logins, windowFacade, deviceConfig), new WebCommonNativeFacade(), cryptoFacade, calendarFacade, this.entityClient, this.logins);
            if (isElectronClient()) {
                const desktopInterfaces = createDesktopInterfaces(this.native);
                this.searchTextFacade = desktopInterfaces.searchTextFacade;
                this.interWindowEventSender = desktopInterfaces.interWindowEventSender;
                this.webAuthn = new WebauthnClient(new WebAuthnFacadeSendDispatcher(this.native), this.domainConfigProvider(), isApp());
                if (isDesktop()) {
                    this.desktopSettingsFacade = desktopInterfaces.desktopSettingsFacade;
                    this.desktopSystemFacade = desktopInterfaces.desktopSystemFacade;
                }
            }
            else if (isAndroidApp() || isIOSApp()) {
                const { SystemPermissionHandler } = await import("../../native/main/SystemPermissionHandler.js");
                this.systemPermissionHandler = new SystemPermissionHandler(this.systemFacade);
                this.webAuthn = new WebauthnClient(new WebAuthnFacadeSendDispatcher(this.native), this.domainConfigProvider(), isApp());
            }
        }
        if (this.webAuthn == null) {
            this.webAuthn = new WebauthnClient(new BrowserWebauthn(navigator.credentials, this.domainConfigProvider().getCurrentDomainConfig()), this.domainConfigProvider(), isApp());
        }
        this.secondFactorHandler = new SecondFactorHandler(this.eventController, this.entityClient, this.webAuthn, this.loginFacade, this.domainConfigProvider());
        this.loginListener = new PageContextLoginListener(this.secondFactorHandler);
        this.credentialsProvider = await createCredentialsProvider(deviceEncryptionFacade, this.nativeInterfaces?.native ?? null, sqlCipherFacade, isDesktop() ? this.interWindowEventSender : null);
        this.random = random;
        this.usageTestModel = new UsageTestModel({
            [0 /* StorageBehavior.Persist */]: deviceConfig,
            [1 /* StorageBehavior.Ephemeral */]: new EphemeralUsageTestStorage(),
        }, {
            now() {
                return Date.now();
            },
            timeZone() {
                throw new Error("Not implemented by this provider");
            },
        }, this.serviceExecutor, this.entityClient, this.logins, this.eventController, () => this.usageTestController);
        this.newsModel = new NewsModel(this.serviceExecutor, deviceConfig, async (name) => {
            switch (name) {
                case "usageOptIn":
                    const { UsageOptInNews } = await import("../../misc/news/items/UsageOptInNews.js");
                    return new UsageOptInNews(this.newsModel, this.usageTestModel);
                case "recoveryCode":
                    const { RecoveryCodeNews } = await import("../../misc/news/items/RecoveryCodeNews.js");
                    return new RecoveryCodeNews(this.newsModel, this.logins.getUserController(), this.userManagementFacade);
                case "pinBiometrics":
                    const { PinBiometricsNews } = await import("../../misc/news/items/PinBiometricsNews.js");
                    return new PinBiometricsNews(this.newsModel, this.credentialsProvider, this.logins.getUserController().userId);
                case "referralLink":
                    const { ReferralLinkNews } = await import("../../misc/news/items/ReferralLinkNews.js");
                    const dateProvider = await this.noZoneDateProvider();
                    return new ReferralLinkNews(this.newsModel, dateProvider, this.logins.getUserController());
                case "newPlans":
                    const { NewPlansNews } = await import("../../misc/news/items/NewPlansNews.js");
                    return new NewPlansNews(this.newsModel, this.logins.getUserController());
                case "newPlansOfferEnding":
                    const { NewPlansOfferEndingNews } = await import("../../misc/news/items/NewPlansOfferEndingNews.js");
                    return new NewPlansOfferEndingNews(this.newsModel, this.logins.getUserController());
                default:
                    console.log(`No implementation for news named '${name}'`);
                    return null;
            }
        });
        this.fileController =
            this.nativeInterfaces == null
                ? new FileControllerBrowser(blobFacade, guiDownload)
                : new FileControllerNative(blobFacade, guiDownload, this.nativeInterfaces.fileApp);
        const { ContactModel } = await import("../../contacts/model/ContactModel");
        this.contactModel = new ContactModel(this.searchFacade, this.entityClient, this.logins, this.eventController);
        this.minimizedMailModel = new MinimizedMailEditorViewModel();
        this.usageTestController = new UsageTestController(this.usageTestModel);
    }
    async scheduler() {
        const dateProvider = await this.noZoneDateProvider();
        return new SchedulerImpl(dateProvider, window, window);
    }
    async calendarEventPreviewModel(selectedEvent, calendars) {
        const { findAttendeeInAddresses } = await import("../common/utils/CommonCalendarUtils.js");
        const { getEventType } = await import("../../calendar/gui/CalendarGuiUtils.js");
        const { CalendarEventPreviewViewModel } = await import("../../calendar/gui/eventpopup/CalendarEventPreviewViewModel.js");
        const mailboxDetails = await this.mailModel.getUserMailboxDetails();
        const mailboxProperties = await this.mailModel.getMailboxProperties(mailboxDetails.mailboxGroupRoot);
        const userController = this.logins.getUserController();
        const customer = await userController.loadCustomer();
        const ownMailAddresses = getEnabledMailAddressesWithUser(mailboxDetails, userController.userGroupInfo);
        const ownAttendee = findAttendeeInAddresses(selectedEvent.attendees, ownMailAddresses);
        const eventType = getEventType(selectedEvent, calendars, ownMailAddresses, userController.user);
        const hasBusinessFeature = isCustomizationEnabledForCustomer(customer, FeatureType.BusinessFeatureEnabled) || (await userController.isNewPaidPlan());
        const lazyIndexEntry = async () => (selectedEvent.uid != null ? this.calendarFacade.getEventsByUid(selectedEvent.uid) : null);
        const popupModel = new CalendarEventPreviewViewModel(selectedEvent, await this.calendarModel(), eventType, hasBusinessFeature, ownAttendee, lazyIndexEntry, async (mode) => this.calendarEventModel(mode, selectedEvent, mailboxDetails, mailboxProperties, null));
        // If we have a preview model we want to display the description
        // so makes sense to already sanitize it after building the event
        await popupModel.sanitizeDescription();
        return popupModel;
    }
    // For testing argon2 migration after login. The production server will reject this request.
    // This can be removed when we enable the migration.
    async changeToBycrypt(passphrase) {
        const currentUser = this.logins.getUserController().user;
        return this.loginFacade.migrateKdfType(KdfType.Bcrypt, passphrase, currentUser);
    }
}
export const locator = new MainLocator();
if (typeof window !== "undefined") {
    window.tutao.locator = locator;
}
//# sourceMappingURL=MainLocator.js.map