import { UserTypeRef } from "../entities/sys/TypeRefs.js";
import { isAdminClient, isTest } from "../common/Env.js";
import { isSameTypeRefByAttr } from "@tutao/tutanota-utils";
import { isSameId } from "../common/utils/EntityUtils.js";
/** A bit of glue to distribute event bus events across the app. */
export class EventBusEventCoordinator {
    constructor(worker, connectivityListener, mailFacade, indexer, userFacade, entityClient, eventController, configurationDatabase) {
        this.worker = worker;
        this.connectivityListener = connectivityListener;
        this.mailFacade = mailFacade;
        this.indexer = indexer;
        this.userFacade = userFacade;
        this.entityClient = entityClient;
        this.eventController = eventController;
        this.configurationDatabase = configurationDatabase;
    }
    onWebsocketStateChanged(state) {
        this.connectivityListener.updateWebSocketState(state);
    }
    async onEntityEventsReceived(events, batchId, groupId) {
        await this.entityEventsReceived(events);
        await (await this.mailFacade()).entityEventsReceived(events);
        await this.eventController.onEntityUpdateReceived(events, groupId);
        // Call the indexer in this last step because now the processed event is stored and the indexer has a separate event queue that
        // shall not receive the event twice.
        if (!isTest() && !isAdminClient()) {
            const queuedBatch = { groupId, batchId, events };
            const configurationDatabase = await this.configurationDatabase();
            await configurationDatabase.onEntityEventsReceived(queuedBatch);
            const indexer = await this.indexer();
            indexer.addBatchesToQueue([queuedBatch]);
            indexer.startProcessing();
        }
    }
    /**
     * @param markers only phishing (not spam) marker will be sent as websocket updates
     */
    async onPhishingMarkersReceived(markers) {
        ;
        (await this.mailFacade()).phishingMarkersUpdateReceived(markers);
    }
    onError(tutanotaError) {
        this.worker.sendError(tutanotaError);
    }
    onLeaderStatusChanged(leaderStatus) {
        this.connectivityListener.onLeaderStatusChanged(leaderStatus);
    }
    onCounterChanged(counter) {
        this.eventController.onCountersUpdateReceived(counter);
    }
    async entityEventsReceived(data) {
        // This is a compromise to not add entityClient to UserFacade which would introduce a circular dep.
        for (const update of data) {
            const user = this.userFacade.getUser();
            if (user != null &&
                update.operation === "1" /* OperationType.UPDATE */ &&
                isSameTypeRefByAttr(UserTypeRef, update.application, update.type) &&
                isSameId(user._id, update.instanceId)) {
                this.userFacade.updateUser(await this.entityClient.load(UserTypeRef, user._id));
            }
        }
    }
}
//# sourceMappingURL=EventBusEventCoordinator.js.map