import { arrayEquals, assertNotNull, base64ToUint8Array, downcast, isSameTypeRef, isSameTypeRefByAttr, neverNull, ofClass, promiseMap, stringToUtf8Uint8Array, TypeRef, uint8ArrayToBase64, uint8ArrayToHex, } from "@tutao/tutanota-utils";
import { AccountType, assertEnumValue, CryptoProtocolVersion, EncryptionAuthStatus, GroupType, SYSTEM_GROUP_MAIL_ADDRESS, } from "../../common/TutanotaConstants";
import { resolveTypeReference } from "../../common/EntityFunctions";
import { BucketKeyTypeRef, BucketPermissionTypeRef, createInstanceSessionKey, createPublicKeyGetIn, createPublicKeyPutIn, createUpdatePermissionKeyData, GroupInfoTypeRef, GroupTypeRef, PermissionTypeRef, PushIdentifierTypeRef, } from "../../entities/sys/TypeRefs.js";
import { ContactTypeRef, createEncryptTutanotaPropertiesData, createInternalRecipientKeyData, createSymEncInternalRecipientKeyData, FileTypeRef, MailTypeRef, TutanotaPropertiesTypeRef, } from "../../entities/tutanota/TypeRefs.js";
import { typeRefToPath } from "../rest/EntityRestClient";
import { LockedError, NotFoundError, PayloadTooLargeError, TooManyRequestsError } from "../../common/error/RestError";
import { SessionKeyNotFoundError } from "../../common/error/SessionKeyNotFoundError";
import { birthdayToIsoDate, oldBirthdayToBirthday } from "../../common/utils/BirthdayUtils";
import { assertWorkerOrNode } from "../../common/Env";
import { aes256RandomKey, aesEncrypt, bitArrayToUint8Array, bytesToKyberPublicKey, decryptKey, ENABLE_MAC, encryptEccKey, encryptKey, generateEccKeyPair, hexToRsaPublicKey, isPqKeyPairs, isPqPublicKey, isRsaEccKeyPair, isRsaOrRsaEccKeyPair, isRsaPublicKey, IV_BYTE_LENGTH, KeyPairType, random, sha256Hash, uint8ArrayToBitArray, } from "@tutao/tutanota-crypto";
import { RecipientNotResolvedError } from "../../common/error/RecipientNotResolvedError";
import { EncryptTutanotaPropertiesService } from "../../entities/tutanota/Services";
import { PublicKeyService, UpdatePermissionKeyService } from "../../entities/sys/Services";
import { elementIdPart, getElementId, getListId } from "../../common/utils/EntityUtils.js";
import { decodePQMessage, encodePQMessage } from "../facades/PQMessage.js";
import { CryptoError } from "@tutao/tutanota-crypto/error.js";
assertWorkerOrNode();
export function encryptBytes(sk, value) {
    return aesEncrypt(sk, value, random.generateRandomData(IV_BYTE_LENGTH), true, ENABLE_MAC);
}
export function encryptString(sk, value) {
    return aesEncrypt(sk, stringToUtf8Uint8Array(value), random.generateRandomData(IV_BYTE_LENGTH), true, ENABLE_MAC);
}
/**
 * Encrypts the key with the encryptingKey and return the encrypted key and the version of the encryptingKey.
 * @param encryptingKey the encrypting key.
 * @param key the key to be encrypted.
 */
export function encryptKeyWithVersionedKey(encryptingKey, key) {
    return {
        encryptingKeyVersion: encryptingKey.version,
        key: encryptKey(encryptingKey.object, key),
    };
}
export class CryptoFacade {
    constructor(userFacade, entityClient, restClient, rsa, serviceExecutor, instanceMapper, ownerEncSessionKeysUpdateQueue, pq, cache, keyLoaderFacade) {
        this.userFacade = userFacade;
        this.entityClient = entityClient;
        this.restClient = restClient;
        this.rsa = rsa;
        this.serviceExecutor = serviceExecutor;
        this.instanceMapper = instanceMapper;
        this.ownerEncSessionKeysUpdateQueue = ownerEncSessionKeysUpdateQueue;
        this.pq = pq;
        this.cache = cache;
        this.keyLoaderFacade = keyLoaderFacade;
    }
    async applyMigrationsForInstance(decryptedInstance) {
        const instanceType = downcast(decryptedInstance)._type;
        if (isSameTypeRef(instanceType, ContactTypeRef)) {
            const contact = downcast(decryptedInstance);
            try {
                if (!contact.birthdayIso && contact.oldBirthdayAggregate) {
                    contact.birthdayIso = birthdayToIsoDate(contact.oldBirthdayAggregate);
                    contact.oldBirthdayAggregate = null;
                    contact.oldBirthdayDate = null;
                    await this.entityClient.update(contact);
                }
                else if (!contact.birthdayIso && contact.oldBirthdayDate) {
                    contact.birthdayIso = birthdayToIsoDate(oldBirthdayToBirthday(contact.oldBirthdayDate));
                    contact.oldBirthdayDate = null;
                    await this.entityClient.update(contact);
                }
                else if (contact.birthdayIso && (contact.oldBirthdayAggregate || contact.oldBirthdayDate)) {
                    contact.oldBirthdayAggregate = null;
                    contact.oldBirthdayDate = null;
                    await this.entityClient.update(contact);
                }
            }
            catch (e) {
                if (!(e instanceof LockedError)) {
                    throw e;
                }
            }
        }
        return decryptedInstance;
    }
    async resolveSessionKeyForInstance(instance) {
        const typeModel = await resolveTypeReference(instance._type);
        return this.resolveSessionKey(typeModel, instance);
    }
    /** Helper for the rare cases when we needed it on the client side. */
    async resolveSessionKeyForInstanceBinary(instance) {
        const key = await this.resolveSessionKeyForInstance(instance);
        return key == null ? null : bitArrayToUint8Array(key);
    }
    /** Resolve a session key an {@param instance} using an already known {@param ownerKey}. */
    resolveSessionKeyWithOwnerKey(instance, ownerKey) {
        let key = instance._ownerEncSessionKey;
        if (typeof key === "string") {
            key = base64ToUint8Array(key);
        }
        return decryptKey(ownerKey, key);
    }
    async decryptSessionKey(instance, ownerEncSessionKey) {
        const gk = await this.keyLoaderFacade.loadSymGroupKey(instance._ownerGroup, ownerEncSessionKey.encryptingKeyVersion);
        return decryptKey(gk, ownerEncSessionKey.key);
    }
    /**
     * Returns the session key for the provided type/instance:
     * * null, if the instance is unencrypted
     * * the decrypted _ownerEncSessionKey, if it is available
     * * the public decrypted session key, otherwise
     *
     * @param typeModel the type model of the instance
     * @param instance The unencrypted (client-side) instance or encrypted (server-side) object literal
     */
    async resolveSessionKey(typeModel, instance) {
        try {
            if (!typeModel.encrypted) {
                return null;
            }
            if (instance.bucketKey) {
                // if we have a bucket key, then we need to cache the session keys stored in the bucket key for details, files, etc.
                // we need to do this BEFORE we check the owner enc session key
                const bucketKey = await this.convertBucketKeyToInstanceIfNecessary(instance.bucketKey);
                const resolvedSessionKeys = await this.resolveWithBucketKey(bucketKey, instance, typeModel);
                return resolvedSessionKeys.resolvedSessionKeyForInstance;
            }
            else if (instance._ownerEncSessionKey && this.userFacade.isFullyLoggedIn() && this.userFacade.hasGroup(instance._ownerGroup)) {
                const gk = await this.keyLoaderFacade.loadSymGroupKey(instance._ownerGroup, Number(instance._ownerKeyVersion ?? 0));
                return this.resolveSessionKeyWithOwnerKey(instance, gk);
            }
            else if (instance.ownerEncSessionKey) {
                // Likely a DataTransferType, so this is a service.
                const gk = await this.keyLoaderFacade.loadSymGroupKey(this.userFacade.getGroupId(GroupType.Mail), Number(instance.ownerKeyVersion ?? 0));
                return this.resolveSessionKeyWithOwnerKey(instance, gk);
            }
            else {
                // See PermissionType jsdoc for more info on permissions
                const permissions = await this.entityClient.loadAll(PermissionTypeRef, instance._permissions);
                return (await this.trySymmetricPermission(permissions)) ?? (await this.resolveWithPublicOrExternalPermission(permissions, instance, typeModel));
            }
        }
        catch (e) {
            if (e instanceof CryptoError) {
                console.log("failed to resolve session key", e);
                throw new SessionKeyNotFoundError("Crypto error while resolving session key for instance " + instance._id);
            }
            else {
                throw e;
            }
        }
    }
    /**
     * Takes a freshly JSON-parsed, unmapped object and apply migrations as necessary
     * @param typeRef
     * @param data
     * @return the unmapped and still encrypted instance
     */
    async applyMigrations(typeRef, data) {
        if (isSameTypeRef(typeRef, GroupInfoTypeRef) && data._ownerGroup == null) {
            return this.applyCustomerGroupOwnershipToGroupInfo(data);
        }
        else if (isSameTypeRef(typeRef, TutanotaPropertiesTypeRef) && data._ownerEncSessionKey == null) {
            return this.encryptTutanotaProperties(data);
        }
        else if (isSameTypeRef(typeRef, PushIdentifierTypeRef) && data._ownerEncSessionKey == null) {
            return this.addSessionKeyToPushIdentifier(data);
        }
        else {
            return data;
        }
    }
    /**
     * In case the given bucketKey is a literal the literal will be converted to an instance and return. In case the BucketKey is already an instance the instance is returned.
     * @param bucketKeyInstanceOrLiteral The bucket key as literal or instance
     */
    async convertBucketKeyToInstanceIfNecessary(bucketKeyInstanceOrLiteral) {
        if (!this.isLiteralInstance(bucketKeyInstanceOrLiteral)) {
            // bucket key was already decoded from base 64
            return bucketKeyInstanceOrLiteral;
        }
        else {
            // decryptAndMapToInstance is misleading here, but we want to map the BucketKey aggregate and its session key from a literal to an instance
            // to have the encrypted keys in binary format and not as base 64. There is actually no decryption ongoing, just mapToInstance.
            const bucketKeyTypeModel = await resolveTypeReference(BucketKeyTypeRef);
            return (await this.instanceMapper.decryptAndMapToInstance(bucketKeyTypeModel, bucketKeyInstanceOrLiteral, null));
        }
    }
    async resolveWithBucketKey(bucketKey, instance, typeModel) {
        const instanceElementId = this.getElementIdFromInstance(instance);
        let decBucketKey;
        let unencryptedSenderAuthStatus = null;
        let pqMessageSenderKey = null;
        if (bucketKey.keyGroup && bucketKey.pubEncBucketKey) {
            // bucket key is encrypted with public key for internal recipient
            const { decryptedBucketKey, pqMessageSenderIdentityPubKey } = await this.decryptBucketKeyWithKeyPairOfGroupAndPrepareAuthentication(bucketKey.keyGroup, bucketKey.pubEncBucketKey, Number(bucketKey.recipientKeyVersion));
            decBucketKey = decryptedBucketKey;
            pqMessageSenderKey = pqMessageSenderIdentityPubKey;
        }
        else if (bucketKey.groupEncBucketKey) {
            // received as secure external recipient or reply from secure external sender
            let keyGroup;
            const groupKeyVersion = Number(bucketKey.recipientKeyVersion);
            if (bucketKey.keyGroup) {
                // 1. Uses when receiving confidential replies from external users.
                // 2. legacy code path for old external clients that used to encrypt bucket keys with user group keys.
                keyGroup = bucketKey.keyGroup;
            }
            else {
                // by default, we try to decrypt the bucket key with the ownerGroupKey (e.g. secure external recipient)
                keyGroup = neverNull(instance._ownerGroup);
            }
            decBucketKey = await this.resolveWithGroupReference(keyGroup, groupKeyVersion, bucketKey.groupEncBucketKey);
            unencryptedSenderAuthStatus = EncryptionAuthStatus.AES_NO_AUTHENTICATION;
        }
        else {
            throw new SessionKeyNotFoundError(`encrypted bucket key not set on instance ${typeModel.name}`);
        }
        const resolvedSessionKeys = await this.collectAllInstanceSessionKeysAndAuthenticate(bucketKey, decBucketKey, instanceElementId, instance, typeModel, unencryptedSenderAuthStatus, pqMessageSenderKey);
        this.ownerEncSessionKeysUpdateQueue.updateInstanceSessionKeys(resolvedSessionKeys.instanceSessionKeys);
        // for symmetrically encrypted instances _ownerEncSessionKey is sent from the server.
        // in this case it is not yet and we need to set it because the rest of the app expects it.
        const groupKey = await this.keyLoaderFacade.getCurrentSymGroupKey(instance._ownerGroup); // get current key for encrypting
        this.setOwnerEncSessionKeyUnmapped(instance, encryptKeyWithVersionedKey(groupKey, resolvedSessionKeys.resolvedSessionKeyForInstance));
        return resolvedSessionKeys;
    }
    /**
     * Calculates the SHA-256 checksum of a string value as UTF-8 bytes and returns it as a base64-encoded string
     */
    async sha256(value) {
        return uint8ArrayToBase64(sha256Hash(stringToUtf8Uint8Array(value)));
    }
    /**
     * Decrypts the given encrypted bucket key with the group key of the given group. In case the current user is not
     * member of the key group the function tries to resolve the group key using the adminEncGroupKey.
     * This is necessary for resolving the BucketKey when receiving a reply from an external Mailbox.
     * @param keyGroup The group that holds the encryption key.
     * @param groupKeyVersion the version of the key from the keyGroup
     * @param groupEncBucketKey The group key encrypted bucket key.
     */
    async resolveWithGroupReference(keyGroup, groupKeyVersion, groupEncBucketKey) {
        if (this.userFacade.hasGroup(keyGroup)) {
            // the logged-in user (most likely external) is a member of that group. Then we have the group key from the memberships
            const groupKey = await this.keyLoaderFacade.loadSymGroupKey(keyGroup, groupKeyVersion);
            return decryptKey(groupKey, groupEncBucketKey);
        }
        else {
            // internal user receiving a mail from secure external:
            // internal user group key -> external user group key -> external mail group key -> bucket key
            const externalMailGroupId = keyGroup;
            const externalMailGroupKeyVersion = groupKeyVersion;
            const externalMailGroup = await this.entityClient.load(GroupTypeRef, externalMailGroupId);
            const externalUserGroupdId = externalMailGroup.admin;
            if (!externalUserGroupdId) {
                throw new SessionKeyNotFoundError("no admin group on key group: " + externalMailGroupId);
            }
            const externalUserGroupKeyVersion = Number(externalMailGroup.adminGroupKeyVersion ?? 0);
            const externalUserGroup = await this.entityClient.load(GroupTypeRef, externalUserGroupdId);
            const internalUserGroupId = externalUserGroup.admin;
            const internalUserGroupKeyVersion = Number(externalUserGroup.adminGroupKeyVersion ?? 0);
            if (!(internalUserGroupId && this.userFacade.hasGroup(internalUserGroupId))) {
                throw new SessionKeyNotFoundError("no admin group or no membership of admin group: " + internalUserGroupId);
            }
            const internalUserGroupKey = await this.keyLoaderFacade.loadSymGroupKey(internalUserGroupId, internalUserGroupKeyVersion);
            const currentExternalUserGroupKey = decryptKey(internalUserGroupKey, assertNotNull(externalUserGroup.adminGroupEncGKey));
            const externalUserGroupKey = await this.keyLoaderFacade.loadSymGroupKey(externalUserGroupdId, externalUserGroupKeyVersion, {
                object: currentExternalUserGroupKey,
                version: Number(externalUserGroup.groupKeyVersion),
            });
            const currentExternalMailGroupKey = decryptKey(externalUserGroupKey, assertNotNull(externalMailGroup.adminGroupEncGKey));
            const externalMailGroupKey = await this.keyLoaderFacade.loadSymGroupKey(externalMailGroupId, externalMailGroupKeyVersion, {
                object: currentExternalMailGroupKey,
                version: Number(externalMailGroup.groupKeyVersion),
            });
            return decryptKey(externalMailGroupKey, groupEncBucketKey);
        }
    }
    async addSessionKeyToPushIdentifier(data) {
        const userGroupKey = this.userFacade.getCurrentUserGroupKey();
        // set sessionKey for allowing encryption when old instance (< v43) is updated
        const typeModel = await resolveTypeReference(PushIdentifierTypeRef);
        await this.updateOwnerEncSessionKey(typeModel, data, userGroupKey, aes256RandomKey());
        return data;
    }
    async encryptTutanotaProperties(data) {
        const userGroupKey = this.userFacade.getCurrentUserGroupKey();
        // EncryptTutanotaPropertiesService could be removed and replaced with a Migration that writes the key
        const groupEncSessionKey = encryptKeyWithVersionedKey(userGroupKey, aes256RandomKey());
        this.setOwnerEncSessionKeyUnmapped(data, groupEncSessionKey, this.userFacade.getUserGroupId());
        const migrationData = createEncryptTutanotaPropertiesData({
            properties: data._id,
            symKeyVersion: String(groupEncSessionKey.encryptingKeyVersion),
            symEncSessionKey: groupEncSessionKey.key,
        });
        await this.serviceExecutor.post(EncryptTutanotaPropertiesService, migrationData);
        return data;
    }
    async applyCustomerGroupOwnershipToGroupInfo(data) {
        const customerGroupMembership = assertNotNull(this.userFacade.getLoggedInUser().memberships.find((g) => g.groupType === GroupType.Customer));
        const listPermissions = await this.entityClient.loadAll(PermissionTypeRef, data._id[0]);
        const customerGroupPermission = listPermissions.find((p) => p.group === customerGroupMembership.group);
        if (!customerGroupPermission)
            throw new SessionKeyNotFoundError("Permission not found, could not apply OwnerGroup migration");
        const customerGroupKeyVersion = Number(customerGroupPermission.symKeyVersion ?? 0);
        const customerGroupKey = await this.keyLoaderFacade.loadSymGroupKey(customerGroupMembership.group, customerGroupKeyVersion);
        const versionedCustomerGroupKey = { object: customerGroupKey, version: customerGroupKeyVersion };
        const listKey = decryptKey(customerGroupKey, assertNotNull(customerGroupPermission.symEncSessionKey));
        const groupInfoSk = decryptKey(listKey, base64ToUint8Array(data._listEncSessionKey));
        this.setOwnerEncSessionKeyUnmapped(data, encryptKeyWithVersionedKey(versionedCustomerGroupKey, groupInfoSk), customerGroupMembership.group);
        return data;
    }
    setOwnerEncSessionKeyUnmapped(unmappedInstance, key, ownerGroup) {
        unmappedInstance._ownerEncSessionKey = uint8ArrayToBase64(key.key);
        unmappedInstance._ownerKeyVersion = key.encryptingKeyVersion.toString();
        if (ownerGroup) {
            unmappedInstance._ownerGroup;
        }
    }
    setOwnerEncSessionKey(instance, key) {
        instance._ownerEncSessionKey = key.key;
        instance._ownerKeyVersion = key.encryptingKeyVersion.toString();
    }
    isLiteralInstance(elementOrLiteral) {
        return typeof elementOrLiteral._type === "undefined";
    }
    async trySymmetricPermission(listPermissions) {
        const symmetricPermission = listPermissions.find((p) => (p.type === "2" /* PermissionType.Public_Symmetric */ || p.type === "1" /* PermissionType.Symmetric */) &&
            p._ownerGroup &&
            this.userFacade.hasGroup(p._ownerGroup)) ?? null;
        if (symmetricPermission) {
            const gk = await this.keyLoaderFacade.loadSymGroupKey(assertNotNull(symmetricPermission._ownerGroup), Number(symmetricPermission._ownerKeyVersion ?? 0));
            return decryptKey(gk, assertNotNull(symmetricPermission._ownerEncSessionKey));
        }
        else {
            return null;
        }
    }
    /**
     * Resolves the session key for the provided instance and collects all other instances'
     * session keys in order to update them.
     */
    async collectAllInstanceSessionKeysAndAuthenticate(bucketKey, decBucketKey, instanceElementId, instance, typeModel, encryptionAuthStatus, pqMessageSenderKey) {
        let resolvedSessionKeyForInstance = undefined;
        const instanceSessionKeys = await promiseMap(bucketKey.bucketEncSessionKeys, async (instanceSessionKey) => {
            const decryptedSessionKey = decryptKey(decBucketKey, instanceSessionKey.symEncSessionKey);
            const groupKey = await this.keyLoaderFacade.getCurrentSymGroupKey(instance._ownerGroup);
            const ownerEncSessionKey = encryptKeyWithVersionedKey(groupKey, decryptedSessionKey);
            const instanceSessionKeyWithOwnerEncSessionKey = createInstanceSessionKey(instanceSessionKey);
            if (instanceElementId == instanceSessionKey.instanceId) {
                resolvedSessionKeyForInstance = decryptedSessionKey;
                // we can only authenticate once we have the session key
                // because we need to check if the confidential flag is set, which is encrypted still
                // we need to do it here at the latest because we must write the flag when updating the session key on the instance
                await this.authenticateMainInstance(typeModel, encryptionAuthStatus, pqMessageSenderKey, bucketKey.protocolVersion === CryptoProtocolVersion.TUTA_CRYPT ? Number(bucketKey.senderKeyVersion ?? 0) : null, instance, resolvedSessionKeyForInstance, instanceSessionKeyWithOwnerEncSessionKey, decryptedSessionKey);
            }
            instanceSessionKeyWithOwnerEncSessionKey.symEncSessionKey = ownerEncSessionKey.key;
            instanceSessionKeyWithOwnerEncSessionKey.symKeyVersion = String(ownerEncSessionKey.encryptingKeyVersion);
            return instanceSessionKeyWithOwnerEncSessionKey;
        });
        if (resolvedSessionKeyForInstance) {
            return { resolvedSessionKeyForInstance, instanceSessionKeys };
        }
        else {
            throw new SessionKeyNotFoundError("no session key for instance " + instance._id);
        }
    }
    async authenticateMainInstance(typeModel, encryptionAuthStatus, pqMessageSenderKey, pqMessageSenderKeyVersion, instance, resolvedSessionKeyForInstance, instanceSessionKeyWithOwnerEncSessionKey, decryptedSessionKey) {
        // we only authenticate mail instances
        const isMailInstance = isSameTypeRefByAttr(MailTypeRef, typeModel.app, typeModel.name);
        if (isMailInstance) {
            if (!encryptionAuthStatus) {
                if (!pqMessageSenderKey) {
                    encryptionAuthStatus = EncryptionAuthStatus.RSA_NO_AUTHENTICATION;
                }
                else {
                    const mail = (await this.instanceMapper.decryptAndMapToInstance(typeModel, instance, resolvedSessionKeyForInstance));
                    const senderMailAddress = mail.confidential ? mail.sender.address : SYSTEM_GROUP_MAIL_ADDRESS;
                    encryptionAuthStatus = await this.authenticateSender(senderMailAddress, pqMessageSenderKey, assertNotNull(pqMessageSenderKeyVersion));
                }
            }
            instanceSessionKeyWithOwnerEncSessionKey.encryptionAuthStatus = aesEncrypt(decryptedSessionKey, stringToUtf8Uint8Array(encryptionAuthStatus));
        }
    }
    async resolveWithPublicOrExternalPermission(listPermissions, instance, typeModel) {
        const pubOrExtPermission = listPermissions.find((p) => p.type === "0" /* PermissionType.Public */ || p.type === "5" /* PermissionType.External */) ?? null;
        if (pubOrExtPermission == null) {
            const typeName = `${typeModel.app}/${typeModel.name}`;
            throw new SessionKeyNotFoundError(`could not find permission for instance of type ${typeName} with id ${this.getElementIdFromInstance(instance)}`);
        }
        const bucketPermissions = await this.entityClient.loadAll(BucketPermissionTypeRef, assertNotNull(pubOrExtPermission.bucket).bucketPermissions);
        const bucketPermission = bucketPermissions.find((bp) => (bp.type === "2" /* BucketPermissionType.Public */ || bp.type === "3" /* BucketPermissionType.External */) && pubOrExtPermission._ownerGroup === bp._ownerGroup);
        // find the bucket permission with the same group as the permission and public type
        if (bucketPermission == null) {
            throw new SessionKeyNotFoundError("no corresponding bucket permission found");
        }
        if (bucketPermission.type === "3" /* BucketPermissionType.External */) {
            return this.decryptWithExternalBucket(bucketPermission, pubOrExtPermission, instance);
        }
        else {
            return this.decryptWithPublicBucket(bucketPermission, instance, pubOrExtPermission, typeModel);
        }
    }
    async decryptWithExternalBucket(bucketPermission, pubOrExtPermission, instance) {
        let bucketKey;
        if (bucketPermission.ownerEncBucketKey != null) {
            const ownerGroupKey = await this.keyLoaderFacade.loadSymGroupKey(neverNull(bucketPermission._ownerGroup), Number(bucketPermission.ownerKeyVersion ?? 0));
            bucketKey = decryptKey(ownerGroupKey, bucketPermission.ownerEncBucketKey);
        }
        else if (bucketPermission.symEncBucketKey) {
            // legacy case: for very old email sent to external user we used symEncBucketKey on the bucket permission.
            // The bucket key is encrypted with the user group key of the external user.
            // We maintain this code as we still have some old BucketKeys in some external mailboxes.
            // Can be removed if we finished mail details migration or when we do cleanup of external mailboxes.
            const userGroupKey = await this.keyLoaderFacade.loadSymUserGroupKey(Number(bucketPermission.symKeyVersion ?? 0));
            bucketKey = decryptKey(userGroupKey, bucketPermission.symEncBucketKey);
        }
        else {
            throw new SessionKeyNotFoundError(`BucketEncSessionKey is not defined for Permission ${pubOrExtPermission._id.toString()} (Instance: ${JSON.stringify(instance)})`);
        }
        return decryptKey(bucketKey, neverNull(pubOrExtPermission.bucketEncSessionKey));
    }
    async decryptBucketKeyWithKeyPairOfGroupAndPrepareAuthentication(keyPairGroupId, pubEncBucketKey, recipientKeyVersion) {
        const keyPair = await this.keyLoaderFacade.loadKeypair(keyPairGroupId, recipientKeyVersion);
        const algo = keyPair.keyPairType;
        if (isPqKeyPairs(keyPair)) {
            const pqMessage = decodePQMessage(pubEncBucketKey);
            const decryptedBucketKey = await this.pq.decapsulate(pqMessage, keyPair);
            return {
                decryptedBucketKey: uint8ArrayToBitArray(decryptedBucketKey),
                pqMessageSenderIdentityPubKey: pqMessage.senderIdentityPubKey,
            };
        }
        else if (isRsaOrRsaEccKeyPair(keyPair)) {
            const privateKey = keyPair.privateKey;
            const decryptedBucketKey = await this.rsa.decrypt(privateKey, pubEncBucketKey);
            return {
                decryptedBucketKey: uint8ArrayToBitArray(decryptedBucketKey),
                pqMessageSenderIdentityPubKey: null,
            };
        }
        else {
            throw new CryptoError("unknown key pair type: " + algo);
        }
    }
    async decryptWithPublicBucket(bucketPermission, instance, pubOrExtPermission, typeModel) {
        const pubEncBucketKey = bucketPermission.pubEncBucketKey;
        if (pubEncBucketKey == null) {
            throw new SessionKeyNotFoundError(`PubEncBucketKey is not defined for BucketPermission ${bucketPermission._id.toString()} (Instance: ${JSON.stringify(instance)})`);
        }
        const bucketEncSessionKey = pubOrExtPermission.bucketEncSessionKey;
        if (bucketEncSessionKey == null) {
            throw new SessionKeyNotFoundError(`BucketEncSessionKey is not defined for Permission ${pubOrExtPermission._id.toString()} (Instance: ${JSON.stringify(instance)})`);
        }
        const { decryptedBucketKey } = await this.decryptBucketKeyWithKeyPairOfGroupAndPrepareAuthentication(bucketPermission.group, pubEncBucketKey, Number(bucketPermission.pubKeyVersion ?? 0));
        const sk = decryptKey(decryptedBucketKey, bucketEncSessionKey);
        if (bucketPermission._ownerGroup) {
            // is not defined for some old AccountingInfos
            let bucketPermissionOwnerGroupKey = await this.keyLoaderFacade.getCurrentSymGroupKey(neverNull(bucketPermission._ownerGroup)); // get current key for encrypting
            await this.updateWithSymPermissionKey(typeModel, instance, pubOrExtPermission, bucketPermission, bucketPermissionOwnerGroupKey, sk).catch(ofClass(NotFoundError, () => {
                console.log("w> could not find instance to update permission");
            }));
        }
        return sk;
    }
    /**
     * Returns the session key for the provided service response:
     * * null, if the instance is unencrypted
     * * the decrypted _ownerPublicEncSessionKey, if it is available
     * @param instance The unencrypted (client-side) or encrypted (server-side) instance
     *
     */
    async resolveServiceSessionKey(instance) {
        if (instance._ownerPublicEncSessionKey) {
            // we assume the server uses the current key pair of the recipient
            const keypair = await this.keyLoaderFacade.loadCurrentKeyPair(instance._ownerGroup);
            // we do not authenticate as we could remove data transfer type encryption altogether and only rely on tls
            return this.unauthenticatedDecryptPubEncSessionKey(base64ToUint8Array(instance._ownerPublicEncSessionKey), assertEnumValue(CryptoProtocolVersion, instance._publicCryptoProtocolVersion), keypair.object);
        }
        return null;
    }
    async encryptPubSymKey(symKey, recipientPublicKeys, senderGroupId) {
        let pubEncSymKey, cryptoProtocolVersion;
        const recipientPublicKey = this.getRecipientPublicKey(recipientPublicKeys.object);
        const algo = recipientPublicKey.keyPairType;
        let senderKeyVersion;
        if (isPqPublicKey(recipientPublicKey)) {
            const senderKeyPair = await this.keyLoaderFacade.loadCurrentKeyPair(senderGroupId);
            const senderEccKeyPair = await this.getOrMakeSenderIdentityKeyPair(senderKeyPair.object, senderGroupId);
            const ephemeralKeyPair = generateEccKeyPair();
            pubEncSymKey = encodePQMessage(await this.pq.encapsulate(senderEccKeyPair, ephemeralKeyPair, recipientPublicKey, bitArrayToUint8Array(symKey)));
            cryptoProtocolVersion = CryptoProtocolVersion.TUTA_CRYPT;
            senderKeyVersion = senderKeyPair.version.toString();
        }
        else if (isRsaPublicKey(recipientPublicKey)) {
            pubEncSymKey = await this.rsa.encrypt(recipientPublicKey, bitArrayToUint8Array(symKey));
            cryptoProtocolVersion = CryptoProtocolVersion.RSA;
            senderKeyVersion = null;
        }
        else {
            throw new CryptoError("unknown public key type: " + algo);
        }
        return { pubEncSymKey, cryptoProtocolVersion, senderKeyVersion, recipientKeyVersion: String(recipientPublicKeys.version) };
    }
    /**
     * Decrypts the pubEncSymKey with the keyPair.
     * @param pubEncSessionKey the asymmetrically encrypted session key
     * @param cryptoProtocolVersion asymmetric protocol to decrypt pubEncSessionKey (RSA or TutaCrypt)
     * @param keyPair the current keyPair
     */
    async unauthenticatedDecryptPubEncSessionKey(pubEncSessionKey, cryptoProtocolVersion, keyPair) {
        let decryptedBytes;
        switch (cryptoProtocolVersion) {
            case CryptoProtocolVersion.RSA: {
                if (!isRsaOrRsaEccKeyPair(keyPair)) {
                    throw new CryptoError("wrong key type. expecte rsa. got " + keyPair.keyPairType);
                }
                decryptedBytes = await this.rsa.decrypt(keyPair.privateKey, pubEncSessionKey);
                break;
            }
            case CryptoProtocolVersion.TUTA_CRYPT: {
                if (!isPqKeyPairs(keyPair)) {
                    throw new CryptoError("wrong key type. expected tuta-crypt. got " + keyPair.keyPairType);
                }
                decryptedBytes = await this.pq.decapsulateEncoded(pubEncSessionKey, keyPair);
                break;
            }
            default:
                throw new CryptoError("invalid cryptoProtocolVersion: " + cryptoProtocolVersion);
        }
        return uint8ArrayToBitArray(decryptedBytes);
    }
    /**
     * Creates a new _ownerEncSessionKey and assigns it to the provided entity
     * the entity must already have an _ownerGroup
     * @returns the generated key
     */
    async setNewOwnerEncSessionKey(model, entity, keyToEncryptSessionKey) {
        if (!entity._ownerGroup) {
            throw new Error(`no owner group set  ${JSON.stringify(entity)}`);
        }
        if (model.encrypted) {
            if (entity._ownerEncSessionKey) {
                throw new Error(`ownerEncSessionKey already set ${JSON.stringify(entity)}`);
            }
            const sessionKey = aes256RandomKey();
            const effectiveKeyToEncryptSessionKey = keyToEncryptSessionKey ?? (await this.keyLoaderFacade.getCurrentSymGroupKey(entity._ownerGroup));
            const encryptedSessionKey = encryptKeyWithVersionedKey(effectiveKeyToEncryptSessionKey, sessionKey);
            this.setOwnerEncSessionKey(entity, encryptedSessionKey);
            return sessionKey;
        }
        else {
            return null;
        }
    }
    async encryptBucketKeyForInternalRecipient(senderUserGroupId, bucketKey, recipientMailAddress, notFoundRecipients) {
        let keyData = createPublicKeyGetIn({
            mailAddress: recipientMailAddress,
            version: null,
        });
        try {
            const publicKeyGetOut = await this.serviceExecutor.get(PublicKeyService, keyData);
            // We do not create any key data in case there is one not found recipient, but we want to
            // collect ALL not found recipients when iterating a recipient list.
            if (notFoundRecipients.length !== 0) {
                return null;
            }
            const isExternalSender = this.userFacade.getUser()?.accountType === AccountType.EXTERNAL;
            // we only encrypt symmetric as external sender if the recipient supports tuta-crypt.
            // Clients need to support symmetric decryption from external users. We can always encrypt symmetricly when old clients are deactivated that don't support tuta-crypt.
            if (publicKeyGetOut.pubKyberKey && isExternalSender) {
                return this.createSymEncInternalRecipientKeyData(recipientMailAddress, bucketKey);
            }
            else {
                return this.createPubEncInternalRecipientKeyData(bucketKey, recipientMailAddress, publicKeyGetOut, senderUserGroupId);
            }
        }
        catch (e) {
            if (e instanceof NotFoundError) {
                notFoundRecipients.push(recipientMailAddress);
                return null;
            }
            else if (e instanceof TooManyRequestsError) {
                throw new RecipientNotResolvedError("");
            }
            else {
                throw e;
            }
        }
    }
    async createPubEncInternalRecipientKeyData(bucketKey, recipientMailAddress, publicKeyGetOut, senderGroupId) {
        const recipientPublicKeys = {
            object: {
                pubRsaKey: publicKeyGetOut.pubRsaKey,
                pubKyberKey: publicKeyGetOut.pubKyberKey,
                pubEccKey: publicKeyGetOut.pubEccKey,
            },
            version: Number(publicKeyGetOut.pubKeyVersion),
        };
        const pubEncBucketKey = await this.encryptPubSymKey(bucketKey, recipientPublicKeys, senderGroupId);
        return createInternalRecipientKeyData({
            mailAddress: recipientMailAddress,
            pubEncBucketKey: pubEncBucketKey.pubEncSymKey,
            recipientKeyVersion: pubEncBucketKey.recipientKeyVersion,
            senderKeyVersion: pubEncBucketKey.senderKeyVersion,
            protocolVersion: pubEncBucketKey.cryptoProtocolVersion,
        });
    }
    async createSymEncInternalRecipientKeyData(recipientMailAddress, bucketKey) {
        const keyGroup = this.userFacade.getGroupId(GroupType.Mail);
        const externalMailGroupKey = await this.keyLoaderFacade.getCurrentSymGroupKey(keyGroup);
        return createSymEncInternalRecipientKeyData({
            mailAddress: recipientMailAddress,
            symEncBucketKey: encryptKey(externalMailGroupKey.object, bucketKey),
            keyGroup,
            symKeyVersion: String(externalMailGroupKey.version),
        });
    }
    /**
     * Returns the SenderIdentityKeyPair that is either already on the KeyPair that is being passed in,
     * or creates a new one and writes it to the respective Group.
     * @param senderKeyPair
     * @param keyGroupId Id for the Group that Public Key Service might write a new IdentityKeyPair for.
     * 						This is necessary as a User might send an E-Mail from a shared mailbox,
     * 						for which the KeyPair should be created.
     */
    async getOrMakeSenderIdentityKeyPair(senderKeyPair, keyGroupId) {
        const algo = senderKeyPair.keyPairType;
        if (isPqKeyPairs(senderKeyPair)) {
            return senderKeyPair.eccKeyPair;
        }
        else if (isRsaEccKeyPair(senderKeyPair)) {
            return { publicKey: senderKeyPair.publicEccKey, privateKey: senderKeyPair.privateEccKey };
        }
        else if (isRsaOrRsaEccKeyPair(senderKeyPair)) {
            const symGroupKey = await this.keyLoaderFacade.getCurrentSymGroupKey(keyGroupId);
            const newIdentityKeyPair = generateEccKeyPair();
            const symEncPrivEccKey = encryptEccKey(symGroupKey.object, newIdentityKeyPair.privateKey);
            const data = createPublicKeyPutIn({ pubEccKey: newIdentityKeyPair.publicKey, symEncPrivEccKey, keyGroup: keyGroupId });
            await this.serviceExecutor.put(PublicKeyService, data);
            return newIdentityKeyPair;
        }
        else {
            throw new CryptoError("unknow key pair type: " + algo);
        }
    }
    async authenticateSender(mailSenderAddress, senderIdentityPubKey, senderKeyVersion) {
        let keyData = createPublicKeyGetIn({
            mailAddress: mailSenderAddress,
            version: senderKeyVersion.toString(),
        });
        try {
            const publicKeyGetOut = await this.serviceExecutor.get(PublicKeyService, keyData);
            return publicKeyGetOut.pubEccKey != null && arrayEquals(publicKeyGetOut.pubEccKey, senderIdentityPubKey)
                ? EncryptionAuthStatus.PQ_AUTHENTICATION_SUCCEEDED
                : EncryptionAuthStatus.PQ_AUTHENTICATION_FAILED;
        }
        catch (e) {
            console.error("Could not authenticate sender", e);
            return EncryptionAuthStatus.PQ_AUTHENTICATION_FAILED;
        }
    }
    /**
     * Updates the given public permission with the given symmetric key for faster access if the client is the leader and otherwise does nothing.
     * @param typeModel The type model of the instance
     * @param instance The unencrypted (client-side) or encrypted (server-side) instance
     * @param permission The permission.
     * @param bucketPermission The bucket permission.
     * @param permissionOwnerGroupKey The symmetric group key for the owner group on the permission.
     * @param sessionKey The symmetric session key.
     */
    async updateWithSymPermissionKey(typeModel, instance, permission, bucketPermission, permissionOwnerGroupKey, sessionKey) {
        if (!this.isLiteralInstance(instance) || !this.userFacade.isLeader()) {
            // do not update the session key in case of an unencrypted (client-side) instance
            // or in case we are not the leader client
            return;
        }
        if (!instance._ownerEncSessionKey && permission._ownerGroup === instance._ownerGroup) {
            return this.updateOwnerEncSessionKey(typeModel, instance, permissionOwnerGroupKey, sessionKey);
        }
        else {
            // instances shared via permissions (e.g. body)
            const encryptedKey = encryptKeyWithVersionedKey(permissionOwnerGroupKey, sessionKey);
            let updateService = createUpdatePermissionKeyData({
                ownerKeyVersion: String(encryptedKey.encryptingKeyVersion),
                ownerEncSessionKey: encryptedKey.key,
                permission: permission._id,
                bucketPermission: bucketPermission._id,
            });
            await this.serviceExecutor.post(UpdatePermissionKeyService, updateService);
        }
    }
    /**
     * Resolves the ownerEncSessionKey of a mail. This might be needed if it wasn't updated yet
     * by the OwnerEncSessionKeysUpdateQueue but the file is already downloaded.
     * @param mainInstance the instance that has the bucketKey
     * @param childInstances the files that belong to the mainInstance
     */
    async enforceSessionKeyUpdateIfNeeded(mainInstance, childInstances) {
        if (!childInstances.some((f) => f._ownerEncSessionKey == null)) {
            return childInstances;
        }
        const typeModel = await resolveTypeReference(mainInstance._type);
        const outOfSyncInstances = childInstances.filter((f) => f._ownerEncSessionKey == null);
        if (mainInstance.bucketKey) {
            // invoke updateSessionKeys service in case a bucket key is still available
            const bucketKey = await this.convertBucketKeyToInstanceIfNecessary(mainInstance.bucketKey);
            const resolvedSessionKeys = await this.resolveWithBucketKey(bucketKey, mainInstance, typeModel);
            await this.ownerEncSessionKeysUpdateQueue.postUpdateSessionKeysService(resolvedSessionKeys.instanceSessionKeys);
        }
        else {
            console.warn("files are out of sync refreshing", outOfSyncInstances.map((f) => f._id).join(", "));
        }
        for (const childInstance of outOfSyncInstances) {
            await this.cache?.deleteFromCacheIfExists(FileTypeRef, getListId(childInstance), getElementId(childInstance));
        }
        // we have a caching entity client, so this re-inserts the deleted instances
        return await this.entityClient.loadMultiple(FileTypeRef, getListId(childInstances[0]), childInstances.map((childInstance) => getElementId(childInstance)));
    }
    updateOwnerEncSessionKey(typeModel, instance, ownerGroupKey, sessionKey) {
        this.setOwnerEncSessionKeyUnmapped(instance, encryptKeyWithVersionedKey(ownerGroupKey, sessionKey));
        // we have to call the rest client directly because instance is still the encrypted server-side version
        const path = typeRefToPath(new TypeRef(typeModel.app, typeModel.name)) + "/" + (instance._id instanceof Array ? instance._id.join("/") : instance._id);
        const headers = this.userFacade.createAuthHeaders();
        headers.v = typeModel.version;
        return this.restClient
            .request(path, "PUT" /* HttpMethod.PUT */, {
            headers,
            body: JSON.stringify(instance),
            queryParams: { updateOwnerEncSessionKey: "true" },
        })
            .catch(ofClass(PayloadTooLargeError, (e) => {
            console.log("Could not update owner enc session key - PayloadTooLargeError", e);
        }));
    }
    getElementIdFromInstance(instance) {
        if (typeof instance._id === "string") {
            return instance._id;
        }
        else {
            const idTuple = instance._id;
            return elementIdPart(idTuple);
        }
    }
    getRecipientPublicKey(publicKeys) {
        if (publicKeys.pubRsaKey) {
            // we ignore ecc keys as this is only used for the recipient keys
            return hexToRsaPublicKey(uint8ArrayToHex(publicKeys.pubRsaKey));
        }
        else if (publicKeys.pubKyberKey && publicKeys.pubEccKey) {
            var eccPublicKey = publicKeys.pubEccKey;
            var kyberPublicKey = bytesToKyberPublicKey(publicKeys.pubKyberKey);
            return { keyPairType: KeyPairType.TUTA_CRYPT, eccPublicKey, kyberPublicKey };
        }
        else {
            throw new Error("Inconsistent Keypair");
        }
    }
}
if (!("toJSON" in Error.prototype)) {
    Object.defineProperty(Error.prototype, "toJSON", {
        value: function () {
            const alt = {};
            for (let key of Object.getOwnPropertyNames(this)) {
                alt[key] = this[key];
            }
            return alt;
        },
        configurable: true,
        writable: true,
    });
}
//# sourceMappingURL=CryptoFacade.js.map