import { assertWorkerOrNode } from "../../common/Env";
import { BlobAccessTokenService } from "../../entities/storage/Services";
import { createBlobAccessTokenPostIn, createBlobReadData, createBlobWriteData, createInstanceId } from "../../entities/storage/TypeRefs";
import { resolveTypeReference } from "../../common/EntityFunctions.js";
assertWorkerOrNode();
/**
 * The BlobAccessTokenFacade requests blobAccessTokens from the BlobAccessTokenService to get or post to the BlobService (binary blobs)
 * or DefaultBlobElementResource (instances).
 *
 * All tokens are cached.
 */
export class BlobAccessTokenFacade {
    constructor(serviceExecutor, dateProvider, authDataProvider) {
        this.serviceExecutor = serviceExecutor;
        this.dateProvider = dateProvider;
        this.authDataProvider = authDataProvider;
        this.readArchiveCache = new BlobAccessTokenCache(dateProvider);
        this.readBlobCache = new BlobAccessTokenCache(dateProvider);
        this.writeCache = new BlobAccessTokenCache(dateProvider);
    }
    /**
     * Requests a token that allows uploading blobs for the given ArchiveDataType and ownerGroup.
     * @param archiveDataType The type of data that should be stored.
     * @param ownerGroupId The ownerGroup were the data belongs to (e.g. group of type mail)
     */
    async requestWriteToken(archiveDataType, ownerGroupId) {
        const requestNewToken = async () => {
            const tokenRequest = createBlobAccessTokenPostIn({
                archiveDataType,
                write: createBlobWriteData({
                    archiveOwnerGroup: ownerGroupId,
                }),
                read: null,
            });
            const { blobAccessInfo } = await this.serviceExecutor.post(BlobAccessTokenService, tokenRequest);
            return blobAccessInfo;
        };
        const key = this.makeWriteCacheKey(ownerGroupId, archiveDataType);
        return this.writeCache.getToken(key, requestNewToken);
    }
    makeWriteCacheKey(ownerGroupId, archiveDataType) {
        return ownerGroupId + archiveDataType;
    }
    /**
     * Remove a given write token from the cache.
     * @param archiveDataType
     * @param ownerGroupId
     */
    evictWriteToken(archiveDataType, ownerGroupId) {
        const key = this.makeWriteCacheKey(ownerGroupId, archiveDataType);
        this.writeCache.evict(key);
    }
    /**
     * Requests a token that grants read access to all blobs that are referenced by the given instance.
     * A user must be owner of the instance but must not be owner of the archive were the blobs are stored in.
     * @param archiveDataType specify the data type
     * @param referencingInstance the instance that references the blobs
     */
    async requestReadTokenBlobs(archiveDataType, referencingInstance) {
        const requestNewToken = async () => {
            const archiveId = this.getArchiveId(referencingInstance.blobs);
            const instanceListId = referencingInstance.listId;
            const instanceId = referencingInstance.elementId;
            const instanceIds = [createInstanceId({ instanceId })];
            const tokenRequest = createBlobAccessTokenPostIn({
                archiveDataType,
                read: createBlobReadData({
                    archiveId,
                    instanceListId,
                    instanceIds,
                }),
                write: null,
            });
            const { blobAccessInfo } = await this.serviceExecutor.post(BlobAccessTokenService, tokenRequest);
            return blobAccessInfo;
        };
        return this.readBlobCache.getToken(referencingInstance.elementId, requestNewToken);
    }
    /**
     * Remove a given read blobs token from the cache.
     * @param referencingInstance
     */
    evictReadBlobsToken(referencingInstance) {
        this.readBlobCache.evict(referencingInstance.elementId);
    }
    /**
     * Requests a token that grants access to all blobs stored in the given archive. The user must own the archive (member of group)
     * @param archiveId ID for the archive to read blobs from
     */
    async requestReadTokenArchive(archiveId) {
        const requestNewToken = async () => {
            const tokenRequest = createBlobAccessTokenPostIn({
                archiveDataType: null,
                read: createBlobReadData({
                    archiveId,
                    instanceIds: [],
                    instanceListId: null,
                }),
                write: null,
            });
            const { blobAccessInfo } = await this.serviceExecutor.post(BlobAccessTokenService, tokenRequest);
            return blobAccessInfo;
        };
        return this.readArchiveCache.getToken(archiveId, requestNewToken);
    }
    /**
     * Remove a given read archive token from the cache.
     * @param archiveId
     */
    evictArchiveToken(archiveId) {
        this.readArchiveCache.evict(archiveId);
    }
    getArchiveId(blobs) {
        if (blobs.length == 0) {
            throw new Error("must pass blobs");
        }
        let archiveIds = new Set(blobs.map((b) => b.archiveId));
        if (archiveIds.size != 1) {
            throw new Error(`only one archive id allowed, but was ${archiveIds}`);
        }
        return blobs[0].archiveId;
    }
    /**
     *
     * @param blobServerAccessInfo
     * @param additionalRequestParams
     * @param typeRef the typeRef that shall be used to determine the correct model version
     */
    async createQueryParams(blobServerAccessInfo, additionalRequestParams, typeRef) {
        const typeModel = await resolveTypeReference(typeRef);
        return Object.assign(additionalRequestParams, {
            blobAccessToken: blobServerAccessInfo.blobAccessToken,
            v: typeModel.version,
        }, this.authDataProvider.createAuthHeaders());
    }
}
/**
 * Checks if the given access token can be used for another blob service requests.
 * @param blobServerAccessInfo
 * @param dateProvider
 */
function canBeUsedForAnotherRequest(blobServerAccessInfo, dateProvider) {
    return blobServerAccessInfo.expires.getTime() > dateProvider.now();
}
class BlobAccessTokenCache {
    constructor(dateProvider) {
        this.cache = new Map();
        this.dateProvider = dateProvider;
    }
    async getToken(key, loader) {
        const cached = this.cache.get(key);
        if (cached && canBeUsedForAnotherRequest(cached, this.dateProvider)) {
            return cached;
        }
        else {
            const newToken = await loader();
            this.cache.set(key, newToken);
            return newToken;
        }
    }
    evict(key) {
        this.cache.delete(key);
    }
}
//# sourceMappingURL=BlobAccessTokenFacade.js.map