import { KeyRotationTypeRef, UserGroupRootTypeRef } from "../../entities/sys/TypeRefs.js";
import { GroupType } from "../../common/TutanotaConstants.js";
import { isEmpty } from "@tutao/tutanota-utils";
/**
 * Facade to handle key rotation requests. Maintains and processes @PendingKeyRotation
 */
export class KeyRotationFacade {
    constructor(entityClient) {
        this.entityClient = entityClient;
        this.pendingKeyRotations = {
            pwKey: null,
            adminOrUserGroupKeyRotation: null,
            otherGroupsKeyRotation: [],
        };
    }
    /**
     * Queries the server for pending key rotations for a given user and saves them and optionally the given password key (in case an admin or user group needs to be rotated).
     * Needs to be called during login when the password key is still available.
     *
     * Note that this function currently makes 2 server requests to load the key rotation list and check if a key rotation is needed.
     * This routine should be optimized in the future by saving a flag on the user to determine whether a key rotation is required or not.
     */
    async loadPendingKeyRotations(user, pwKey) {
        const userGroupRoot = await this.entityClient.load(UserGroupRootTypeRef, user.userGroup.group);
        if (userGroupRoot.keyRotations != null) {
            const pendingKeyRotations = await this.entityClient.loadAll(KeyRotationTypeRef, userGroupRoot.keyRotations.list);
            const adminOrUserGroupRotation = pendingKeyRotations.find((keyRotation) => keyRotation.groupType === GroupType.Admin || keyRotation.groupType === GroupType.User);
            this.pendingKeyRotations = {
                pwKey: adminOrUserGroupRotation ? pwKey : null,
                adminOrUserGroupKeyRotation: adminOrUserGroupRotation || null,
                otherGroupsKeyRotation: pendingKeyRotations.filter((keyRotation) => keyRotation.groupType !== GroupType.User && keyRotation.groupType !== GroupType.Admin),
            };
        }
    }
    /**
     * Processes the internal list of @PendingKeyRotation. Key rotations and (if existent) password keys are deleted after processing.
     */
    async processPendingKeyRotation() {
        // first admin, then user and then user area
        if (this.pendingKeyRotations.adminOrUserGroupKeyRotation && this.pendingKeyRotations.pwKey) {
            if (this.pendingKeyRotations.adminOrUserGroupKeyRotation.groupType == GroupType.Admin) {
                await this.rotateAdminGroup(this.pendingKeyRotations.pwKey);
            }
            else if (this.pendingKeyRotations.adminOrUserGroupKeyRotation.groupType == GroupType.User) {
                await this.rotateUserGroup(this.pendingKeyRotations.pwKey);
            }
            this.pendingKeyRotations.adminOrUserGroupKeyRotation = null;
            this.pendingKeyRotations.pwKey = null;
        }
        if (!isEmpty(this.pendingKeyRotations.otherGroupsKeyRotation)) {
            await this.rotateOtherGroups();
            this.pendingKeyRotations.otherGroupsKeyRotation = [];
        }
    }
    async rotateAdminGroup(pwKey) {
        console.log("TODO Rotating the admin group key");
    }
    async rotateUserGroup(pwKey) {
        console.log("TODO Rotating the user group key");
    }
    async rotateOtherGroups() {
        console.log("TODO Rotating other group keys");
    }
}
//# sourceMappingURL=KeyRotationFacade.js.map