import { GroupType } from "../../common/TutanotaConstants";
import { decryptKey } from "@tutao/tutanota-crypto";
import { assertNotNull } from "@tutao/tutanota-utils";
import { ProgrammingError } from "../../common/error/ProgrammingError";
import { createWebsocketLeaderStatus } from "../../entities/sys/TypeRefs";
import { LoginIncompleteError } from "../../common/error/LoginIncompleteError";
import { isSameId } from "../../common/utils/EntityUtils.js";
/** Holder for the user and session-related data on the worker side. */
export class UserFacade {
    constructor() {
        this.user = null;
        this.accessToken = null;
        this.currentUserGroupKey = null;
        this.reset();
    }
    // Login process is somehow multi-step and we don't use a separate network stack for it. So we have to break up setters.
    // 1. We need to download user. For that we need to set access token already (to authenticate the request for the server as its passed in headers).
    // 2. We need to get group keys. For that we need to unlock userGroupKey with userPasspharseKey
    // so this leads to this steps in UserFacade:
    // 1. Access token is set
    // 2. User is set
    // 3. UserGroupKey is unlocked
    setAccessToken(accessToken) {
        this.accessToken = accessToken;
    }
    getAccessToken() {
        return this.accessToken;
    }
    setUser(user) {
        if (this.accessToken == null) {
            throw new ProgrammingError("invalid state: no access token");
        }
        this.user = user;
    }
    unlockUserGroupKey(userPassphraseKey) {
        if (this.user == null) {
            throw new ProgrammingError("Invalid state: no user");
        }
        const userGroupMembership = this.user.userGroup;
        this.currentUserGroupKey = {
            version: Number(userGroupMembership.groupKeyVersion),
            object: decryptKey(userPassphraseKey, userGroupMembership.symEncGKey),
        };
    }
    updateUser(user) {
        if (this.user == null) {
            throw new ProgrammingError("Update user is called without logging in. This function is not for you.");
        }
        this.user = user;
    }
    getUser() {
        return this.user;
    }
    /**
     * @return The map which contains authentication data for the logged in user.
     */
    createAuthHeaders() {
        return this.accessToken
            ? {
                accessToken: this.accessToken,
            }
            : {};
    }
    getUserGroupId() {
        return this.getLoggedInUser().userGroup.group;
    }
    getAllGroupIds() {
        let groups = this.getLoggedInUser().memberships.map((membership) => membership.group);
        groups.push(this.getLoggedInUser().userGroup.group);
        return groups;
    }
    getCurrentUserGroupKey() {
        // the userGroupKey is always written after the login to this.currentUserGroupKey
        //if the user has only logged in offline this has not happened
        if (this.currentUserGroupKey == null) {
            if (this.isPartiallyLoggedIn()) {
                throw new LoginIncompleteError("userGroupKey not available");
            }
            else {
                throw new ProgrammingError("Invalid state: userGroupKey is not available");
            }
        }
        return this.currentUserGroupKey;
    }
    getMembership(groupId) {
        let membership = this.getLoggedInUser().memberships.find((g) => isSameId(g.group, groupId));
        if (!membership) {
            throw new Error(`No group with groupId ${groupId} found!`);
        }
        return membership;
    }
    hasGroup(groupId) {
        if (!this.user) {
            return false;
        }
        else {
            return groupId === this.user.userGroup.group || this.user.memberships.some((m) => m.group === groupId);
        }
    }
    getGroupId(groupType) {
        if (groupType === GroupType.User) {
            return this.getUserGroupId();
        }
        else {
            let membership = this.getLoggedInUser().memberships.find((m) => m.groupType === groupType);
            if (!membership) {
                throw new Error("could not find groupType " + groupType + " for user " + this.getLoggedInUser()._id);
            }
            return membership.group;
        }
    }
    getGroupIds(groupType) {
        return this.getLoggedInUser()
            .memberships.filter((m) => m.groupType === groupType)
            .map((gm) => gm.group);
    }
    isPartiallyLoggedIn() {
        return this.user != null;
    }
    isFullyLoggedIn() {
        // We have userGroupKey and we can decrypt any other key - we are good to go
        return this.currentUserGroupKey != null;
    }
    getLoggedInUser() {
        return assertNotNull(this.user);
    }
    setLeaderStatus(status) {
        this.leaderStatus = status;
        console.log("New leader status set:", status.leaderStatus);
    }
    isLeader() {
        return this.leaderStatus.leaderStatus;
    }
    reset() {
        this.user = null;
        this.accessToken = null;
        this.currentUserGroupKey = null;
        this.leaderStatus = createWebsocketLeaderStatus({
            leaderStatus: false,
        });
    }
}
//# sourceMappingURL=UserFacade.js.map