import { AccountType, asKdfType, Const, CounterType, DEFAULT_KDF_TYPE, GroupType } from "../../../common/TutanotaConstants.js";
import { createMembershipAddData, createRecoverCode, createResetPasswordData, createUserDataDelete, GroupTypeRef, RecoverCodeTypeRef, } from "../../../entities/sys/TypeRefs.js";
import { encryptBytes, encryptKeyWithVersionedKey, encryptString } from "../../crypto/CryptoFacade.js";
import { assertNotNull, neverNull, uint8ArrayToHex } from "@tutao/tutanota-utils";
import { createUserAccountCreateData, createUserAccountUserData } from "../../../entities/tutanota/TypeRefs.js";
import { assertWorkerOrNode } from "../../../common/Env.js";
import { aes256RandomKey, bitArrayToUint8Array, createAuthVerifier, createAuthVerifierAsBase64Url, decryptKey, encryptKey, generateRandomSalt, random, uint8ArrayToKey, } from "@tutao/tutanota-crypto";
import { MembershipService, ResetPasswordService, SystemKeysService, UserService } from "../../../entities/sys/Services.js";
import { UserAccountService } from "../../../entities/tutanota/Services.js";
import { freshVersioned } from "@tutao/tutanota-utils/dist/Utils.js";
assertWorkerOrNode();
export class UserManagementFacade {
    constructor(userFacade, groupManagement, counters, entityClient, serviceExecutor, operationProgressTracker, loginFacade, pqFacade, keyLoaderFacade) {
        this.userFacade = userFacade;
        this.groupManagement = groupManagement;
        this.counters = counters;
        this.entityClient = entityClient;
        this.serviceExecutor = serviceExecutor;
        this.operationProgressTracker = operationProgressTracker;
        this.loginFacade = loginFacade;
        this.pqFacade = pqFacade;
        this.keyLoaderFacade = keyLoaderFacade;
    }
    async changeUserPassword(user, newPassword) {
        const userGroupKey = await this.groupManagement.getCurrentGroupKeyViaAdminEncGKey(user.userGroup.group);
        const salt = generateRandomSalt();
        const kdfType = DEFAULT_KDF_TYPE;
        const passwordKey = await this.loginFacade.deriveUserPassphraseKey({ kdfType, passphrase: newPassword, salt });
        const pwEncUserGroupKey = encryptKey(passwordKey, userGroupKey.object);
        const passwordVerifier = createAuthVerifier(passwordKey);
        const data = createResetPasswordData({
            user: user._id,
            salt,
            verifier: passwordVerifier,
            pwEncUserGroupKey,
            kdfVersion: kdfType,
        });
        await this.serviceExecutor.post(ResetPasswordService, data);
    }
    async changeAdminFlag(user, admin) {
        const adminGroupId = this.userFacade.getGroupId(GroupType.Admin);
        const userGroup = await this.entityClient.load(GroupTypeRef, user.userGroup.group);
        const requiredAdminKeyVersion = Number(userGroup.adminGroupKeyVersion ?? 0);
        const requiredAdminGroupKey = await this.keyLoaderFacade.loadSymGroupKey(adminGroupId, requiredAdminKeyVersion);
        const userGroupKey = { object: decryptKey(requiredAdminGroupKey, neverNull(userGroup.adminGroupEncGKey)), version: Number(userGroup.groupKeyVersion) };
        if (admin) {
            await this.groupManagement.addUserToGroup(user, adminGroupId);
            if (user.accountType !== AccountType.SYSTEM) {
                const keyData = await this._getAccountKeyData();
                const userEncAccountGroupKey = encryptKeyWithVersionedKey(userGroupKey, keyData.accountGroupKey);
                // we can not use addUserToGroup here because the admin is not admin of the account group
                const addAccountGroup = createMembershipAddData({
                    user: user._id,
                    group: keyData.accountGroup,
                    symEncGKey: userEncAccountGroupKey.key,
                    symKeyVersion: userEncAccountGroupKey.encryptingKeyVersion.toString(),
                    groupKeyVersion: keyData.accountGroupKeyVersion,
                });
                await this.serviceExecutor.post(MembershipService, addAccountGroup);
            }
        }
        else {
            await this.groupManagement.removeUserFromGroup(user._id, adminGroupId);
            if (user.accountType !== AccountType.SYSTEM) {
                const keyData = await this._getAccountKeyData();
                return this.groupManagement.removeUserFromGroup(user._id, keyData.accountGroup);
            }
        }
    }
    /**
     * Get key and id of premium group.
     * @throws Error if account type is not paid
     *
     * @private
     */
    async _getAccountKeyData() {
        const keysReturn = await this.serviceExecutor.get(SystemKeysService, null);
        const user = this.userFacade.getLoggedInUser();
        if (user.accountType === AccountType.PAID) {
            return {
                accountGroup: neverNull(keysReturn.premiumGroup),
                accountGroupKey: uint8ArrayToKey(keysReturn.premiumGroupKey),
                accountGroupKeyVersion: keysReturn.premiumGroupKeyVersion,
            };
        }
        else {
            throw new Error(`Trying to get keyData for user with account type ${user.accountType}`);
        }
    }
    async readUsedUserStorage(user) {
        const counterValue = await this.counters.readCounterValue(CounterType.UserStorageLegacy, neverNull(user.customer), user.userGroup.group);
        return Number(counterValue);
    }
    async deleteUser(user, restore) {
        const data = createUserDataDelete({
            user: user._id,
            restore,
            date: Const.CURRENT_DATE,
        });
        await this.serviceExecutor.delete(UserService, data);
    }
    async createUser(name, mailAddress, password, userIndex, overallNbrOfUsersToCreate, operationId) {
        let adminGroupIds = this.userFacade.getGroupIds(GroupType.Admin);
        if (adminGroupIds.length === 0) {
            adminGroupIds = this.userFacade.getGroupIds(GroupType.LocalAdmin);
        }
        const adminGroupId = adminGroupIds[0];
        const adminGroupKey = await this.keyLoaderFacade.getCurrentSymGroupKey(adminGroupId);
        const customerGroupKey = await this.keyLoaderFacade.getCurrentSymGroupKey(this.userFacade.getGroupId(GroupType.Customer));
        const userGroupKey = freshVersioned(aes256RandomKey());
        const userGroupInfoSessionKey = aes256RandomKey();
        const keyPair = await this.pqFacade.generateKeyPairs();
        const userGroupData = this.groupManagement.generateInternalGroupData(keyPair, userGroupKey.object, userGroupInfoSessionKey, adminGroupId, adminGroupKey, customerGroupKey);
        await this.operationProgressTracker.onProgress(operationId, ((userIndex + 0.8) / overallNbrOfUsersToCreate) * 100);
        let data = createUserAccountCreateData({
            date: Const.CURRENT_DATE,
            userGroupData: userGroupData,
            userData: await this.generateUserAccountData(userGroupKey, userGroupInfoSessionKey, customerGroupKey, mailAddress, password, name, this.generateRecoveryCode(userGroupKey)),
        });
        await this.serviceExecutor.post(UserAccountService, data);
        return this.operationProgressTracker.onProgress(operationId, ((userIndex + 1) / overallNbrOfUsersToCreate) * 100);
    }
    async generateUserAccountData(userGroupKey, userGroupInfoSessionKey, customerGroupKey, mailAddress, passphrase, userName, recoverData) {
        const kdfType = DEFAULT_KDF_TYPE;
        const salt = generateRandomSalt();
        const userPassphraseKey = await this.loginFacade.deriveUserPassphraseKey({ kdfType, passphrase, salt });
        const mailGroupKey = freshVersioned(aes256RandomKey());
        const contactGroupKey = freshVersioned(aes256RandomKey());
        const fileGroupKey = freshVersioned(aes256RandomKey());
        const mailboxSessionKey = aes256RandomKey();
        const contactListSessionKey = aes256RandomKey();
        const fileSystemSessionKey = aes256RandomKey();
        const mailGroupInfoSessionKey = aes256RandomKey();
        const contactGroupInfoSessionKey = aes256RandomKey();
        const fileGroupInfoSessionKey = aes256RandomKey();
        const tutanotaPropertiesSessionKey = aes256RandomKey();
        const userEncCustomerGroupKey = encryptKeyWithVersionedKey(userGroupKey, customerGroupKey.object);
        const userEncMailGroupKey = encryptKeyWithVersionedKey(userGroupKey, mailGroupKey.object);
        const userEncContactGroupKey = encryptKeyWithVersionedKey(userGroupKey, contactGroupKey.object);
        const userEncFileGroupKey = encryptKeyWithVersionedKey(userGroupKey, fileGroupKey.object);
        const userEncTutanotaPropertiesSessionKey = encryptKeyWithVersionedKey(userGroupKey, tutanotaPropertiesSessionKey);
        const userEncEntropy = encryptBytes(userGroupKey.object, random.generateRandomData(32));
        const customerEncMailGroupInfoSessionKey = encryptKeyWithVersionedKey(customerGroupKey, mailGroupInfoSessionKey);
        const customerEncContactGroupInfoSessionKey = encryptKeyWithVersionedKey(customerGroupKey, contactGroupInfoSessionKey);
        const customerEncFileGroupInfoSessionKey = encryptKeyWithVersionedKey(customerGroupKey, fileGroupInfoSessionKey);
        const contactEncContactListSessionKey = encryptKeyWithVersionedKey(contactGroupKey, contactListSessionKey);
        const fileEncFileSystemSessionKey = encryptKeyWithVersionedKey(fileGroupKey, fileSystemSessionKey);
        const mailEncMailBoxSessionKey = encryptKeyWithVersionedKey(mailGroupKey, mailboxSessionKey);
        return createUserAccountUserData({
            mailAddress: mailAddress,
            encryptedName: encryptString(userGroupInfoSessionKey, userName),
            salt: salt,
            kdfVersion: kdfType,
            verifier: createAuthVerifier(userPassphraseKey),
            pwEncUserGroupKey: encryptKey(userPassphraseKey, userGroupKey.object),
            userEncCustomerGroupKey: userEncCustomerGroupKey.key,
            userEncMailGroupKey: userEncMailGroupKey.key,
            userEncContactGroupKey: userEncContactGroupKey.key,
            userEncFileGroupKey: userEncFileGroupKey.key,
            userEncEntropy: userEncEntropy,
            userEncTutanotaPropertiesSessionKey: userEncTutanotaPropertiesSessionKey.key,
            contactEncContactListSessionKey: contactEncContactListSessionKey.key,
            fileEncFileSystemSessionKey: fileEncFileSystemSessionKey.key,
            mailEncMailBoxSessionKey: mailEncMailBoxSessionKey.key,
            customerEncMailGroupInfoSessionKey: customerEncMailGroupInfoSessionKey.key,
            customerEncContactGroupInfoSessionKey: customerEncContactGroupInfoSessionKey.key,
            customerEncFileGroupInfoSessionKey: customerEncFileGroupInfoSessionKey.key,
            customerKeyVersion: customerEncContactGroupInfoSessionKey.encryptingKeyVersion.toString(),
            recoverCodeEncUserGroupKey: recoverData.recoverCodeEncUserGroupKey,
            recoverCodeVerifier: recoverData.recoveryCodeVerifier,
            userEncRecoverCode: recoverData.userEncRecoverCode,
        });
    }
    generateRecoveryCode(currentUserGroupKey) {
        const recoveryCode = aes256RandomKey();
        const userEncRecoverCode = encryptKey(currentUserGroupKey.object, recoveryCode);
        const recoverCodeEncUserGroupKey = encryptKey(recoveryCode, currentUserGroupKey.object);
        const recoveryCodeVerifier = createAuthVerifier(recoveryCode);
        return {
            userEncRecoverCode,
            userKeyVersion: currentUserGroupKey.version,
            recoverCodeEncUserGroupKey,
            hexCode: uint8ArrayToHex(bitArrayToUint8Array(recoveryCode)),
            recoveryCodeVerifier,
        };
    }
    async getRecoverCode(passphrase) {
        const user = this.userFacade.getLoggedInUser();
        const recoverCodeId = user.auth?.recoverCode;
        if (recoverCodeId == null) {
            throw new Error("Auth is missing");
        }
        const passphraseKeyData = {
            kdfType: asKdfType(user.kdfVersion),
            passphrase,
            salt: assertNotNull(user.salt),
        };
        const passphraseKey = await this.loginFacade.deriveUserPassphraseKey(passphraseKeyData);
        const extraHeaders = {
            authVerifier: createAuthVerifierAsBase64Url(passphraseKey),
        };
        const recoveryCodeEntity = await this.entityClient.load(RecoverCodeTypeRef, recoverCodeId, undefined, extraHeaders);
        const userGroupKey = await this.keyLoaderFacade.loadSymUserGroupKey(Number(recoveryCodeEntity.userKeyVersion));
        return uint8ArrayToHex(bitArrayToUint8Array(decryptKey(userGroupKey, recoveryCodeEntity.userEncRecoverCode)));
    }
    async createRecoveryCode(passphrase) {
        const user = this.userFacade.getUser();
        if (user == null || user.auth == null) {
            throw new Error("Invalid state: no user or no user.auth");
        }
        const { userEncRecoverCode, userKeyVersion, recoverCodeEncUserGroupKey, hexCode, recoveryCodeVerifier } = this.generateRecoveryCode(this.userFacade.getCurrentUserGroupKey());
        const recoverPasswordEntity = createRecoverCode({
            userEncRecoverCode: userEncRecoverCode,
            userKeyVersion: String(userKeyVersion),
            recoverCodeEncUserGroupKey: recoverCodeEncUserGroupKey,
            _ownerGroup: this.userFacade.getUserGroupId(),
            verifier: recoveryCodeVerifier,
        });
        const passphraseKeyData = {
            kdfType: asKdfType(user.kdfVersion),
            passphrase,
            salt: assertNotNull(user.salt),
        };
        const pwKey = await this.loginFacade.deriveUserPassphraseKey(passphraseKeyData);
        const authVerifier = createAuthVerifierAsBase64Url(pwKey);
        await this.entityClient.setup(null, recoverPasswordEntity, {
            authVerifier,
        });
        return hexCode;
    }
}
//# sourceMappingURL=UserManagementFacade.js.map