import { boldFontWidths, PdfStreamEncoding, regularFontWidths } from "./PdfConstants.js";
import { Deflater } from "./Deflater.js";
import { stringToUtf8Uint8Array } from "@tutao/tutanota-utils";
export var PDF_FONTS;
(function (PDF_FONTS) {
    PDF_FONTS[PDF_FONTS["REGULAR"] = 1] = "REGULAR";
    PDF_FONTS[PDF_FONTS["BOLD"] = 2] = "BOLD";
})(PDF_FONTS || (PDF_FONTS = {}));
export var PDF_IMAGES;
(function (PDF_IMAGES) {
    PDF_IMAGES[PDF_IMAGES["TUTA_LOGO"] = 1] = "TUTA_LOGO";
})(PDF_IMAGES || (PDF_IMAGES = {}));
export const MARGIN_TOP = 20;
export const MARGIN_LEFT = 25;
export const TABLE_VERTICAL_SPACING = 5;
const TEXT_VERTICAL_SPACING = 2;
const PAPER_HEIGHT = 297;
const PAPER_WIDTH = 210;
const ORIGIN_POSITION = [0, 0];
// Transform matrix to set origin point top-left
const TRANSFORM_MATRIX = `1 0 0 -1 0 ${mmToPSPoint(PAPER_HEIGHT)}`;
// 1 InvoiceItem = 2 Table rows (first row item info, second row dates)
// Amount of table rows that can fit on the first page
const ROWS_FIRST_PAGE_SINGLE = 4; // 2 InvoiceItems
// Amount of table rows that can fit on the first page if a second is rendered too
const ROWS_FIRST_PAGE_MULTIPLE = 24; // 12 InvoiceItems
// Amount of table rows that can fit on any n-th page that isn't the first
const ROWS_N_PAGE = 50;
/**
 * Object which manages the high-level creation of a PDF document by parsing function instructions into PDF streams.
 * Use this object to create PDF documents. For the low-level functionality see "PdfWriter" class
 *
 * Once instantiated, use the public methods starting with "add...()" or "change...()" to describe the document.
 * Generate the PDF by calling the "create()" method any time. Do not utilize the same PdfDocument instance again once a PDF file has been created with it.
 * Private "render...()" methods are only used internally to interact with the low-level PdfWriter
 */
export class PdfDocument {
    constructor(pdfWriter) {
        this.pageCount = 0;
        this.textStream = "";
        this.graphicsStream = "";
        this.currentFont = PDF_FONTS.REGULAR;
        this.currentFontSize = 12;
        this.pageList = [];
        this.pdfWriter = pdfWriter;
        this.pdfWriter.setupDefaultObjects();
        this.deflater = new Deflater();
    }
    /**
     * Create the document: commit all streams to objects and let the PdfWriter write the file
     */
    async create() {
        // Write all open streams and add the page tree with all pages of the document to the PDF
        await this.renderText();
        await this.renderGraphics();
        this.pdfWriter.createObject(new Map([
            ["Type", "/Pages"],
            ["Parent", { refId: "CATALOG" }],
            ["Kids", this.pageList],
            ["Count", `${this.pageCount}`],
        ]), "PAGES");
        return await this.pdfWriter.writePdfFile();
    }
    /**
     * Closes the current textStream and writes it into an object
     */
    async renderText() {
        const encodedTextStream = await this.deflater.deflate(stringToUtf8Uint8Array(`BT q ${TRANSFORM_MATRIX} cm /F${this.currentFont} ${this.currentFontSize} Tf ` + this.textStream + ` Q ET`));
        this.pdfWriter.createStreamObject(new Map(), encodedTextStream, PdfStreamEncoding.FLATE, `TEXT_${this.pageCount}`);
        this.textStream = "";
    }
    /**
     * Closes the current graphicsStream and writes it into an object
     */
    async renderGraphics() {
        const encodedGraphicsStream = await this.deflater.deflate(stringToUtf8Uint8Array(`q ${TRANSFORM_MATRIX} cm ` + this.graphicsStream + ` Q`));
        this.pdfWriter.createStreamObject(new Map(), encodedGraphicsStream, PdfStreamEncoding.FLATE, `GRAPHICS_${this.pageCount}`);
        this.graphicsStream = "";
    }
    /**
     * Append a new page to the PDF document
     */
    async addPage() {
        // When adding a new page, all content streams must be rendered on the previous page. When creating the first page, all streams are still empty
        if (this.pageCount > 0) {
            await this.renderText();
            await this.renderGraphics();
        }
        this.pageCount++;
        // Create new page object
        const pageRefId = `PAGE_${this.pageCount}`;
        this.pdfWriter.createObject(new Map([
            ["Type", "/Page"],
            ["Parent", { refId: "PAGES" }],
            ["MediaBox", `[ 0 0 ${mmToPSPoint(PAPER_WIDTH)} ${mmToPSPoint(PAPER_HEIGHT)}]`],
            ["Resources", { refId: "RESOURCES" }],
            ["Contents", [{ refId: `TEXT_${this.pageCount}` }, { refId: `GRAPHICS_${this.pageCount}` }]],
        ]), pageRefId);
        // Add new page to page-tree
        this.pageList.push({ refId: pageRefId });
        return this;
    }
    /**
     * Add a text string at the given coordinates in millimeters
     * The coordinate field is in the fourth quadrant, i.e. the point of origin is top-left
     * @param text The text to place
     * @param position Coordinates [x,y] where to place the text, can be omitted to keep the current position
     */
    addText(text, position = ORIGIN_POSITION) {
        if (text === "")
            return this;
        // If no position was specified, don't update the text cursor position
        if (position !== ORIGIN_POSITION) {
            this.textStream += `1 0 0 -1 ${mmToPSPoint(position[0])} ${mmToPSPoint(position[1]) + this.currentFontSize} Tm <${toUnicodePoint(text).join("")}> Tj `;
        }
        else {
            this.textStream += `<${toUnicodePoint(text).join("")}> Tj `;
        }
        return this;
    }
    /**
     *
     *
     * Add a text string at the given coordinates in millimeters and align it to the right of its "container"
     * The container is specified by a width in millimeters.
     * @param text The text to place
     * @param position Coordinates [x,y] where to place the text, can be omitted to keep the current position
     * @param containerWidth The width in millimeters of the "container" in which the text is to be right aligned in
     */
    addTextRightAlign(text, position, containerWidth) {
        if (text === "")
            return this;
        const unicodePoints = toUnicodePoint(text);
        this.textStream += `1 0 0 -1 ${mmToPSPoint(position[0]) + mmToPSPoint(containerWidth) - getWordLengthInPoints(unicodePoints, this.currentFont, this.currentFontSize)} ${mmToPSPoint(position[1]) + this.currentFontSize} Tm <${unicodePoints.join("")}> Tj `;
        return this;
    }
    /**
     * Add a linebreak in the text
     */
    addLineBreak() {
        this.textStream += "T* ";
        return this;
    }
    /**
     * Add an image at the given coordinates in millimeters
     * The coordinate field is in the fourth quadrant, i.e. the point of origin is top-left
     * @param image The image to add which is an enum of readily embeddable images
     * @param position Coordinates [x,y] where to place the image
     * @param dimensions Dimensions [width, height] of the image. Image will be stretched to fit the given dimensions
     */
    addImage(image, position, dimensions) {
        // Image placement demands two matrix transformations, so it must make its own graphic state to not affect graphic elements which need no transform (drawLine)
        this.graphicsStream += `Q q ${TRANSFORM_MATRIX} cm ${mmToPSPoint(dimensions[0])} 0 0 -${mmToPSPoint(dimensions[1])} ${mmToPSPoint(position[0])} ${mmToPSPoint(position[1])} cm /Im${image} Do Q q ${TRANSFORM_MATRIX} cm `;
        return this;
    }
    /**
     * Draw black line "fromPos" "toPos"
     * @param fromPos Starting point of the line [x,y]
     * @param toPos Ending point of the line [x,y]
     */
    addDrawnLine(fromPos, toPos) {
        this.graphicsStream += `${mmToPSPoint(fromPos[0])} ${mmToPSPoint(fromPos[1])} m ${mmToPSPoint(toPos[0])} ${mmToPSPoint(toPos[1])} l s `;
        return this;
    }
    /**
     * Change the currently used text font
     * @param font Font to change to which is an enum of readily embeddable fonts
     * @param points Font size in PostScript points
     */
    changeFont(font, points) {
        this.textStream += `/F${font} ${points} Tf ${points + TEXT_VERTICAL_SPACING} TL `;
        this.currentFont = font;
        this.currentFontSize = points;
        return this;
    }
    /**
     * Changes the current text cursor position to the specified position
     * @param position The position to place the cursor at
     */
    changeTextCursorPosition(position) {
        this.textStream += `1 0 0 -1 ${mmToPSPoint(position[0])} ${mmToPSPoint(position[1]) + this.currentFontSize} Tm `;
        return this;
    }
    /**
     * Change font size of the currently used font
     * @param points Font size in PostScript points
     */
    changeFontSize(points) {
        this.textStream += `/F${this.currentFont} ${points} Tf ${points + TEXT_VERTICAL_SPACING} TL `;
        this.currentFontSize = points;
        return this;
    }
    /**
     * Change the grayscale of the current text
     * @param grayScale Float between 0 and 1 by which the text lightness shall be adjusted. 0 = white, 1 = black
     */
    changeTextGrayscale(grayScale) {
        grayScale = Math.max(Math.min(grayScale, 1), 0);
        this.textStream += `${grayScale} g `;
        return this;
    }
    /**
     * Render a table with the input of a two-dimensional array. Produces automatic page breaks.
     * Returns the table's height on the last page where it is still rendered, allowing that value to be used to continue placing objects after the table
     * @param position Coordinates [x,y] where to place the table's origin point
     * @param tableWidth The width of the table
     * @param columns Array of ColumnObjects, specifying the header name and width of each column in percent of the total tableWidth { headerName: string, columnWidth: number  }
     * @param data Two-dimensional array of strings, specifying the data for every row : [ //row1 [a,b,c] //row2 [x,y,z]...   ]. The inner arrays (rows) must have the same length as the columns array!
     */
    async addTable(position, tableWidth, columns, data) {
        this.addTableHeader(position, tableWidth, columns);
        // If all entries fit on the first page, then have "ITEMS_FIRST_PAGE_SINGLE" amount of entries, else "ROWS_FIRST_PAGE_MULTIPLE"
        const entriesOnFirstPage = data.length > ROWS_FIRST_PAGE_SINGLE ? ROWS_FIRST_PAGE_MULTIPLE : ROWS_FIRST_PAGE_SINGLE;
        // Render the first page, save the height of the table
        let tableHeight = this.addTablePage(position, tableWidth, columns, data.slice(0, entriesOnFirstPage));
        let entryCounter = entriesOnFirstPage;
        // Keep writing pages of entries until all data is exhausted
        while (entryCounter < data.length) {
            await this.addPage();
            position = [position[0], MARGIN_TOP];
            tableHeight = this.addTablePage(position, tableWidth, columns, data.slice(entryCounter, entryCounter + ROWS_N_PAGE));
            entryCounter += ROWS_N_PAGE;
        }
        const lastPageCannotFitRemainingRows = (entryCounter - entriesOnFirstPage) % ROWS_N_PAGE <= ROWS_FIRST_PAGE_MULTIPLE;
        const insufficientSpaceBelowTable = entryCounter == ROWS_FIRST_PAGE_MULTIPLE;
        if (!lastPageCannotFitRemainingRows || insufficientSpaceBelowTable) {
            await this.addPage();
            tableHeight = MARGIN_TOP;
        }
        this.addDrawnLine([position[0], tableHeight], [position[0] + tableWidth, tableHeight]);
        return tableHeight;
    }
    /**
     * Render the actual visible table, starting with a header and all subsequent entries inside the "chunk"
     */
    addTablePage(position, tableWidth, columns, chunk) {
        this.addTableHeader(position, tableWidth, columns);
        let previousRowOffset = TABLE_VERTICAL_SPACING;
        for (const row of chunk) {
            this.addTableRow([position[0], position[1] + previousRowOffset], columns, row);
            previousRowOffset += TABLE_VERTICAL_SPACING;
        }
        return position[1] + previousRowOffset;
    }
    /**
     * Render the table header
     */
    addTableHeader(position, tableWidth, columns) {
        this.changeFont(PDF_FONTS.BOLD, 11);
        this.addTableRow(position, columns, columns.flatMap((column) => column.headerName));
        this.addDrawnLine([position[0], position[1] + 5], [position[0] + tableWidth, position[1] + 5]);
        this.changeFont(PDF_FONTS.REGULAR, 11);
    }
    /**
     * Render a table row
     */
    addTableRow(position, columnInfo, rowItems) {
        if (rowItems.length !== columnInfo.length)
            console.error("Amount of items in table row not equal to amount of columns!");
        let previousWidthOffset = 0;
        for (let i = 0; i < rowItems.length; i++) {
            if (i >= 2) {
                this.addTextRightAlign(rowItems[i], [position[0] + previousWidthOffset, position[1]], columnInfo[i].columnWidth);
            }
            else {
                this.addText(rowItems[i], [position[0] + previousWidthOffset, position[1]]);
            }
            previousWidthOffset += columnInfo[i].columnWidth;
        }
    }
}
/**
 * Convert a text string into a string where each character is replaced by its 1-byte unicode point
 */
export function toUnicodePoint(input) {
    const out = [];
    for (let i = 0; i < input.length; i++) {
        const codePoint = input.codePointAt(i);
        if (codePoint && codePoint < 256) {
            out.push(codePoint.toString(16));
        }
        else {
            console.warn("Tried printing a character longer than one byte! Ignoring it...");
        }
    }
    return out;
}
export function getWordLengthInPoints(codePoints, font, fontSize) {
    const widthsArray = font === PDF_FONTS.REGULAR ? regularFontWidths : boldFontWidths;
    let total = 0;
    for (let i = 0; i < codePoints.length; i++) {
        let index = parseInt(codePoints[i], 16) - 32;
        total += 1 / (1000 / widthsArray[index]);
    }
    return total * fontSize;
}
/**
 * Convert millimeters to PostScript points
 */
function mmToPSPoint(mm) {
    return mm * 2.834645688;
}
//# sourceMappingURL=PdfDocument.js.map