import { typeRefToPath } from "./EntityRestClient.js";
import { firstBiggerThanSecond, getElementId, getListId } from "../../common/utils/EntityUtils.js";
import { assertNotNull, clone, getFromMap, remove } from "@tutao/tutanota-utils";
import { CustomCacheHandlerMap } from "./CustomCacheHandler.js";
import { resolveTypeReference } from "../../common/EntityFunctions.js";
import { Type as TypeId } from "../../common/EntityConstants.js";
import { ProgrammingError } from "../../common/error/ProgrammingError.js";
export class EphemeralCacheStorage {
    constructor() {
        /** Path to id to entity map. */
        this.entities = new Map();
        this.lists = new Map();
        this.blobEntities = new Map();
        this.customCacheHandlerMap = new CustomCacheHandlerMap();
        this.lastUpdateTime = null;
        this.userId = null;
        this.lastBatchIdPerGroup = new Map();
    }
    init({ userId }) {
        this.userId = userId;
    }
    deinit() {
        this.userId = null;
        this.entities.clear();
        this.lists.clear();
        this.blobEntities.clear();
        this.lastUpdateTime = null;
        this.lastBatchIdPerGroup.clear();
    }
    /**
     * Get a given entity from the cache, expects that you have already checked for existence
     */
    async get(typeRef, listId, id) {
        // We downcast because we can't prove that map has correct entity on the type level
        const path = typeRefToPath(typeRef);
        const typeModel = await resolveTypeReference(typeRef);
        switch (typeModel.type) {
            case TypeId.Element:
                return clone(this.entities.get(path)?.get(id) ?? null);
            case TypeId.ListElement:
                return clone(this.lists.get(path)?.get(assertNotNull(listId))?.elements.get(id) ?? null);
            case TypeId.BlobElement:
                return clone(this.blobEntities.get(path)?.get(assertNotNull(listId))?.elements.get(id) ?? null);
            default:
                throw new ProgrammingError("must be a persistent type");
        }
    }
    async deleteIfExists(typeRef, listId, id) {
        const path = typeRefToPath(typeRef);
        let typeModel;
        try {
            typeModel = await resolveTypeReference(typeRef);
        }
        catch (e) {
            // prevent failed lookup for BlobToFileMapping - this catch block can be removed after May 2023
            console.log("couldn't resolve typeRef ", typeRef);
            return;
        }
        switch (typeModel.type) {
            case TypeId.Element:
                this.entities.get(path)?.delete(id);
                break;
            case TypeId.ListElement:
                const cache = this.lists.get(path)?.get(assertNotNull(listId));
                if (cache != null) {
                    cache.elements.delete(id);
                    remove(cache.allRange, id);
                }
                break;
            case TypeId.BlobElement:
                this.blobEntities.get(path)?.get(assertNotNull(listId))?.elements.delete(id);
                break;
            default:
                throw new ProgrammingError("must be a persistent type");
        }
    }
    addElementEntity(typeRef, id, entity) {
        getFromMap(this.entities, typeRefToPath(typeRef), () => new Map()).set(id, entity);
    }
    async isElementIdInCacheRange(typeRef, listId, id) {
        const cache = this.lists.get(typeRefToPath(typeRef))?.get(listId);
        return cache != null && !firstBiggerThanSecond(id, cache.upperRangeId) && !firstBiggerThanSecond(cache.lowerRangeId, id);
    }
    async put(originalEntity) {
        const entity = clone(originalEntity);
        const typeRef = entity._type;
        const typeModel = await resolveTypeReference(typeRef);
        switch (typeModel.type) {
            case TypeId.Element:
                const elementEntity = entity;
                this.addElementEntity(elementEntity._type, elementEntity._id, elementEntity);
                break;
            case TypeId.ListElement:
                const listElementEntity = entity;
                const listElementTypeRef = typeRef;
                await this.putListElement(listElementEntity, listElementTypeRef);
                break;
            case TypeId.BlobElement:
                const blobElementEntity = entity;
                const blobTypeRef = typeRef;
                await this.putBlobElement(blobElementEntity, blobTypeRef);
                break;
            default:
                throw new ProgrammingError("must be a persistent type");
        }
    }
    async putBlobElement(entity, typeRef) {
        const listId = getListId(entity);
        const elementId = getElementId(entity);
        const cache = this.blobEntities.get(typeRefToPath(typeRef))?.get(listId);
        if (cache == null) {
            // first element in this list
            const newCache = {
                elements: new Map([[elementId, entity]]),
            };
            getFromMap(this.blobEntities, typeRefToPath(typeRef), () => new Map()).set(listId, newCache);
        }
        else {
            // if the element already exists in the cache, overwrite it
            cache.elements.set(elementId, entity);
        }
    }
    async putListElement(entity, typeRef) {
        const listId = getListId(entity);
        const elementId = getElementId(entity);
        const cache = this.lists.get(typeRefToPath(typeRef))?.get(listId);
        if (cache == null) {
            // first element in this list
            const newCache = {
                allRange: [elementId],
                lowerRangeId: elementId,
                upperRangeId: elementId,
                elements: new Map([[elementId, entity]]),
            };
            getFromMap(this.lists, typeRefToPath(typeRef), () => new Map()).set(listId, newCache);
        }
        else {
            // if the element already exists in the cache, overwrite it
            // add new element to existing list if necessary
            cache.elements.set(elementId, entity);
            if (await this.isElementIdInCacheRange(typeRef, listId, elementId)) {
                this.insertIntoRange(cache.allRange, elementId);
            }
        }
    }
    insertIntoRange(allRange, elementId) {
        for (let i = 0; i < allRange.length; i++) {
            const rangeElement = allRange[i];
            if (firstBiggerThanSecond(rangeElement, elementId)) {
                allRange.splice(i, 0, elementId);
                return;
            }
            if (rangeElement === elementId) {
                return;
            }
        }
        allRange.push(elementId);
    }
    async provideFromRange(typeRef, listId, start, count, reverse) {
        const listCache = this.lists.get(typeRefToPath(typeRef))?.get(listId);
        if (listCache == null) {
            return [];
        }
        let range = listCache.allRange;
        let ids = [];
        if (reverse) {
            let i;
            for (i = range.length - 1; i >= 0; i--) {
                if (firstBiggerThanSecond(start, range[i])) {
                    break;
                }
            }
            if (i >= 0) {
                let startIndex = i + 1 - count;
                if (startIndex < 0) {
                    // start index may be negative if more elements have been requested than available when getting elements reverse.
                    startIndex = 0;
                }
                ids = range.slice(startIndex, i + 1);
                ids.reverse();
            }
            else {
                ids = [];
            }
        }
        else {
            const i = range.findIndex((id) => firstBiggerThanSecond(id, start));
            ids = range.slice(i, i + count);
        }
        let result = [];
        for (let a = 0; a < ids.length; a++) {
            result.push(clone(listCache.elements.get(ids[a])));
        }
        return result;
    }
    async getRangeForList(typeRef, listId) {
        const listCache = this.lists.get(typeRefToPath(typeRef))?.get(listId);
        if (listCache == null) {
            return null;
        }
        return { lower: listCache.lowerRangeId, upper: listCache.upperRangeId };
    }
    async setUpperRangeForList(typeRef, listId, id) {
        const listCache = this.lists.get(typeRefToPath(typeRef))?.get(listId);
        if (listCache == null) {
            throw new Error("list does not exist");
        }
        listCache.upperRangeId = id;
    }
    async setLowerRangeForList(typeRef, listId, id) {
        const listCache = this.lists.get(typeRefToPath(typeRef))?.get(listId);
        if (listCache == null) {
            throw new Error("list does not exist");
        }
        listCache.lowerRangeId = id;
    }
    /**
     * Creates a new list cache if there is none. Resets everything but elements.
     * @param typeRef
     * @param listId
     * @param lower
     * @param upper
     */
    async setNewRangeForList(typeRef, listId, lower, upper) {
        const listCache = this.lists.get(typeRefToPath(typeRef))?.get(listId);
        if (listCache == null) {
            getFromMap(this.lists, typeRefToPath(typeRef), () => new Map()).set(listId, {
                allRange: [],
                lowerRangeId: lower,
                upperRangeId: upper,
                elements: new Map(),
            });
        }
        else {
            listCache.lowerRangeId = lower;
            listCache.upperRangeId = upper;
            listCache.allRange = [];
        }
    }
    async getIdsInRange(typeRef, listId) {
        return this.lists.get(typeRefToPath(typeRef))?.get(listId)?.allRange ?? [];
    }
    async getLastBatchIdForGroup(groupId) {
        return this.lastBatchIdPerGroup.get(groupId) ?? null;
    }
    async putLastBatchIdForGroup(groupId, batchId) {
        this.lastBatchIdPerGroup.set(groupId, batchId);
    }
    purgeStorage() {
        return Promise.resolve();
    }
    async getLastUpdateTime() {
        return this.lastUpdateTime ? { type: "recorded", time: this.lastUpdateTime } : { type: "never" };
    }
    async putLastUpdateTime(value) {
        this.lastUpdateTime = value;
    }
    async getWholeList(typeRef, listId) {
        const listCache = this.lists.get(typeRefToPath(typeRef))?.get(listId);
        if (listCache == null) {
            return [];
        }
        return listCache.allRange.map((id) => clone(listCache.elements.get(id)));
    }
    getCustomCacheHandlerMap(entityRestClient) {
        return this.customCacheHandlerMap;
    }
    getUserId() {
        return assertNotNull(this.userId, "No user id, not initialized?");
    }
    async deleteAllOwnedBy(owner) {
        for (const typeMap of this.entities.values()) {
            for (const [id, entity] of typeMap.entries()) {
                if (entity._ownerGroup === owner) {
                    typeMap.delete(id);
                }
            }
        }
        for (const cacheForType of this.lists.values()) {
            this.deleteAllOwnedByFromCache(cacheForType, owner);
        }
        for (const cacheForType of this.blobEntities.values()) {
            this.deleteAllOwnedByFromCache(cacheForType, owner);
        }
        this.lastBatchIdPerGroup.delete(owner);
    }
    deleteAllOwnedByFromCache(cacheForType, owner) {
        // If we find at least one element in the list that is owned by our target owner, we delete the entire list.
        // This is OK in most cases because the vast majority of lists are single owner.
        // For the other cases, we are just clearing the cache a bit sooner than needed.
        const listIdsToDelete = [];
        for (const [listId, listCache] of cacheForType.entries()) {
            for (const [id, element] of listCache.elements.entries()) {
                if (element._ownerGroup === owner) {
                    listIdsToDelete.push(listId);
                    break;
                }
            }
        }
        for (const listId of listIdsToDelete) {
            cacheForType.delete(listId);
        }
    }
    clearExcludedData() {
        return Promise.resolve();
    }
    /**
     * We want to lock the access to the "ranges" db when updating / reading the
     * offline available mail list ranges for each mail list (referenced using the listId)
     * @param listId the mail list that we want to lock
     */
    lockRangesDbAccess(listId) {
        return Promise.resolve();
    }
    /**
     * This is the counterpart to the function "lockRangesDbAccess(listId)"
     * @param listId the mail list that we want to unlock
     */
    unlockRangesDbAccess(listId) {
        return Promise.resolve();
    }
}
//# sourceMappingURL=EphemeralCacheStorage.js.map