/**
 * This file contains the functions used to set up and tear down edit dialogs for calendar events.
 *
 * they're not responsible for upholding invariants or ensure valid events (CalendarEventModel.editModels
 * and CalendarEventEditView do that), but know what additional information to ask the user before saving
 * and which methods to call to save the changes.
 */
import { Dialog } from "../../../gui/base/Dialog.js";
import { lang } from "../../../misc/LanguageViewModel.js";
import { Keys } from "../../../api/common/TutanotaConstants.js";
import { getStartOfTheWeekOffsetForUser, getTimeFormatForUser } from "../../date/CalendarUtils.js";
import { client } from "../../../misc/ClientDetector.js";
import { assertNotNull, noOp } from "@tutao/tutanota-utils";
import { locator } from "../../../api/main/MainLocator.js";
import { CalendarEventEditView } from "./CalendarEventEditView.js";
import { askIfShouldSendCalendarUpdatesToAttendees } from "../CalendarGuiUtils.js";
import { UserError } from "../../../api/main/UserError.js";
import { showUserError } from "../../../misc/ErrorHandlerImpl.js";
import { ProgrammingError } from "../../../api/common/error/ProgrammingError.js";
import { UpgradeRequiredError } from "../../../api/main/UpgradeRequiredError.js";
import { showPlanUpgradeRequiredDialog } from "../../../misc/SubscriptionDialogs.js";
import { convertTextToHtml } from "../../../misc/Formatter.js";
/**
 * the generic way to open any calendar edit dialog. the caller should know what to do after the
 * dialog is closed.
 */
async function showCalendarEventEditDialog(model, responseMail, handler) {
    const recipientsSearch = await locator.recipientsSearchModel();
    const { HtmlEditor } = await import("../../../gui/editor/HtmlEditor.js");
    const groupColors = locator.logins.getUserController().userSettingsGroupRoot.groupSettings.reduce((acc, gc) => {
        acc.set(gc.group, gc.color);
        return acc;
    }, new Map());
    const descriptionText = convertTextToHtml(model.editModels.description.content);
    const descriptionEditor = new HtmlEditor("description_label")
        .setMinHeight(400)
        .showBorders()
        .setEnabled(true)
        // We only set it once, we don't viewModel on every change, that would be slow
        .setValue(descriptionText)
        .setToolbarOptions({
        alignmentEnabled: false,
        fontSizeEnabled: false,
    })
        .enableToolbar();
    const okAction = (dom) => {
        model.editModels.description.content = descriptionEditor.getTrimmedValue();
        handler(dom.getBoundingClientRect(), () => dialog.close());
    };
    let headerDom = null;
    const heading = () => {
        const summary = model.editModels.summary.content;
        return summary.trim().length > 0 ? summary : lang.get("createEvent_label");
    };
    const dialogHeaderBarAttrs = {
        left: [
            {
                label: "cancel_action",
                click: () => dialog.close(),
                type: "secondary" /* ButtonType.Secondary */,
            },
        ],
        middle: heading,
        right: [
            {
                label: "save_action",
                click: (event, dom) => okAction(dom),
                type: "primary" /* ButtonType.Primary */,
            },
        ],
        create: (dom) => {
            headerDom = dom;
        },
    };
    const dialog = Dialog.editDialog(dialogHeaderBarAttrs, CalendarEventEditView, {
        model,
        recipientsSearch,
        descriptionEditor,
        startOfTheWeekOffset: getStartOfTheWeekOffsetForUser(locator.logins.getUserController().userSettingsGroupRoot),
        timeFormat: getTimeFormatForUser(locator.logins.getUserController().userSettingsGroupRoot),
        groupColors,
    })
        .addShortcut({
        key: Keys.ESC,
        exec: () => dialog.close(),
        help: "close_alt",
    })
        .addShortcut({
        key: Keys.S,
        ctrlOrCmd: true,
        exec: () => okAction(assertNotNull(headerDom, "headerDom was null")),
        help: "save_action",
    });
    if (client.isMobileDevice()) {
        // Prevent focusing text field automatically on mobile. It opens keyboard and you don't see all details.
        dialog.setFocusOnLoadFunction(noOp);
    }
    dialog.show();
}
/**
 * show an edit dialog for an event that does not exist on the server yet (or anywhere else)
 *
 * will unconditionally send invites on save.
 * @param model the calendar event model used to edit and save the event
 */
export async function showNewCalendarEventEditDialog(model) {
    let finished = false;
    const okAction = async (posRect, finish) => {
        /** new event, so we always want to send invites. */
        model.editModels.whoModel.shouldSendUpdates = true;
        if (finished || (await askUserIfInsecurePasswordsAreOk(model)) === 0 /* ConfirmationResult.Cancel */) {
            return;
        }
        try {
            const result = await model.apply();
            if (result === 0 /* EventSaveResult.Saved */) {
                finished = true;
                finish();
            }
        }
        catch (e) {
            if (e instanceof UserError) {
                // noinspection ES6MissingAwait
                showUserError(e);
            }
            else if (e instanceof UpgradeRequiredError) {
                await showPlanUpgradeRequiredDialog(e.plans);
            }
            else {
                throw e;
            }
        }
    };
    return showCalendarEventEditDialog(model, null, okAction);
}
/**
 * show a dialog that allows to edit a calendar event that already exists.
 *
 * on save, will validate external passwords, account type and user intent before actually saving and sending updates/invites/cancellations.
 *
 * @param model the calendar event model used to edit & save the event
 * @param identity the identity of the event to edit
 * @param responseMail a mail containing an invite and/or update for this event in case we need to reply to the organizer
 */
export async function showExistingCalendarEventEditDialog(model, identity, responseMail = null) {
    let finished = false;
    if (identity.uid == null) {
        throw new ProgrammingError("tried to edit existing event without uid, this is impossible for certain edit operations.");
    }
    const okAction = async (posRect, finish) => {
        if (finished ||
            (await askUserIfUpdatesAreNeededOrCancel(model)) === 0 /* ConfirmationResult.Cancel */ ||
            (await askUserIfInsecurePasswordsAreOk(model)) === 0 /* ConfirmationResult.Cancel */) {
            return;
        }
        try {
            const result = await model.apply();
            if (result === 0 /* EventSaveResult.Saved */ || result === 2 /* EventSaveResult.NotFound */) {
                finished = true;
                finish();
                // Inform the user that the event was deleted, avoiding misunderstanding that the event was saved
                if (result === 2 /* EventSaveResult.NotFound */)
                    Dialog.message("eventNoLongerExists_msg");
            }
        }
        catch (e) {
            if (e instanceof UserError) {
                // noinspection ES6MissingAwait
                showUserError(e);
            }
            else if (e instanceof UpgradeRequiredError) {
                await showPlanUpgradeRequiredDialog(e.plans);
            }
            else {
                throw e;
            }
        }
    };
    await showCalendarEventEditDialog(model, responseMail, okAction);
}
/** if there are update worthy changes on the model, ask the user what to do with them.
 * @returns {ConfirmationResult} Cancel if the whole process should be cancelled, Continue if the user selected whether to send updates and the saving
 * should proceed.
 * */
async function askUserIfUpdatesAreNeededOrCancel(model) {
    if (model.isAskingForUpdatesNeeded()) {
        switch (await askIfShouldSendCalendarUpdatesToAttendees()) {
            case "yes":
                model.editModels.whoModel.shouldSendUpdates = true;
                break;
            case "no":
                break;
            case "cancel":
                console.log("not saving event: user cancelled update sending.");
                return 0 /* ConfirmationResult.Cancel */;
        }
    }
    return 1 /* ConfirmationResult.Continue */;
}
/** if {@param model} is set to send updates but has insecure external passwords, ask the user what the action to resolve this should be.
 * @returns {ConfirmationResult} Cancel if the dialog should stay open, Continue if the save action should proceed despite insecure passwords. */
async function askUserIfInsecurePasswordsAreOk(model) {
    if (model.editModels.whoModel.shouldSendUpdates && // we want to send updates
        model.editModels.whoModel.hasInsecurePasswords() && // the model declares some of the passwords insecure
        !(await Dialog.confirm("presharedPasswordNotStrongEnough_msg")) // and the user is not OK with that
    ) {
        console.log("not saving event: insecure passwords.");
        return 0 /* ConfirmationResult.Cancel */;
    }
    return 1 /* ConfirmationResult.Continue */;
}
//# sourceMappingURL=CalendarEventEditDialog.js.map