import m from "mithril";
import { ExpanderButton, ExpanderPanel } from "../../../gui/base/Expander.js";
import { AttendeeListEditor } from "./AttendeeListEditor.js";
import { locator } from "../../../api/main/MainLocator.js";
import { EventTimeEditor } from "./EventTimeEditor.js";
import { RepeatRuleEditor } from "./RepeatRuleEditor.js";
import { TextField } from "../../../gui/base/TextField.js";
import { defaultCalendarColor } from "../../../api/common/TutanotaConstants.js";
import { lang } from "../../../misc/LanguageViewModel.js";
import { DropDownSelector } from "../../../gui/base/DropDownSelector.js";
import { AlarmIntervalUnit } from "../../date/CalendarUtils.js";
import { IconButton } from "../../../gui/base/IconButton.js";
import { attachDropdown } from "../../../gui/base/Dropdown.js";
import { InfoBanner } from "../../../gui/base/InfoBanner.js";
import { Dialog } from "../../../gui/base/Dialog.js";
import { getSharedGroupName } from "../../../sharing/GroupUtils.js";
import { createAlarmIntervalItems, createCustomRepeatRuleUnitValues, humanDescriptionForAlarmInterval } from "../CalendarGuiUtils.js";
/**
 * combines several semi-related editor components into a full editor for editing calendar events
 * to be displayed in a dialog.
 *
 * controls the enabling/disabling of certain editor components and the display of additional info
 * in the dialog depending on the type of the event being edited.
 */
export class CalendarEventEditView {
    constructor(vnode) {
        this.attendeesExpanded = false;
        this.timeFormat = vnode.attrs.timeFormat;
        this.startOfTheWeekOffset = vnode.attrs.startOfTheWeekOffset;
        this.attendeesExpanded = vnode.attrs.model.editModels.whoModel.canModifyGuests && vnode.attrs.model.editModels.whoModel.guests.length > 0;
        this.recipientsSearch = vnode.attrs.recipientsSearch;
    }
    view(vnode) {
        return m(".pb", {
            style: {
                // The date picker dialogs have position: fixed, and they are fixed relative to the most recent ancestor with
                // a transform. So doing a no-op transform will make the dropdowns scroll with the dialog
                // without this, then the date picker dialogs will show at the same place on the screen regardless of whether the
                // editor has scrolled or not.
                // Ideally we could do this inside DatePicker itself, but the rendering breaks and the dialog appears below it's siblings
                // We also don't want to do this for all dialogs because it could potentially cause other issues
                transform: "translate(0)",
            },
        }, [
            this.renderReadonlyMessage(vnode.attrs),
            this.renderHeading(vnode.attrs),
            this.renderAttendees(vnode.attrs),
            this.renderEventTimeEditor(vnode.attrs),
            this.renderRepeatRuleEditor(vnode.attrs),
            m(".flex", [this.renderCalendarPicker(vnode), this.renderRemindersEditor(vnode)]),
            this.renderLocationField(vnode),
            this.renderDescriptionEditor(vnode),
        ]);
    }
    renderHeading(attrs) {
        const { model } = attrs;
        return m(TextField, {
            label: "title_placeholder",
            value: model.editModels.summary.content,
            oninput: (v) => (model.editModels.summary.content = v),
            disabled: !model.isFullyWritable(),
            class: "big-input pt flex-grow",
            injectionsRight: () => this.renderGuestsExpanderButton(attrs),
        });
    }
    renderGuestsExpanderButton(attrs) {
        if (!attrs.model.editModels.whoModel.canModifyGuests && attrs.model.editModels.whoModel.guests.length === 0)
            return null;
        return m(".mr-s", m(ExpanderButton, {
            label: "guests_label",
            expanded: this.attendeesExpanded,
            onExpandedChange: (v) => (this.attendeesExpanded = v),
            style: {
                paddingTop: 0,
            },
        }));
    }
    renderReadonlyMessage(attrs) {
        const { model } = attrs;
        const makeMessage = (message) => m(".pt-s", m(InfoBanner, {
            message: () => m(".small.selectable", lang.get(message)),
            icon: "People" /* Icons.People */,
            type: "info" /* BannerType.Info */,
            buttons: [],
        }));
        switch (model.getReadonlyReason()) {
            case 0 /* ReadonlyReason.SHARED */:
                return makeMessage("cannotEditFullEvent_msg");
            case 1 /* ReadonlyReason.SINGLE_INSTANCE */:
                return makeMessage("cannotEditSingleInstance_msg");
            case 2 /* ReadonlyReason.NOT_ORGANIZER */:
                return makeMessage("cannotEditNotOrganizer_msg");
            case 3 /* ReadonlyReason.UNKNOWN */:
                return makeMessage("cannotEditEvent_msg");
            case 4 /* ReadonlyReason.NONE */:
                return null;
        }
    }
    renderAttendees(attrs) {
        const { model } = attrs;
        return m(".mb.rel", m(ExpanderPanel, { expanded: this.attendeesExpanded }, m(AttendeeListEditor, {
            model,
            recipientsSearch: this.recipientsSearch,
            logins: locator.logins,
        })));
    }
    renderEventTimeEditor(attrs) {
        return m(EventTimeEditor, {
            editModel: attrs.model.editModels.whenModel,
            timeFormat: this.timeFormat,
            startOfTheWeekOffset: this.startOfTheWeekOffset,
            disabled: !attrs.model.isFullyWritable(),
        });
    }
    renderRepeatRuleEditor({ model }) {
        return m(RepeatRuleEditor, {
            model: model.editModels.whenModel,
            startOfTheWeekOffset: this.startOfTheWeekOffset,
            disabled: !model.canEditSeries(),
        });
    }
    renderCalendarPicker(vnode) {
        const { model } = vnode.attrs;
        const availableCalendars = model.editModels.whoModel.getAvailableCalendars();
        return m(".flex-half.pr-s", m(DropDownSelector, {
            label: "calendar_label",
            items: availableCalendars.map((calendarInfo) => {
                return {
                    name: getSharedGroupName(calendarInfo.groupInfo, model.userController, calendarInfo.shared),
                    value: calendarInfo,
                };
            }),
            selectedValue: model.editModels.whoModel.selectedCalendar,
            selectionChangedHandler: (v) => (model.editModels.whoModel.selectedCalendar = v),
            icon: "Expand" /* BootIcons.Expand */,
            disabled: !model.canChangeCalendar() || availableCalendars.length < 2,
            helpLabel: () => this.renderCalendarColor(model.editModels.whoModel.selectedCalendar, vnode.attrs.groupColors),
        }));
    }
    renderCalendarColor(selectedCalendar, groupColors) {
        const color = selectedCalendar ? groupColors.get(selectedCalendar.groupInfo.group) ?? defaultCalendarColor : null;
        return m(".mt-xs", {
            style: {
                width: "100px",
                height: "10px",
                background: color ? "#" + color : "transparent",
            },
        });
    }
    renderRemindersEditor(vnode) {
        if (!vnode.attrs.model.editModels.alarmModel.canEditReminders)
            return null;
        const { alarmModel } = vnode.attrs.model.editModels;
        const textFieldAttrs = alarmModel.alarms.map((a) => ({
            value: humanDescriptionForAlarmInterval(a, lang.languageTag),
            label: "emptyString_msg",
            isReadOnly: true,
            injectionsRight: () => m(IconButton, {
                title: "delete_action",
                icon: "Cancel" /* Icons.Cancel */,
                click: () => alarmModel.removeAlarm(a),
            }),
        }));
        textFieldAttrs.push({
            value: lang.get("add_action"),
            label: "emptyString_msg",
            isReadOnly: true,
            injectionsRight: () => m(IconButton, attachDropdown({
                mainButtonAttrs: {
                    title: "add_action",
                    icon: "Add" /* Icons.Add */,
                },
                childAttrs: () => [
                    ...createAlarmIntervalItems(lang.languageTag).map((i) => ({
                        label: () => i.name,
                        click: () => alarmModel.addAlarm(i.value),
                    })),
                    {
                        label: () => lang.get("calendarReminderIntervalDropdownCustomItem_label"),
                        click: () => {
                            this.showCustomReminderIntervalDialog((value, unit) => {
                                alarmModel.addAlarm({
                                    value,
                                    unit,
                                });
                            });
                        },
                    },
                ],
            })),
        });
        textFieldAttrs[0].label = "reminderBeforeEvent_label";
        return m(".flex.col.flex-half.pl-s", textFieldAttrs.map((a) => m(TextField, a)));
    }
    renderLocationField(vnode) {
        const { model } = vnode.attrs;
        return m(TextField, {
            label: "location_label",
            value: model.editModels.location.content,
            oninput: (v) => (model.editModels.location.content = v),
            disabled: !model.isFullyWritable(),
            class: "pt-s", // override default pt with pt-s because calendar color indicator takes up some space
            injectionsRight: () => {
                let address = encodeURIComponent(model.editModels.location.content);
                if (address === "") {
                    return null;
                }
                return m(IconButton, {
                    title: "showAddress_alt",
                    icon: "Pin" /* Icons.Pin */,
                    size: 1 /* ButtonSize.Compact */,
                    click: () => {
                        window.open(`https://www.openstreetmap.org/search?query=${address}`, "_blank");
                    },
                });
            },
        });
    }
    renderDescriptionEditor(vnode) {
        const { model } = vnode.attrs;
        vnode.attrs.descriptionEditor.setReadOnly(!model.isFullyWritable());
        return m(vnode.attrs.descriptionEditor);
    }
    showCustomReminderIntervalDialog(onAddAction) {
        let timeReminderValue = 0;
        let timeReminderUnit = AlarmIntervalUnit.MINUTE;
        Dialog.showActionDialog({
            title: () => lang.get("calendarReminderIntervalCustomDialog_title"),
            allowOkWithReturn: true,
            child: {
                view: () => {
                    const unitItems = createCustomRepeatRuleUnitValues() ?? [];
                    return m(".flex full-width pt-s", [
                        m(TextField, {
                            type: "number" /* TextFieldType.Number */,
                            min: 0,
                            label: "calendarReminderIntervalValue_label",
                            value: timeReminderValue.toString(),
                            oninput: (v) => {
                                const time = Number.parseInt(v);
                                const isEmpty = v === "";
                                if (!Number.isNaN(time) || isEmpty)
                                    timeReminderValue = isEmpty ? 0 : time;
                            },
                            class: "flex-half no-appearance", //Removes the up/down arrow from input number. Pressing arrow up/down key still working
                        }),
                        m(DropDownSelector, {
                            label: "emptyString_msg",
                            selectedValue: timeReminderUnit,
                            items: unitItems,
                            class: "flex-half pl-s",
                            selectionChangedHandler: (selectedValue) => (timeReminderUnit = selectedValue),
                            disabled: false,
                        }),
                    ]);
                },
            },
            okActionTextId: "add_action",
            okAction: (dialog) => {
                onAddAction(timeReminderValue, timeReminderUnit);
                dialog.close();
            },
        });
    }
}
//# sourceMappingURL=CalendarEventEditView.js.map