import m from "mithril";
import { TextField } from "../../../gui/base/TextField.js";
import { lang } from "../../../misc/LanguageViewModel.js";
import { DropDownSelector } from "../../../gui/base/DropDownSelector.js";
import { DatePicker } from "../pickers/DatePicker.js";
import { IconButton } from "../../../gui/base/IconButton.js";
import { renderTwoColumnsIfFits } from "../../../gui/base/GuiUtils.js";
import { createIntervalValues, createRepeatRuleEndTypeValues, createRepeatRuleFrequencyValues } from "../CalendarGuiUtils.js";
export class RepeatRuleEditor {
    view({ attrs }) {
        const { model, disabled } = attrs;
        return [
            renderTwoColumnsIfFits([
                m(".flex-grow.pr-s", this.renderRepeatPeriod(attrs)),
                m(".flex-grow.pl-s" + (model.repeatPeriod != null ? "" : ".hidden"), this.renderRepeatInterval({
                    ...attrs,
                    disabled: model.repeatPeriod == null ? true : attrs.disabled,
                })),
            ], this.renderEndCondition(attrs)),
            renderTwoColumnsIfFits(this.renderExclusionCount(attrs), null),
        ];
    }
    renderEndCondition(attrs) {
        const { model } = attrs;
        if (model.repeatPeriod == null) {
            return null;
        }
        return [m(".flex-grow.pr-s", this.renderEndType(attrs)), m(".flex-grow.pl-s", this.renderEndValue(attrs))];
    }
    renderExclusionCount({ model, disabled }) {
        if (model.repeatPeriod == null || model.excludedDates.length === 0) {
            return null;
        }
        return [
            m(".flex-grow.pr-s", m(TextField, {
                label: "emptyString_msg",
                value: lang.get("someRepetitionsDeleted_msg"),
                injectionsRight: () => (disabled ? null : this.renderDeleteExclusionButton(model)),
                isReadOnly: true,
            })),
        ];
    }
    /**
     * how frequently the event repeats (Never, daily, annually etc)
     * @private
     */
    renderRepeatPeriod({ model, disabled }) {
        const repeatValues = createRepeatRuleFrequencyValues();
        return m(DropDownSelector, {
            label: "calendarRepeating_label",
            items: repeatValues,
            selectedValue: model.repeatPeriod,
            selectionChangedHandler: (period) => (model.repeatPeriod = period),
            icon: "Expand" /* BootIcons.Expand */,
            disabled,
        });
    }
    /** Repeat interval: every day, every second day etc
     * @private
     */
    renderRepeatInterval({ model, disabled }) {
        const intervalValues = createIntervalValues();
        return m(DropDownSelector, {
            label: "interval_title",
            items: intervalValues,
            selectedValue: model.repeatInterval,
            selectionChangedHandler: (interval) => (model.repeatInterval = interval),
            icon: "Expand" /* BootIcons.Expand */,
            disabled,
        });
    }
    /**
     * if and how the event stops repeating, like after number of occurrences or after a date.
     * @param model
     * @private
     */
    renderEndType({ model, disabled }) {
        const endTypeValues = createRepeatRuleEndTypeValues();
        return m(DropDownSelector, {
            label: () => lang.get("calendarRepeatStopCondition_label"),
            items: endTypeValues,
            selectedValue: model.repeatEndType,
            selectionChangedHandler: (end) => (model.repeatEndType = end),
            icon: "Expand" /* BootIcons.Expand */,
            disabled,
        });
    }
    /**
     * the value of the end condition - a number for ending after number of occurrences, a date for ending after date.
     * @private
     */
    renderEndValue(attrs) {
        const { model, startOfTheWeekOffset, disabled } = attrs;
        const intervalValues = createIntervalValues();
        if (model.repeatPeriod == null || model.repeatEndType === "0" /* EndType.Never */) {
            return null;
        }
        else if (model.repeatEndType === "1" /* EndType.Count */) {
            return m(DropDownSelector, {
                label: "emptyString_msg",
                items: intervalValues,
                selectedValue: model.repeatEndOccurrences,
                selectionChangedHandler: (endValue) => (model.repeatEndOccurrences = endValue),
                icon: "Expand" /* BootIcons.Expand */,
                disabled,
            });
        }
        else if (model.repeatEndType === "2" /* EndType.UntilDate */) {
            return m(DatePicker, {
                date: model.repeatEndDateForDisplay,
                onDateSelected: (date) => (model.repeatEndDateForDisplay = date),
                startOfTheWeekOffset,
                label: "emptyString_msg",
                nullSelectionText: "emptyString_msg",
                // When the guests expander is expanded and the dialog has overflow, then the scrollbar will overlap the date picker popup
                // to fix this we could either:
                // * reorganize the layout so it doesn't go over the right edge
                // * change the alignment so that it goes to the left (this is what we do)
                rightAlignDropdown: true,
                disabled,
            });
        }
        else {
            return null;
        }
    }
    renderDeleteExclusionButton(model) {
        return m(IconButton, {
            title: "restoreExcludedRecurrences_action",
            click: () => model.deleteExcludedDates(),
            icon: "Cancel" /* Icons.Cancel */,
        });
    }
}
//# sourceMappingURL=RepeatRuleEditor.js.map