import m from "mithril";
import { px } from "../../../gui/size.js";
import { modal } from "../../../gui/base/Modal.js";
import { EventPreviewView } from "./EventPreviewView.js";
import { Dialog } from "../../../gui/base/Dialog.js";
import { createAsyncDropdown, DROPDOWN_MARGIN, showDropdown } from "../../../gui/base/Dropdown.js";
import { Keys } from "../../../api/common/TutanotaConstants.js";
import { prepareCalendarDescription } from "../../date/CalendarUtils.js";
import { IconButton } from "../../../gui/base/IconButton.js";
import { convertTextToHtml } from "../../../misc/Formatter.js";
/**
 * small modal displaying all relevant information about an event in a compact fashion. offers limited editing capabilities to participants in the
 * form of quick response buttons to set attendance and adding exclusions.
 *
 * It is the first line of defense against invalid edit operations since it's the only way to open a
 * calendar editor for an event that's not brand new or to delete an event/parts of it.
 *
 * it is also responsible for divining the users intent before starting any operations so we can make sure we have the right
 * information available, like when editing all event occurrences or only some.
 */
export class CalendarEventPopup {
    /**
     * @param model
     * @param eventBubbleRect the rect where the event bubble was displayed that was clicked (if any)
     * @param htmlSanitizer
     */
    constructor(model, eventBubbleRect, htmlSanitizer) {
        this.model = model;
        this.eventBubbleRect = eventBubbleRect;
        this._shortcuts = [];
        this.dom = null;
        this.handleDeleteButtonClick = async (ev, receiver) => {
            if (await this.model.isRepeatingForDeleting()) {
                createAsyncDropdown({
                    lazyButtons: () => Promise.resolve([
                        {
                            label: "deleteSingleEventRecurrence_action",
                            click: async () => {
                                await this.model.deleteSingle();
                                this.close();
                            },
                        },
                        {
                            label: "deleteAllEventRecurrence_action",
                            click: () => this.confirmDeleteClose(),
                        },
                    ]),
                    width: 300,
                })(ev, receiver);
            }
            else {
                // noinspection JSIgnoredPromiseFromCall
                this.confirmDeleteClose();
            }
        };
        this.handleEditButtonClick = (ev, receiver) => {
            if (this.model.isRepeatingForEditing) {
                createAsyncDropdown({
                    lazyButtons: () => Promise.resolve([
                        {
                            label: "updateOneCalendarEvent_action",
                            click: () => {
                                // noinspection JSIgnoredPromiseFromCall
                                this.model.editSingle();
                                this.close();
                            },
                        },
                        {
                            label: "updateAllCalendarEvents_action",
                            click: () => {
                                // noinspection JSIgnoredPromiseFromCall
                                this.model.editAll();
                                this.close();
                            },
                        },
                    ]),
                    width: 300,
                })(ev, receiver);
            }
            else {
                // noinspection JSIgnoredPromiseFromCall
                this.model.editAll();
                this.close();
            }
        };
        // we handle askForUpdates here to avoid making a request if not necessary
        this.handleSendUpdatesClick = async () => {
            const confirmed = await Dialog.confirm("sendUpdates_msg");
            if (confirmed)
                await this.model.sendUpdates();
            this.close();
        };
        // We receive the HtmlSanitizer from outside and do the sanitization inside, so that we don't have to just assume it was already done
        this.sanitizedDescription = prepareCalendarDescription(model.calendarEvent.description, (s) => htmlSanitizer.sanitizeHTML(convertTextToHtml(s), {
            blockExternalContent: true,
        }).html);
        this.setupShortcuts();
        this.view = this.view.bind(this);
    }
    view() {
        return m(".abs.elevated-bg.plr.pb.border-radius.dropdown-shadow.flex.flex-column", {
            style: {
                // minus margin, need to apply it now to not overflow later
                width: px(Math.min(window.innerWidth - DROPDOWN_MARGIN * 2, 400)),
                // see hack description below
                opacity: "0",
                // because calendar event bubbles have 1px border, we want to align
                margin: "1px",
            },
            oncreate: (vnode) => {
                this.dom = vnode.dom;
                // This is a hack to get "natural" view size but render it without opacity first and then show dropdown with inferred
                // size.
                setTimeout(() => showDropdown(this.eventBubbleRect, this.dom, this.dom.offsetHeight, 400), 24);
            },
        }, [
            m(".flex.flex-end", [this.renderSendUpdateButton(), this.renderEditButton(), this.renderDeleteButton(), this.renderCloseButton()]),
            m(".flex-grow.scroll.visible-scrollbar", [
                m(EventPreviewView, {
                    event: this.model.calendarEvent,
                    sanitizedDescription: this.sanitizedDescription,
                    // closing this since repeated edit operations from the popup would always
                    // use the version of the event the popup was opened with, which means the next
                    // click uses an outdated version.
                    participation: this.model.getParticipationSetterAndThen(() => this.close()),
                }),
            ]),
        ]);
    }
    renderEditButton() {
        if (!this.model.canEdit)
            return null;
        return m(IconButton, { title: "edit_action", icon: "Edit" /* Icons.Edit */, click: this.handleEditButtonClick });
    }
    renderDeleteButton() {
        if (!this.model.canDelete)
            return null;
        return m(IconButton, { title: "delete_action", icon: "Trash" /* Icons.Trash */, click: this.handleDeleteButtonClick });
    }
    renderSendUpdateButton() {
        if (!this.model.canSendUpdates)
            return null;
        return m(IconButton, {
            title: "sendUpdates_label",
            click: () => this.handleSendUpdatesClick(),
            icon: "Mail" /* BootIcons.Mail */,
        });
    }
    renderCloseButton() {
        return m(IconButton, {
            title: "close_alt",
            click: () => this.close(),
            icon: "Cancel" /* Icons.Cancel */,
        });
    }
    show() {
        modal.display(this, false);
    }
    close() {
        modal.remove(this);
    }
    backgroundClick(e) {
        modal.remove(this);
    }
    hideAnimation() {
        return Promise.resolve();
    }
    onClose() {
        this.close();
    }
    shortcuts() {
        return this._shortcuts;
    }
    popState(e) {
        modal.remove(this);
        return false;
    }
    setupShortcuts() {
        const close = {
            key: Keys.ESC,
            exec: () => this.close(),
            help: "close_alt",
        };
        const edit = {
            key: Keys.E,
            exec: () => this.handleEditButtonClick(new MouseEvent("click", {}), this.dom),
            help: "edit_action",
        };
        const sendUpdates = {
            key: Keys.R,
            exec: this.handleSendUpdatesClick,
            help: "sendUpdates_label",
        };
        const remove = {
            key: Keys.DELETE,
            exec: () => this.handleDeleteButtonClick(new MouseEvent("click", {}), this.dom),
            help: "delete_action",
        };
        this._shortcuts.push(close);
        if (this.model.canSendUpdates) {
            this._shortcuts.push(sendUpdates);
        }
        if (this.model.canEdit) {
            this._shortcuts.push(edit);
        }
        if (this.model.canDelete) {
            this._shortcuts.push(remove);
        }
    }
    async confirmDeleteClose() {
        if (!(await Dialog.confirm("deleteEventConfirmation_msg")))
            return;
        await this.model.deleteAll();
        this.close();
    }
}
//# sourceMappingURL=CalendarEventPopup.js.map