import m from "mithril";
import { TextField } from "../../../gui/base/TextField.js";
import { theme } from "../../../gui/theme.js";
import { client } from "../../../misc/ClientDetector.js";
import { Keys } from "../../../api/common/TutanotaConstants.js";
import { timeStringFromParts } from "../../../misc/Formatter.js";
import { Time } from "../../date/Time.js";
import { isKeyPressed } from "../../../misc/KeyManager.js";
export class TimePicker {
    constructor({ attrs }) {
        this.focused = false;
        this.value = "";
        this.amPm = attrs.timeFormat === "1" /* TimeFormat.TWELVE_HOURS */;
        const times = [];
        for (let hour = 0; hour < 24; hour++) {
            for (let minute = 0; minute < 60; minute += 30) {
                times.push(timeStringFromParts(hour, minute, this.amPm));
            }
        }
        this.oldValue = attrs.time?.toString(false) ?? "--";
        this.values = times;
    }
    view({ attrs }) {
        if (attrs.time) {
            const timeAsString = attrs.time?.toString(this.amPm) ?? "";
            this.selectedIndex = this.values.indexOf(timeAsString);
            if (!this.focused) {
                this.value = timeAsString;
            }
        }
        if (client.isMobileDevice()) {
            return this.renderNativeTimePicker(attrs);
        }
        else {
            return this.renderCustomTimePicker(attrs);
        }
    }
    renderNativeTimePicker(attrs) {
        if (this.oldValue !== attrs.time?.toString(false)) {
            this.onSelected(attrs);
        }
        // input[type=time] wants time in 24h format, no matter what is actually displayed. Otherwise it will be empty.
        const timeAsString = attrs.time?.toString(false) ?? "";
        this.oldValue = timeAsString;
        this.value = timeAsString;
        const displayTime = attrs.time?.toString(this.amPm) ?? "";
        return [
            m(TextField, {
                class: "time-picker pt",
                label: "emptyString_msg",
                value: this.value,
                type: "time" /* TextFieldType.Time */,
                oninput: (value) => {
                    if (this.value === value) {
                        return;
                    }
                    this.value = value;
                    attrs.onTimeSelected(Time.parseFromString(value));
                },
                disabled: attrs.disabled,
            }),
            // A 'fake' display that overlays over the real time input that allows us to show 12 or 24 time format regardless of browser locale
            m(".time-picker-fake-display.rel.no-hover", displayTime),
        ];
    }
    renderCustomTimePicker(attrs) {
        return [this.renderInputField(attrs), this.focused ? this.renderTimeSelector(attrs) : null];
    }
    renderInputField(attrs) {
        return m(TextField, {
            label: "emptyString_msg",
            value: this.value,
            oninput: (v) => (this.value = v),
            disabled: attrs.disabled,
            onfocus: (dom, input) => {
                this.focused = true;
                input.select();
            },
            onblur: (e) => {
                if (this.focused) {
                    this.onSelected(attrs);
                }
                e.redraw = false;
            },
            keyHandler: (key) => {
                if (isKeyPressed(key.key, Keys.RETURN)) {
                    this.onSelected(attrs);
                    const active = document.activeElement;
                    active?.blur();
                }
                return true;
            },
        });
    }
    renderTimeSelector(attrs) {
        return m(".fixed.flex.col.mt-s.menu-shadow", {
            oncreate: (vnode) => this.setScrollTop(attrs, vnode),
            onupdate: (vnode) => this.setScrollTop(attrs, vnode),
            style: {
                width: "100px",
                height: "400px",
                "z-index": "3",
                background: theme.content_bg,
                overflow: "auto",
            },
        }, this.values.map((time, i) => m("pr-s.pl-s.darker-hover", {
            key: time,
            style: {
                "background-color": this.selectedIndex === i ? theme.list_bg : theme.list_alternate_bg,
                flex: "1 0 auto",
                "line-height": "44px",
            },
            onmousedown: () => {
                this.focused = false;
                attrs.onTimeSelected(Time.parseFromString(time));
            },
        }, time)));
    }
    onSelected(attrs) {
        this.focused = false;
        attrs.onTimeSelected(Time.parseFromString(this.value));
    }
    setScrollTop(attrs, vnode) {
        if (this.selectedIndex !== -1) {
            requestAnimationFrame(() => {
                vnode.dom.scrollTop = 44 * this.selectedIndex;
            });
        }
    }
}
//# sourceMappingURL=TimePicker.js.map