import m from "mithril";
import { Header } from "../../gui/Header.js";
import { ViewColumn } from "../../gui/base/ViewColumn";
import { lang } from "../../misc/LanguageViewModel";
import { ViewSlider } from "../../gui/nav/ViewSlider.js";
import { isKeyPressed, keyManager } from "../../misc/KeyManager";
import { downcast, getStartOfDay, isSameDayOfDate, ofClass } from "@tutao/tutanota-utils";
import { createGroupSettings } from "../../api/entities/tutanota/TypeRefs.js";
import { defaultCalendarColor, GroupType, Keys, reverse } from "../../api/common/TutanotaConstants";
import { locator } from "../../api/main/MainLocator";
import { getStartOfTheWeekOffset, getStartOfTheWeekOffsetForUser, getTimeZone, getWeekNumber } from "../date/CalendarUtils";
import { CalendarMonthView } from "./CalendarMonthView";
import { DateTime } from "luxon";
import { NotFoundError } from "../../api/common/error/RestError";
import { CalendarAgendaView } from "./CalendarAgendaView";
import { showEditCalendarDialog } from "../gui/EditCalendarDialog.js";
import { styles } from "../../gui/styles";
import { MultiDayCalendarView } from "./MultiDayCalendarView";
import { Dialog } from "../../gui/base/Dialog";
import { isApp } from "../../api/common/Env";
import { px, size } from "../../gui/size";
import { FolderColumnView } from "../../gui/FolderColumnView.js";
import { deviceConfig } from "../../misc/DeviceConfig";
import { exportCalendar, showCalendarImportDialog } from "../export/CalendarImporterDialog";
import { showNotAvailableForFreeDialog } from "../../misc/SubscriptionDialogs";
import { getSharedGroupName, hasCapabilityOnGroup, loadGroupMembers } from "../../sharing/GroupUtils";
import { showGroupSharingDialog } from "../../sharing/view/GroupSharingDialog";
import { GroupInvitationFolderRow } from "../../sharing/view/GroupInvitationFolderRow";
import { SidebarSection } from "../../gui/SidebarSection";
import { ProgrammingError } from "../../api/common/error/ProgrammingError";
import { calendarNavConfiguration, CalendarViewType, calendarWeek, daysHaveEvents, getGroupColors, shouldDefaultToAmPmTimeFormat, } from "../gui/CalendarGuiUtils.js";
import { showNewCalendarEventEditDialog } from "../gui/eventeditor-view/CalendarEventEditDialog.js";
import { CalendarEventPopup } from "../gui/eventpopup/CalendarEventPopup.js";
import { showProgressDialog } from "../../gui/dialogs/ProgressDialog";
import { IconButton } from "../../gui/base/IconButton.js";
import { createDropdown } from "../../gui/base/Dropdown.js";
import { BottomNav } from "../../gui/nav/BottomNav.js";
import { BaseTopLevelView } from "../../gui/BaseTopLevelView.js";
import { getEventWithDefaultTimes } from "../../api/common/utils/CommonCalendarUtils.js";
import { BackgroundColumnLayout } from "../../gui/BackgroundColumnLayout.js";
import { theme } from "../../gui/theme.js";
import { CalendarMobileHeader } from "./CalendarMobileHeader.js";
import { CalendarDesktopToolbar } from "./CalendarDesktopToolbar.js";
import { LazySearchBar } from "../../misc/LazySearchBar.js";
import { Time } from "../date/Time.js";
import { DaySelectorSidebar } from "../gui/day-selector/DaySelectorSidebar.js";
import { DaySelectorPopup } from "../gui/day-selector/DaySelectorPopup.js";
const CalendarViewTypeByValue = reverse(CalendarViewType);
export class CalendarView extends BaseTopLevelView {
    constructor({ attrs }) {
        super();
        this.redrawIntervalId = null;
        this.redrawTimeoutId = null;
        const userId = locator.logins.getUserController().user._id;
        this.viewModel = attrs.calendarViewModel;
        this.currentViewType = deviceConfig.getDefaultCalendarView(userId) || CalendarViewType.MONTH;
        this.htmlSanitizer = import("../../misc/HtmlSanitizer").then((m) => m.htmlSanitizer);
        this.sidebarColumn = new ViewColumn({
            view: () => m(FolderColumnView, {
                drawer: attrs.drawerAttrs,
                button: styles.isDesktopLayout()
                    ? {
                        label: "newEvent_action",
                        click: () => this.createNewEventDialog(),
                    }
                    : null,
                content: [
                    styles.isDesktopLayout()
                        ? m(DaySelectorSidebar, {
                            selectedDate: this.viewModel.selectedDate(),
                            onDateSelected: (date) => {
                                this.setUrl(this.currentViewType, date);
                                m.redraw();
                            },
                            startOfTheWeekOffset: getStartOfTheWeekOffset(downcast(locator.logins.getUserController().userSettingsGroupRoot.startOfTheWeek)),
                            showDaySelection: this.currentViewType !== CalendarViewType.MONTH && this.currentViewType !== CalendarViewType.WEEK,
                            highlightToday: true,
                            highlightSelectedWeek: this.currentViewType === CalendarViewType.WEEK,
                            hasEventsOn: (date) => this.hasEventsOn(date),
                        })
                        : null,
                    m(SidebarSection, {
                        name: "yourCalendars_label",
                        button: m(IconButton, {
                            title: "addCalendar_action",
                            colors: "nav" /* ButtonColor.Nav */,
                            click: () => this.onPressedAddCalendar(),
                            icon: "Add" /* Icons.Add */,
                            size: 1 /* ButtonSize.Compact */,
                        }),
                    }, this.renderCalendars(false)),
                    m(SidebarSection, {
                        name: "otherCalendars_label",
                    }, this.renderCalendars(true)),
                    this.viewModel.calendarInvitations().length > 0
                        ? m(SidebarSection, {
                            name: "calendarInvitations_label",
                        }, this.viewModel.calendarInvitations().map((invitation) => m(GroupInvitationFolderRow, {
                            invitation,
                        })))
                        : null,
                ],
                ariaLabel: "calendar_label",
            }),
        }, 0 /* ColumnType.Foreground */, {
            minWidth: size.calendar_first_col_min_width,
            maxWidth: size.first_col_max_width,
            headerCenter: () => (this.currentViewType === CalendarViewType.WEEK ? lang.get("month_label") : lang.get("calendar_label")),
        });
        this.contentColumn = new ViewColumn({
            view: () => {
                const groupColors = getGroupColors(locator.logins.getUserController().userSettingsGroupRoot);
                switch (this.currentViewType) {
                    case CalendarViewType.MONTH:
                        return m(BackgroundColumnLayout, {
                            backgroundColor: theme.navigation_bg,
                            desktopToolbar: () => this.renderDesktopToolbar(),
                            mobileHeader: () => this.renderMobileHeader(attrs.header),
                            columnLayout: m(CalendarMonthView, {
                                temporaryEvents: this.viewModel.temporaryEvents,
                                eventsForDays: this.viewModel.eventsForDays,
                                getEventsOnDaysToRender: this.viewModel.getEventsOnDaysToRender.bind(this.viewModel),
                                onEventClicked: (calendarEvent, domEvent) => this.onEventSelected(calendarEvent, domEvent, this.htmlSanitizer),
                                onNewEvent: (date) => {
                                    this.createNewEventDialog(date);
                                },
                                selectedDate: this.viewModel.selectedDate(),
                                onDateSelected: (date, calendarViewType) => {
                                    this.setUrl(calendarViewType, date);
                                },
                                onChangeMonth: (next) => this.viewPeriod(CalendarViewType.MONTH, next),
                                amPmFormat: locator.logins.getUserController().userSettingsGroupRoot.timeFormat === "1" /* TimeFormat.TWELVE_HOURS */,
                                startOfTheWeek: downcast(locator.logins.getUserController().userSettingsGroupRoot.startOfTheWeek),
                                groupColors,
                                hiddenCalendars: this.viewModel.hiddenCalendars,
                                dragHandlerCallbacks: this.viewModel,
                            }),
                        });
                    case CalendarViewType.DAY:
                        return m(BackgroundColumnLayout, {
                            backgroundColor: theme.navigation_bg,
                            desktopToolbar: () => this.renderDesktopToolbar(),
                            mobileHeader: () => this.renderMobileHeader(attrs.header),
                            columnLayout: m(MultiDayCalendarView, {
                                temporaryEvents: this.viewModel.temporaryEvents,
                                getEventsOnDays: this.viewModel.getEventsOnDaysToRender.bind(this.viewModel),
                                daysInPeriod: 1,
                                onEventClicked: (event, domEvent) => this.onEventSelected(event, domEvent, this.htmlSanitizer),
                                onNewEvent: (date) => {
                                    this.createNewEventDialog(date);
                                },
                                selectedDate: this.viewModel.selectedDate(),
                                onDateSelected: (date) => {
                                    this.setUrl(CalendarViewType.DAY, date);
                                },
                                groupColors,
                                onChangeViewPeriod: (next) => this.viewPeriod(CalendarViewType.DAY, next),
                                startOfTheWeek: downcast(locator.logins.getUserController().userSettingsGroupRoot.startOfTheWeek),
                                dragHandlerCallbacks: this.viewModel,
                                isDaySelectorExpanded: this.viewModel.isDaySelectorExpanded(),
                                weekIndicator: calendarWeek(this.viewModel.selectedDate(), this.viewModel.weekStart),
                                selectedTime: this.viewModel.selectedTime,
                            }),
                        });
                    case CalendarViewType.WEEK:
                        return m(BackgroundColumnLayout, {
                            backgroundColor: theme.navigation_bg,
                            desktopToolbar: () => this.renderDesktopToolbar(),
                            mobileHeader: () => this.renderMobileHeader(attrs.header),
                            columnLayout: m(MultiDayCalendarView, {
                                temporaryEvents: this.viewModel.temporaryEvents,
                                getEventsOnDays: this.viewModel.getEventsOnDaysToRender.bind(this.viewModel),
                                daysInPeriod: 7,
                                onEventClicked: (event, domEvent) => this.onEventSelected(event, domEvent, this.htmlSanitizer),
                                onNewEvent: (date) => {
                                    this.createNewEventDialog(date);
                                },
                                selectedDate: this.viewModel.selectedDate(),
                                onDateSelected: (date, viewType) => {
                                    this.viewModel.selectedDate(date);
                                    this.setUrl(viewType ?? CalendarViewType.WEEK, date);
                                },
                                startOfTheWeek: downcast(locator.logins.getUserController().userSettingsGroupRoot.startOfTheWeek),
                                groupColors,
                                onChangeViewPeriod: (next) => this.viewPeriod(CalendarViewType.WEEK, next),
                                dragHandlerCallbacks: this.viewModel,
                                isDaySelectorExpanded: this.viewModel.isDaySelectorExpanded(),
                                weekIndicator: calendarWeek(this.viewModel.selectedDate(), this.viewModel.weekStart),
                                selectedTime: this.viewModel.selectedTime,
                            }),
                        });
                    case CalendarViewType.AGENDA:
                        return m(BackgroundColumnLayout, {
                            backgroundColor: theme.navigation_bg,
                            desktopToolbar: () => this.renderDesktopToolbar(),
                            mobileHeader: () => this.renderMobileHeader(attrs.header),
                            columnLayout: m(CalendarAgendaView, {
                                selectedDate: this.viewModel.selectedDate(),
                                selectedTime: this.viewModel.selectedTime,
                                eventsForDays: this.viewModel.eventsForDays,
                                amPmFormat: shouldDefaultToAmPmTimeFormat(),
                                onEventClicked: (event, domEvent) => {
                                    if (styles.isDesktopLayout()) {
                                        this.viewModel.updatePreviewedEvent(event);
                                    }
                                    else {
                                        this.onEventSelected(event, domEvent, this.htmlSanitizer);
                                    }
                                },
                                groupColors,
                                hiddenCalendars: this.viewModel.hiddenCalendars,
                                startOfTheWeekOffset: getStartOfTheWeekOffsetForUser(locator.logins.getUserController().userSettingsGroupRoot),
                                isDaySelectorExpanded: this.viewModel.isDaySelectorExpanded(),
                                onDateSelected: (date) => this.setUrl(CalendarViewType.AGENDA, date),
                                onShowDate: (date) => this.setUrl(CalendarViewType.DAY, date),
                                eventPreviewModel: this.viewModel.eventPreviewModel,
                            }),
                        });
                    default:
                        throw new ProgrammingError(`invalid CalendarViewType: "${this.currentViewType}"`);
                }
            },
        }, 1 /* ColumnType.Background */, {
            minWidth: size.calendar_first_col_min_width + size.third_col_min_width,
            maxWidth: size.third_col_max_width,
        });
        this.viewSlider = new ViewSlider([this.sidebarColumn, this.contentColumn]);
        const shortcuts = this.setupShortcuts();
        const streamListeners = [];
        this.oncreate = () => {
            keyManager.registerShortcuts(shortcuts);
            // do both a timeout and interval to ensure the time indicator is done on the minute rather than some delay afterwards
            if (!this.redrawIntervalId && !this.redrawTimeoutId) {
                const timeToNextMinute = (60 - new Date().getSeconds()) * 1000;
                this.redrawTimeoutId = window.setTimeout(() => {
                    this.redrawIntervalId = window.setInterval(m.redraw, 1000 * 60);
                    this.redrawTimeoutId = null;
                    m.redraw();
                }, timeToNextMinute);
            }
            streamListeners.push(this.viewModel.calendarInvitations.map(() => {
                m.redraw();
            }));
            streamListeners.push(this.viewModel.redraw.map(m.redraw));
        };
        this.onremove = () => {
            keyManager.unregisterShortcuts(shortcuts);
            if (this.redrawTimeoutId) {
                window.clearTimeout(this.redrawTimeoutId);
                this.redrawTimeoutId = null;
            }
            if (this.redrawIntervalId) {
                window.clearInterval(this.redrawIntervalId);
                this.redrawIntervalId = null;
            }
            for (let listener of streamListeners) {
                listener.end(true);
            }
        };
    }
    renderDesktopToolbar() {
        return m(CalendarDesktopToolbar, {
            navConfig: calendarNavConfiguration(this.currentViewType, this.viewModel.selectedDate(), this.viewModel.weekStart, "detailed", (viewType, next) => this.viewPeriod(viewType, next)),
            viewType: this.currentViewType,
            onToday: () => {
                // in case it has been set, when onToday is called we definitely do not want the time to be ignored
                this.viewModel.ignoreNextValidTimeSelection = false;
                this.setUrl(m.route.param("view"), new Date());
            },
            onViewTypeSelected: (viewType) => this.setUrl(viewType, this.viewModel.selectedDate()),
        });
    }
    renderMobileHeader(header) {
        return m(CalendarMobileHeader, {
            ...header,
            viewType: this.currentViewType,
            viewSlider: this.viewSlider,
            showExpandIcon: !styles.isDesktopLayout() && this.currentViewType !== CalendarViewType.MONTH,
            isDaySelectorExpanded: this.viewModel.isDaySelectorExpanded(),
            navConfiguration: calendarNavConfiguration(this.currentViewType, this.viewModel.selectedDate(), this.viewModel.weekStart, "short", (viewType, next) => this.viewPeriod(viewType, next)),
            onCreateEvent: () => this.createNewEventDialog(),
            onToday: () => {
                // in case it has been set, when onToday is called we definitely do not want the time to be ignored
                this.viewModel.ignoreNextValidTimeSelection = false;
                this.setUrl(m.route.param("view"), new Date());
            },
            onViewTypeSelected: (viewType) => this.setUrl(viewType, this.viewModel.selectedDate()),
            onTap: (_event, dom) => {
                if (this.currentViewType !== CalendarViewType.MONTH && styles.isSingleColumnLayout()) {
                    this.viewModel.setDaySelectorExpanded(!this.viewModel.isDaySelectorExpanded());
                    return;
                }
                if (!styles.isDesktopLayout() && this.currentViewType !== CalendarViewType.MONTH) {
                    if (this.viewModel.isDaySelectorExpanded()) {
                        this.viewModel.setDaySelectorExpanded(false);
                    }
                    this.showCalendarPopup(dom);
                }
            },
        });
    }
    setupShortcuts() {
        return [
            {
                key: Keys.ONE,
                exec: () => this.setUrl(CalendarViewType.WEEK, this.viewModel.selectedDate()),
                help: "switchWeekView_action",
            },
            {
                key: Keys.TWO,
                exec: () => this.setUrl(CalendarViewType.MONTH, this.viewModel.selectedDate()),
                help: "switchMonthView_action",
            },
            {
                key: Keys.THREE,
                exec: () => this.setUrl(CalendarViewType.AGENDA, this.viewModel.selectedDate()),
                help: "switchAgendaView_action",
            },
            {
                key: Keys.T,
                exec: () => this.setUrl(m.route.param("view"), new Date()),
                help: "viewToday_action",
            },
            {
                key: Keys.J,
                enabled: () => this.currentViewType !== CalendarViewType.AGENDA,
                exec: () => this.viewPeriod(this.currentViewType, true),
                help: "viewNextPeriod_action",
            },
            {
                key: Keys.K,
                enabled: () => this.currentViewType !== CalendarViewType.AGENDA,
                exec: () => this.viewPeriod(this.currentViewType, false),
                help: "viewPrevPeriod_action",
            },
            {
                key: Keys.N,
                exec: () => {
                    this.createNewEventDialog();
                },
                help: "newEvent_action",
            },
        ];
    }
    async createNewEventDialog(date = null) {
        const dateToUse = date ?? this.viewModel.selectedDate();
        // Disallow creation of events when there is no existing calendar
        let calendarInfos = this.viewModel.getCalendarInfosCreateIfNeeded();
        if (calendarInfos instanceof Promise) {
            await showProgressDialog("pleaseWait_msg", calendarInfos);
        }
        const mailboxDetails = await locator.mailModel.getUserMailboxDetails();
        const mailboxProperties = await locator.mailModel.getMailboxProperties(mailboxDetails.mailboxGroupRoot);
        const model = await locator.calendarEventModel(0 /* CalendarOperation.Create */, getEventWithDefaultTimes(dateToUse), mailboxDetails, mailboxProperties, null);
        if (model) {
            await showNewCalendarEventEditDialog(model);
        }
    }
    viewPeriod(viewType, next) {
        let duration;
        let unit;
        switch (viewType) {
            case CalendarViewType.MONTH:
                duration = {
                    month: 1,
                };
                unit = "month";
                break;
            case CalendarViewType.WEEK:
                duration = {
                    week: 1,
                };
                unit = "week";
                break;
            case CalendarViewType.DAY:
                duration = {
                    day: 1,
                };
                unit = "day";
                break;
            case CalendarViewType.AGENDA:
                duration = styles.isDesktopLayout()
                    ? { day: 1 }
                    : {
                        week: this.viewModel.isDaySelectorExpanded() ? 0 : 1,
                        month: this.viewModel.isDaySelectorExpanded() ? 1 : 0,
                    };
                unit = "day";
                break;
            default:
                throw new ProgrammingError("Invalid CalendarViewType: " + viewType);
        }
        const dateTime = DateTime.fromJSDate(this.viewModel.selectedDate());
        const newDate = next ? dateTime.plus(duration).startOf(unit).toJSDate() : dateTime.minus(duration).startOf(unit).toJSDate();
        this.viewModel.selectedDate(newDate);
        // ignoreNextTimeSelection is set to true here, as viewPeriod is only called when changing the view by swiping (or using previous/next buttons)
        // and we don't want jarring time jumps when doing that
        this.viewModel.ignoreNextValidTimeSelection = true;
        this.setUrl(viewType, newDate, false);
        m.redraw();
    }
    onPressedAddCalendar() {
        if (locator.logins.getUserController().getCalendarMemberships().length === 0) {
            this.showCreateCalendarDialog();
        }
        else {
            import("../../misc/SubscriptionDialogs")
                .then((SubscriptionDialogUtils) => SubscriptionDialogUtils.checkPaidSubscription())
                .then((ok) => {
                if (ok) {
                    this.showCreateCalendarDialog();
                }
            });
        }
    }
    showCreateCalendarDialog() {
        showEditCalendarDialog({
            name: "",
            color: Math.random().toString(16).slice(-6),
        }, "add_action", false, async (dialog, properties) => {
            const calendarModel = await locator.calendarModel();
            await calendarModel.createCalendar(properties.name, properties.color);
            dialog.close();
        }, "save_action");
    }
    renderCalendars(shared) {
        const setHidden = (viewModel, groupRootId) => {
            const newHiddenCalendars = new Set(viewModel.hiddenCalendars);
            viewModel.hiddenCalendars.has(groupRootId) ? newHiddenCalendars.delete(groupRootId) : newHiddenCalendars.add(groupRootId);
            viewModel.setHiddenCalendars(newHiddenCalendars);
        };
        const calendarInfos = this.viewModel.calendarInfos;
        return calendarInfos.size > 0
            ? Array.from(calendarInfos.values())
                .filter((calendarInfo) => calendarInfo.shared === shared)
                .map((calendarInfo) => {
                const { userSettingsGroupRoot } = locator.logins.getUserController();
                const existingGroupSettings = userSettingsGroupRoot.groupSettings.find((gc) => gc.group === calendarInfo.groupInfo.group) ?? null;
                const colorValue = "#" + (existingGroupSettings ? existingGroupSettings.color : defaultCalendarColor);
                const groupRootId = calendarInfo.groupRoot._id;
                const groupName = getSharedGroupName(calendarInfo.groupInfo, locator.logins.getUserController(), shared);
                const isHidden = this.viewModel.hiddenCalendars.has(groupRootId);
                return m(".folder-row.flex-start.plr-button", [
                    m(".flex.flex-grow.center-vertically.button-height", [
                        m(".calendar-checkbox", {
                            role: "checkbox",
                            title: groupName,
                            tabindex: "0" /* TabIndex.Default */,
                            "aria-checked": (!isHidden).toString(),
                            "aria-label": groupName,
                            onclick: () => setHidden(this.viewModel, groupRootId),
                            onkeydown: (e) => {
                                if (isKeyPressed(e.key, Keys.SPACE, Keys.RETURN)) {
                                    setHidden(this.viewModel, groupRootId);
                                    e.preventDefault();
                                }
                            },
                            style: {
                                "border-color": colorValue,
                                background: isHidden ? "" : colorValue,
                                transition: "all 0.3s",
                                cursor: "pointer",
                                marginLeft: px(size.hpad_button),
                            },
                        }),
                        m(".pl-m.b.flex-grow.text-ellipsis", {
                            style: {
                                width: 0,
                            },
                        }, groupName),
                    ]),
                    this.createCalendarActionDropdown(calendarInfo, colorValue, existingGroupSettings, userSettingsGroupRoot, shared),
                ]);
            })
            : null;
    }
    createCalendarActionDropdown(calendarInfo, colorValue, existingGroupSettings, userSettingsGroupRoot, sharedCalendar) {
        const { group, groupInfo, groupRoot } = calendarInfo;
        const user = locator.logins.getUserController().user;
        return m(IconButton, {
            title: "more_label",
            colors: "nav" /* ButtonColor.Nav */,
            icon: "More" /* Icons.More */,
            size: 1 /* ButtonSize.Compact */,
            click: createDropdown({
                lazyButtons: () => [
                    {
                        label: "edit_action",
                        icon: "Edit" /* Icons.Edit */,
                        size: 1 /* ButtonSize.Compact */,
                        click: () => this.onPressedEditCalendar(groupInfo, colorValue, existingGroupSettings, userSettingsGroupRoot, sharedCalendar),
                    },
                    {
                        label: "sharing_label",
                        icon: "ContactImport" /* Icons.ContactImport */,
                        click: () => {
                            if (locator.logins.getUserController().isFreeAccount()) {
                                showNotAvailableForFreeDialog();
                            }
                            else {
                                showGroupSharingDialog(groupInfo, sharedCalendar);
                            }
                        },
                    },
                    !isApp() && group.type === GroupType.Calendar && hasCapabilityOnGroup(user, group, "1" /* ShareCapability.Write */)
                        ? {
                            label: "import_action",
                            icon: "Import" /* Icons.Import */,
                            click: () => showCalendarImportDialog(groupRoot),
                        }
                        : null,
                    !isApp() && group.type === GroupType.Calendar && hasCapabilityOnGroup(user, group, "0" /* ShareCapability.Read */)
                        ? {
                            label: "export_action",
                            icon: "Export" /* Icons.Export */,
                            click: () => {
                                const alarmInfoList = user.alarmInfoList;
                                alarmInfoList &&
                                    exportCalendar(getSharedGroupName(groupInfo, locator.logins.getUserController(), sharedCalendar), groupRoot, alarmInfoList.alarms, new Date(), getTimeZone());
                            },
                        }
                        : null,
                    !sharedCalendar
                        ? {
                            label: "delete_action",
                            icon: "Trash" /* Icons.Trash */,
                            click: () => this.confirmDeleteCalendar(calendarInfo),
                        }
                        : null,
                ],
            }),
        });
    }
    confirmDeleteCalendar(calendarInfo) {
        const calendarName = getSharedGroupName(calendarInfo.groupInfo, locator.logins.getUserController(), false);
        loadGroupMembers(calendarInfo.group, locator.entityClient).then((members) => {
            const ownerMail = locator.logins.getUserController().userGroupInfo.mailAddress;
            const otherMembers = members.filter((member) => member.info.mailAddress !== ownerMail);
            Dialog.confirm(() => (otherMembers.length > 0
                ? lang.get("deleteSharedCalendarConfirm_msg", {
                    "{calendar}": calendarName,
                }) + " "
                : "") +
                lang.get("deleteCalendarConfirm_msg", {
                    "{calendar}": calendarName,
                })).then((confirmed) => {
                if (confirmed) {
                    this.viewModel.deleteCalendar(calendarInfo).catch(ofClass(NotFoundError, () => console.log("Calendar to be deleted was not found.")));
                }
            });
        });
    }
    onPressedEditCalendar(groupInfo, colorValue, existingGroupSettings, userSettingsGroupRoot, shared) {
        showEditCalendarDialog({
            name: getSharedGroupName(groupInfo, locator.logins.getUserController(), shared),
            color: colorValue.substring(1),
        }, "edit_action", shared, (dialog, properties) => {
            if (!shared) {
                groupInfo.name = properties.name;
                locator.entityClient.update(groupInfo);
            }
            // color always set for existing calendar
            if (existingGroupSettings) {
                existingGroupSettings.color = properties.color;
                existingGroupSettings.name = shared && properties.name !== groupInfo.name ? properties.name : null;
            }
            else {
                const newGroupSettings = createGroupSettings({
                    group: groupInfo.group,
                    color: properties.color,
                    name: shared && properties.name !== groupInfo.name ? properties.name : null,
                });
                userSettingsGroupRoot.groupSettings.push(newGroupSettings);
            }
            locator.entityClient.update(userSettingsGroupRoot);
            dialog.close();
        }, "save_action");
    }
    view({ attrs }) {
        return m(".main-view", m(this.viewSlider, {
            header: m(Header, {
                searchBar: () => m(LazySearchBar, {
                    placeholder: lang.get("searchCalendar_placeholder"),
                }),
                ...attrs.header,
            }),
            bottomNav: m(BottomNav),
        }));
    }
    onNewUrl(args) {
        if (!args.view) {
            this.setUrl(this.currentViewType, this.viewModel.selectedDate(), true);
        }
        else {
            this.currentViewType = CalendarViewTypeByValue[args.view] ? args.view : CalendarViewType.MONTH;
            const urlDateParam = args.date;
            if (urlDateParam) {
                // Unlike JS Luxon assumes local time zone when parsing and not UTC. That's what we want
                const luxonDate = DateTime.fromISO(urlDateParam);
                let date = new Date();
                if (luxonDate.isValid) {
                    date = luxonDate.toJSDate();
                }
                if (this.viewModel.selectedDate().getTime() !== date.getTime()) {
                    this.viewModel.selectedDate(date);
                    m.redraw();
                }
                const today = new Date();
                if (isSameDayOfDate(today, date) || (args.view === "week" && getWeekNumber(date) === getWeekNumber(today))) {
                    const time = Time.fromDate(today);
                    this.viewModel.setSelectedTime(time);
                }
                else {
                    this.viewModel.setSelectedTime(undefined);
                }
            }
            deviceConfig.setDefaultCalendarView(locator.logins.getUserController().user._id, this.currentViewType);
        }
    }
    getViewSlider() {
        return this.viewSlider;
    }
    setUrl(view, date, replace = false) {
        const dateString = DateTime.fromJSDate(date).toISODate();
        m.route.set("/calendar/:view/:date", {
            view,
            date: dateString,
        }, {
            replace,
        });
    }
    async onEventSelected(selectedEvent, domEvent, htmlSanitizerPromise) {
        const domTarget = domEvent.currentTarget;
        if (domTarget == null || !(domTarget instanceof HTMLElement)) {
            return;
        }
        const x = domEvent.clientX;
        const y = domEvent.clientY;
        // We want the popup to show at the users mouse
        const rect = {
            bottom: y,
            height: 0,
            width: 0,
            top: y,
            left: x,
            right: x,
        };
        const calendars = await this.viewModel.getCalendarInfosCreateIfNeeded();
        const [popupModel, htmlSanitizer] = await Promise.all([locator.calendarEventPreviewModel(selectedEvent, calendars), htmlSanitizerPromise]);
        new CalendarEventPopup(popupModel, rect, htmlSanitizer).show();
    }
    showCalendarPopup(dom) {
        // When the user clicks the month name in the header, the target can be the month's name or the icon on the right
        // side of month's name, so we hardcoded the left spacing to be the same used by the month name, so doesn't matter
        // if the user clicks on month's name or on the icon
        // noinspection JSSuspiciousNameCombination
        const elementRect = { ...dom.getBoundingClientRect(), left: size.button_height };
        const selector = new DaySelectorPopup(elementRect, {
            selectedDate: getStartOfDay(this.viewModel.selectedDate()),
            onDateSelected: (date) => {
                this.viewModel.selectedDate(date);
                this.setUrl(this.currentViewType, date);
                selector.close();
            },
            startOfTheWeekOffset: getStartOfTheWeekOffset(locator.logins.getUserController().userSettingsGroupRoot.startOfTheWeek),
            highlightToday: true,
            highlightSelectedWeek: this.currentViewType === CalendarViewType.WEEK,
            hasEventsOn: (date) => this.hasEventsOn(date),
        });
        selector.show();
    }
    hasEventsOn(date) {
        return daysHaveEvents(this.viewModel.getEventsOnDaysToRender([date]));
    }
}
//# sourceMappingURL=CalendarView.js.map