import { locator } from "../api/main/MainLocator.js";
import { Dialog } from "../gui/base/Dialog.js";
import m from "mithril";
import { TextField } from "../gui/base/TextField.js";
import { px, size } from "../gui/size.js";
import { IconButton } from "../gui/base/IconButton.js";
import { MailRecipientsTextField } from "../gui/MailRecipientsTextField.js";
import { noOp } from "@tutao/tutanota-utils";
import { lang } from "../misc/LanguageViewModel.js";
import { isSameId } from "../api/common/utils/EntityUtils.js";
import { Keys } from "../api/common/TutanotaConstants.js";
import { isMailAddress } from "../misc/FormatValidator.js";
import { cleanMailAddress } from "../api/common/utils/CommonCalendarUtils.js";
export async function showContactListEditor(contactListGroupRoot, headerText, save, addressesOnList) {
    let showNameInput = true;
    const recipientsSearch = await locator.recipientsSearchModel();
    if (contactListGroupRoot) {
        showNameInput = false;
        recipientsSearch.setFilter((item) => {
            // Exclude the list that we are editing to not show up in suggestions.
            // It is valid to include other lists to copy them into the current one.
            return !(item.type === "contactlist" && isSameId(item.value.groupRoot._id, contactListGroupRoot._id));
        });
    }
    const editorModel = new ContactListEditorModel(addressesOnList ?? []);
    const dialogCloseAction = () => {
        dialog.close();
    };
    let headerBarAttrs = {
        left: [
            {
                label: "cancel_action",
                click: dialogCloseAction,
                type: "secondary" /* ButtonType.Secondary */,
            },
        ],
        right: [
            {
                label: "save_action",
                click: () => {
                    save(editorModel.name, editorModel.newAddresses);
                    dialog.close();
                },
                type: "primary" /* ButtonType.Primary */,
            },
        ],
        middle: () => headerText,
    };
    const dialog = Dialog.editDialog(headerBarAttrs, ContactListEditor, {
        model: editorModel,
        contactSearch: recipientsSearch,
        showNameInput,
    }).addShortcut({
        key: Keys.ESC,
        exec: () => dialog.close(),
        help: "close_alt",
    });
    dialog.show();
}
export async function showContactListNameEditor(name, save) {
    let nameInput = name;
    let form = () => [
        m(TextField, {
            label: "name_label",
            value: nameInput,
            oninput: (newInput) => {
                nameInput = newInput;
            },
        }),
    ];
    const okAction = async (dialog) => {
        dialog.close();
        save(nameInput);
    };
    Dialog.showActionDialog({
        title: lang.get("editContactList_action"),
        child: form,
        allowOkWithReturn: true,
        okAction: okAction,
    });
}
export class ContactListEditorModel {
    constructor(addresses) {
        this.name = "";
        this.newAddresses = [];
        this.currentAddresses = addresses;
    }
    addRecipient(address) {
        this.newAddresses = [address, ...this.newAddresses];
    }
    removeRecipient(address) {
        this.newAddresses = this.newAddresses.filter((a) => address !== a);
    }
}
class ContactListEditor {
    constructor(vnode) {
        this.newAddress = "";
        this.showNameInput = true;
        this.model = vnode.attrs.model;
        this.search = vnode.attrs.contactSearch;
        this.showNameInput = vnode.attrs.showNameInput ?? true;
    }
    view() {
        let helpLabel = null;
        if (this.newAddress.trim().length > 0 && !isMailAddress(this.newAddress.trim(), false)) {
            helpLabel = () => lang.get("invalidInputFormat_msg");
        }
        else if (this.model.currentAddresses.includes(cleanMailAddress(this.newAddress)) ||
            this.model.newAddresses.includes(cleanMailAddress(this.newAddress))) {
            helpLabel = () => lang.get("addressAlreadyExistsOnList_msg");
        }
        return m("", [
            this.showNameInput
                ? m(TextField, {
                    label: "name_label",
                    class: "big-input pt flex-grow",
                    value: this.model.name,
                    oninput: (name) => (this.model.name = name),
                })
                : null,
            m(MailRecipientsTextField, {
                label: "addEntries_action",
                text: this.newAddress,
                onTextChanged: (v) => (this.newAddress = v),
                // we don't show bubbles, we just want the search dropdown
                recipients: [],
                disabled: false,
                onRecipientAdded: (address) => {
                    if (!this.model.newAddresses.includes(address) && !this.model.currentAddresses.includes(address)) {
                        this.model.addRecipient(address);
                    }
                    m.redraw();
                },
                // do nothing because we don't have any bubbles here
                onRecipientRemoved: noOp,
                search: this.search,
                helpLabel,
            }),
            this.model.newAddresses.map((address) => this.renderAddress(address)),
        ]);
    }
    renderAddress(address) {
        return m(".flex", {
            style: {
                height: px(size.button_height),
                borderBottom: "1px transparent",
                marginTop: px(size.vpad),
            },
        }, [
            m(".flex.col.flex-grow.overflow-hidden.flex-no-grow-shrink-auto", [address]),
            m(".flex-grow"),
            m(IconButton, {
                title: "remove_action",
                icon: "Cancel" /* Icons.Cancel */,
                click: () => this.model.removeRecipient(address),
            }),
        ]);
    }
}
//# sourceMappingURL=ContactListEditor.js.map