import { ListModel } from "../../misc/ListModel.js";
import { ContactTypeRef } from "../../api/entities/tutanota/TypeRefs.js";
import { compareContacts } from "./ContactGuiUtils.js";
import { assertNotNull, lazyMemoized } from "@tutao/tutanota-utils";
import { GENERATED_MAX_ID, getElementId } from "../../api/common/utils/EntityUtils.js";
import { isUpdateForTypeRef } from "../../api/common/utils/EntityUpdateUtils.js";
import { ListAutoSelectBehavior } from "../../misc/DeviceConfig.js";
/** ViewModel for the overall contact view. */
export class ContactViewModel {
    constructor(contactModel, entityClient, eventController, router, updateUi) {
        this.contactModel = contactModel;
        this.entityClient = entityClient;
        this.eventController = eventController;
        this.router = router;
        this.updateUi = updateUi;
        /** id of the contact we are trying to load based on the url */
        this.targetContactId = null;
        this.sortByFirstName = true;
        this.listModelStateStream = null;
        this.listModel = new ListModel({
            topId: GENERATED_MAX_ID,
            fetch: async () => {
                const items = await this.entityClient.loadAll(ContactTypeRef, this.contactListId);
                return { items, complete: true };
            },
            loadSingle: async (elementId) => {
                const listId = await this.contactModel.getContactListId();
                if (listId == null)
                    return null;
                return this.entityClient.load(ContactTypeRef, [listId, elementId]);
            },
            sortCompare: (c1, c2) => compareContacts(c1, c2, this.sortByFirstName),
            autoSelectBehavior: () => ListAutoSelectBehavior.NONE,
        });
        this.initOnce = lazyMemoized(() => {
            this.eventController.addEntityListener(this.entityListener);
            this.listModelStateStream = this.listModel.stateStream.map(() => {
                this.updateUi();
                this.updateUrl();
            });
        });
        this.entityListener = async (updates) => {
            for (const update of updates) {
                if (isUpdateForTypeRef(ContactTypeRef, update) && update.instanceListId === this.contactListId) {
                    await this.listModel.entityEventReceived(update.instanceId, update.operation);
                }
            }
        };
    }
    async init(contactListId, contactId) {
        // update url if the view was just opened
        if (contactListId == null && contactId == null)
            this.updateUrl();
        if (this.contactListId)
            return;
        this.contactListId = assertNotNull(await this.contactModel.getContactListId(), "not available for external users");
        this.targetContactId = contactId ?? null;
        this.listModel.loadInitial().then(() => {
            // we are loading all contacts at once anyway so we are not worried about starting parallel loads for target
            typeof contactId === "string" &&
                this.loadAndSelect(contactId).finally(() => {
                    this.targetContactId = null;
                });
        });
        this.initOnce();
    }
    updateUrl() {
        const contactId = this.targetContactId != null
            ? this.targetContactId
            : !this.listModel.state.inMultiselect && this.listModel.getSelectedAsArray().length === 1
                ? getElementId(this.listModel.getSelectedAsArray()[0])
                : null;
        if (contactId) {
            this.router.routeTo(`/contact/:listId/:contactId`, { listId: this.contactListId, contactId: contactId });
        }
        else {
            this.router.routeTo(`/contact/:listId`, { listId: this.contactListId });
        }
    }
    async loadAndSelect(contactId) {
        const listId = this.contactListId;
        await this.listModel.loadAndSelect(contactId, () => this.contactListId !== listId);
    }
    setSortByFirstName(sorting) {
        this.sortByFirstName = sorting;
        this.listModel.sort();
    }
    listState() {
        return this.listModel.state;
    }
    dispose() {
        this.eventController.removeEntityListener(this.entityListener);
        this.listModelStateStream?.end(true);
        this.listModelStateStream = null;
    }
}
//# sourceMappingURL=ContactViewModel.js.map