import m from "mithril";
import stream from "mithril/stream";
import { assertNotNull, groupBy, lazyMemoized, neverNull, noOp, ofClass, promiseMap, splitInChunks } from "@tutao/tutanota-utils";
import { createMailAddressProperties, createMailboxProperties, MailboxGroupRootTypeRef, MailboxPropertiesTypeRef, MailBoxTypeRef, MailFolderTypeRef, MailTypeRef, } from "../../api/entities/tutanota/TypeRefs.js";
import { GroupInfoTypeRef, GroupTypeRef } from "../../api/entities/sys/TypeRefs.js";
import { FeatureType, MailFolderType, MAX_NBR_MOVE_DELETE_MAIL_SERVICE } from "../../api/common/TutanotaConstants";
import { lang } from "../../misc/LanguageViewModel";
import { elementIdPart, GENERATED_MAX_ID, getElementId, getListId, isSameId, listIdPart } from "../../api/common/utils/EntityUtils";
import { LockedError, NotFoundError, PreconditionFailedError } from "../../api/common/error/RestError";
import { getEnabledMailAddressesWithUser } from "./MailUtils.js";
import { ProgrammingError } from "../../api/common/error/ProgrammingError.js";
import { FolderSystem } from "../../api/common/mail/FolderSystem.js";
import { UserError } from "../../api/main/UserError.js";
import { assertSystemFolderOfType, isSpamOrTrashFolder } from "../../api/common/mail/CommonMailUtils.js";
import { containsEventOfType, isUpdateForTypeRef } from "../../api/common/utils/EntityUpdateUtils.js";
export class MailModel {
    constructor(notifications, eventController, connectivityModel, mailFacade, entityClient, logins, inboxRuleHandler) {
        this.notifications = notifications;
        this.eventController = eventController;
        this.connectivityModel = connectivityModel;
        this.mailFacade = mailFacade;
        this.entityClient = entityClient;
        this.logins = logins;
        this.inboxRuleHandler = inboxRuleHandler;
        /** Empty stream until init() is finished, exposed mostly for map()-ing, use getMailboxDetails to get a promise */
        this.mailboxDetails = stream();
        this.mailboxCounters = stream({});
        this.initialization = null;
        /**
         * Map from MailboxGroupRoot id to MailboxProperties
         * A way to avoid race conditions in case we try to create mailbox properties from multiple places.
         *
         */
        this.mailboxPropertiesPromises = new Map();
        // only init listeners once
        this.initListeners = lazyMemoized(() => {
            this.eventController.addEntityListener((updates, eventOwnerGroupId) => this.entityEventsReceived(updates, eventOwnerGroupId));
            this.eventController.getCountersStream().map((update) => {
                this._mailboxCountersUpdates(update);
            });
        });
    }
    init() {
        // if we are in the process of loading do not start another one in parallel
        if (this.initialization) {
            return this.initialization;
        }
        this.initListeners();
        return this._init();
    }
    _init() {
        const mailGroupMemberships = this.logins.getUserController().getMailGroupMemberships();
        const mailBoxDetailsPromises = mailGroupMemberships.map((m) => this.mailboxDetailsFromMembership(m));
        this.initialization = Promise.all(mailBoxDetailsPromises).then((details) => {
            this.mailboxDetails(details);
        });
        return this.initialization.catch((e) => {
            console.warn("mail model initialization failed!", e);
            this.initialization = null;
            throw e;
        });
    }
    /**
     * load mailbox details from a mailgroup membership
     */
    async mailboxDetailsFromMembership(membership) {
        const [mailboxGroupRoot, mailGroupInfo, mailGroup] = await Promise.all([
            this.entityClient.load(MailboxGroupRootTypeRef, membership.group),
            this.entityClient.load(GroupInfoTypeRef, membership.groupInfo),
            this.entityClient.load(GroupTypeRef, membership.group),
        ]);
        const mailbox = await this.entityClient.load(MailBoxTypeRef, mailboxGroupRoot.mailbox);
        const folders = await this.loadFolders(neverNull(mailbox.folders).folders);
        return {
            mailbox,
            folders: new FolderSystem(folders),
            mailGroupInfo,
            mailGroup,
            mailboxGroupRoot,
        };
    }
    loadFolders(folderListId) {
        return this.entityClient.loadAll(MailFolderTypeRef, folderListId).then((folders) => {
            return folders.filter((f) => {
                // We do not show spam or archive for external users
                if (!this.logins.isInternalUserLoggedIn() && (f.folderType === MailFolderType.SPAM || f.folderType === MailFolderType.ARCHIVE)) {
                    return false;
                }
                else if (this.logins.isEnabled(FeatureType.InternalCommunication) && f.folderType === MailFolderType.SPAM) {
                    return false;
                }
                else {
                    return true;
                }
            });
        });
    }
    /**
     * Get the list of MailboxDetails that this user has access to from their memberships.
     *
     * Will wait for successful initialization.
     */
    async getMailboxDetails() {
        // If details are there, use them
        if (this.mailboxDetails()) {
            return this.mailboxDetails();
        }
        else {
            // If they are not there, trigger loading again (just in case) but do not fail and wait until we actually have the details.
            // This is so that the rest of the app is not in the broken state if details fail to load but is just waiting until the success.
            return new Promise((resolve) => {
                this.init();
                const end = this.mailboxDetails.map((details) => {
                    resolve(details);
                    end.end(true);
                });
            });
        }
    }
    getMailboxDetailsForMail(mail) {
        return this.getMailboxDetailsForMailListId(mail._id[0]);
    }
    async getMailboxDetailsForMailListId(mailListId) {
        const mailboxDetails = await this.getMailboxDetails();
        const detail = mailboxDetails.find((md) => md.folders.getFolderByMailListId(mailListId)) ?? null;
        if (detail == null) {
            console.warn("Mailbox detail for mail list does not exist", mailListId);
        }
        return detail;
    }
    async getMailboxDetailsForMailGroup(mailGroupId) {
        const mailboxDetails = await this.getMailboxDetails();
        return assertNotNull(mailboxDetails.find((md) => mailGroupId === md.mailGroup._id), "No mailbox details for mail group");
    }
    async getUserMailboxDetails() {
        const userMailGroupMembership = this.logins.getUserController().getUserMailGroupMembership();
        const mailboxDetails = await this.getMailboxDetails();
        return assertNotNull(mailboxDetails.find((md) => md.mailGroup._id === userMailGroupMembership.group));
    }
    getMailboxFolders(mail) {
        return this.getMailboxDetailsForMail(mail).then((md) => md && md.folders);
    }
    getMailFolder(mailListId) {
        const mailboxDetails = this.mailboxDetails() || [];
        for (let detail of mailboxDetails) {
            const f = detail.folders.getFolderByMailListId(mailListId);
            if (f) {
                return f;
            }
        }
        return null;
    }
    /**
     * Sends the given folder and all its descendants to the spam folder, reporting mails (if applicable) and removes any empty folders
     */
    async sendFolderToSpam(folder) {
        const mailboxDetail = await this.getMailboxDetailsForMailListId(folder.mails);
        if (mailboxDetail == null) {
            return;
        }
        let deletedFolder = await this.removeAllEmpty(mailboxDetail, folder);
        if (!deletedFolder) {
            return this.mailFacade.updateMailFolderParent(folder, assertSystemFolderOfType(mailboxDetail.folders, MailFolderType.SPAM)._id);
        }
    }
    async reportMails(reportType, mails) {
        for (const mail of mails) {
            await this.mailFacade.reportMail(mail, reportType).catch(ofClass(NotFoundError, (e) => console.log("mail to be reported not found", e)));
        }
    }
    /**
     * Finally deletes all given mails. Caller must ensure that mails are only from one folder
     */
    async _moveMails(mails, targetMailFolder) {
        let moveMails = mails.filter((m) => m._id[0] !== targetMailFolder.mails && targetMailFolder._ownerGroup === m._ownerGroup); // prevent moving mails between mail boxes.
        // Do not move if target is the same as the current mailFolder
        const sourceMailFolder = this.getMailFolder(getListId(mails[0]));
        if (moveMails.length > 0 && sourceMailFolder && !isSameId(targetMailFolder._id, sourceMailFolder._id)) {
            const mailChunks = splitInChunks(MAX_NBR_MOVE_DELETE_MAIL_SERVICE, mails.map((m) => m._id));
            for (const mailChunk of mailChunks) {
                await this.mailFacade.moveMails(mailChunk, targetMailFolder._id);
            }
        }
    }
    /**
     * Preferably use moveMails() in MailGuiUtils.js which has built-in error handling
     * @throws PreconditionFailedError or LockedError if operation is locked on the server
     */
    async moveMails(mails, targetMailFolder) {
        const mailsPerFolder = groupBy(mails, (mail) => {
            return getListId(mail);
        });
        for (const [listId, mails] of mailsPerFolder) {
            const sourceMailFolder = this.getMailFolder(listId);
            if (sourceMailFolder) {
                await this._moveMails(mails, targetMailFolder);
            }
            else {
                console.log("Move mail: no mail folder for list id", listId);
            }
        }
    }
    isMovingMailsAllowed() {
        return this.logins.getUserController().isInternalUser();
    }
    isExportingMailsAllowed() {
        return !this.logins.isEnabled(FeatureType.DisableMailExport);
    }
    async markMails(mails, unread) {
        await promiseMap(mails, async (mail) => {
            if (mail.unread !== unread) {
                mail.unread = unread;
                return this.entityClient.update(mail).catch(ofClass(NotFoundError, noOp)).catch(ofClass(LockedError, noOp));
            }
        }, { concurrency: 5 });
    }
    /**
     * Finally deletes the given mails if they are already in the trash or spam folders,
     * otherwise moves them to the trash folder.
     * A deletion confirmation must have been show before.
     */
    async deleteMails(mails) {
        const mailsPerFolder = groupBy(mails, (mail) => {
            return getListId(mail);
        });
        if (mails.length === 0) {
            return;
        }
        const folders = await this.getMailboxFolders(mails[0]);
        if (folders == null) {
            return;
        }
        const trashFolder = assertNotNull(folders.getSystemFolderByType(MailFolderType.TRASH));
        for (const [listId, mails] of mailsPerFolder) {
            const sourceMailFolder = this.getMailFolder(listId);
            if (sourceMailFolder) {
                if (isSpamOrTrashFolder(folders, sourceMailFolder)) {
                    await this._finallyDeleteMails(mails);
                }
                else {
                    await this._moveMails(mails, trashFolder);
                }
            }
            else {
                console.log("Delete mail: no mail folder for list id", listId);
            }
        }
    }
    /**
     * Finally deletes all given mails. Caller must ensure that mails are only from one folder and the folder must allow final delete operation.
     */
    async _finallyDeleteMails(mails) {
        if (!mails.length)
            return Promise.resolve();
        const mailFolder = neverNull(this.getMailFolder(getListId(mails[0])));
        const mailIds = mails.map((m) => m._id);
        const mailChunks = splitInChunks(MAX_NBR_MOVE_DELETE_MAIL_SERVICE, mailIds);
        for (const mailChunk of mailChunks) {
            await this.mailFacade.deleteMails(mailChunk, mailFolder._id);
        }
    }
    async entityEventsReceived(updates, eventOwnerGroupId) {
        for (const update of updates) {
            if (isUpdateForTypeRef(MailFolderTypeRef, update)) {
                await this._init();
                m.redraw();
            }
            else if (isUpdateForTypeRef(GroupInfoTypeRef, update)) {
                if (update.operation === "1" /* OperationType.UPDATE */) {
                    await this._init();
                    m.redraw;
                }
            }
            else if (this.logins.getUserController().isUpdateForLoggedInUserInstance(update, eventOwnerGroupId)) {
                let newMemberships = this.logins.getUserController().getMailGroupMemberships();
                const mailboxDetails = await this.getMailboxDetails();
                if (newMemberships.length !== mailboxDetails.length) {
                    await this._init();
                    m.redraw();
                }
            }
            else if (isUpdateForTypeRef(MailTypeRef, update) && update.operation === "0" /* OperationType.CREATE */) {
                const folder = this.getMailFolder(update.instanceListId);
                if (folder && folder.folderType === MailFolderType.INBOX && !containsEventOfType(updates, "2" /* OperationType.DELETE */, update.instanceId)) {
                    // If we don't find another delete operation on this email in the batch, then it should be a create operation,
                    // otherwise it's a move
                    const mailId = [update.instanceListId, update.instanceId];
                    const mail = await this.entityClient.load(MailTypeRef, mailId);
                    await this.getMailboxDetailsForMailListId(update.instanceListId)
                        .then((mailboxDetail) => {
                        // We only apply rules on server if we are the leader in case of incoming messages
                        return mailboxDetail && this.inboxRuleHandler.findAndApplyMatchingRule(mailboxDetail, mail, this.connectivityModel.isLeader());
                    })
                        .then((newId) => this._showNotification(newId || mailId))
                        .catch(noOp);
                }
            }
        }
    }
    _mailboxCountersUpdates(counters) {
        const normalized = this.mailboxCounters() || {};
        const group = normalized[counters.mailGroup] || {};
        for (const value of counters.counterValues) {
            group[value.mailListId] = Number(value.count) || 0;
        }
        normalized[counters.mailGroup] = group;
        this.mailboxCounters(normalized);
    }
    _showNotification(mailId) {
        this.notifications.showNotification(lang.get("newMails_msg"), {
            actions: [],
        }, (_) => {
            m.route.set(`/mail/${listIdPart(mailId)}/${elementIdPart(mailId)}`);
            window.focus();
        });
    }
    getCounterValue(listId) {
        return this.getMailboxDetailsForMailListId(listId)
            .then((mailboxDetails) => {
            if (mailboxDetails == null) {
                return null;
            }
            else {
                const counters = this.mailboxCounters();
                const mailGroupCounter = counters[mailboxDetails.mailGroup._id];
                return mailGroupCounter && mailGroupCounter[listId];
            }
        })
            .catch(() => null);
    }
    checkMailForPhishing(mail, links) {
        return this.mailFacade.checkMailForPhishing(mail, links);
    }
    /**
     * Sends the given folder and all its descendants to the trash folder, removes any empty folders
     */
    async trashFolderAndSubfolders(folder) {
        const mailboxDetail = await this.getMailboxDetailsForMailListId(folder.mails);
        if (mailboxDetail == null) {
            return;
        }
        let deletedFolder = await this.removeAllEmpty(mailboxDetail, folder);
        if (!deletedFolder) {
            const trash = assertSystemFolderOfType(mailboxDetail.folders, MailFolderType.TRASH);
            return this.mailFacade.updateMailFolderParent(folder, trash._id);
        }
    }
    /**
     * This is called when moving a folder to SPAM or TRASH, which do not allow empty folders (since only folders that contain mail are allowed)
     */
    async removeAllEmpty(mailboxDetail, folder) {
        // sort descendants deepest first so that we can clean them up before checking their ancestors
        const descendants = mailboxDetail.folders.getDescendantFoldersOfParent(folder._id).sort((l, r) => r.level - l.level);
        // we completely delete empty folders
        let someNonEmpty = false;
        // we don't update folder system quickly enough so we keep track of deleted folders here and consider them "empty" when all their children are here
        const deleted = new Set();
        for (const descendant of descendants) {
            // Only load one mail, if there is even one we won't remove
            if ((await this.entityClient.loadRange(MailTypeRef, descendant.folder.mails, GENERATED_MAX_ID, 1, true)).length === 0 &&
                mailboxDetail.folders.getCustomFoldersOfParent(descendant.folder._id).every((f) => deleted.has(getElementId(f)))) {
                deleted.add(getElementId(descendant.folder));
                await this.finallyDeleteCustomMailFolder(descendant.folder);
            }
            else {
                someNonEmpty = true;
            }
        }
        // Only load one mail, if there is even one we won't remove
        if ((await this.entityClient.loadRange(MailTypeRef, folder.mails, GENERATED_MAX_ID, 1, true)).length === 0 &&
            mailboxDetail.folders.getCustomFoldersOfParent(folder._id).every((f) => deleted.has(getElementId(f))) &&
            !someNonEmpty) {
            await this.finallyDeleteCustomMailFolder(folder);
            return true;
        }
        else {
            return false;
        }
    }
    async finallyDeleteCustomMailFolder(folder) {
        if (folder.folderType !== MailFolderType.CUSTOM) {
            throw new ProgrammingError("Cannot delete non-custom folder: " + String(folder._id));
        }
        return await this.mailFacade
            .deleteFolder(folder._id)
            .catch(ofClass(NotFoundError, () => console.log("mail folder already deleted")))
            .catch(ofClass(PreconditionFailedError, () => {
            throw new UserError("operationStillActive_msg");
        }));
    }
    async fixupCounterForMailList(listId, unreadMails) {
        const mailboxDetails = await this.getMailboxDetailsForMailListId(listId);
        mailboxDetails && (await this.mailFacade.fixupCounterForMailList(mailboxDetails.mailGroup._id, listId, unreadMails));
    }
    async clearFolder(folder) {
        await this.mailFacade.clearFolder(folder._id);
    }
    async unsubscribe(mail, recipient, headers) {
        await this.mailFacade.unsubscribe(mail._id, recipient, headers);
    }
    async getMailboxProperties(mailboxGroupRoot) {
        // MailboxProperties is an encrypted instance that is created lazily. When we create it the reference is automatically written to the MailboxGroupRoot.
        // Unfortunately we will only get updated new MailboxGroupRoot with the next EntityUpdate.
        // To prevent parallel creation attempts we do two things:
        //  - we save the loading promise to avoid calling setup() twice in parallel
        //  - we set mailboxProperties reference manually (we could save the id elsewhere but it's easier this way)
        // If we are already loading/creating, just return it to avoid races
        const existingPromise = this.mailboxPropertiesPromises.get(mailboxGroupRoot._id);
        if (existingPromise) {
            return existingPromise;
        }
        const promise = this.loadOrCreateMailboxProperties(mailboxGroupRoot);
        this.mailboxPropertiesPromises.set(mailboxGroupRoot._id, promise);
        return promise.finally(() => this.mailboxPropertiesPromises.delete(mailboxGroupRoot._id));
    }
    async loadOrCreateMailboxProperties(mailboxGroupRoot) {
        if (!mailboxGroupRoot.mailboxProperties) {
            mailboxGroupRoot.mailboxProperties = await this.entityClient
                .setup(null, createMailboxProperties({
                _ownerGroup: mailboxGroupRoot._ownerGroup ?? "",
                reportMovedMails: "0",
                mailAddressProperties: [],
            }))
                .catch(ofClass(PreconditionFailedError, (e) => {
                // We try to prevent race conditions but they can still happen with multiple clients trying ot create mailboxProperties at the same time.
                // We send special precondition from the server with an existing id.
                if (e.data && e.data.startsWith("exists:")) {
                    const existingId = e.data.substring("exists:".length);
                    console.log("mailboxProperties already exists", existingId);
                    return existingId;
                }
                else {
                    throw new ProgrammingError(`Could not create mailboxProperties, precondition: ${e.data}`);
                }
            }));
        }
        const mailboxProperties = await this.entityClient.load(MailboxPropertiesTypeRef, mailboxGroupRoot.mailboxProperties);
        if (mailboxProperties.mailAddressProperties.length === 0) {
            await this.migrateFromOldSenderName(mailboxGroupRoot, mailboxProperties);
        }
        return mailboxProperties;
    }
    /** If there was no sender name configured before take the user's name and assign it to all email addresses. */
    async migrateFromOldSenderName(mailboxGroupRoot, mailboxProperties) {
        const userGroupInfo = this.logins.getUserController().userGroupInfo;
        const legacySenderName = userGroupInfo.name;
        const mailboxDetails = await this.getMailboxDetailsForMailGroup(mailboxGroupRoot._id);
        const mailAddresses = getEnabledMailAddressesWithUser(mailboxDetails, userGroupInfo);
        for (const mailAddress of mailAddresses) {
            mailboxProperties.mailAddressProperties.push(createMailAddressProperties({
                mailAddress,
                senderName: legacySenderName,
            }));
        }
        await this.entityClient.update(mailboxProperties);
    }
    async saveReportMovedMails(mailboxGroupRoot, reportMovedMails) {
        const mailboxProperties = await this.loadOrCreateMailboxProperties(mailboxGroupRoot);
        mailboxProperties.reportMovedMails = reportMovedMails;
        await this.entityClient.update(mailboxProperties);
        return mailboxProperties;
    }
}
//# sourceMappingURL=MailModel.js.map