import m from "mithril";
import { locator } from "../../api/main/MainLocator.js";
import { SidebarSection } from "../../gui/SidebarSection.js";
import { IconButton } from "../../gui/base/IconButton.js";
import { getElementId } from "../../api/common/utils/EntityUtils.js";
import { isSelectedPrefix } from "../../gui/base/NavButton.js";
import { getFolderIcon, getFolderName, MAX_FOLDER_INDENT_LEVEL } from "../model/MailUtils.js";
import { MAIL_PREFIX } from "../../misc/RouteChange.js";
import { MailFolderRow } from "./MailFolderRow.js";
import { last, noOp } from "@tutao/tutanota-utils";
import { attachDropdown } from "../../gui/base/Dropdown.js";
import { MailFolderType } from "../../api/common/TutanotaConstants.js";
import { isSpamOrTrashFolder } from "../../api/common/mail/CommonMailUtils.js";
import { px, size } from "../../gui/size.js";
import { RowButton } from "../../gui/base/buttons/RowButton.js";
/** Displays a tree of all folders. */
export class MailFoldersView {
    constructor() {
        // Contains the id of the visible row
        this.visibleRow = null;
    }
    view({ attrs }) {
        const { mailboxDetail } = attrs;
        const groupCounters = locator.mailModel.mailboxCounters()[mailboxDetail.mailGroup._id] || {};
        // Important: this array is keyed so each item must have a key and `null` cannot be in the array
        // So instead we push or not push into array
        const customSystems = mailboxDetail.folders.customSubtrees;
        const systemSystems = mailboxDetail.folders.systemSubtrees;
        const children = [];
        const selectedFolder = mailboxDetail.folders
            .getIndentedList()
            .map((f) => f.folder)
            .find((f) => isSelectedPrefix(MAIL_PREFIX + "/" + f.mails));
        const path = selectedFolder ? mailboxDetail.folders.getPathToFolder(selectedFolder._id) : [];
        const isInternalUser = locator.logins.isInternalUserLoggedIn();
        const systemChildren = this.renderFolderTree(systemSystems, groupCounters, attrs, path, isInternalUser);
        if (systemChildren) {
            children.push(...systemChildren.children);
        }
        if (isInternalUser) {
            children.push(m(SidebarSection, {
                name: "yourFolders_action",
                button: attrs.inEditMode ? this.renderCreateFolderAddButton(null, attrs) : this.renderEditFoldersButton(attrs),
                key: "yourFolders", // we need to set a key because folder rows also have a key.
            }, this.renderFolderTree(customSystems, groupCounters, attrs, path, isInternalUser).children));
            children.push(this.renderAddFolderButtonRow(attrs));
        }
        return children;
    }
    renderFolderTree(subSystems, groupCounters, attrs, path, isInternalUser, indentationLevel = 0) {
        // we need to keep track of how many rows we've drawn so far for this subtree so that we can draw hierarchy lines correctly
        const result = { children: [], numRows: 0 };
        for (let system of subSystems) {
            const id = getElementId(system.folder);
            const button = {
                label: () => getFolderName(system.folder),
                href: () => {
                    if (attrs.inEditMode) {
                        return m.route.get();
                    }
                    else {
                        const mailId = attrs.mailListToSelectedMail.get(system.folder.mails);
                        if (mailId) {
                            return `/mail/${system.folder.mails}/${mailId}`;
                        }
                        else {
                            return `/mail/${system.folder.mails}`;
                        }
                    }
                },
                isSelectedPrefix: attrs.inEditMode ? false : MAIL_PREFIX + "/" + system.folder.mails,
                colors: "nav" /* NavButtonColor.Nav */,
                click: () => attrs.onFolderClick(system.folder),
                dropHandler: (droppedMailId) => attrs.onFolderDrop(droppedMailId, system.folder),
                disableHoverBackground: true,
                disabled: attrs.inEditMode,
            };
            const currentExpansionState = attrs.inEditMode ? true : attrs.expandedFolders.has(getElementId(system.folder)) ?? false; //default is false
            const hasChildren = system.children.length > 0;
            const summedCount = !currentExpansionState && hasChildren ? this.getTotalFolderCounter(groupCounters, system) : groupCounters[system.folder.mails];
            const childResult = hasChildren && currentExpansionState
                ? this.renderFolderTree(system.children, groupCounters, attrs, path, isInternalUser, indentationLevel + 1)
                : { children: null, numRows: 0 };
            const isTrashOrSpam = system.folder.folderType === MailFolderType.TRASH || system.folder.folderType === MailFolderType.SPAM;
            const isRightButtonVisible = this.visibleRow === id;
            const rightButton = isInternalUser && !isTrashOrSpam && (isRightButtonVisible || attrs.inEditMode)
                ? this.createFolderMoreButton(system.folder, attrs, () => {
                    this.visibleRow = null;
                })
                : null;
            const render = m.fragment({
                key: id,
            }, [
                m(MailFolderRow, {
                    count: attrs.inEditMode ? 0 : summedCount,
                    button,
                    icon: getFolderIcon(system.folder),
                    rightButton,
                    expanded: hasChildren ? currentExpansionState : null,
                    indentationLevel: Math.min(indentationLevel, MAX_FOLDER_INDENT_LEVEL),
                    onExpanderClick: hasChildren ? () => attrs.onFolderExpanded(system.folder, currentExpansionState) : noOp,
                    hasChildren,
                    onSelectedPath: path.includes(system.folder),
                    numberOfPreviousRows: result.numRows,
                    isLastSibling: last(subSystems) === system,
                    editMode: attrs.inEditMode,
                    onHover: () => {
                        this.visibleRow = id;
                    },
                }),
                childResult.children,
            ]);
            result.numRows += childResult.numRows + 1;
            result.children.push(render);
        }
        return result;
    }
    renderAddFolderButtonRow(attrs) {
        // This button needs to fill the whole row, but is not a navigation button (so IconButton or NavButton weren't appropriate)
        return m(RowButton, {
            label: "addFolder_action",
            key: "addFolder",
            icon: "Add" /* Icons.Add */,
            class: "folder-row mlr-button border-radius-small",
            style: {
                width: `calc(100% - ${px(size.hpad_button * 2)})`,
            },
            onclick: () => {
                attrs.onShowFolderAddEditDialog(attrs.mailboxDetail.mailGroup._id, null, null);
            },
        });
    }
    getTotalFolderCounter(counters, system) {
        return (counters[system.folder.mails] ?? 0) + system.children.reduce((acc, child) => acc + this.getTotalFolderCounter(counters, child), 0);
    }
    createFolderMoreButton(folder, attrs, onClose) {
        return attachDropdown({
            mainButtonAttrs: {
                title: "more_label",
                icon: "More" /* Icons.More */,
                colors: "nav" /* ButtonColor.Nav */,
                size: 1 /* ButtonSize.Compact */,
            },
            childAttrs: () => {
                return folder.folderType === MailFolderType.CUSTOM
                    ? // cannot add new folder to custom folder in spam or trash folder
                        isSpamOrTrashFolder(attrs.mailboxDetail.folders, folder)
                            ? [this.editButtonAttrs(attrs, folder), this.deleteButtonAttrs(attrs, folder)]
                            : [this.editButtonAttrs(attrs, folder), this.addButtonAttrs(attrs, folder), this.deleteButtonAttrs(attrs, folder)]
                    : [this.addButtonAttrs(attrs, folder)];
            },
            onClose,
        });
    }
    deleteButtonAttrs(attrs, folder) {
        return {
            label: "delete_action",
            icon: "Trash" /* Icons.Trash */,
            click: () => {
                attrs.onDeleteCustomMailFolder(folder);
            },
        };
    }
    addButtonAttrs(attrs, folder) {
        return {
            label: "addFolder_action",
            icon: "Add" /* Icons.Add */,
            click: () => {
                attrs.onShowFolderAddEditDialog(attrs.mailboxDetail.mailGroup._id, null, folder);
            },
        };
    }
    editButtonAttrs(attrs, folder) {
        return {
            label: "edit_action",
            icon: "Edit" /* Icons.Edit */,
            click: () => {
                attrs.onShowFolderAddEditDialog(attrs.mailboxDetail.mailGroup._id, folder, folder.parentFolder ? attrs.mailboxDetail.folders.getFolderById(folder.parentFolder) : null);
            },
        };
    }
    renderCreateFolderAddButton(parentFolder, attrs) {
        return m(IconButton, {
            title: "addFolder_action",
            click: () => {
                return attrs.onShowFolderAddEditDialog(attrs.mailboxDetail.mailGroup._id, null, parentFolder);
            },
            icon: "Add" /* Icons.Add */,
            size: 1 /* ButtonSize.Compact */,
        });
    }
    renderEditFoldersButton(attrs) {
        return m(IconButton, {
            title: "edit_action",
            click: () => attrs.onEditMailbox(),
            icon: "Edit" /* Icons.Edit */,
            size: 1 /* ButtonSize.Compact */,
        });
    }
}
//# sourceMappingURL=MailFoldersView.js.map