import { createMail } from "../../api/entities/tutanota/TypeRefs.js";
import { LockedError, PreconditionFailedError } from "../../api/common/error/RestError";
import { Dialog } from "../../gui/base/Dialog";
import { locator } from "../../api/main/MainLocator";
import { getFolderIcon, getFolderName, getIndentedFolderNameForDropdown, getMoveTargetFolderSystems } from "../model/MailUtils";
import { isApp, isDesktop } from "../../api/common/Env";
import { assertNotNull, neverNull, noOp, promiseMap } from "@tutao/tutanota-utils";
import { MailFolderType } from "../../api/common/TutanotaConstants";
import { getElementId, getListId } from "../../api/common/utils/EntityUtils";
import { reportMailsAutomatically } from "./MailReportDialog";
import { lang } from "../../misc/LanguageViewModel";
import { DomRectReadOnlyPolyfilled, Dropdown } from "../../gui/base/Dropdown.js";
import { modal } from "../../gui/base/Modal.js";
import { assertSystemFolderOfType, isOfTypeOrSubfolderOf, isSpamOrTrashFolder } from "../../api/common/mail/CommonMailUtils.js";
import { size } from "../../gui/size.js";
import { PinchZoom } from "../../gui/PinchZoom.js";
export async function showDeleteConfirmationDialog(mails) {
    let trashMails = [];
    let moveMails = [];
    for (let mail of mails) {
        const folder = locator.mailModel.getMailFolder(mail._id[0]);
        const mailboxDetail = await locator.mailModel.getMailboxDetailsForMailListId(getListId(mail));
        if (mailboxDetail == null) {
            continue;
        }
        const isFinalDelete = folder && isSpamOrTrashFolder(mailboxDetail.folders, folder);
        isFinalDelete ? trashMails.push(mail) : moveMails.push(mail);
    }
    let confirmationTextId = null;
    if (trashMails.length > 0) {
        if (moveMails.length > 0) {
            confirmationTextId = "finallyDeleteSelectedEmails_msg";
        }
        else {
            confirmationTextId = "finallyDeleteEmails_msg";
        }
    }
    if (confirmationTextId != null) {
        return Dialog.confirm(confirmationTextId, "ok_action");
    }
    else {
        return Promise.resolve(true);
    }
}
/**
 * @return whether emails were deleted
 */
export function promptAndDeleteMails(mailModel, mails, onConfirm) {
    return showDeleteConfirmationDialog(mails).then((confirmed) => {
        if (confirmed) {
            onConfirm();
            return mailModel
                .deleteMails(mails)
                .then(() => true)
                .catch((e) => {
                //LockedError should no longer be thrown!?!
                if (e instanceof PreconditionFailedError || e instanceof LockedError) {
                    return Dialog.message("operationStillActive_msg").then(() => false);
                }
                else {
                    throw e;
                }
            });
        }
        else {
            return Promise.resolve(false);
        }
    });
}
/**
 * Moves the mails and reports them as spam if the user or settings allow it.
 * @return whether mails were actually moved
 */
export async function moveMails({ mailModel, mails, targetMailFolder, isReportable = true }) {
    const details = await mailModel.getMailboxDetailsForMailListId(targetMailFolder.mails);
    if (details == null) {
        return false;
    }
    const system = details.folders;
    return mailModel
        .moveMails(mails, targetMailFolder)
        .then(async () => {
        if (isOfTypeOrSubfolderOf(system, targetMailFolder, MailFolderType.SPAM) && isReportable) {
            const reportableMails = mails.map((mail) => {
                // mails have just been moved
                const reportableMail = createMail(mail);
                reportableMail._id = [targetMailFolder.mails, getElementId(mail)];
                return reportableMail;
            });
            const mailboxDetails = await mailModel.getMailboxDetailsForMailGroup(assertNotNull(targetMailFolder._ownerGroup));
            await reportMailsAutomatically("1" /* MailReportType.SPAM */, mailModel, mailboxDetails, reportableMails);
        }
        return true;
    })
        .catch((e) => {
        //LockedError should no longer be thrown!?!
        if (e instanceof LockedError || e instanceof PreconditionFailedError) {
            return Dialog.message("operationStillActive_msg").then(() => false);
        }
        else {
            throw e;
        }
    });
}
export function archiveMails(mails) {
    if (mails.length > 0) {
        // assume all mails in the array belong to the same Mailbox
        return locator.mailModel.getMailboxFolders(mails[0]).then((folders) => {
            folders &&
                moveMails({
                    mailModel: locator.mailModel,
                    mails: mails,
                    targetMailFolder: assertSystemFolderOfType(folders, MailFolderType.ARCHIVE),
                });
        });
    }
    else {
        return Promise.resolve();
    }
}
export function moveToInbox(mails) {
    if (mails.length > 0) {
        // assume all mails in the array belong to the same Mailbox
        return locator.mailModel.getMailboxFolders(mails[0]).then((folders) => {
            folders &&
                moveMails({
                    mailModel: locator.mailModel,
                    mails: mails,
                    targetMailFolder: assertSystemFolderOfType(folders, MailFolderType.INBOX),
                });
        });
    }
    else {
        return Promise.resolve();
    }
}
export function getMailFolderIcon(mail) {
    let folder = locator.mailModel.getMailFolder(mail._id[0]);
    if (folder) {
        return getFolderIcon(folder);
    }
    else {
        return "Folder" /* Icons.Folder */;
    }
}
export function replaceCidsWithInlineImages(dom, inlineImages, onContext) {
    // all image tags which have cid attribute. The cid attribute has been set by the sanitizer for adding a default image.
    const imageElements = Array.from(dom.querySelectorAll("img[cid]"));
    if (dom.shadowRoot) {
        const shadowImageElements = Array.from(dom.shadowRoot.querySelectorAll("img[cid]"));
        imageElements.push(...shadowImageElements);
    }
    const elementsWithCid = [];
    for (const imageElement of imageElements) {
        const cid = imageElement.getAttribute("cid");
        if (cid) {
            const inlineImage = inlineImages.get(cid);
            if (inlineImage) {
                elementsWithCid.push(imageElement);
                imageElement.setAttribute("src", inlineImage.objectUrl);
                imageElement.classList.remove("tutanota-placeholder");
                if (isApp()) {
                    // Add long press action for apps
                    let timeoutId;
                    let startCoords;
                    imageElement.addEventListener("touchstart", (e) => {
                        const touch = e.touches[0];
                        if (!touch)
                            return;
                        startCoords = {
                            x: touch.clientX,
                            y: touch.clientY,
                        };
                        timeoutId && clearTimeout(timeoutId);
                        timeoutId = setTimeout(() => {
                            onContext(inlineImage.cid, e, imageElement);
                        }, 800);
                    });
                    imageElement.addEventListener("touchmove", (e) => {
                        const touch = e.touches[0];
                        if (!touch || !startCoords || !timeoutId)
                            return;
                        if (Math.abs(touch.clientX - startCoords.x) > PinchZoom.DRAG_THRESHOLD ||
                            Math.abs(touch.clientY - startCoords.y) > PinchZoom.DRAG_THRESHOLD) {
                            clearTimeout(timeoutId);
                            timeoutId = null;
                        }
                    });
                    imageElement.addEventListener("touchend", () => {
                        timeoutId && clearTimeout(timeoutId);
                        timeoutId = null;
                    });
                }
                if (isDesktop()) {
                    // add right click action for desktop apps
                    imageElement.addEventListener("contextmenu", (e) => {
                        onContext(inlineImage.cid, e, imageElement);
                        e.preventDefault();
                    });
                }
            }
        }
    }
    return elementsWithCid;
}
export function replaceInlineImagesWithCids(dom) {
    const domClone = dom.cloneNode(true);
    const inlineImages = Array.from(domClone.querySelectorAll("img[cid]"));
    for (const inlineImage of inlineImages) {
        const cid = inlineImage.getAttribute("cid");
        inlineImage.setAttribute("src", "cid:" + (cid || ""));
        inlineImage.removeAttribute("cid");
    }
    return domClone;
}
export function createInlineImage(file) {
    const cid = Math.random().toString(30).substring(2);
    file.cid = cid;
    return createInlineImageReference(file, cid);
}
function createInlineImageReference(file, cid) {
    const blob = new Blob([file.data], {
        type: file.mimeType,
    });
    const objectUrl = URL.createObjectURL(blob);
    return {
        cid,
        objectUrl,
        blob,
    };
}
export function cloneInlineImages(inlineImages) {
    const newMap = new Map();
    for (const [k, v] of inlineImages.entries()) {
        const blob = new Blob([v.blob]);
        const objectUrl = URL.createObjectURL(blob);
        newMap.set(k, {
            cid: v.cid,
            objectUrl,
            blob,
        });
    }
    return newMap;
}
export function revokeInlineImages(inlineImages) {
    for (const v of inlineImages.values()) {
        URL.revokeObjectURL(v.objectUrl);
    }
}
export async function loadInlineImages(fileController, attachments, referencedCids) {
    const filesToLoad = getReferencedAttachments(attachments, referencedCids);
    const inlineImages = new Map();
    return promiseMap(filesToLoad, async (file) => {
        let dataFile = await fileController.getAsDataFile(file);
        const { htmlSanitizer } = await import("../../misc/HtmlSanitizer");
        dataFile = htmlSanitizer.sanitizeInlineAttachment(dataFile);
        const inlineImageReference = createInlineImageReference(dataFile, neverNull(file.cid));
        inlineImages.set(inlineImageReference.cid, inlineImageReference);
    }).then(() => inlineImages);
}
export function getReferencedAttachments(attachments, referencedCids) {
    return attachments.filter((file) => referencedCids.find((rcid) => file.cid === rcid));
}
export async function showMoveMailsDropdown(model, origin, mails, opts) {
    const { width = 300, withBackground = false, onSelected = noOp } = opts ?? {};
    const folders = await getMoveTargetFolderSystems(model, mails);
    if (folders.length === 0)
        return;
    const folderButtons = folders.map((f) => ({
        // We need to pass in the raw folder name to avoid including it in searches
        label: () => lang.get("folderDepth_label", { "{folderName}": getFolderName(f.folder), "{depth}": f.level }),
        text: () => getIndentedFolderNameForDropdown(f),
        click: () => {
            onSelected();
            moveMails({ mailModel: model, mails: mails, targetMailFolder: f.folder });
        },
        icon: getFolderIcon(f.folder),
    }));
    const dropdown = new Dropdown(() => folderButtons, width);
    dropdown.setOrigin(new DomRectReadOnlyPolyfilled(origin.left, origin.top, origin.width, origin.height));
    modal.displayUnique(dropdown, withBackground);
}
export function getConversationTitle(conversationViewModel) {
    if (!conversationViewModel.isFinished()) {
        return lang.get("loading_msg");
    }
    const numberOfEmails = conversationViewModel.conversationItems().length;
    if (numberOfEmails === 1) {
        return lang.get("oneEmail_label");
    }
    else {
        return lang.get("nbrOrEmails_label", { "{number}": numberOfEmails });
    }
}
export function getMoveMailBounds() {
    // just putting the move mail dropdown in the left side of the viewport with a bit of margin
    return new DomRectReadOnlyPolyfilled(size.hpad_large, size.vpad_large, 0, 0);
}
//# sourceMappingURL=MailGuiUtils.js.map