import m from "mithril";
import { ViewSlider } from "../../gui/nav/ViewSlider.js";
import { ViewColumn } from "../../gui/base/ViewColumn";
import { lang } from "../../misc/LanguageViewModel";
import { Dialog } from "../../gui/base/Dialog";
import { FeatureType, Keys, MailFolderType } from "../../api/common/TutanotaConstants";
import { Header } from "../../gui/Header.js";
import { noOp, ofClass } from "@tutao/tutanota-utils";
import { MailListView } from "./MailListView";
import { assertMainOrNode, isApp } from "../../api/common/Env";
import { keyManager } from "../../misc/KeyManager";
import { getMailSelectionMessage, MultiItemViewer } from "./MultiItemViewer.js";
import { showProgressDialog } from "../../gui/dialogs/ProgressDialog";
import { canDoDragAndDropExport, getFolderName, getMailboxName } from "../model/MailUtils";
import { locator } from "../../api/main/MainLocator";
import { PermissionError } from "../../api/common/error/PermissionError";
import { styles } from "../../gui/styles";
import { px, size } from "../../gui/size";
import { UserError } from "../../api/main/UserError";
import { showUserError } from "../../misc/ErrorHandlerImpl";
import { archiveMails, getConversationTitle, getMoveMailBounds, moveMails, moveToInbox, promptAndDeleteMails, showMoveMailsDropdown } from "./MailGuiUtils";
import { getElementId, isSameId } from "../../api/common/utils/EntityUtils";
import { isNewMailActionAvailable } from "../../gui/nav/NavFunctions";
import { CancelledError } from "../../api/common/error/CancelledError";
import { readLocalFiles } from "../../file/FileController.js";
import { BottomNav } from "../../gui/nav/BottomNav.js";
import { MobileMailActionBar } from "./MobileMailActionBar.js";
import { deviceConfig } from "../../misc/DeviceConfig.js";
import { BaseTopLevelView } from "../../gui/BaseTopLevelView.js";
import { showEditFolderDialog } from "./EditFolderDialog.js";
import { MailFoldersView } from "./MailFoldersView.js";
import { isSpamOrTrashFolder } from "../../api/common/mail/CommonMailUtils.js";
import { FolderColumnView } from "../../gui/FolderColumnView.js";
import { SidebarSection } from "../../gui/SidebarSection.js";
import { EditFoldersDialog } from "./EditFoldersDialog.js";
import { conversationCardMargin, ConversationViewer } from "./ConversationViewer.js";
import { IconButton } from "../../gui/base/IconButton.js";
import { BackgroundColumnLayout } from "../../gui/BackgroundColumnLayout.js";
import { MailViewerActions } from "./MailViewerToolbar.js";
import { theme } from "../../gui/theme.js";
import { MobileMailMultiselectionActionBar } from "./MobileMailMultiselectionActionBar.js";
import { SelectAllCheckbox } from "../../gui/SelectAllCheckbox.js";
import { DesktopListToolbar, DesktopViewerToolbar } from "../../gui/DesktopToolbars.js";
import { MobileHeader } from "../../gui/MobileHeader.js";
import { LazySearchBar } from "../../misc/LazySearchBar.js";
import { MultiselectMobileHeader } from "../../gui/MultiselectMobileHeader.js";
import { selectionAttrsForList } from "../../misc/ListModel.js";
import { ListLoadingState } from "../../gui/base/List.js";
import { EnterMultiselectIconButton } from "../../gui/EnterMultiselectIconButton.js";
import { MailFilterButton } from "./MailFilterButton.js";
import { listSelectionKeyboardShortcuts } from "../../gui/base/ListUtils.js";
assertMainOrNode();
/**
 * Top-level view for displaying mailboxes.
 */
export class MailView extends BaseTopLevelView {
    get conversationViewModel() {
        return this.mailViewModel.getConversationViewModel();
    }
    constructor(vnode) {
        super();
        this.countersStream = null;
        this.desktopSystemFacade = vnode.attrs.desktopSystemFacade;
        this.expandedState = new Set(deviceConfig.getExpandedFolders(locator.logins.getUserController().userId));
        this.cache = vnode.attrs.cache;
        this.folderColumn = this.createFolderColumn(null, vnode.attrs.drawerAttrs);
        this.mailViewModel = vnode.attrs.mailViewModel;
        this.listColumn = new ViewColumn({
            view: () => {
                const listId = this.mailViewModel.getListId();
                return m(BackgroundColumnLayout, {
                    backgroundColor: theme.navigation_bg,
                    desktopToolbar: () => m(DesktopListToolbar, m(SelectAllCheckbox, selectionAttrsForList(this.mailViewModel.listModel)), this.renderFilterButton()),
                    columnLayout: listId
                        ? m("", {
                            style: {
                                marginBottom: px(conversationCardMargin),
                            },
                        }, m(MailListView, {
                            key: listId,
                            mailViewModel: this.mailViewModel,
                            listId: listId,
                            onSingleSelection: (mail) => {
                                if (!this.mailViewModel.listModel?.state.inMultiselect) {
                                    this.viewSlider.focus(this.mailColumn);
                                    // Make sure that we mark mail as read if you select the mail again, even if it was selected before.
                                    // Do it in the next even loop to not rely on what is called first, listModel or us. ListModel changes are
                                    // sync so this should be enough.
                                    Promise.resolve().then(() => {
                                        const conversationViewModel = this.mailViewModel.getConversationViewModel();
                                        if (conversationViewModel && isSameId(mail._id, conversationViewModel.primaryMail._id)) {
                                            conversationViewModel?.primaryViewModel().setUnread(false);
                                        }
                                    });
                                }
                            },
                            onClearFolder: async () => {
                                const folder = this.mailViewModel.getSelectedFolder();
                                if (folder == null) {
                                    console.warn("Cannot delete folder, no folder is selected");
                                    return;
                                }
                                const confirmed = await Dialog.confirm(() => lang.get("confirmDeleteFinallySystemFolder_msg", { "{1}": getFolderName(folder) }));
                                if (confirmed) {
                                    showProgressDialog("progressDeleting_msg", this.mailViewModel.finallyDeleteAllMailsInSelectedFolder(folder));
                                }
                            },
                        }))
                        : null,
                    mobileHeader: () => this.mailViewModel.listModel?.state.inMultiselect
                        ? m(MultiselectMobileHeader, {
                            ...selectionAttrsForList(this.mailViewModel.listModel),
                            message: getMailSelectionMessage(this.mailViewModel.listModel.getSelectedAsArray()),
                        })
                        : m(MobileHeader, {
                            ...vnode.attrs.header,
                            title: this.listColumn.getTitle(),
                            columnType: "first",
                            actions: [
                                this.renderFilterButton(),
                                m(EnterMultiselectIconButton, {
                                    clickAction: () => {
                                        this.mailViewModel.listModel?.enterMultiselect();
                                    },
                                }),
                            ],
                            primaryAction: () => this.renderHeaderRightView(),
                            backAction: () => this.viewSlider.focusPreviousColumn(),
                        }),
                });
            },
        }, 1 /* ColumnType.Background */, {
            minWidth: size.second_col_min_width,
            maxWidth: size.second_col_max_width,
            headerCenter: () => {
                const selectedFolder = this.mailViewModel.getSelectedFolder();
                return selectedFolder ? getFolderName(selectedFolder) : "";
            },
        });
        this.mailColumn = new ViewColumn({
            view: () => {
                const viewModel = this.conversationViewModel;
                if (viewModel) {
                    return this.renderSingleMailViewer(vnode.attrs.header, viewModel);
                }
                else {
                    return this.renderMultiMailViewer(vnode.attrs.header);
                }
            },
        }, 1 /* ColumnType.Background */, {
            minWidth: size.third_col_min_width,
            maxWidth: size.third_col_max_width,
            ariaLabel: () => lang.get("email_label"),
        });
        this.viewSlider = new ViewSlider([this.folderColumn, this.listColumn, this.mailColumn]);
        this.viewSlider.focusedColumn = this.listColumn;
        const shortcuts = this.getShortcuts();
        vnode.attrs.mailViewModel.init();
        this.oncreate = () => {
            this.countersStream = locator.mailModel.mailboxCounters.map(m.redraw);
            keyManager.registerShortcuts(shortcuts);
            this.cache.conversationViewPreference = deviceConfig.getConversationViewShowOnlySelectedMail();
        };
        this.onremove = () => {
            // cancel the loading if we are destroyed
            this.mailViewModel.listModel?.cancelLoadAll();
            this.countersStream?.end(true);
            this.countersStream = null;
            keyManager.unregisterShortcuts(shortcuts);
        };
    }
    renderFilterButton() {
        return m(MailFilterButton, { filter: this.mailViewModel.filterType, setFilter: (filter) => this.mailViewModel.setFilter(filter) });
    }
    mailViewerSingleActions(viewModel) {
        return m(MailViewerActions, {
            mailModel: viewModel.primaryViewModel().mailModel,
            mailViewerViewModel: viewModel.primaryViewModel(),
            mails: [viewModel.primaryMail],
        });
    }
    renderSingleMailViewer(header, viewModel) {
        return m(BackgroundColumnLayout, {
            backgroundColor: theme.navigation_bg,
            desktopToolbar: () => m(DesktopViewerToolbar, this.mailViewerSingleActions(viewModel)),
            mobileHeader: () => m(MobileHeader, {
                ...header,
                backAction: () => {
                    this.viewSlider.focusPreviousColumn();
                },
                columnType: "other",
                actions: null,
                multicolumnActions: () => this.mailViewerSingleActions(viewModel),
                primaryAction: () => this.renderHeaderRightView(),
                title: getConversationTitle(viewModel),
            }),
            columnLayout: m(ConversationViewer, {
                // Re-create the whole viewer and its vnode tree if email has changed
                key: getElementId(viewModel.primaryMail),
                viewModel: viewModel,
                // this assumes that the viewSlider focus animation is already started
                delayBodyRendering: this.viewSlider.waitForAnimation(),
            }),
        });
    }
    mailViewerMultiActions() {
        return m(MailViewerActions, {
            mailModel: locator.mailModel,
            mails: this.mailViewModel.listModel?.getSelectedAsArray() ?? [],
            selectNone: () => this.mailViewModel.listModel?.selectNone(),
        });
    }
    renderMultiMailViewer(header) {
        return m(BackgroundColumnLayout, {
            backgroundColor: theme.navigation_bg,
            desktopToolbar: () => m(DesktopViewerToolbar, this.mailViewerMultiActions()),
            mobileHeader: () => m(MobileHeader, {
                actions: this.mailViewerMultiActions(),
                primaryAction: () => this.renderHeaderRightView(),
                backAction: () => this.viewSlider.focusPreviousColumn(),
                ...header,
                columnType: "other",
            }),
            columnLayout: m(MultiItemViewer, {
                selectedEntities: this.mailViewModel.listModel?.getSelectedAsArray() ?? [],
                selectNone: () => {
                    this.mailViewModel.listModel?.selectNone();
                },
                loadAll: () => this.mailViewModel.listModel?.loadAll(),
                stopLoadAll: () => this.mailViewModel.listModel?.cancelLoadAll(),
                loadingAll: this.mailViewModel.listModel?.state.loadingAll
                    ? "loading"
                    : this.mailViewModel.listModel?.state.loadingStatus === ListLoadingState.Done
                        ? "loaded"
                        : "can_load",
                getSelectionMessage: (selected) => getMailSelectionMessage(selected),
            }),
        });
    }
    view({ attrs }) {
        return m("#mail.main-view", {
            ondragover: (ev) => {
                // do not check the data transfer here because it is not always filled, e.g. in Safari
                ev.stopPropagation();
                ev.preventDefault();
            },
            ondrop: (ev) => {
                if (isNewMailActionAvailable() && ev.dataTransfer?.files && ev.dataTransfer.files.length > 0) {
                    Promise.all([
                        this.mailViewModel.getMailboxDetails(),
                        readLocalFiles(ev.dataTransfer.files),
                        import("../signature/Signature"),
                        import("../editor/MailEditor"),
                    ])
                        .then(([mailbox, dataFiles, { appendEmailSignature }, { newMailEditorFromTemplate }]) => {
                        mailbox &&
                            newMailEditorFromTemplate(mailbox, {}, "", appendEmailSignature("", locator.logins.getUserController().props), dataFiles).then((dialog) => dialog.show());
                    })
                        .catch(ofClass(PermissionError, noOp))
                        .catch(ofClass(UserError, showUserError));
                }
                // prevent in any case because firefox tries to open
                // dataTransfer as a URL otherwise.
                ev.stopPropagation();
                ev.preventDefault();
            },
        }, m(this.viewSlider, {
            header: m(Header, {
                rightView: this.renderHeaderRightView(),
                searchBar: () => 
                // not showing search for external users
                locator.logins.isInternalUserLoggedIn()
                    ? m(LazySearchBar, {
                        placeholder: lang.get("searchEmails_placeholder"),
                        disabled: !locator.logins.isFullyLoggedIn(),
                    })
                    : null,
                ...attrs.header,
            }),
            bottomNav: styles.isSingleColumnLayout() && this.viewSlider.focusedColumn === this.mailColumn && this.conversationViewModel
                ? m(MobileMailActionBar, { viewModel: this.conversationViewModel.primaryViewModel() })
                : styles.isSingleColumnLayout() && this.mailViewModel.listModel?.state.inMultiselect
                    ? m(MobileMailMultiselectionActionBar, {
                        mails: this.mailViewModel.listModel.getSelectedAsArray(),
                        selectNone: () => this.mailViewModel.listModel?.selectNone(),
                        mailModel: locator.mailModel,
                    })
                    : m(BottomNav),
        }));
    }
    getViewSlider() {
        return this.viewSlider;
    }
    handleBackButton() {
        const listModel = this.mailViewModel.listModel;
        if (listModel && listModel.state.inMultiselect) {
            listModel.selectNone();
            return true;
        }
        return false;
    }
    renderHeaderRightView() {
        return isNewMailActionAvailable()
            ? [
                m(IconButton, {
                    title: "newMail_action",
                    click: () => this.showNewMailDialog().catch(ofClass(PermissionError, noOp)),
                    icon: "PencilSquare" /* Icons.PencilSquare */,
                }),
            ]
            : null;
    }
    getShortcuts() {
        return [
            ...listSelectionKeyboardShortcuts(1 /* MultiselectMode.Enabled */, () => this.mailViewModel.listModel ?? null),
            {
                key: Keys.N,
                exec: () => {
                    this.showNewMailDialog().catch(ofClass(PermissionError, noOp));
                },
                enabled: () => !!this.mailViewModel.getSelectedFolder() && isNewMailActionAvailable(),
                help: "newMail_action",
            },
            {
                key: Keys.DELETE,
                exec: () => {
                    this.mailViewModel.listModel && this.deleteMails(this.mailViewModel.listModel.getSelectedAsArray());
                },
                help: "deleteEmails_action",
            },
            {
                key: Keys.BACKSPACE,
                exec: () => {
                    this.mailViewModel.listModel && this.deleteMails(this.mailViewModel.listModel.getSelectedAsArray());
                },
                help: "deleteEmails_action",
            },
            {
                key: Keys.A,
                exec: () => {
                    this.mailViewModel.listModel && archiveMails(this.mailViewModel.listModel.getSelectedAsArray());
                    return true;
                },
                help: "archive_action",
                enabled: () => locator.logins.isInternalUserLoggedIn(),
            },
            {
                key: Keys.I,
                exec: () => {
                    this.mailViewModel.listModel && moveToInbox(this.mailViewModel.listModel.getSelectedAsArray());
                    return true;
                },
                help: "moveToInbox_action",
            },
            {
                key: Keys.V,
                exec: () => {
                    this.moveMails();
                    return true;
                },
                help: "move_action",
            },
            {
                key: Keys.U,
                exec: () => {
                    this.mailViewModel.listModel && this.toggleUnreadMails(this.mailViewModel.listModel.getSelectedAsArray());
                },
                help: "toggleUnread_action",
            },
            {
                key: Keys.ONE,
                exec: () => {
                    this.mailViewModel.switchToFolder(MailFolderType.INBOX);
                    return true;
                },
                help: "switchInbox_action",
            },
            {
                key: Keys.TWO,
                exec: () => {
                    this.mailViewModel.switchToFolder(MailFolderType.DRAFT);
                    return true;
                },
                help: "switchDrafts_action",
            },
            {
                key: Keys.THREE,
                exec: () => {
                    this.mailViewModel.switchToFolder(MailFolderType.SENT);
                    return true;
                },
                help: "switchSentFolder_action",
            },
            {
                key: Keys.FOUR,
                exec: () => {
                    this.mailViewModel.switchToFolder(MailFolderType.TRASH);
                    return true;
                },
                help: "switchTrash_action",
            },
            {
                key: Keys.FIVE,
                exec: () => {
                    this.mailViewModel.switchToFolder(MailFolderType.ARCHIVE);
                    return true;
                },
                enabled: () => locator.logins.isInternalUserLoggedIn(),
                help: "switchArchive_action",
            },
            {
                key: Keys.SIX,
                exec: () => {
                    this.mailViewModel.switchToFolder(MailFolderType.SPAM);
                    return true;
                },
                enabled: () => locator.logins.isInternalUserLoggedIn() && !locator.logins.isEnabled(FeatureType.InternalCommunication),
                help: "switchSpam_action",
            },
            {
                key: Keys.CTRL,
                exec: () => false,
                enabled: canDoDragAndDropExport,
                help: "dragAndDrop_action",
            },
            {
                key: Keys.P,
                exec: () => {
                    new Promise(async (resolve) => {
                        const { openPressReleaseEditor } = await import("../press/PressReleaseEditor");
                        const mailboxDetails = await this.mailViewModel.getMailboxDetails();
                        mailboxDetails && openPressReleaseEditor(mailboxDetails);
                        resolve(null);
                    });
                    return true;
                },
                help: "emptyString_msg",
                enabled: () => locator.logins.isEnabled(FeatureType.Newsletter),
            },
        ];
    }
    moveMails() {
        const mailList = this.mailViewModel.listModel;
        if (mailList == null) {
            return;
        }
        const selectedMails = mailList.getSelectedAsArray();
        showMoveMailsDropdown(locator.mailModel, getMoveMailBounds(), selectedMails);
    }
    createFolderColumn(editingFolderForMailGroup = null, drawerAttrs) {
        return new ViewColumn({
            view: () => {
                return m(FolderColumnView, {
                    drawer: drawerAttrs,
                    button: editingFolderForMailGroup
                        ? null
                        : !styles.isUsingBottomNavigation() && isNewMailActionAvailable()
                            ? {
                                label: "newMail_action",
                                click: () => this.showNewMailDialog().catch(ofClass(PermissionError, noOp)),
                            }
                            : null,
                    content: this.renderFolders(editingFolderForMailGroup),
                    ariaLabel: "folderTitle_label",
                });
            },
        }, editingFolderForMailGroup ? 1 /* ColumnType.Background */ : 0 /* ColumnType.Foreground */, {
            minWidth: size.first_col_min_width,
            maxWidth: size.first_col_max_width,
            headerCenter: () => lang.get("folderTitle_label"),
        });
    }
    renderFolders(editingFolderForMailGroup) {
        const details = locator.mailModel.mailboxDetails() ?? [];
        return [
            ...details.map((mailboxDetail) => {
                const inEditMode = editingFolderForMailGroup === mailboxDetail.mailGroup._id;
                // Only show folders for mailbox in which edit was selected
                if (editingFolderForMailGroup && !inEditMode) {
                    return null;
                }
                else {
                    return m(SidebarSection, {
                        name: () => getMailboxName(locator.logins, mailboxDetail),
                    }, this.createMailboxFolderItems(mailboxDetail, inEditMode, () => {
                        EditFoldersDialog.showEdit(() => this.renderFolders(mailboxDetail.mailGroup._id));
                    }));
                }
            }),
        ];
    }
    createMailboxFolderItems(mailboxDetail, inEditMode, onEditMailbox) {
        return m(MailFoldersView, {
            mailboxDetail,
            expandedFolders: this.expandedState,
            mailListToSelectedMail: this.mailViewModel.getMailListToSelectedMail(),
            onFolderClick: () => {
                if (!inEditMode) {
                    this.viewSlider.focus(this.listColumn);
                }
            },
            onFolderExpanded: (folder, state) => this.setExpandedState(folder, state),
            onShowFolderAddEditDialog: (...args) => this.showFolderAddEditDialog(...args),
            onDeleteCustomMailFolder: (folder) => this.deleteCustomMailFolder(mailboxDetail, folder),
            onFolderDrop: (mailId, folder) => this.handleFolderDrop(mailId, folder),
            inEditMode,
            onEditMailbox,
        });
    }
    setExpandedState(folder, currentExpansionState) {
        currentExpansionState ? this.expandedState.delete(getElementId(folder)) : this.expandedState.add(getElementId(folder));
        deviceConfig.setExpandedFolders(locator.logins.getUserController().userId, [...this.expandedState]);
    }
    onNewUrl(args, requestedPath) {
        if (requestedPath.startsWith("/mailto")) {
            if (location.hash.length > 5) {
                let url = location.hash.substring(5);
                let decodedUrl = decodeURIComponent(url);
                Promise.all([locator.mailModel.getUserMailboxDetails(), import("../editor/MailEditor")]).then(([mailboxDetails, { newMailtoUrlMailEditor }]) => {
                    newMailtoUrlMailEditor(decodedUrl, false, mailboxDetails)
                        .then((editor) => editor.show())
                        .catch(ofClass(CancelledError, noOp));
                    history.pushState("", document.title, window.location.pathname); // remove # from url
                });
            }
        }
        else if (args.action === "supportMail" && locator.logins.isGlobalAdminUserLoggedIn()) {
            import("../editor/MailEditor").then(({ writeSupportMail }) => writeSupportMail());
        }
        if (isApp()) {
            let userGroupInfo = locator.logins.getUserController().userGroupInfo;
            locator.pushService.closePushNotification(userGroupInfo.mailAddressAliases.map((alias) => alias.mailAddress).concat(userGroupInfo.mailAddress || []));
        }
        if (styles.isSingleColumnLayout() && !args.mailId && this.viewSlider.focusedColumn === this.mailColumn) {
            this.viewSlider.focusPreviousColumn();
        }
        this.mailViewModel.showMail(args.listId, args.mailId);
    }
    async handleFolderDrop(droppedMailId, folder) {
        if (!this.mailViewModel.listModel) {
            return;
        }
        let mailsToMove = [];
        // the dropped mail is among the selected mails, move all selected mails
        if (this.mailViewModel.listModel.isItemSelected(droppedMailId)) {
            mailsToMove = this.mailViewModel.listModel.getSelectedAsArray();
        }
        else {
            const entity = this.mailViewModel.listModel.state.items.find((item) => getElementId(item) === droppedMailId);
            if (entity) {
                mailsToMove.push(entity);
            }
        }
        moveMails({ mailModel: locator.mailModel, mails: mailsToMove, targetMailFolder: folder });
    }
    async showNewMailDialog() {
        const mailboxDetails = await this.mailViewModel.getMailboxDetails();
        if (mailboxDetails == null) {
            return;
        }
        const { newMailEditor } = await import("../editor/MailEditor");
        const dialog = await newMailEditor(mailboxDetails);
        dialog.show();
    }
    async deleteCustomMailFolder(mailboxDetail, folder) {
        if (folder.folderType !== MailFolderType.CUSTOM) {
            throw new Error("Cannot delete non-custom folder: " + String(folder._id));
        }
        // remove any selection to avoid that the next mail is loaded and selected for each deleted mail event
        this.mailViewModel?.listModel?.selectNone();
        if (isSpamOrTrashFolder(mailboxDetail.folders, folder)) {
            const confirmed = await Dialog.confirm(() => lang.get("confirmDeleteFinallyCustomFolder_msg", {
                "{1}": getFolderName(folder),
            }));
            if (!confirmed)
                return;
            await locator.mailModel.finallyDeleteCustomMailFolder(folder);
        }
        else {
            const confirmed = await Dialog.confirm(() => lang.get("confirmDeleteCustomFolder_msg", {
                "{1}": getFolderName(folder),
            }));
            if (!confirmed)
                return;
            await locator.mailModel.trashFolderAndSubfolders(folder);
        }
    }
    logout() {
        m.route.set("/");
    }
    async toggleUnreadMails(mails) {
        if (mails.length == 0) {
            return;
        }
        // set all selected emails to the opposite of the first email's unread state
        await locator.mailModel.markMails(mails, !mails[0].unread);
    }
    deleteMails(mails) {
        return promptAndDeleteMails(locator.mailModel, mails, noOp);
    }
    async showFolderAddEditDialog(mailGroupId, folder, parentFolder) {
        const mailboxDetail = await locator.mailModel.getMailboxDetailsForMailGroup(mailGroupId);
        await showEditFolderDialog(mailboxDetail, folder, parentFolder);
    }
}
//# sourceMappingURL=MailView.js.map