import { loadMailDetails } from "../model/MailUtils";
import { ALLOWED_IMAGE_FORMATS, Keys, MAX_BASE64_IMAGE_SIZE } from "../../api/common/TutanotaConstants";
import { ofClass, uint8ArrayToBase64 } from "@tutao/tutanota-utils";
import { lang } from "../../misc/LanguageViewModel";
import { Dialog } from "../../gui/base/Dialog";
import { showFileChooser } from "../../file/FileController.js";
import m from "mithril";
import { Button } from "../../gui/base/Button.js";
import { progressIcon } from "../../gui/base/Icon.js";
import { checkApprovalStatus } from "../../misc/LoginUtils.js";
import { locator } from "../../api/main/MainLocator.js";
import { UserError } from "../../api/main/UserError.js";
import { showUserError } from "../../misc/ErrorHandlerImpl.js";
import { client } from "../../misc/ClientDetector.js";
import { showProgressDialog } from "../../gui/dialogs/ProgressDialog.js";
import { LockedError } from "../../api/common/error/RestError.js";
import { ifAllowedTutaLinks } from "../../gui/base/GuiUtils.js";
import { ExternalLink } from "../../gui/base/ExternalLink.js";
import { SourceCodeViewer } from "./SourceCodeViewer.js";
export function insertInlineImageB64ClickHandler(ev, handler) {
    showFileChooser(true, ALLOWED_IMAGE_FORMATS).then((files) => {
        const tooBig = [];
        for (let file of files) {
            if (file.size > MAX_BASE64_IMAGE_SIZE) {
                tooBig.push(file);
            }
            else {
                const b64 = uint8ArrayToBase64(file.data);
                const dataUrlString = `data:${file.mimeType};base64,${b64}`;
                handler.insertImage(dataUrlString, {
                    style: "max-width: 100%",
                });
            }
        }
        if (tooBig.length > 0) {
            Dialog.message(() => lang.get("tooBigInlineImages_msg", {
                "{size}": MAX_BASE64_IMAGE_SIZE / 1024,
            }));
        }
    });
}
export async function showHeaderDialog(headersPromise) {
    let state = { state: "loading" };
    headersPromise.then((headers) => {
        state = { state: "loaded", headers };
        m.redraw();
    });
    let mailHeadersDialog;
    const closeHeadersAction = () => {
        mailHeadersDialog?.close();
    };
    mailHeadersDialog = Dialog.largeDialog({
        right: [
            {
                label: "ok_action",
                click: closeHeadersAction,
                type: "secondary" /* ButtonType.Secondary */,
            },
        ],
        middle: () => lang.get("mailHeaders_title"),
    }, {
        view: () => m(".white-space-pre.pt.pb.selectable", state.state === "loading" ? m(".center", progressIcon()) : state.headers ?? m(".center", lang.get("noEntries_msg"))),
    })
        .addShortcut({
        key: Keys.ESC,
        exec: closeHeadersAction,
        help: "close_alt",
    })
        .setCloseHandler(closeHeadersAction)
        .show();
}
export async function editDraft(viewModel) {
    const sendAllowed = await checkApprovalStatus(locator.logins, false);
    if (sendAllowed) {
        // check if to be opened draft has already been minimized, iff that is the case, re-open it
        const minimizedEditor = locator.minimizedMailModel.getEditorForDraft(viewModel.mail);
        if (minimizedEditor) {
            locator.minimizedMailModel.reopenMinimizedEditor(minimizedEditor);
        }
        else {
            try {
                const [mailboxDetails, { newMailEditorFromDraft }] = await Promise.all([
                    viewModel.mailModel.getMailboxDetailsForMail(viewModel.mail),
                    import("../editor/MailEditor"),
                ]);
                if (mailboxDetails == null) {
                    return;
                }
                const editorDialog = await newMailEditorFromDraft(viewModel.getAttachments(), await loadMailDetails(locator.mailFacade, locator.entityClient, viewModel.mail), viewModel.isBlockingExternalImages(), viewModel.getLoadedInlineImages(), mailboxDetails);
                editorDialog.show();
            }
            catch (e) {
                if (e instanceof UserError) {
                    await showUserError(e);
                }
                else {
                    throw e;
                }
            }
        }
    }
}
export async function showSourceDialog(rawHtml) {
    return Dialog.viewerDialog("emailSourceCode_title", SourceCodeViewer, { rawHtml });
}
export function mailViewerMoreActions(viewModel, showReadButton = true) {
    const moreButtons = [];
    if (showReadButton) {
        if (viewModel.isUnread()) {
            moreButtons.push({
                label: "markRead_action",
                click: () => viewModel.setUnread(false),
                icon: "Eye" /* Icons.Eye */,
            });
        }
        else {
            moreButtons.push({
                label: "markUnread_action",
                click: () => viewModel.setUnread(true),
                icon: "NoEye" /* Icons.NoEye */,
            });
        }
    }
    if (viewModel.canPersistBlockingStatus() && viewModel.isShowingExternalContent()) {
        moreButtons.push({
            label: "disallowExternalContent_action",
            click: () => viewModel.setContentBlockingStatus("0" /* ContentBlockingStatus.Block */),
            icon: "Picture" /* Icons.Picture */,
        });
    }
    if (viewModel.canPersistBlockingStatus() && viewModel.isBlockingExternalImages()) {
        moreButtons.push({
            label: "showImages_action",
            click: () => viewModel.setContentBlockingStatus("1" /* ContentBlockingStatus.Show */),
            icon: "Picture" /* Icons.Picture */,
        });
    }
    if (viewModel.isListUnsubscribe()) {
        moreButtons.push({
            label: "unsubscribe_action",
            click: () => unsubscribe(viewModel),
            icon: "Cancel" /* Icons.Cancel */,
        });
    }
    if (!client.isMobileDevice() && viewModel.canExport()) {
        moreButtons.push({
            label: "export_action",
            click: () => showProgressDialog("pleaseWait_msg", viewModel.exportMail()),
            icon: "Export" /* Icons.Export */,
        });
    }
    if (!client.isMobileDevice() && typeof window.print === "function" && viewModel.canPrint()) {
        moreButtons.push({
            label: "print_action",
            click: () => window.print(),
            icon: "Print" /* Icons.Print */,
        });
    }
    if (viewModel.canShowHeaders()) {
        moreButtons.push({
            label: "showHeaders_action",
            click: () => showHeaderDialog(viewModel.getHeaders()),
            icon: "ListUnordered" /* Icons.ListUnordered */,
        });
    }
    if (viewModel.canReport()) {
        moreButtons.push({
            label: "reportEmail_action",
            click: () => reportMail(viewModel),
            icon: "Warning" /* Icons.Warning */,
        });
    }
    // adding more optional buttons? put them above the report action so the new button
    // is not sometimes where the report action usually sits.
    return moreButtons;
}
function unsubscribe(viewModel) {
    return showProgressDialog("pleaseWait_msg", viewModel.unsubscribe())
        .then((success) => {
        if (success) {
            return Dialog.message("unsubscribeSuccessful_msg");
        }
    })
        .catch((e) => {
        if (e instanceof LockedError) {
            return Dialog.message("operationStillActive_msg");
        }
        else {
            return Dialog.message("unsubscribeFailed_msg");
        }
    });
}
function reportMail(viewModel) {
    const sendReport = (reportType) => {
        viewModel
            .reportMail(reportType)
            .catch(ofClass(LockedError, () => Dialog.message("operationStillActive_msg")))
            .finally(m.redraw);
    };
    const dialog = Dialog.showActionDialog({
        title: lang.get("reportEmail_action"),
        child: () => m(".flex.col.mt-m", {
            // So that space below buttons doesn't look huge
            style: {
                marginBottom: "-10px",
            },
        }, [
            m("div", lang.get("phishingReport_msg")),
            ifAllowedTutaLinks(locator.logins, "https://tuta.com/faq#phishing" /* InfoLink.Phishing */, (link) => m(ExternalLink, {
                href: link,
                text: lang.get("whatIsPhishing_msg"),
                isCompanySite: true,
                class: "mt-s",
            })),
            m(".flex-wrap.flex-end", [
                m(Button, {
                    label: "reportPhishing_action",
                    click: () => {
                        sendReport("0" /* MailReportType.PHISHING */);
                        dialog.close();
                    },
                    type: "secondary" /* ButtonType.Secondary */,
                }),
                m(Button, {
                    label: "reportSpam_action",
                    click: () => {
                        sendReport("1" /* MailReportType.SPAM */);
                        dialog.close();
                    },
                    type: "secondary" /* ButtonType.Secondary */,
                }),
            ]),
        ]),
        okAction: null,
    });
}
//# sourceMappingURL=MailViewerUtils.js.map