/**
 * Interface for obtaining the key that is used to encrypt credentials. Any access to that key should always be done using this interface
 * rather than directly accessing device storage.
 */
export class CredentialsKeyProvider {
    constructor(nativeCredentials, credentialsStorage, deviceEncryptionFacade) {
        this.nativeCredentials = nativeCredentials;
        this.credentialsStorage = credentialsStorage;
        this.deviceEncryptionFacade = deviceEncryptionFacade;
    }
    /**
     * Return the key that is used for encrypting credentials on the device. If no key exists on the device, a new key will be created
     * and also stored in the device's credentials storage.
     */
    async getCredentialsKey() {
        const encryptedCredentialsKey = this.credentialsStorage.getCredentialsEncryptionKey();
        if (encryptedCredentialsKey) {
            const credentialsKey = await this.nativeCredentials.decryptUsingKeychain(encryptedCredentialsKey, this.getEncryptionMode());
            return credentialsKey;
        }
        else {
            const credentialsKey = await this.deviceEncryptionFacade.generateKey();
            const encryptedCredentialsKey = await this.nativeCredentials.encryptUsingKeychain(credentialsKey, this.getEncryptionMode());
            this.credentialsStorage.setCredentialsEncryptionKey(encryptedCredentialsKey);
            return credentialsKey;
        }
    }
    getEncryptionMode() {
        const encryptionMode = this.credentialsStorage.getCredentialEncryptionMode();
        if (!encryptionMode) {
            throw new Error("Encryption mode not set");
        }
        return encryptionMode;
    }
}
//# sourceMappingURL=CredentialsKeyProvider.js.map