import { assertNotNull } from "@tutao/tutanota-utils";
/**
 * Main entry point to interact with credentials, i.e. storing and retrieving credentials from/to persistence.
 */
export class CredentialsProvider {
    constructor(credentialsEncryption, storage, keyMigrator, databaseKeyFactory, sqliteCipherFacade, interWindowEventSender) {
        this.credentialsEncryption = credentialsEncryption;
        this.storage = storage;
        this.keyMigrator = keyMigrator;
        this.databaseKeyFactory = databaseKeyFactory;
        this.sqliteCipherFacade = sqliteCipherFacade;
        this.interWindowEventSender = interWindowEventSender;
    }
    /**
     * Stores credentials. If credentials already exist for login, they will be overwritten.
     * Also creates a database key
     */
    async store(credentialsAndKey) {
        const encryptedCredentials = await this.credentialsEncryption.encrypt(credentialsAndKey);
        this.storage.store(encryptedCredentials);
    }
    /**
     * Change the encrypted password for the stored credentials.
     */
    async replacePassword(credentials, encryptedPassword) {
        const encryptedCredentials = this.storage.loadByUserId(credentials.userId);
        if (encryptedCredentials == null) {
            throw new Error(`Trying to replace password for credentials but credentials are not persisted: ${credentials.userId}`);
        }
        // Encrypted password is encrypted with the session key and is the same for encrypted and decrypted credentials, no additional logic is needed.
        const newEncryptedCredentials = { ...encryptedCredentials, encryptedPassword };
        this.storage.store(newEncryptedCredentials);
    }
    storeRaw(persistentCredentials) {
        this.storage.store(persistentCredentials);
    }
    async getCredentialsInfoByUserId(userId) {
        const persistentCredentials = this.storage.loadByUserId(userId);
        return persistentCredentials?.credentialInfo ?? null;
    }
    /**
     * Returns the full credentials for the userId passed in.
     * @param userId
     */
    async getCredentialsByUserId(userId) {
        const persistentCredentials = this.storage.loadByUserId(userId);
        if (persistentCredentials == null) {
            return null;
        }
        const decrypted = await this.credentialsEncryption.decrypt(persistentCredentials);
        if (decrypted.databaseKey == null) {
            // When offline mode is first released, there will be users who have saved credentials but no database key.
            // In the future, we will never save credentials without it, but we need to create one here
            decrypted.databaseKey = await this.databaseKeyFactory.generateKey();
            if (decrypted.databaseKey != null) {
                const reEncrypted = await this.credentialsEncryption.encrypt(decrypted);
                this.storage.store(reEncrypted);
            }
        }
        return decrypted;
    }
    /**
     * Returns the stored credentials infos of all internal users, i.e. users that have a "real" tutanota account and not the ones that
     * have a secure external mailbox.
     */
    async getInternalCredentialsInfos() {
        const allCredentials = this.storage.loadAll().map((persistentCredentials) => persistentCredentials.credentialInfo);
        return allCredentials.filter((credential) => {
            return credential.type === "internal";
        });
    }
    /**
     * Deletes stored credentials with specified userId.
     * No-op if credentials are not there.
     * @param opts.deleteOfflineDb whether to delete offline database. Will delete by default.
     */
    async deleteByUserId(userId, opts = { deleteOfflineDb: true }) {
        await this.interWindowEventSender?.localUserDataInvalidated(userId);
        if (opts.deleteOfflineDb) {
            await this.sqliteCipherFacade?.deleteDb(userId);
        }
        this.storage.deleteByUserId(userId);
    }
    /**
     * Sets the credentials encryption mode, i.e. how the intermediate key used for encrypting credentials is protected.
     * @param encryptionMode
     * @throws KeyPermanentlyInvalidatedError
     * @throws CredentialAuthenticationError
     */
    async setCredentialsEncryptionMode(encryptionMode) {
        if (encryptionMode === this.getCredentialsEncryptionMode()) {
            return;
        }
        const oldKeyEncrypted = this.storage.getCredentialsEncryptionKey();
        if (oldKeyEncrypted) {
            // if we have a key we must have a method
            const oldEncryptionMode = assertNotNull(this.storage.getCredentialEncryptionMode());
            const newlyEncryptedKey = await this.keyMigrator.migrateCredentialsKey(oldKeyEncrypted, oldEncryptionMode, encryptionMode);
            this.storage.setCredentialsEncryptionKey(newlyEncryptedKey);
        }
        this.storage.setCredentialEncryptionMode(encryptionMode);
        this.interWindowEventSender?.reloadDeviceConfig();
    }
    /**
     * Returns the credentials encryption mode, i.e. how the intermediate key used for encrypting credentials is protected.
     */
    getCredentialsEncryptionMode() {
        return this.storage.getCredentialEncryptionMode();
    }
    /**
     * Returns all credentials encryption modes that are supported by the device.
     */
    async getSupportedEncryptionModes() {
        return await this.credentialsEncryption.getSupportedEncryptionModes();
    }
    /**
     * Removes all stored credentials as well as any settings associated with credentials encryption.
     */
    async clearCredentials(reason) {
        console.warn("clearing all stored credentials:", reason);
        const storedCredentials = this.storage.loadAll();
        for (let storedCredential of storedCredentials) {
            await this.deleteByUserId(storedCredential.credentialInfo.userId);
        }
        this.storage.setCredentialsEncryptionKey(null);
        this.storage.setCredentialEncryptionMode(null);
    }
}
//# sourceMappingURL=CredentialsProvider.js.map