import m from "mithril";
import { px, size } from "../gui/size";
import stream from "mithril/stream";
import { displayOverlay } from "../gui/base/Overlay";
import { CalendarEventTypeRef, ContactTypeRef, MailTypeRef } from "../api/entities/tutanota/TypeRefs.js";
import { isKeyPressed, keyManager } from "../misc/KeyManager";
import { encodeCalendarSearchKey, getRestriction } from "./model/SearchUtils";
import { locator } from "../api/main/MainLocator";
import { Dialog } from "../gui/base/Dialog";
import { FULL_INDEXED_TIMESTAMP, Keys } from "../api/common/TutanotaConstants";
import { assertMainOrNode, isApp } from "../api/common/Env";
import { styles } from "../gui/styles";
import { client } from "../misc/ClientDetector";
import { debounce, downcast, isSameTypeRef, memoized, mod, ofClass } from "@tutao/tutanota-utils";
import { hasMoreResults } from "./model/SearchModel";
import { SearchBarOverlay } from "./SearchBarOverlay";
import { IndexingNotSupportedError } from "../api/common/error/IndexingNotSupportedError";
import { assertIsEntity, getElementId } from "../api/common/utils/EntityUtils";
import { compareContacts } from "../contacts/view/ContactGuiUtils";
import { BaseSearchBar } from "../gui/base/BaseSearchBar.js";
import { SearchRouter } from "./view/SearchRouter.js";
import { PageSize } from "../gui/base/ListUtils.js";
import { generateCalendarInstancesInRange } from "../calendar/date/CalendarUtils.js";
import { loadMultipleFromLists } from "../api/common/EntityClient.js";
assertMainOrNode();
const MAX_SEARCH_PREVIEW_RESULTS = 10;
// create our own copy which is not perfect because we don't benefit from the shared cache but currently there's no way to get async dependencies into
// singletons like this (without top-level await at least)
// once SearchBar is rewritten this should be removed
const searchRouter = new SearchRouter(locator.throttledRouter());
export class SearchBar {
    constructor() {
        this.focused = false;
        this.busy = false;
        this.lastSelectedWhitelabelChildrenInfoResult = stream();
        this.closeOverlayFunction = null;
        this.confirmDialogShown = false;
        this.indexStateStream = null;
        this.stateStream = null;
        this.lastQueryStream = null;
        /**
         * this reacts to URL changes by clearing the suggestions - the selected item may have changed (in the mail view maybe)
         * that shouldn't clear our current state, but if the URL changed in a way that makes the previous state outdated, we clear it.
         */
        this.onPathChange = memoized((newPath) => {
            if (locator.search.isNewSearch(this.state().query, getRestriction(newPath))) {
                this.updateState({
                    searchResult: null,
                    selected: null,
                    entities: [],
                });
            }
        });
        this.onkeydown = (e) => {
            const { selected, entities } = this.state();
            const keyHandlers = [
                {
                    key: Keys.F1,
                    exec: () => keyManager.openF1Help(),
                },
                {
                    key: Keys.ESC,
                    exec: () => this.clear(),
                },
                {
                    key: Keys.RETURN,
                    exec: () => {
                        if (selected) {
                            this.selectResult(selected);
                        }
                        else {
                            this.search();
                        }
                    },
                },
                {
                    key: Keys.UP,
                    exec: () => {
                        if (entities.length > 0) {
                            let oldSelected = selected || entities[0];
                            this.updateState({
                                selected: entities[mod(entities.indexOf(oldSelected) - 1, entities.length)],
                            });
                        }
                    },
                },
                {
                    key: Keys.DOWN,
                    exec: () => {
                        if (entities.length > 0) {
                            let newSelected = selected || entities[0];
                            this.updateState({
                                selected: entities[mod(entities.indexOf(newSelected) + 1, entities.length)],
                            });
                        }
                    },
                },
            ];
            let keyHandler = keyHandlers.find((handler) => isKeyPressed(e.key, handler.key));
            if (keyHandler) {
                keyHandler.exec();
                e.preventDefault();
            }
            // disable shortcuts
            e.stopPropagation();
            return true;
        };
        this.shortcuts = [
            {
                key: Keys.F,
                enabled: () => true,
                exec: () => {
                    this.onFocus();
                    m.redraw();
                },
                help: "search_label",
            },
        ];
        this.doSearch = debounce(300, (query, restriction, cb) => {
            if (!this.isQuickSearch()) {
                // if we're already on the search view, we don't want to wait until there's a new result to update the
                // UI. we can directly go to the URL and let the SearchViewModel do its thing from there.
                searchRouter.routeTo(query, restriction);
                return cb();
            }
            let useSuggestions = m.route.get().startsWith("/settings");
            // We don't limit contacts because we need to download all of them to sort them. They should be cached anyway.
            const limit = isSameTypeRef(MailTypeRef, restriction.type) ? (this.isQuickSearch() ? MAX_SEARCH_PREVIEW_RESULTS : PageSize) : null;
            locator.search
                .search({
                query: query ?? "",
                restriction,
                minSuggestionCount: useSuggestions ? 10 : 0,
                maxResults: limit,
            }, locator.progressTracker)
                .then((result) => this.loadAndDisplayResult(query, result ? result : null, limit))
                .finally(() => cb());
        });
        this.state = stream({
            query: "",
            searchResult: null,
            indexState: locator.search.indexState(),
            entities: [],
            selected: null,
        });
        this.overlayContentComponent = {
            view: () => {
                return m(SearchBarOverlay, {
                    state: this.state(),
                    isQuickSearch: this.isQuickSearch(),
                    isFocused: this.focused,
                    selectResult: (selected) => this.selectResult(selected),
                });
            },
        };
        this.view = this.view.bind(this);
        this.oncreate = this.oncreate.bind(this);
        this.onremove = this.onremove.bind(this);
    }
    view(vnode) {
        this.onPathChange(m.route.get());
        return m(BaseSearchBar, {
            placeholder: vnode.attrs.placeholder,
            text: this.state().query,
            busy: this.busy,
            disabled: vnode.attrs.disabled,
            onInput: (text) => this.search(text),
            onSearchClick: () => this.handleSearchClick(),
            onClear: () => {
                this.clear();
            },
            onWrapperCreated: (dom) => {
                this.domWrapper = dom;
                this.showOverlay();
            },
            onInputCreated: (dom) => {
                this.domInput = dom;
            },
            onFocus: () => (this.focused = true),
            onBlur: () => this.onBlur(),
            onKeyDown: (e) => this.onkeydown(e),
        });
    }
    oncreate() {
        keyManager.registerShortcuts(this.shortcuts);
        this.indexStateStream = locator.search.indexState.map((indexState) => {
            // When we finished indexing, search again forcibly to not confuse anyone with old results
            const currentResult = this.state().searchResult;
            if (!indexState.failedIndexingUpTo &&
                currentResult &&
                this.state().indexState.progress !== 0 &&
                indexState.progress === 0 &&
                //if period is changed from search view a new search is triggered there,  and we do not want to overwrite its result
                !this.timePeriodHasChanged(currentResult.restriction.end, indexState.aimedMailIndexTimestamp)) {
                this.doSearch(this.state().query, currentResult.restriction, m.redraw);
            }
            this.updateState({
                indexState,
            });
        });
        this.stateStream = this.state.map((state) => m.redraw());
        this.lastQueryStream = locator.search.lastQueryString.map((value) => {
            // Set value from the model when it's set from the URL e.g. reloading the page on the search screen
            if (value) {
                this.updateState({
                    query: value,
                });
            }
        });
    }
    onremove() {
        this.focused = false;
        this.shortcuts && keyManager.unregisterShortcuts(this.shortcuts);
        this.stateStream?.end(true);
        this.lastQueryStream?.end(true);
        this.indexStateStream?.end(true);
        this.closeOverlay();
    }
    timePeriodHasChanged(oldEnd, aimedEnd) {
        return oldEnd !== aimedEnd;
    }
    /**
     * Ensure that overlay exists in DOM
     */
    showOverlay() {
        if (this.closeOverlayFunction == null && this.domWrapper != null) {
            this.closeOverlayFunction = displayOverlay(() => this.makeOverlayRect(), this.overlayContentComponent, undefined, undefined, "dropdown-shadow border-radius");
        }
        else {
            m.redraw();
        }
    }
    closeOverlay() {
        if (this.closeOverlayFunction) {
            this.closeOverlayFunction();
            this.closeOverlayFunction = null;
        }
    }
    makeOverlayRect() {
        // note: this is called on every render which probably thrashes our layout constantly.
        // we should at least not do it while we don't have anything to show
        let overlayRect;
        const domRect = this.domWrapper.getBoundingClientRect();
        if (styles.isDesktopLayout()) {
            overlayRect = {
                top: px(domRect.bottom + 5),
                right: px(window.innerWidth - domRect.right),
                width: px(350),
                zIndex: 100 /* LayerType.LowPriorityOverlay */,
            };
        }
        else if (window.innerWidth < 500) {
            overlayRect = {
                top: px(size.navbar_height_mobile + 6),
                left: px(16),
                right: px(16),
                zIndex: 100 /* LayerType.LowPriorityOverlay */,
            };
        }
        else {
            overlayRect = {
                top: px(size.navbar_height_mobile + 6),
                left: px(domRect.left),
                right: px(window.innerWidth - domRect.right),
                zIndex: 100 /* LayerType.LowPriorityOverlay */,
            };
        }
        return overlayRect;
    }
    selectResult(result) {
        const { query } = this.state();
        if (result != null) {
            let type = "_type" in result ? result._type : null;
            if (!type) {
                // click on SHOW MORE button
                if (result.allowShowMore) {
                    this.updateSearchUrl(query);
                }
            }
            else if (isSameTypeRef(MailTypeRef, type)) {
                this.updateSearchUrl(query, downcast(result));
            }
            else if (isSameTypeRef(ContactTypeRef, type)) {
                this.updateSearchUrl(query, downcast(result));
            }
            else if (isSameTypeRef(CalendarEventTypeRef, type)) {
                this.updateSearchUrl(query, downcast(result));
            }
        }
    }
    handleSearchClick() {
        if (!this.focused) {
            this.onFocus();
        }
        else {
            this.search();
        }
    }
    getRestriction() {
        return getRestriction(m.route.get());
    }
    updateSearchUrl(query, selected) {
        if (selected && assertIsEntity(selected, CalendarEventTypeRef)) {
            searchRouter.routeTo(query, this.getRestriction(), selected && encodeCalendarSearchKey(selected));
        }
        else {
            searchRouter.routeTo(query, this.getRestriction(), selected && getElementId(selected));
        }
    }
    search(query) {
        let oldQuery = this.state().query;
        if (query != null) {
            this.updateState({
                query,
            });
        }
        else {
            query = oldQuery;
        }
        let restriction = this.getRestriction();
        if (!locator.search.indexState().mailIndexEnabled && restriction && isSameTypeRef(restriction.type, MailTypeRef) && !this.confirmDialogShown) {
            this.focused = false;
            this.confirmDialogShown = true;
            Dialog.confirm("enableSearchMailbox_msg", "search_label")
                .then((confirmed) => {
                if (confirmed) {
                    locator.indexerFacade
                        .enableMailIndexing()
                        .then(() => {
                        this.search();
                        this.onFocus();
                    })
                        .catch(ofClass(IndexingNotSupportedError, () => {
                        Dialog.message(isApp() ? "searchDisabledApp_msg" : "searchDisabled_msg");
                    }));
                }
            })
                .finally(() => (this.confirmDialogShown = false));
        }
        else {
            // Skip the search if the user is trying to bypass the search dialog
            if (!locator.search.indexState().mailIndexEnabled && isSameTypeRef(restriction.type, MailTypeRef)) {
                return;
            }
            if (!locator.search.isNewSearch(query, restriction) && oldQuery === query) {
                const result = locator.search.result();
                if (this.isQuickSearch() && result) {
                    this.showResultsInOverlay(result);
                }
                this.busy = false;
            }
            else {
                if (query.trim() !== "") {
                    this.busy = true;
                }
                this.doSearch(query, restriction, () => {
                    this.busy = false;
                    m.redraw();
                });
            }
        }
    }
    /** Given the result from the search load additional results if needed and then display them or set URL. */
    loadAndDisplayResult(query, result, limit) {
        const safeResult = result, safeLimit = limit;
        this.updateState({
            searchResult: safeResult,
        });
        if (!safeResult || locator.search.isNewSearch(query, safeResult.restriction)) {
            return;
        }
        if (this.isQuickSearch()) {
            if (safeLimit && hasMoreResults(safeResult) && safeResult.results.length < safeLimit) {
                locator.searchFacade.getMoreSearchResults(safeResult, safeLimit - safeResult.results.length).then((moreResults) => {
                    if (locator.search.isNewSearch(query, moreResults.restriction)) {
                        return;
                    }
                    else {
                        this.loadAndDisplayResult(query, moreResults, limit);
                    }
                });
            }
            else {
                this.showResultsInOverlay(safeResult);
            }
        }
        else {
            // instances will be displayed as part of the list of the search view, when the search view is displayed
            searchRouter.routeTo(query, safeResult.restriction);
        }
    }
    clear() {
        if (m.route.get().startsWith("/search")) {
            // this needs to happen in this order, otherwise the list's result subscription will override our
            // routing.
            this.updateSearchUrl("");
            locator.search.result(null);
        }
        this.updateState({
            query: "",
            entities: [],
            selected: null,
            searchResult: null,
        });
    }
    async showResultsInOverlay(result) {
        const entries = await loadMultipleFromLists(result.restriction.type, locator.entityClient, result.results);
        // If there was no new search while we've been downloading the result
        if (!locator.search.isNewSearch(result.query, result.restriction)) {
            const { filteredEntries, couldShowMore } = this.filterResults(entries, result.restriction);
            if (result.query.trim() !== "" &&
                (filteredEntries.length === 0 || hasMoreResults(result) || couldShowMore || result.currentIndexTimestamp !== FULL_INDEXED_TIMESTAMP)) {
                const moreEntry = {
                    resultCount: result.results.length,
                    shownCount: filteredEntries.length,
                    indexTimestamp: result.currentIndexTimestamp,
                    allowShowMore: true,
                };
                filteredEntries.push(moreEntry);
            }
            this.updateState({
                entities: filteredEntries,
                selected: filteredEntries[0],
            });
        }
    }
    isQuickSearch() {
        return !m.route.get().startsWith("/search");
    }
    filterResults(instances, restriction) {
        if (isSameTypeRef(restriction.type, ContactTypeRef)) {
            // Sort contacts by name
            return {
                filteredEntries: instances
                    .slice() // we can't modify the given array
                    .sort((o1, o2) => compareContacts(o1, o2))
                    .slice(0, MAX_SEARCH_PREVIEW_RESULTS),
                couldShowMore: instances.length > MAX_SEARCH_PREVIEW_RESULTS,
            };
        }
        else if (isSameTypeRef(restriction.type, CalendarEventTypeRef)) {
            const range = { start: restriction.start ?? 0, end: restriction.end ?? 0 };
            const generatedInstances = generateCalendarInstancesInRange(downcast(instances), range, MAX_SEARCH_PREVIEW_RESULTS + 1);
            return {
                filteredEntries: generatedInstances.slice(0, MAX_SEARCH_PREVIEW_RESULTS),
                couldShowMore: generatedInstances.length > MAX_SEARCH_PREVIEW_RESULTS,
            };
        }
        return { filteredEntries: instances.slice(0, MAX_SEARCH_PREVIEW_RESULTS), couldShowMore: instances.length > MAX_SEARCH_PREVIEW_RESULTS };
    }
    onFocus() {
        if (!locator.search.indexingSupported) {
            Dialog.message(isApp() ? "searchDisabledApp_msg" : "searchDisabled_msg");
        }
        else if (!this.focused) {
            this.focused = true;
            // setTimeout to fix bug in current Safari with losing focus
            setTimeout(() => {
                this.domInput.focus();
                this.search();
            }, client.browser === "Safari" /* BrowserType.SAFARI */ ? 200 : 0);
        }
    }
    onBlur() {
        this.focused = false;
        if (this.state().query === "") {
            if (m.route.get().startsWith("/search")) {
                const restriction = searchRouter.getRestriction();
                searchRouter.routeTo("", restriction);
            }
        }
        m.redraw();
    }
    updateState(update) {
        const newState = Object.assign({}, this.state(), update);
        this.state(newState);
        return newState;
    }
}
// Should be changed to not be a singleton and be proper component (instantiated by mithril).
// We need to extract some state of it into some kind of viewModel, pluggable depending on the current view but this requires complete rewrite of SearchBar.
export const searchBar = new SearchBar();
//# sourceMappingURL=SearchBar.js.map