import m from "mithril";
import { isApp, isDesktop } from "../api/common/Env";
import { lang } from "../misc/LanguageViewModel";
import { assertNotNull, neverNull, noOp, ofClass } from "@tutao/tutanota-utils";
import { createPushIdentifier, PushIdentifierTypeRef } from "../api/entities/sys/TypeRefs.js";
import { showProgressDialog } from "../gui/dialogs/ProgressDialog";
import { Dialog } from "../gui/base/Dialog";
import { NotFoundError } from "../api/common/error/RestError";
import { attachDropdown } from "../gui/base/Dropdown.js";
import { ExpanderButton, ExpanderPanel } from "../gui/base/Expander";
import stream from "mithril/stream";
import { TextField } from "../gui/base/TextField.js";
import { showNotAvailableForFreeDialog } from "../misc/SubscriptionDialogs";
import { getCleanedMailAddress } from "../misc/parsing/MailAddressParser";
import { locator } from "../api/main/MainLocator";
import { IconButton } from "../gui/base/IconButton.js";
import { isUpdateForTypeRef } from "../api/common/utils/EntityUpdateUtils.js";
class IdentifierRow {
    view(vnode) {
        const dropdownAttrs = attachDropdown({
            mainButtonAttrs: {
                title: "edit_action",
                icon: "More" /* Icons.More */,
                size: 1 /* ButtonSize.Compact */,
            },
            childAttrs: () => [
                {
                    label: () => lang.get(vnode.attrs.disabled ? "activate_action" : "deactivate_action"),
                    click: vnode.attrs.disableClicked,
                },
                {
                    label: "delete_action",
                    click: vnode.attrs.removeClicked,
                },
            ],
        });
        return m(".flex.flex-column.full-width", [
            m(".flex.items-center.selectable", [
                m("span" + (vnode.attrs.current ? ".b" : ""), vnode.attrs.name),
                vnode.attrs.disabled ? m(".mlr", `(${lang.get("notificationsDisabled_label")})`) : null,
                m(".flex-grow"),
                m(IconButton, dropdownAttrs),
            ]),
            this._identifier(vnode),
        ]);
    }
    _identifier(vnode) {
        const identifierText = vnode.attrs.formatIdentifier
            ? neverNull(vnode.attrs.identifier.match(/.{2}/g)).map((el, i) => m("span.pr-s" + (i % 2 === 0 ? ".b" : ""), el))
            : vnode.attrs.identifier;
        return m(".text-break.small.monospace.mt-negative-hpad-button.selectable", identifierText);
    }
}
export class IdentifierListViewer {
    constructor(user) {
        this._currentIdentifier = null;
        this._expanded = stream(false);
        this._identifiers = [];
        this._user = user;
        this._loadPushIdentifiers();
    }
    _disableIdentifier(identifier) {
        identifier.disabled = !identifier.disabled;
        locator.entityClient.update(identifier).then(m.redraw);
    }
    view() {
        const pushIdentifiersExpanderAttrs = {
            label: "show_action",
            expanded: this._expanded(),
            onExpandedChange: this._expanded,
        };
        const expanderContent = {
            view: () => {
                const buttonAddAttrs = {
                    title: "emailPushNotification_action",
                    click: () => this._showAddNotificationEmailAddressDialog(this._user),
                    icon: "Add" /* Icons.Add */,
                    size: 1 /* ButtonSize.Compact */,
                };
                const rowAdd = m(".full-width.flex-space-between.items-center.mb-s", [lang.get("emailPushNotification_action"), m(IconButton, buttonAddAttrs)]);
                const rows = this._identifiers
                    .map((identifier) => {
                    const isCurrentDevice = (isApp() || isDesktop()) && identifier.identifier === this._currentIdentifier;
                    return m(IdentifierRow, {
                        name: this._identifierDisplayName(isCurrentDevice, identifier.pushServiceType, identifier.displayName),
                        disabled: identifier.disabled,
                        identifier: identifier.identifier,
                        current: isCurrentDevice,
                        removeClicked: () => {
                            locator.entityClient.erase(identifier).catch(ofClass(NotFoundError, noOp));
                        },
                        formatIdentifier: identifier.pushServiceType !== "2" /* PushServiceType.EMAIL */,
                        disableClicked: () => this._disableIdentifier(identifier),
                    });
                })
                    .sort((l, r) => +r.attrs.current - +l.attrs.current);
                return m(".flex.flex-column.items-end.mb", [rowAdd].concat(rows));
            },
        };
        return [
            m(".flex-space-between.items-center.mt-l.mb-s", [
                m(".h4", lang.get("notificationSettings_action")),
                m(ExpanderButton, pushIdentifiersExpanderAttrs),
            ]),
            m(ExpanderPanel, {
                expanded: this._expanded(),
            }, m(expanderContent)),
            m(".small", lang.get("pushIdentifierInfoMessage_msg")),
        ];
    }
    _identifierDisplayName(current, type, displayName) {
        if (current) {
            return lang.get("pushIdentifierCurrentDevice_label");
        }
        else if (displayName) {
            return displayName;
        }
        else {
            return ["Android FCM", "iOS", lang.get("adminEmailSettings_action"), "Android Tutanota"][Number(type)];
        }
    }
    async _loadPushIdentifiers() {
        if (!this._user) {
            return;
        }
        this._currentIdentifier = this.getCurrentIdentifier();
        const list = this._user.pushIdentifierList;
        if (list) {
            this._identifiers = await locator.entityClient.loadAll(PushIdentifierTypeRef, list.list);
            m.redraw();
        }
    }
    getCurrentIdentifier() {
        return isApp() || isDesktop() ? locator.pushService.getLoadedPushIdentifier() : null;
    }
    _showAddNotificationEmailAddressDialog(user) {
        if (locator.logins.getUserController().isFreeAccount()) {
            showNotAvailableForFreeDialog();
        }
        else {
            let mailAddress = "";
            Dialog.showActionDialog({
                title: lang.get("notificationSettings_action"),
                child: {
                    view: () => [
                        m(TextField, {
                            label: "mailAddress_label",
                            value: mailAddress,
                            type: "email" /* TextFieldType.Email */,
                            oninput: (newValue) => (mailAddress = newValue),
                        }),
                        m(".small.mt-s", lang.get("emailPushNotification_msg")),
                    ],
                },
                validator: () => this._validateAddNotificationEmailAddressInput(mailAddress),
                allowOkWithReturn: true,
                okAction: (dialog) => {
                    this.createNotificationEmail(mailAddress, user);
                    dialog.close();
                },
            });
        }
    }
    createNotificationEmail(mailAddress, user) {
        const pushIdentifier = createPushIdentifier({
            _area: "0", // legacy
            _owner: user.userGroup.group, // legacy
            _ownerGroup: user.userGroup.group,
            displayName: lang.get("adminEmailSettings_action"),
            identifier: assertNotNull(getCleanedMailAddress(mailAddress)),
            language: lang.code,
            pushServiceType: "2" /* PushServiceType.EMAIL */,
            lastUsageTime: new Date(),
            lastNotificationDate: null,
            disabled: false,
        });
        let p = locator.entityClient.setup(assertNotNull(user.pushIdentifierList).list, pushIdentifier);
        showProgressDialog("pleaseWait_msg", p);
    }
    _validateAddNotificationEmailAddressInput(emailAddress) {
        return getCleanedMailAddress(emailAddress) == null ? "mailAddressInvalid_msg" : null; // TODO check if it is a Tutanota mail address
    }
    async entityEventReceived(update) {
        if (isUpdateForTypeRef(PushIdentifierTypeRef, update)) {
            await this._loadPushIdentifiers();
        }
    }
}
//# sourceMappingURL=IdentifierListViewer.js.map