import m from "mithril";
import { KnowledgeBaseEntryTypeRef } from "../api/entities/tutanota/TypeRefs.js";
import { lang } from "../misc/LanguageViewModel";
import { size } from "../gui/size";
import { GENERATED_MAX_ID, isSameId, listIdPart } from "../api/common/utils/EntityUtils";
import { hasCapabilityOnGroup } from "../sharing/GroupUtils";
import { ListColumnWrapper } from "../gui/ListColumnWrapper";
import { KnowledgeBaseEntryView } from "../knowledgebase/view/KnowledgeBaseEntryView";
import { memoized, NBSP, noOp } from "@tutao/tutanota-utils";
import { assertMainOrNode } from "../api/common/Env";
import { SelectableRowContainer } from "../gui/SelectableRowContainer.js";
import { ListModel } from "../misc/ListModel.js";
import { listSelectionKeyboardShortcuts, onlySingleSelection } from "../gui/base/ListUtils.js";
import { List } from "../gui/base/List.js";
import { BaseSearchBar } from "../gui/base/BaseSearchBar.js";
import { IconButton } from "../gui/base/IconButton.js";
import ColumnEmptyMessageBox from "../gui/base/ColumnEmptyMessageBox.js";
import { theme } from "../gui/theme.js";
import { knowledgeBaseSearch } from "../knowledgebase/model/KnowledgeBaseSearchFilter.js";
import { showKnowledgeBaseEditor } from "./KnowledgeBaseEditor.js";
import { keyManager } from "../misc/KeyManager.js";
import { isUpdateForTypeRef } from "../api/common/utils/EntityUpdateUtils.js";
import { ListAutoSelectBehavior } from "../misc/DeviceConfig.js";
assertMainOrNode();
/**
 *  List that is rendered within the knowledgeBase Settings
 */
export class KnowledgeBaseListView {
    constructor(entityClient, logins, templateGroupRoot, templateGroup, updateDetailsViewer, focusDetailsViewer) {
        this.entityClient = entityClient;
        this.logins = logins;
        this.templateGroupRoot = templateGroupRoot;
        this.templateGroup = templateGroup;
        this.updateDetailsViewer = updateDetailsViewer;
        this.focusDetailsViewer = focusDetailsViewer;
        this.searchQuery = "";
        this.resultItemIds = [];
        this.listStateSubscription = null;
        this.renderConfig = {
            itemHeight: size.list_row_height,
            multiselectionAllowed: 0 /* MultiselectMode.Disabled */,
            swipe: null,
            createElement: (dom) => {
                const knowledgebaseRow = new KnowledgeBaseRow();
                m.render(dom, knowledgebaseRow.render());
                return knowledgebaseRow;
            },
        };
        this.shortcuts = listSelectionKeyboardShortcuts(0 /* MultiselectMode.Disabled */, () => this.listModel);
        this.onSelectionChanged = memoized((item) => {
            const detailsViewer = item == null ? null : new KnowledgeBaseSettingsDetailsViewer(item, !this.userCanEdit());
            this.updateDetailsViewer(detailsViewer);
        });
        this.listModel = this.makeListModel();
        this.listModel.loadInitial();
        // hacks for old components
        this.view = this.view.bind(this);
        this.oncreate = this.oncreate.bind(this);
        this.onremove = this.onremove.bind(this);
    }
    oncreate() {
        keyManager.registerShortcuts(this.shortcuts);
    }
    onremove() {
        keyManager.unregisterShortcuts(this.shortcuts);
    }
    makeListModel() {
        const listModel = new ListModel({
            topId: GENERATED_MAX_ID,
            sortCompare: (a, b) => {
                var titleA = a.title.toUpperCase();
                var titleB = b.title.toUpperCase();
                return titleA < titleB ? -1 : titleA > titleB ? 1 : 0;
            },
            fetch: async (startId, count) => {
                // fetch works like in ContactListView, because we have a custom sort order there too
                if (startId === GENERATED_MAX_ID) {
                    // load all entries at once to apply custom sort order
                    const allEntries = await this.entityClient.loadAll(KnowledgeBaseEntryTypeRef, this.getListId());
                    return { items: allEntries, complete: true };
                }
                else {
                    throw new Error("fetch knowledgeBase entry called for specific start id");
                }
            },
            loadSingle: (elementId) => {
                return this.entityClient.load(KnowledgeBaseEntryTypeRef, [this.getListId(), elementId]);
            },
            autoSelectBehavior: () => ListAutoSelectBehavior.OLDER,
        });
        listModel.setFilter((item) => this.queryFilter(item));
        this.listStateSubscription?.end(true);
        this.listStateSubscription = listModel.stateStream.map((state) => {
            this.onSelectionChanged(onlySingleSelection(state));
            m.redraw();
        });
        return listModel;
    }
    view() {
        return m(ListColumnWrapper, {
            headerContent: m(".flex.flex-space-between.center-vertically.plr-l", m(BaseSearchBar, {
                text: this.searchQuery,
                onInput: (text) => this.updateQuery(text),
                busy: false,
                onKeyDown: (e) => e.stopPropagation(),
                onClear: () => {
                    this.searchQuery = "";
                    this.resultItemIds = [];
                    this.listModel.reapplyFilter();
                },
                placeholder: lang.get("searchKnowledgebase_placeholder"),
            }), this.userCanEdit()
                ? m(".mr-negative-s", m(IconButton, {
                    title: "addEntry_label",
                    icon: "Add" /* Icons.Add */,
                    click: () => {
                        showKnowledgeBaseEditor(null, this.templateGroupRoot);
                    },
                }))
                : null),
        }, this.listModel.isEmptyAndDone()
            ? m(ColumnEmptyMessageBox, {
                color: theme.list_message_bg,
                icon: "Book" /* Icons.Book */,
                message: "noEntries_msg",
            })
            : m(List, {
                renderConfig: this.renderConfig,
                state: this.listModel.state,
                onLoadMore: () => this.listModel.loadMore(),
                onRetryLoading: () => this.listModel.retryLoading(),
                onStopLoading: () => this.listModel.stopLoading(),
                onSingleSelection: (item) => {
                    this.listModel.onSingleSelection(item);
                    this.focusDetailsViewer();
                },
                onSingleTogglingMultiselection: noOp,
                onRangeSelectionTowards: noOp,
            }));
    }
    async entityEventsReceived(updates) {
        for (const update of updates) {
            if (isUpdateForTypeRef(KnowledgeBaseEntryTypeRef, update) && isSameId(this.getListId(), update.instanceListId)) {
                await this.listModel.entityEventReceived(update.instanceId, update.operation);
            }
        }
        // we need to make another search in case items have changed
        this.updateQuery(this.searchQuery);
        m.redraw();
    }
    queryFilter(item) {
        return this.searchQuery === "" ? true : this.resultItemIds.includes(item._id);
    }
    updateQuery(query) {
        this.searchQuery = query;
        this.resultItemIds = knowledgeBaseSearch(this.searchQuery, this.listModel.getUnfilteredAsArray()).map((item) => item._id);
        this.listModel.reapplyFilter();
    }
    userCanEdit() {
        return hasCapabilityOnGroup(this.logins.getUserController().user, this.templateGroup, "1" /* ShareCapability.Write */);
    }
    getListId() {
        return this.templateGroupRoot.knowledgeBase;
    }
}
export class KnowledgeBaseRow {
    constructor() {
        this.top = 0;
        this.domElement = null;
        this.entity = null;
    }
    update(entry, selected) {
        this.entity = entry;
        this.selectionUpdater(selected, false);
        this.entryTitleDom.textContent = entry.title;
    }
    render() {
        return m(SelectableRowContainer, {
            onSelectedChangeRef: (updater) => (this.selectionUpdater = updater),
        }, m(".flex.col", [
            m(".text-ellipsis.badge-line-height", {
                oncreate: (vnode) => (this.entryTitleDom = vnode.dom),
            }),
            // to create a second row
            m(".smaller.mt-xxs", NBSP),
        ]));
    }
}
export class KnowledgeBaseSettingsDetailsViewer {
    constructor(entry, readonly) {
        this.entry = entry;
        this.readonly = readonly;
    }
    renderView() {
        return m(".plr-l", m(KnowledgeBaseEntryView, {
            entry: this.entry,
            onTemplateSelected: (templateId) => m.route.set(`/settings/templates/${listIdPart(templateId)}}`),
            readonly: this.readonly,
        }));
    }
    entityEventsReceived(updates) {
        return Promise.resolve();
    }
}
//# sourceMappingURL=KnowledgeBaseListView.js.map