import m from "mithril";
import { TextField } from "../gui/base/TextField.js";
import { CompletenessIndicator } from "../gui/CompletenessIndicator.js";
import { getPasswordStrength, isSecurePassword, scaleToVisualPasswordStrength } from "../misc/passwords/PasswordUtils";
import { lang } from "../misc/LanguageViewModel";
import { StatusField } from "../gui/base/StatusField";
import { assertMainOrNode } from "../api/common/Env";
import { getEnabledMailAddressesForGroupInfo } from "../api/common/utils/GroupUtils.js";
import { showPasswordGeneratorDialog } from "../misc/passwords/PasswordGeneratorDialog";
import { theme } from "../gui/theme";
import { px, size } from "../gui/size.js";
import { ToggleButton } from "../gui/base/buttons/ToggleButton.js";
assertMainOrNode();
export class PasswordModel {
    constructor(usageTestController, logins, config, mailValid) {
        this.usageTestController = usageTestController;
        this.logins = logins;
        this.config = config;
        this.newPassword = "";
        this.oldPassword = "";
        this.repeatedPassword = "";
        this.revealOldPassword = false;
        this.revealNewPassword = false;
        this.revealConfirmPassword = false;
        this.passwordStrength = this.calculatePasswordStrength();
        this.__mailValid = mailValid;
        this.__signupFreeTest = this.usageTestController.getTest("signup.free");
        this.__signupPaidTest = this.usageTestController.getTest("signup.paid");
    }
    _checkBothValidAndSendPing() {
        if (this.getNewPasswordStatus().type === "valid" && this.getRepeatedPasswordStatus().type === "valid") {
            // Password entry (both passwords entered and valid)
            // Only the started test's (either free or paid clicked) stage is completed here
            this.__signupFreeTest?.getStage(3).complete();
            this.__signupPaidTest?.getStage(2).complete();
        }
    }
    getNewPassword() {
        return this.newPassword;
    }
    setNewPassword(newPassword) {
        if (this.__mailValid && this.__mailValid()) {
            // Email address selection finished (email address is available and clicked in password field)
            // Only the started test's (either free or paid clicked) stage is completed here
            this.__signupFreeTest?.getStage(2).complete();
            this.__signupPaidTest?.getStage(1).complete();
        }
        this.newPassword = newPassword;
        this.recalculatePasswordStrength();
    }
    /**
     * Might be needed when reserved strings change in the config
     */
    recalculatePasswordStrength() {
        this.passwordStrength = this.calculatePasswordStrength();
        this._checkBothValidAndSendPing();
    }
    getOldPassword() {
        return this.oldPassword;
    }
    setOldPassword(oldPassword) {
        this.oldPassword = oldPassword;
        this.passwordStrength = this.calculatePasswordStrength();
    }
    getRepeatedPassword() {
        return this.repeatedPassword;
    }
    setRepeatedPassword(repeatedPassword) {
        this.repeatedPassword = repeatedPassword;
        this.passwordStrength = this.calculatePasswordStrength();
        this._checkBothValidAndSendPing();
    }
    clear() {
        this.newPassword = "";
        this.oldPassword = "";
        this.repeatedPassword = "";
        this.passwordStrength = this.calculatePasswordStrength();
    }
    getErrorMessageId() {
        return (this.getErrorFromStatus(this.getOldPasswordStatus()) ??
            this.getErrorFromStatus(this.getNewPasswordStatus()) ??
            this.getErrorFromStatus(this.getRepeatedPasswordStatus()));
    }
    getOldPasswordStatus() {
        if (this.config.checkOldPassword && this.oldPassword === "") {
            return {
                type: "neutral",
                text: "oldPasswordNeutral_msg",
            };
        }
        else {
            return {
                type: "valid",
                text: "emptyString_msg",
            };
        }
    }
    getNewPasswordStatus() {
        if (this.newPassword === "") {
            return {
                type: "neutral",
                text: "password1Neutral_msg",
            };
        }
        else if (this.config.checkOldPassword && this.oldPassword === this.newPassword) {
            return {
                type: "invalid",
                text: "password1InvalidSame_msg",
            };
        }
        else if (this.isPasswordInsecure()) {
            if (this.config.enforceStrength) {
                return {
                    type: "invalid",
                    text: "password1InvalidUnsecure_msg",
                };
            }
            else {
                return {
                    type: "valid",
                    text: "password1InvalidUnsecure_msg",
                };
            }
        }
        else {
            return {
                type: "valid",
                text: "passwordValid_msg",
            };
        }
    }
    getRepeatedPasswordStatus() {
        if (this.config.hideConfirmation) {
            return {
                type: "valid",
                text: "passwordValid_msg",
            };
        }
        const repeatedPassword = this.repeatedPassword;
        const newPassword = this.newPassword;
        if (repeatedPassword === "") {
            return {
                type: "neutral",
                text: "password2Neutral_msg",
            };
        }
        else if (repeatedPassword !== newPassword) {
            return {
                type: "invalid",
                text: "password2Invalid_msg",
            };
        }
        else {
            return {
                type: "valid",
                text: "passwordValid_msg",
            };
        }
    }
    isPasswordInsecure() {
        return !isSecurePassword(this.getPasswordStrength());
    }
    getPasswordStrength() {
        return this.passwordStrength;
    }
    getErrorFromStatus(status) {
        if (!status)
            return null;
        return status.type !== "valid" ? status.text : null;
    }
    calculatePasswordStrength() {
        let reserved = this.config.reservedStrings ? this.config.reservedStrings() : [];
        if (this.logins.isUserLoggedIn()) {
            reserved = reserved
                .concat(getEnabledMailAddressesForGroupInfo(this.logins.getUserController().userGroupInfo))
                .concat(this.logins.getUserController().userGroupInfo.name);
        }
        // 80% strength is minimum. we expand it to 100%, so the password indicator if completely filled when the password is strong enough
        return getPasswordStrength(this.newPassword, reserved);
    }
    toggleRevealPassword(type) {
        switch (type) {
            case 0 /* PasswordFieldType.Old */:
                this.revealOldPassword = !this.revealOldPassword;
                break;
            case 1 /* PasswordFieldType.New */:
                this.revealNewPassword = !this.revealNewPassword;
                break;
            case 2 /* PasswordFieldType.Confirm */:
                this.revealConfirmPassword = !this.revealConfirmPassword;
                break;
        }
    }
    isPasswordRevealed(type) {
        switch (type) {
            case 0 /* PasswordFieldType.Old */:
                return this.revealOldPassword;
            case 1 /* PasswordFieldType.New */:
                return this.revealNewPassword;
            case 2 /* PasswordFieldType.Confirm */:
                return this.revealConfirmPassword;
        }
    }
}
/**
 * A form for entering a new password. Optionally it allows to enter the old password for validation and/or to repeat the new password.
 * showChangeOwnPasswordDialog() and showChangeUserPasswordAsAdminDialog() show this form as dialog.
 */
export class PasswordForm {
    view({ attrs }) {
        return m("", {
            onremove: () => attrs.model.clear(),
        }, [
            attrs.model.config.checkOldPassword
                ? m(TextField, {
                    label: "oldPassword_label",
                    value: attrs.model.getOldPassword(),
                    helpLabel: () => m(StatusField, { status: attrs.model.getOldPasswordStatus() }),
                    oninput: (input) => attrs.model.setOldPassword(input),
                    autocompleteAs: "current-password" /* Autocomplete.currentPassword */,
                    fontSize: px(size.font_size_smaller),
                    type: attrs.model.isPasswordRevealed(0 /* PasswordFieldType.Old */) ? "text" /* TextFieldType.Text */ : "password" /* TextFieldType.Password */,
                    injectionsRight: () => this.renderRevealIcon(attrs, 0 /* PasswordFieldType.Old */),
                })
                : null,
            m(TextField, {
                label: "newPassword_label",
                value: attrs.model.getNewPassword(),
                helpLabel: () => m(".flex.col.mt-xs", [
                    m(".flex.items-center", [
                        m(".mr-s", m(CompletenessIndicator, {
                            percentageCompleted: scaleToVisualPasswordStrength(attrs.model.getPasswordStrength()),
                        })),
                        m(StatusField, { status: attrs.model.getNewPasswordStatus() }),
                    ]),
                    this.renderPasswordGeneratorHelp(attrs),
                ]),
                oninput: (input) => attrs.model.setNewPassword(input),
                autocompleteAs: "new-password" /* Autocomplete.newPassword */,
                fontSize: px(size.font_size_smaller),
                type: attrs.model.isPasswordRevealed(1 /* PasswordFieldType.New */) ? "text" /* TextFieldType.Text */ : "password" /* TextFieldType.Password */,
                injectionsRight: () => this.renderRevealIcon(attrs, 1 /* PasswordFieldType.New */),
            }),
            attrs.model.config.hideConfirmation
                ? null
                : m(TextField, {
                    label: "repeatedPassword_label",
                    value: attrs.model.getRepeatedPassword(),
                    autocompleteAs: "new-password" /* Autocomplete.newPassword */,
                    helpLabel: () => m(StatusField, {
                        status: attrs.model.getRepeatedPasswordStatus(),
                    }),
                    oninput: (input) => attrs.model.setRepeatedPassword(input),
                    fontSize: px(size.font_size_smaller),
                    type: attrs.model.isPasswordRevealed(2 /* PasswordFieldType.Confirm */) ? "text" /* TextFieldType.Text */ : "password" /* TextFieldType.Password */,
                    injectionsRight: () => this.renderRevealIcon(attrs, 2 /* PasswordFieldType.Confirm */),
                }),
            attrs.passwordInfoKey ? m(".small.mt-s", lang.get(attrs.passwordInfoKey)) : null,
        ]);
    }
    renderPasswordGeneratorHelp(attrs) {
        return m("", [
            m(".b.mr-xs.hover.click.darkest-hover.mt-xs", {
                style: { display: "inline-block", color: theme.navigation_button_selected },
                onclick: async () => {
                    attrs.model.setNewPassword(await showPasswordGeneratorDialog());
                    m.redraw();
                },
            }, lang.get("generatePassphrase_action")),
        ]);
    }
    renderRevealIcon(attrs, passwordType) {
        return m(ToggleButton, {
            title: "revealPassword_action",
            toggled: attrs.model.isPasswordRevealed(passwordType),
            onToggled: (_, e) => {
                attrs.model.toggleRevealPassword(passwordType);
                e.stopPropagation();
            },
            icon: attrs.model.isPasswordRevealed(passwordType) ? "NoEye" /* Icons.NoEye */ : "Eye" /* Icons.Eye */,
            size: 1 /* ButtonSize.Compact */,
        });
    }
}
//# sourceMappingURL=PasswordForm.js.map