import m from "mithril";
import { lang } from "../misc/LanguageViewModel";
import { isMailAddress } from "../misc/FormatValidator";
import { AccessDeactivatedError } from "../api/common/error/RestError";
import { debounce, getFirstOrThrow } from "@tutao/tutanota-utils";
import { formatMailAddressFromParts } from "../misc/Formatter";
import { Icon } from "../gui/base/Icon";
import { locator } from "../api/main/MainLocator";
import { assertMainOrNode } from "../api/common/Env";
import { px, size } from "../gui/size.js";
import { inputLineHeight, TextField } from "../gui/base/TextField.js";
import { attachDropdown } from "../gui/base/Dropdown.js";
import { IconButton } from "../gui/base/IconButton.js";
import { SearchDropDown } from "../gui/SearchDropDown.js";
import { isTutanotaMailAddress } from "../api/common/mail/CommonMailUtils.js";
assertMainOrNode();
const VALID_MESSAGE_ID = "mailAddressAvailable_msg";
export class SelectMailAddressFormWithSuggestions {
    constructor({ attrs }) {
        this.focused = false;
        this.mailAvailabilities = [];
        this.selectedMailAddressSuggestionIndex = 0;
        this.suggestionsLoaded = false;
        this.getEmailSuggestions = debounce(1000, async (attrs) => {
            const { onBusyStateChanged } = attrs;
            const { valid } = this.performClientSideEmailValidation(attrs);
            if (!valid) {
                this.suggestionsLoaded = false;
                return;
            }
            let availabilities = [];
            let firstAvailableIndex = 0;
            try {
                availabilities = await locator.mailAddressFacade.areMailAddressesAvailable(this.getCleanMailAddresses());
                this.mailAvailabilities = availabilities;
                firstAvailableIndex = availabilities.findIndex((el) => el.available);
            }
            finally {
                this.onBusyStateChanged(false, onBusyStateChanged);
                this.suggestionsLoaded = true;
                this.selectSuggestion(attrs, firstAvailableIndex);
                m.redraw();
            }
        });
        this.isVerificationBusy = false;
        this.checkAddressTimeout = null;
        this.domain = getFirstOrThrow(attrs.availableDomains);
        this.domains = attrs.availableDomains;
        this.username = "";
        this.messageId = "mailAddressNeutral_msg";
    }
    view({ attrs }) {
        // this is a semi-good hack to reset the username after the user pressed "ok"
        if (attrs.injectionsRightButtonAttrs?.click) {
            const originalCallback = attrs.injectionsRightButtonAttrs.click;
            attrs.injectionsRightButtonAttrs.click = (event, dom) => {
                originalCallback(event, dom);
                this.username = "";
                this.messageId = "mailAddressNeutral_msg";
            };
        }
        return m("", [
            m(TextField, {
                label: "mailAddress_label",
                value: this.username,
                alignRight: true,
                helpLabel: () => this.addressHelpLabel(),
                autocompleteAs: "new-password" /* Autocomplete.newPassword */,
                fontSize: px(size.font_size_smaller),
                oninput: (value) => {
                    this.username = value;
                    this.onBusyStateChanged(true, attrs.onBusyStateChanged);
                    this.getEmailSuggestions(attrs);
                },
                onfocus: () => {
                    this.focused = true;
                },
                onblur: () => {
                    this.focused = false;
                },
                injectionsRight: () => [
                    m(".flex.items-end.align-self-end", {
                        style: {
                            "padding-bottom": "1px",
                            flex: "1 1 auto",
                            fontSize: px(size.font_size_smaller),
                            lineHeight: px(inputLineHeight),
                        },
                    }, `@${this.domain}`),
                    attrs.availableDomains.length > 1
                        ? m(IconButton, attachDropdown({
                            mainButtonAttrs: {
                                title: "domain_label",
                                icon: "Expand" /* BootIcons.Expand */,
                                size: 1 /* ButtonSize.Compact */,
                            },
                            childAttrs: () => attrs.availableDomains.map((domain, index) => this.createDropdownItemAttrs(domain, index, attrs)),
                            showDropdown: () => true,
                            width: 250,
                        }))
                        : attrs.injectionsRightButtonAttrs
                            ? m(IconButton, attrs.injectionsRightButtonAttrs)
                            : null,
                ],
            }),
            this.displaySuggestions() ? this.renderSuggestions(attrs) : null,
        ]);
    }
    displaySuggestions() {
        return this.focused && this.suggestionsLoaded;
    }
    renderSuggestions(attrs) {
        return m(".rel", m(SearchDropDown, {
            suggestions: this.mailAvailabilities.map((availability) => {
                return {
                    firstRow: lang.get(availability.available ? "available_label" : "unavailable_label"),
                    secondRow: availability.mailAddress,
                    display: attrs.displayUnavailableMailAddresses ? true : availability.available,
                };
            }),
            selectedSuggestionIndex: this.selectedMailAddressSuggestionIndex,
            onSuggestionSelected: (sel) => this.selectSuggestion(attrs, sel),
            maxHeight: Math.min(this.mailAvailabilities.filter((mailAvailability) => mailAvailability.available).length, attrs.maxSuggestionsToShow),
        }));
    }
    selectSuggestion(attrs, selection) {
        if (selection === -1) {
            // The email address (local part) is not available on any domains
            this.onValidationFinished("", {
                isValid: false,
                errorId: "mailAddressNA_msg",
            }, attrs.onValidationResult);
            return;
        }
        const selectedMail = this.mailAvailabilities[selection];
        if (!selectedMail.available) {
            // This should never happen as unavailable emails are not clickable
            return;
        }
        this.setDomain(attrs, selection);
        this.onValidationFinished(selectedMail.mailAddress, {
            isValid: true,
            errorId: null,
        }, attrs.onValidationResult);
    }
    setDomain(attrs, domainIndex) {
        this.selectedMailAddressSuggestionIndex = domainIndex;
        this.domain = this.domains[domainIndex];
        attrs.onDomainChanged?.(this.domain);
    }
    getCleanMailAddress() {
        return formatMailAddressFromParts(this.username, this.domain);
    }
    getCleanMailAddresses() {
        return this.domains.map((domain) => formatMailAddressFromParts(this.username, domain));
    }
    addressHelpLabel() {
        return this.isVerificationBusy
            ? m(".flex.items-center.mt-s", [this.progressIcon(), lang.get("mailAddressBusy_msg")])
            : m(".mt-s", lang.get(this.messageId ?? VALID_MESSAGE_ID));
    }
    progressIcon() {
        return m(Icon, {
            icon: "Progress" /* BootIcons.Progress */,
            class: "icon-progress mr-s",
        });
    }
    createDropdownItemAttrs(domain, index, attrs) {
        return {
            label: () => domain,
            click: () => {
                this.setDomain(attrs, index);
                this.verifyMailAddress(attrs);
            },
        };
    }
    onBusyStateChanged(isBusy, onBusyStateChanged) {
        this.isVerificationBusy = isBusy;
        onBusyStateChanged(isBusy);
        m.redraw();
    }
    onValidationFinished(email, validationResult, onValidationResult) {
        this.messageId = validationResult.errorId;
        onValidationResult(email, validationResult);
    }
    performClientSideEmailValidation({ onValidationResult, onBusyStateChanged }) {
        const cleanMailAddress = this.getCleanMailAddress();
        const cleanUsername = this.username.trim().toLowerCase();
        if (cleanUsername === "") {
            this.onValidationFinished(cleanMailAddress, {
                isValid: false,
                errorId: "mailAddressNeutral_msg",
            }, onValidationResult);
            this.onBusyStateChanged(false, onBusyStateChanged);
            return { valid: false };
        }
        else if (!isMailAddress(cleanMailAddress, true) || (isTutanotaMailAddress(cleanMailAddress) && cleanUsername.length < 3)) {
            this.onValidationFinished(cleanMailAddress, {
                isValid: false,
                errorId: "mailAddressInvalid_msg",
            }, onValidationResult);
            this.onBusyStateChanged(false, onBusyStateChanged);
            return { valid: false };
        }
        return { valid: true, cleanMailAddress, cleanUsername };
    }
    verifyMailAddress(attrs) {
        const { onValidationResult, onBusyStateChanged } = attrs;
        this.checkAddressTimeout && clearTimeout(this.checkAddressTimeout);
        const { cleanMailAddress, valid } = this.performClientSideEmailValidation(attrs);
        if (!valid) {
            return;
        }
        this.onBusyStateChanged(true, onBusyStateChanged);
        this.checkAddressTimeout = setTimeout(async () => {
            if (this.getCleanMailAddress() !== cleanMailAddress)
                return;
            let result;
            try {
                const available = await locator.mailAddressFacade.isMailAddressAvailable(cleanMailAddress);
                result = available ? { isValid: true, errorId: null } : { isValid: false, errorId: "mailAddressNA_msg" };
            }
            catch (e) {
                if (e instanceof AccessDeactivatedError) {
                    result = { isValid: false, errorId: "mailAddressDelay_msg" };
                }
                else {
                    throw e;
                }
            }
            finally {
                if (this.getCleanMailAddress() === cleanMailAddress) {
                    this.onBusyStateChanged(false, onBusyStateChanged);
                }
            }
            if (this.getCleanMailAddress() === cleanMailAddress) {
                this.onValidationFinished(cleanMailAddress, result, onValidationResult);
            }
        }, 500);
    }
}
//# sourceMappingURL=SelectMailAddressFormWithSuggestions.js.map