import m from "mithril";
import { BookingItemFeatureType, FeatureType, GroupType } from "../../api/common/TutanotaConstants.js";
import { Dialog } from "../../gui/base/Dialog.js";
import { SelectMailAddressForm } from "../SelectMailAddressForm.js";
import { getGroupTypeDisplayName } from "./GroupDetailsView.js";
import { showProgressDialog } from "../../gui/dialogs/ProgressDialog.js";
import { lang } from "../../misc/LanguageViewModel.js";
import { showBuyDialog } from "../../subscription/BuyDialog.js";
import { PreconditionFailedError } from "../../api/common/error/RestError.js";
import { showPlanUpgradeRequiredDialog } from "../../misc/SubscriptionDialogs.js";
import { TemplateGroupPreconditionFailedReason } from "../../sharing/GroupUtils.js";
import { DropDownSelector } from "../../gui/base/DropDownSelector.js";
import { TextField } from "../../gui/base/TextField.js";
import { getFirstOrThrow, ofClass } from "@tutao/tutanota-utils";
import { locator } from "../../api/main/MainLocator.js";
import { assertMainOrNode } from "../../api/common/Env.js";
import { getAvailableDomains } from "../mailaddress/MailAddressesUtils.js";
import { getAvailablePlansWithTemplates, toFeatureType } from "../../subscription/SubscriptionUtils.js";
import { MoreInfoLink } from "../../misc/news/MoreInfoLink.js";
assertMainOrNode();
export class AddGroupDialog {
    view(vnode) {
        const { availableGroupTypes, groupType, availableDomains, onEmailChanged, onBusyStateChanged, selectedDomain, onDomainChanged } = vnode.attrs;
        return [
            availableGroupTypes.length > 1
                ? m(DropDownSelector, {
                    label: "groupType_label",
                    items: availableGroupTypes.map((t) => {
                        return {
                            name: getGroupTypeDisplayName(t),
                            value: t,
                        };
                    }),
                    selectedValue: groupType,
                    selectionChangedHandler: vnode.attrs.onGroupTypeChanged,
                })
                : null,
            m(TextField, {
                label: "name_label",
                value: vnode.attrs.name,
                oninput: vnode.attrs.onGroupNameChanged,
            }),
            groupType === GroupType.Mail
                ? m("", [
                    m(SelectMailAddressForm, {
                        selectedDomain,
                        availableDomains,
                        onValidationResult: onEmailChanged,
                        onBusyStateChanged,
                        onDomainChanged,
                    }),
                    m(".mt-m", ""),
                    m(MoreInfoLink, { link: "https://tuta.com/support/#shared-mailboxes" /* InfoLink.SharedMailboxes */, isSmall: true }),
                ])
                : m(""),
        ];
    }
}
export class AddGroupDialogViewModel {
    constructor(availableDomains, groupManagementFacade) {
        this.availableDomains = availableDomains;
        this._groupManagementFacade = groupManagementFacade;
        this.groupTypes = this.getAvailableGroupTypes();
        this.groupType = getFirstOrThrow(this.groupTypes);
        this.groupName = "";
        this.mailAddress = "";
        this.errorMessageId = "mailAddressNeutral_msg";
        this.isVerifactionBusy = false;
        this.selectedDomain = getFirstOrThrow(availableDomains);
    }
    createMailGroup() {
        return this._groupManagementFacade.createMailGroup(this.groupName, this.mailAddress);
    }
    validateAddGroupInput() {
        if (this.groupType === GroupType.Mail) {
            return this.errorMessageId;
        }
        else if (this.groupType === GroupType.Template || (this.groupType === GroupType.MailingList && this.groupName.trim() === "")) {
            return "enterName_msg";
        }
        else {
            return null;
        }
    }
    getAvailableGroupTypes() {
        if (locator.logins.isEnabled(FeatureType.WhitelabelChild)) {
            return [];
        }
        else {
            return [GroupType.Mail];
        }
    }
    onEmailChanged(email, validationResult) {
        this.errorMessageId = validationResult.errorId;
        if (validationResult.isValid) {
            this.mailAddress = email;
        }
    }
}
export function show() {
    getAvailableDomains(locator.logins).then((availableDomains) => {
        const viewModel = new AddGroupDialogViewModel(availableDomains, locator.groupManagementFacade);
        if (viewModel.getAvailableGroupTypes().length === 0)
            return Dialog.message("selectionNotAvailable_msg");
        let addGroupOkAction = async (dialog) => {
            if (viewModel.isVerifactionBusy)
                return;
            const errorId = viewModel.validateAddGroupInput();
            if (errorId) {
                Dialog.message(errorId);
                return;
            }
            const userController = locator.logins.getUserController();
            const planType = await userController.getPlanType();
            const useLegacyBookingItem = await userController.useLegacyBookingItem();
            if (viewModel.groupType === GroupType.Mail) {
                showProgressDialog("pleaseWait_msg", showBuyDialog({
                    featureType: useLegacyBookingItem ? toFeatureType(planType) : BookingItemFeatureType.SharedMailGroup,
                    bookingText: "sharedMailbox_label",
                    count: 1,
                    freeAmount: 0,
                    reactivate: false,
                }).then((accepted) => {
                    if (accepted) {
                        dialog.close();
                        return viewModel.createMailGroup();
                    }
                }));
            }
            else if (viewModel.groupType === GroupType.Template) {
                addTemplateGroup(viewModel.groupName).then((success) => {
                    if (success) {
                        dialog.close();
                    }
                });
            }
        };
        Dialog.showActionDialog({
            title: viewModel.groupType == GroupType.Mail ? lang.get("createSharedMailbox_label") : lang.get("addGroup_label"),
            child: () => m(AddGroupDialog, {
                selectedDomain: viewModel.selectedDomain,
                groupType: viewModel.groupType,
                availableDomains: availableDomains,
                availableGroupTypes: viewModel.groupTypes,
                name: viewModel.groupName,
                onGroupNameChanged: (newName) => (viewModel.groupName = newName),
                onGroupTypeChanged: (newType) => (viewModel.groupType = newType),
                onEmailChanged: (mailAddress, validationResult) => viewModel.onEmailChanged(mailAddress, validationResult),
                onBusyStateChanged: (isBusy) => (viewModel.isVerifactionBusy = isBusy),
                onDomainChanged: (domain) => (viewModel.selectedDomain = domain),
            }),
            okAction: addGroupOkAction,
        });
    });
}
/**
 * @returns {Promise<boolean>} true if group was added, false otherwise
 */
function addTemplateGroup(name) {
    return showProgressDialog("pleaseWait_msg", locator.groupManagementFacade
        .createTemplateGroup(name)
        .then(() => true)
        .catch(ofClass(PreconditionFailedError, async (e) => {
        if (e.data === TemplateGroupPreconditionFailedReason.BUSINESS_FEATURE_REQUIRED ||
            e.data === TemplateGroupPreconditionFailedReason.UNLIMITED_REQUIRED) {
            const plans = await getAvailablePlansWithTemplates();
            showPlanUpgradeRequiredDialog(plans);
        }
        else {
            Dialog.message(() => e.message);
        }
        return false;
    })));
}
//# sourceMappingURL=AddGroupDialog.js.map