import m from "mithril";
import { GroupInfoTypeRef, GroupMemberTypeRef } from "../../api/entities/sys/TypeRefs.js";
import { LazyLoaded, memoized, noOp } from "@tutao/tutanota-utils";
import { GroupDetailsView } from "./GroupDetailsView.js";
import * as AddGroupDialog from "./AddGroupDialog.js";
import { Icon } from "../../gui/base/Icon.js";
import { locator } from "../../api/main/MainLocator.js";
import { ListColumnWrapper } from "../../gui/ListColumnWrapper.js";
import { assertMainOrNode } from "../../api/common/Env.js";
import { GroupDetailsModel } from "./GroupDetailsModel.js";
import { SelectableRowContainer, setVisibility } from "../../gui/SelectableRowContainer.js";
import { List } from "../../gui/base/List.js";
import { size } from "../../gui/size.js";
import { GENERATED_MAX_ID } from "../../api/common/utils/EntityUtils.js";
import { ListModel } from "../../misc/ListModel.js";
import { compareGroupInfos } from "../../api/common/utils/GroupUtils.js";
import { NotFoundError } from "../../api/common/error/RestError.js";
import { listSelectionKeyboardShortcuts, onlySingleSelection } from "../../gui/base/ListUtils.js";
import { keyManager } from "../../misc/KeyManager.js";
import { BaseSearchBar } from "../../gui/base/BaseSearchBar.js";
import { lang } from "../../misc/LanguageViewModel.js";
import ColumnEmptyMessageBox from "../../gui/base/ColumnEmptyMessageBox.js";
import { theme } from "../../gui/theme.js";
import { IconButton } from "../../gui/base/IconButton.js";
import { NewPaidPlans } from "../../api/common/TutanotaConstants.js";
import { isUpdateForTypeRef } from "../../api/common/utils/EntityUpdateUtils.js";
import { ListAutoSelectBehavior } from "../../misc/DeviceConfig.js";
assertMainOrNode();
const className = "group-list";
export class GroupListView {
    constructor(updateDetailsViewer, focusDetailsViewer) {
        this.updateDetailsViewer = updateDetailsViewer;
        this.focusDetailsViewer = focusDetailsViewer;
        this.searchQuery = "";
        this.renderConfig = {
            itemHeight: size.list_row_height,
            multiselectionAllowed: 0 /* MultiselectMode.Disabled */,
            swipe: null,
            createElement: (dom) => {
                const groupRow = new GroupRow();
                m.render(dom, groupRow.render());
                return groupRow;
            },
        };
        this.listStateSubscription = null;
        this.shortcuts = listSelectionKeyboardShortcuts(0 /* MultiselectMode.Disabled */, () => this.listModel);
        this.onSelectionChanged = memoized((item) => {
            if (item) {
                const newSelectionModel = new GroupDetailsModel(item, locator.entityClient, m.redraw);
                const detailsViewer = item == null ? null : new GroupDetailsView(newSelectionModel);
                this.updateDetailsViewer(detailsViewer);
            }
        });
        this.groupFilter = (gi) => {
            if (locator.logins.getUserController().isGlobalAdmin()) {
                return true;
            }
            else {
                return !!gi.localAdmin && this.localAdminGroupMemberships.map((gm) => gm.group).includes(gi.localAdmin);
            }
        };
        this.listModel = this.makeListModel();
        this.listId = new LazyLoaded(() => {
            return locator.logins
                .getUserController()
                .loadCustomer()
                .then((customer) => {
                return customer.teamGroups;
            });
        });
        this.localAdminGroupMemberships = locator.logins.getUserController().getLocalAdminGroupMemberships();
        this.listModel.loadInitial();
        this.oncreate = this.oncreate.bind(this);
        this.onremove = this.onremove.bind(this);
        this.view = this.view.bind(this);
    }
    oncreate() {
        keyManager.registerShortcuts(this.shortcuts);
    }
    view() {
        return m(ListColumnWrapper, {
            headerContent: m(".flex.flex-space-between.center-vertically.plr-l", m(BaseSearchBar, {
                text: this.searchQuery,
                onInput: (text) => this.updateQuery(text),
                busy: false,
                onKeyDown: (e) => e.stopPropagation(),
                onClear: () => {
                    this.searchQuery = "";
                    this.listModel.reapplyFilter();
                },
                placeholder: lang.get("searchMailboxes_placeholder"),
            }), m(".mr-negative-s", m(IconButton, {
                title: "createSharedMailbox_label",
                icon: "Add" /* Icons.Add */,
                click: () => this.addButtonClicked(),
            }))),
        }, this.listModel.isEmptyAndDone()
            ? m(ColumnEmptyMessageBox, {
                color: theme.list_message_bg,
                icon: "People" /* Icons.People */,
                message: "noEntries_msg",
            })
            : m(List, {
                renderConfig: this.renderConfig,
                state: this.listModel.state,
                onLoadMore: () => this.listModel.loadMore(),
                onRetryLoading: () => this.listModel.retryLoading(),
                onStopLoading: () => this.listModel.stopLoading(),
                onSingleSelection: (item) => {
                    this.listModel.onSingleSelection(item);
                    this.focusDetailsViewer();
                },
                onSingleTogglingMultiselection: noOp,
                onRangeSelectionTowards: noOp,
            }));
    }
    onremove() {
        keyManager.unregisterShortcuts(this.shortcuts);
        this.listStateSubscription?.end(true);
    }
    async addButtonClicked() {
        if (await locator.logins.getUserController().isNewPaidPlan()) {
            AddGroupDialog.show();
        }
        else {
            const msg = lang.get("newPaidPlanRequired_msg") + " " + lang.get("sharedMailboxesMultiUser_msg");
            const wizard = await import("../../subscription/UpgradeSubscriptionWizard");
            await wizard.showUpgradeWizard(locator.logins, NewPaidPlans, () => msg);
        }
    }
    async entityEventsReceived(updates) {
        for (const update of updates) {
            const { instanceListId, instanceId, operation } = update;
            if (isUpdateForTypeRef(GroupInfoTypeRef, update) && this.listId.getSync() === instanceListId) {
                await this.listModel.entityEventReceived(instanceId, operation);
            }
            else if (isUpdateForTypeRef(GroupMemberTypeRef, update)) {
                this.localAdminGroupMemberships = locator.logins.getUserController().getLocalAdminGroupMemberships();
                this.listModel.reapplyFilter();
            }
            m.redraw();
        }
    }
    makeListModel() {
        const listModel = new ListModel({
            topId: GENERATED_MAX_ID,
            sortCompare: compareGroupInfos,
            fetch: async (startId) => {
                if (startId === GENERATED_MAX_ID) {
                    const listId = await this.listId.getAsync();
                    const allGroupInfos = await locator.entityClient.loadAll(GroupInfoTypeRef, listId);
                    return { items: allGroupInfos, complete: true };
                }
                else {
                    throw new Error("fetch user group infos called for specific start id");
                }
            },
            loadSingle: async (elementId) => {
                const listId = await this.listId.getAsync();
                try {
                    return await locator.entityClient.load(GroupInfoTypeRef, [listId, elementId]);
                }
                catch (e) {
                    if (e instanceof NotFoundError) {
                        // we return null if the GroupInfo does not exist
                        return null;
                    }
                    else {
                        throw e;
                    }
                }
            },
            autoSelectBehavior: () => ListAutoSelectBehavior.OLDER,
        });
        listModel.setFilter((item) => this.groupFilter(item) && this.queryFilter(item));
        this.listStateSubscription?.end(true);
        this.listStateSubscription = listModel.stateStream.map((state) => {
            this.onSelectionChanged(onlySingleSelection(state));
            m.redraw();
        });
        return listModel;
    }
    queryFilter(gi) {
        const lowercaseSearch = this.searchQuery.toLowerCase();
        return gi.name.toLowerCase().includes(lowercaseSearch) || (!!gi.mailAddress && gi.mailAddress?.toLowerCase().includes(lowercaseSearch));
    }
    updateQuery(query) {
        this.searchQuery = query;
        this.listModel.reapplyFilter();
    }
}
export class GroupRow {
    constructor() {
        this.top = 0;
        this.domElement = null; // set from List
        this.entity = null;
    }
    update(groupInfo, selected) {
        this.entity = groupInfo;
        this.selectionUpdater(selected, false);
        this.nameDom.textContent = groupInfo.name;
        this.addressDom.textContent = groupInfo.mailAddress ?? "";
        setVisibility(this.deletedIconDom, groupInfo.deleted != null);
        // mail group or local admin group
        if (groupInfo.mailAddress) {
            this.localAdminIconDom.style.display = "none";
            this.mailIconDom.style.display = "";
        }
        else {
            this.localAdminIconDom.style.display = "";
            this.mailIconDom.style.display = "none";
        }
    }
    /**
     * Only the structure is managed by mithril. We set all contents on our own (see update) in order to avoid the vdom overhead (not negligible on mobiles)
     */
    render() {
        return m(SelectableRowContainer, {
            onSelectedChangeRef: (updater) => (this.selectionUpdater = updater),
        }, m(".flex.col.flex-grow", [
            m(".badge-line-height", [
                m("", {
                    oncreate: (vnode) => (this.nameDom = vnode.dom),
                }),
            ]),
            m(".flex-space-between.mt-xxs", [
                m(".smaller", {
                    oncreate: (vnode) => (this.addressDom = vnode.dom),
                }),
                m(".icons.flex", [
                    m(Icon, {
                        icon: "Trash" /* Icons.Trash */,
                        oncreate: (vnode) => (this.deletedIconDom = vnode.dom),
                        class: "svg-list-accent-fg",
                        style: {
                            display: "none",
                        },
                    }),
                    m(Icon, {
                        icon: "Settings" /* BootIcons.Settings */,
                        oncreate: (vnode) => (this.localAdminIconDom = vnode.dom),
                        class: "svg-list-accent-fg",
                        style: {
                            display: "none",
                        },
                    }),
                    m(Icon, {
                        icon: "Mail" /* BootIcons.Mail */,
                        oncreate: (vnode) => (this.mailIconDom = vnode.dom),
                        class: "svg-list-accent-fg",
                        style: {
                            display: "none",
                        },
                    }),
                ]),
            ]),
        ]));
    }
}
//# sourceMappingURL=GroupListView.js.map