import m from "mithril";
import { assertMainOrNode } from "../../api/common/Env";
import { clear, downcast, LazyLoaded, neverNull, noOp, promiseMap } from "@tutao/tutanota-utils";
import { BookingTypeRef, createStringWrapper, CustomerInfoTypeRef, CustomerPropertiesTypeRef, CustomerTypeRef, WhitelabelConfigTypeRef, } from "../../api/entities/sys/TypeRefs.js";
import { lang } from "../../misc/LanguageViewModel";
import { FeatureType } from "../../api/common/TutanotaConstants";
import { progressIcon } from "../../gui/base/Icon";
import { showProgressDialog } from "../../gui/dialogs/ProgressDialog";
import * as EditNotificationEmailDialog from "../EditNotificationEmailDialog";
import { showBuyOrSetNotificationEmailDialog } from "../EditNotificationEmailDialog";
import { isWhitelabelActive } from "../../subscription/SubscriptionUtils";
import { GENERATED_MAX_ID } from "../../api/common/utils/EntityUtils";
import { WhitelabelBrandingDomainSettings } from "./WhitelabelBrandingDomainSettings";
import { WhitelabelThemeSettings } from "./WhitelabelThemeSettings";
import { WhitelabelImprintAndPrivacySettings } from "./WhitelabelImprintAndPrivacySettings";
import { WhitelabelRegistrationSettings } from "./WhitelabelRegistrationSettings";
import { WhitelabelCustomMetaTagsSettings } from "./WhitelabelCustomMetaTagsSettings";
import { WhitelabelStatusSettings } from "./WhitelabelStatusSettings";
import { WhitelabelNotificationEmailSettings } from "./WhitelabelNotificationEmailSettings";
import { WhitelabelGermanLanguageFileSettings } from "./WhitelabelGermanLanguageFileSettings";
import { getThemeCustomizations } from "../../misc/WhitelabelCustomizations";
import { locator } from "../../api/main/MainLocator";
import { BrandingDomainService } from "../../api/entities/sys/Services";
import { getCustomMailDomains, getWhitelabelDomainInfo } from "../../api/common/utils/CustomerUtils.js";
import { isUpdateForTypeRef } from "../../api/common/utils/EntityUpdateUtils.js";
assertMainOrNode();
export class WhitelabelSettingsViewer {
    constructor(entityClient, logins) {
        this._whitelabelConfig = null;
        this._certificateInfo = null;
        this._whitelabelDomainInfo = null;
        this._customJsonTheme = null;
        this.view = this.view.bind(this);
        this._entityClient = entityClient;
        this._logins = logins;
        this._customer = new LazyLoaded(() => locator.logins.getUserController().loadCustomer());
        this._customerInfo = new LazyLoaded(() => locator.logins.getUserController().loadCustomerInfo());
        this._planConfig = new LazyLoaded(() => locator.logins.getUserController().getPlanConfig());
        this._customerProperties = new LazyLoaded(() => this._customer.getAsync().then((customer) => locator.entityClient.load(CustomerPropertiesTypeRef, neverNull(customer.properties))));
        this._lastBooking = null;
        this._updateFields();
    }
    view() {
        const brandingDomainConfig = this._renderBrandingDomainConfig();
        return m("#global-settings.fill-absolute.scroll.plr-l", brandingDomainConfig
            ? [
                m(".h4.mt-l", lang.get("whitelabel_label")),
                m(".small", lang.get("whitelabelDomainLinkInfo_msg") + " "),
                m("small.text-break", [m(`a[href=${"https://tuta.com/faq#whitelabel" /* InfoLink.Whitelabel */}][target=_blank]`, "https://tuta.com/faq#whitelabel" /* InfoLink.Whitelabel */)]),
                this._renderWhitelabelStatusSettings(),
                this._renderNotificationEmailSettings(),
                m(".h4.mt-l", lang.get("whitelabelDomain_label")),
                brandingDomainConfig,
                this._renderThemeSettings(),
                this._renderCustomMetaTagsSettings(),
                this._renderImprintAndPrivacySettings(),
                this._renderDefaultGermanLanguageFileSettings(),
                this._renderWhitelabelRegistrationSettings(),
                m(".mb-l"),
            ]
            : [m(".flex-center.items-center.button-height.mt-l", progressIcon())]);
    }
    _renderImprintAndPrivacySettings() {
        const whitelabelConfig = this._whitelabelConfig;
        const privacyStatementUrl = whitelabelConfig?.privacyStatementUrl ?? "";
        let onPrivacyStatementUrlChanged = null;
        if (whitelabelConfig) {
            onPrivacyStatementUrlChanged = (privacyStatementUrl) => {
                whitelabelConfig.privacyStatementUrl = privacyStatementUrl;
                this._entityClient.update(whitelabelConfig);
            };
        }
        const imprintUrl = whitelabelConfig?.imprintUrl ?? "";
        let onImprintUrlChanged = null;
        if (whitelabelConfig) {
            onImprintUrlChanged = (imprintUrl) => {
                whitelabelConfig.imprintUrl = imprintUrl;
                this._entityClient.update(whitelabelConfig);
            };
        }
        return m(WhitelabelImprintAndPrivacySettings, {
            privacyStatementUrl,
            onPrivacyStatementUrlChanged,
            imprintUrl,
            onImprintUrlChanged,
        });
    }
    _renderThemeSettings() {
        const customTheme = this._customJsonTheme;
        const whitelabelConfig = this._whitelabelConfig;
        const whitelabelDomainInfo = this._whitelabelDomainInfo;
        const whitelabelThemeSettingsAttrs = {
            whitelabelData: whitelabelConfig && whitelabelDomainInfo && customTheme
                ? {
                    customTheme,
                    whitelabelConfig,
                    whitelabelDomainInfo,
                }
                : null,
        };
        return m(WhitelabelThemeSettings, whitelabelThemeSettingsAttrs);
    }
    _renderWhitelabelRegistrationSettings() {
        if (!this._isWhitelabelRegistrationVisible())
            return null;
        const possibleRegistrationDomains = [
            {
                name: lang.get("deactivated_label"),
                value: null,
            },
        ].concat(getCustomMailDomains(this._customerInfo.getLoaded()).map((d) => {
            return {
                name: d.domain,
                value: d.domain,
            };
        }));
        let onRegistrationDomainSelected = noOp;
        let currentRegistrationDomain = null;
        if (this._whitelabelConfig) {
            onRegistrationDomainSelected = (domain) => {
                clear(neverNull(this._whitelabelConfig).whitelabelRegistrationDomains);
                if (domain) {
                    const domainWrapper = createStringWrapper({
                        value: domain,
                    });
                    neverNull(this._whitelabelConfig).whitelabelRegistrationDomains.push(domainWrapper);
                }
                this._entityClient.update(neverNull(this._whitelabelConfig));
            };
            if (this._whitelabelConfig.whitelabelRegistrationDomains.length > 0) {
                currentRegistrationDomain = this._whitelabelConfig.whitelabelRegistrationDomains[0].value;
            }
        }
        const whitelabelCode = this._whitelabelConfig ? this._whitelabelConfig.whitelabelCode : "";
        let onWhitelabelCodeChanged = noOp;
        if (this._whitelabelConfig) {
            onWhitelabelCodeChanged = (code) => {
                neverNull(this._whitelabelConfig).whitelabelCode = code;
                this._entityClient.update(neverNull(this._whitelabelConfig));
            };
        }
        return m(WhitelabelRegistrationSettings, {
            whitelabelCode,
            onWhitelabelCodeChanged,
            possibleRegistrationDomains,
            currentRegistrationDomain,
            onRegistrationDomainSelected,
        });
    }
    _renderDefaultGermanLanguageFileSettings() {
        if (!this._whitelabelConfig)
            return null;
        if (lang.code !== "de" && lang.code !== "de_sie")
            return null;
        const customGermanLanguageFile = downcast(this._whitelabelConfig.germanLanguageCode);
        const onGermanLanguageFileChanged = (languageFile) => {
            if (languageFile) {
                neverNull(this._whitelabelConfig).germanLanguageCode = languageFile;
                this._entityClient.update(neverNull(this._whitelabelConfig));
                lang.setLanguage({
                    code: languageFile,
                    languageTag: lang.languageTag,
                });
            }
        };
        const whitelabelGermanLanguageFileSettingsAttrs = {
            customGermanLanguageFile,
            onGermanLanguageFileChanged,
        };
        return m(WhitelabelGermanLanguageFileSettings, whitelabelGermanLanguageFileSettingsAttrs);
    }
    _renderCustomMetaTagsSettings() {
        let metaTags = "";
        let onMetaTagsChanged = null;
        if (this._whitelabelConfig) {
            metaTags = this._whitelabelConfig.metaTags;
            onMetaTagsChanged = (metaTags) => {
                neverNull(this._whitelabelConfig).metaTags = metaTags;
                this._entityClient.update(neverNull(this._whitelabelConfig));
            };
        }
        const whitelabelCustomMetaTagsSettingsAttrs = {
            metaTags,
            onMetaTagsChanged,
        };
        return m(WhitelabelCustomMetaTagsSettings, whitelabelCustomMetaTagsSettingsAttrs);
    }
    _renderWhitelabelStatusSettings() {
        const planConfig = this._planConfig.getSync();
        if (planConfig == null)
            return null;
        const whitelabelActive = isWhitelabelActive(this._lastBooking, planConfig);
        const whitelabelStatusSettingsAttrs = {
            isWhitelabelActive: whitelabelActive,
        };
        return m(WhitelabelStatusSettings, whitelabelStatusSettingsAttrs);
    }
    _renderBrandingDomainConfig() {
        const customerInfo = this._customerInfo.getSync();
        const planConfig = this._planConfig.getSync();
        if (!customerInfo || !planConfig)
            return null;
        const certificateInfo = this._certificateInfo;
        const isWhitelabelFeatureEnabled = isWhitelabelActive(this._lastBooking, planConfig);
        const whitelabelDomain = this._whitelabelDomainInfo ? this._whitelabelDomainInfo.domain : "";
        const whitelabelBrandingDomainSettingsAttrs = {
            customerInfo,
            isWhitelabelFeatureEnabled,
            certificateInfo,
            whitelabelDomain,
        };
        return m(WhitelabelBrandingDomainSettings, whitelabelBrandingDomainSettingsAttrs);
    }
    _isWhitelabelRegistrationVisible() {
        return (this._customer.isLoaded() &&
            this._customer.getLoaded().customizations.some((c) => c.feature === FeatureType.WhitelabelParent) &&
            this._customerInfo.isLoaded() &&
            getWhitelabelDomainInfo(this._customerInfo.getLoaded()) != null);
    }
    _tryLoadWhitelabelConfig(domainInfo) {
        if (domainInfo && domainInfo.whitelabelConfig) {
            return Promise.all([
                locator.entityClient.load(WhitelabelConfigTypeRef, domainInfo.whitelabelConfig),
                locator.serviceExecutor.get(BrandingDomainService, null).then((response) => neverNull(response.certificateInfo)),
            ]).then(([whitelabelConfig, certificateInfo]) => ({
                whitelabelConfig,
                certificateInfo,
            }));
        }
        else {
            return Promise.resolve(null);
        }
    }
    async _updateFields() {
        await this._planConfig.getAsync();
        const customerInfo = await this._customerInfo.getAsync();
        this._whitelabelDomainInfo = getWhitelabelDomainInfo(customerInfo);
        const data = await this._tryLoadWhitelabelConfig(this._whitelabelDomainInfo);
        this._whitelabelConfig = data?.whitelabelConfig ?? null;
        this._certificateInfo = data?.certificateInfo ?? null;
        return locator.entityClient.loadRange(BookingTypeRef, neverNull(customerInfo.bookings).items, GENERATED_MAX_ID, 1, true).then((bookings) => {
            this._lastBooking = bookings.length === 1 ? bookings[0] : null;
            this._customJsonTheme = this._whitelabelConfig ? getThemeCustomizations(this._whitelabelConfig) : null;
            m.redraw();
            this._customerProperties.getAsync().then(m.redraw);
        });
    }
    _renderNotificationEmailSettings() {
        const customerProperties = this._customerProperties.getSync();
        if (!customerProperties)
            return null;
        const notificationMailTemplates = customerProperties.notificationMailTemplates;
        const onAddTemplate = () => {
            showBuyOrSetNotificationEmailDialog(this._lastBooking, this._customerProperties);
        };
        const onEditTemplate = (template) => {
            EditNotificationEmailDialog.show(template, this._customerProperties);
        };
        const onRemoveTemplate = (template) => {
            this._removeNotificationMailTemplate(template);
        };
        const whitelabelNotificationEmailSettingsAttrs = {
            notificationMailTemplates,
            onAddTemplate,
            onEditTemplate,
            onRemoveTemplate,
        };
        return m(WhitelabelNotificationEmailSettings, whitelabelNotificationEmailSettingsAttrs);
    }
    _removeNotificationMailTemplate(template) {
        showProgressDialog("pleaseWait_msg", this._customerProperties.getAsync().then((customerProps) => {
            const index = customerProps.notificationMailTemplates.findIndex((t) => t.language === template.language);
            if (index !== -1) {
                customerProps.notificationMailTemplates.splice(index, 1);
                this._entityClient.update(customerProps);
            }
        }));
    }
    entityEventsReceived(updates) {
        return promiseMap(updates, (update) => {
            if (isUpdateForTypeRef(CustomerTypeRef, update) && update.operation === "1" /* OperationType.UPDATE */) {
                this._customer.reset();
                return this._customer.getAsync().then(() => m.redraw());
            }
            else if (isUpdateForTypeRef(CustomerInfoTypeRef, update) && update.operation === "1" /* OperationType.UPDATE */) {
                this._customerInfo.reset();
                return this._updateFields();
            }
            else if (isUpdateForTypeRef(WhitelabelConfigTypeRef, update) && update.operation === "1" /* OperationType.UPDATE */) {
                return this._updateFields();
            }
            else if (isUpdateForTypeRef(CustomerPropertiesTypeRef, update) && update.operation === "1" /* OperationType.UPDATE */) {
                this._customerProperties.reset();
                return this._updateFields();
            }
            else if (isUpdateForTypeRef(BookingTypeRef, update)) {
                return this._updateFields();
            }
        }).then(noOp);
    }
}
//# sourceMappingURL=WhitelabelSettingsViewer.js.map