import stream from "mithril/stream";
import { getElementId, getEtId, isSameId } from "../../api/common/utils/EntityUtils";
import { GroupMemberTypeRef, GroupTypeRef, SentGroupInvitationTypeRef } from "../../api/entities/sys/TypeRefs.js";
import { NotFoundError } from "../../api/common/error/RestError";
import { findAndRemove, noOp, ofClass, promiseMap } from "@tutao/tutanota-utils";
import { getSharedGroupName, hasCapabilityOnGroup, isSharedGroupOwner, loadGroupInfoForMember, loadGroupMembers } from "../GroupUtils";
import { UserError } from "../../api/main/UserError";
import { lang } from "../../misc/LanguageViewModel";
import { RecipientsNotFoundError } from "../../api/common/error/RecipientsNotFoundError";
import { ProgrammingError } from "../../api/common/error/ProgrammingError";
import { ResolveMode } from "../../api/main/RecipientsModel";
import { isUpdateForTypeRef } from "../../api/common/utils/EntityUpdateUtils.js";
export class GroupSharingModel {
    constructor(groupInfo, group, memberInfos, sentGroupInvitations, eventController, entityClient, logins, mailFacade, shareFacade, groupManagementFacade, recipientsModel) {
        this.recipientsModel = recipientsModel;
        this.onEntityEvents = (events, id) => this.entityEventsReceived(events, id);
        this.info = groupInfo;
        this.group = group;
        this.memberInfos = memberInfos;
        this.sentGroupInvitations = sentGroupInvitations;
        this.eventController = eventController;
        this.entityClient = entityClient;
        this.logins = logins;
        this._mailFacade = mailFacade;
        this._shareFacade = shareFacade;
        this._groupManagementFacade = groupManagementFacade;
        this.onEntityUpdate = stream();
        this.eventController.addEntityListener(this.onEntityEvents);
    }
    static newAsync(info, eventController, entityClient, logins, mailFacade, shareFacade, groupManagementFacade, recipientsModel) {
        return entityClient
            .load(GroupTypeRef, info.group)
            .then((group) => Promise.all([entityClient.loadAll(SentGroupInvitationTypeRef, group.invitations), loadGroupMembers(group, entityClient)]).then(([sentGroupInvitations, memberInfos]) => new GroupSharingModel(info, group, memberInfos, sentGroupInvitations, eventController, entityClient, logins, mailFacade, shareFacade, groupManagementFacade, recipientsModel)));
    }
    dispose() {
        this.eventController.removeEntityListener(this.onEntityEvents);
    }
    /**
     * Whether or not a given member can be removed from the group by the current user
     */
    canRemoveGroupMember(member) {
        return ((hasCapabilityOnGroup(this.logins.getUserController().user, this.group, "2" /* ShareCapability.Invite */) || this.memberIsSelf(member)) &&
            !isSharedGroupOwner(this.group, member.user));
    }
    removeGroupMember(member) {
        return this.canRemoveGroupMember(member)
            ? this._groupManagementFacade.removeUserFromGroup(member.user, getEtId(this.group))
            : Promise.reject(new ProgrammingError("User does not have permission to remove this member from the group"));
    }
    /**
     * Whether or not a given invitation can be cancelled by the current user
     * @param group
     * @param sentGroupInvitation
     * @returns {boolean}
     */
    canCancelInvitation(sentGroupInvitation) {
        return (hasCapabilityOnGroup(this.logins.getUserController().user, this.group, "2" /* ShareCapability.Invite */) ||
            isSharedGroupOwner(this.group, this.logins.getUserController().user._id));
    }
    memberIsSelf(member) {
        return isSameId(this.logins.getUserController().user._id, member.user);
    }
    cancelInvitation(invitation) {
        return this.canCancelInvitation(invitation) && invitation.receivedInvitation
            ? this._shareFacade.rejectGroupInvitation(invitation.receivedInvitation)
            : Promise.reject(new Error("User does not have permission to cancel this invitation")); // TODO error type
    }
    async sendGroupInvitation(sharedGroupInfo, recipients, capability) {
        const externalRecipients = [];
        for (let recipient of recipients) {
            const resolved = await this.recipientsModel.resolve(recipient, ResolveMode.Eager).resolved();
            if (resolved.type !== "internal" /* RecipientType.INTERNAL */) {
                externalRecipients.push(resolved.address);
            }
        }
        if (externalRecipients.length) {
            throw new UserError(() => lang.get("featureTutanotaOnly_msg") + " " + lang.get("invalidRecipients_msg") + "\n" + externalRecipients.join("\n"));
        }
        let groupInvitationReturn;
        try {
            groupInvitationReturn = await this._shareFacade.sendGroupInvitation(sharedGroupInfo, getSharedGroupName(sharedGroupInfo, this.logins.getUserController(), false), recipients.map((r) => r.address), capability);
        }
        catch (e) {
            if (e instanceof RecipientsNotFoundError) {
                throw new UserError(() => `${lang.get("tutanotaAddressDoesNotExist_msg")} ${lang.get("invalidRecipients_msg")}\n${e.message}`);
            }
            else {
                throw e;
            }
        }
        if (groupInvitationReturn.existingMailAddresses.length > 0 || groupInvitationReturn.invalidMailAddresses.length > 0) {
            const existingMailAddresses = groupInvitationReturn.existingMailAddresses.map((ma) => ma.address).join("\n");
            const invalidMailAddresses = groupInvitationReturn.invalidMailAddresses.map((ma) => ma.address).join("\n");
            throw new UserError(() => {
                let msg = "";
                msg += existingMailAddresses.length === 0 ? "" : lang.get("existingMailAddress_msg") + "\n" + existingMailAddresses;
                msg += existingMailAddresses.length === 0 && invalidMailAddresses.length === 0 ? "" : "\n\n";
                msg += invalidMailAddresses.length === 0 ? "" : lang.get("invalidMailAddress_msg") + "\n" + invalidMailAddresses;
                return msg;
            });
        }
        return groupInvitationReturn.invitedMailAddresses;
    }
    entityEventsReceived(updates, eventOwnerGroupId) {
        return promiseMap(updates, (update) => {
            if (!isSameId(eventOwnerGroupId, getEtId(this.group))) {
                // ignore events of different group here
                return;
            }
            if (isUpdateForTypeRef(SentGroupInvitationTypeRef, update)) {
                if (update.operation === "0" /* OperationType.CREATE */ && isSameId(update.instanceListId, this.group.invitations)) {
                    return this.entityClient
                        .load(SentGroupInvitationTypeRef, [update.instanceListId, update.instanceId])
                        .then((instance) => {
                        if (instance) {
                            this.sentGroupInvitations.push(instance);
                            this.onEntityUpdate();
                        }
                    })
                        .catch(ofClass(NotFoundError, (e) => console.log("sent invitation not found", update)));
                }
                if (update.operation === "2" /* OperationType.DELETE */) {
                    findAndRemove(this.sentGroupInvitations, (sentGroupInvitation) => isSameId(getElementId(sentGroupInvitation), update.instanceId));
                    this.onEntityUpdate();
                }
            }
            else if (isUpdateForTypeRef(GroupMemberTypeRef, update)) {
                console.log("update received in share dialog", update);
                if (update.operation === "0" /* OperationType.CREATE */ && isSameId(update.instanceListId, this.group.members)) {
                    return this.entityClient
                        .load(GroupMemberTypeRef, [update.instanceListId, update.instanceId])
                        .then((instance) => {
                        if (instance) {
                            return loadGroupInfoForMember(instance, this.entityClient).then((groupMemberInfo) => {
                                console.log("new member", groupMemberInfo);
                                this.memberInfos.push(groupMemberInfo);
                                this.onEntityUpdate();
                            });
                        }
                    })
                        .catch(ofClass(NotFoundError, (e) => console.log("group member not found", update)));
                }
                if (update.operation === "2" /* OperationType.DELETE */) {
                    findAndRemove(this.memberInfos, (memberInfo) => isSameId(getElementId(memberInfo.member), update.instanceId));
                    this.onEntityUpdate();
                }
            }
        }).then(noOp);
    }
}
//# sourceMappingURL=GroupSharingModel.js.map