import m from "mithril";
import { px, size } from "../gui/size";
import { lang } from "../misc/LanguageViewModel";
import { Icon } from "../gui/base/Icon";
import { SegmentControl } from "../gui/base/SegmentControl";
import { Keys } from "../api/common/TutanotaConstants";
import { asPaymentInterval, formatMonthlyPrice, getCountFromPriceData, getPriceFromPriceData } from "./PriceUtils";
import { isKeyPressed } from "../misc/KeyManager.js";
export function getActiveSubscriptionActionButtonReplacement() {
    return () => m(".buyOptionBox.content-accent-fg.center-vertically.text-center", {
        style: {
            "border-radius": px(size.border_radius_small),
        },
    }, lang.get("pricing.currentPlan_label"));
}
export const BOX_MARGIN = 10;
export class BuyOptionDetails {
    constructor() {
        this.featuresExpanded = false;
        this.featureListItemSelector = ".flex";
    }
    onbeforeupdate(vnode, old) {
        // the expand css class renders an animation which is used when the feature list is expanded
        // the animation should only be shown when the user clicked on the feature expansion button which changes the expanded state
        // thus to check whether the button was pressed, the BuyOptionBox before update must not be expanded but the BuyOptionBox after update is
        // otherwise mithril sometimes updates the view and renders the animation even though nothing changed
        if (vnode.attrs.featuresExpanded && !old.attrs.featuresExpanded) {
            this.featureListItemSelector = ".flex.expand";
        }
        else {
            this.featureListItemSelector = ".flex";
        }
    }
    view(vnode) {
        const { attrs } = vnode;
        this.featuresExpanded = attrs.featuresExpanded || false;
        return m(".mt.pl", attrs.categories.map((fc) => {
            return [
                this.renderCategoryTitle(fc, attrs.renderCategoryTitle),
                fc.features
                    .filter((f) => !f.omit || this.featuresExpanded)
                    .map((f) => m(this.featureListItemSelector, { key: f.key }, [
                    f.heart ? m(Icon, { icon: "Heart" /* BootIcons.Heart */ }) : m(Icon, { icon: f.antiFeature ? "Cancel" /* Icons.Cancel */ : "Checkmark" /* Icons.Checkmark */ }),
                    m(".small.text-left.align-self-center.pl-s.button-height.flex-grow.min-width-0.break-word", [m("span", f.text)]),
                    f.toolTip
                        ? //@ts-ignore
                            m(InfoIcon, { text: f.toolTip })
                        : null,
                ])),
                this.renderPlaceholders(fc),
            ];
        }));
    }
    renderCategoryTitle(fc, renderCategoryTitle) {
        if (fc.title && this.featuresExpanded) {
            return [
                m(".b.text-left.align-self-center.pl-s.button-height.flex-grow.min-width-0.break-word", ""),
                m(".b.text-left.align-self-center.pl-s.button-height.flex-grow.min-width-0.break-word", renderCategoryTitle ? fc.title : ""),
            ];
        }
        else {
            return [];
        }
    }
    renderPlaceholders(fc) {
        if (!this.featuresExpanded) {
            return [];
        }
        else {
            const placeholderCount = fc.featureCount.max - fc.features.length;
            return [...Array(placeholderCount)].map(() => m(".button-height", ""));
        }
    }
}
export class BuyOptionBox {
    view(vnode) {
        const { attrs } = vnode;
        return m(".fg-black", {
            style: {
                width: px(attrs.width),
                padding: "10px",
                height: "100%",
            },
        }, [
            m(".buyOptionBox" + (attrs.highlighted ? ".highlighted" : ""), {
                style: {
                    display: "flex",
                    "flex-direction": "column",
                    "min-height": px(attrs.height),
                    "border-radius": "3px",
                    height: "100%",
                },
            }, [
                this.renderRibbon(attrs.bonusMonths),
                typeof attrs.heading === "string" ? this.renderHeading(attrs.heading) : attrs.heading,
                m(".text-center.pt.flex.center-vertically.center-horizontally", m("span.h1", attrs.price)),
                m(".small.text-center", attrs.priceHint ? lang.getMaybeLazy(attrs.priceHint) : lang.get("emptyString_msg")),
                m(".small.text-center.pb-s", lang.getMaybeLazy(attrs.helpLabel)),
                this.renderPaymentIntervalControl(attrs.paymentInterval),
                attrs.actionButton
                    ? m(".button-min-height", {
                        style: {
                            "margin-top": "auto",
                        },
                    }, attrs.actionButton())
                    : null,
            ]),
        ]);
    }
    renderRibbon(bonusMonths) {
        return bonusMonths > 0
            ? m(".ribbon-horizontal", m(".text-center.b", { style: { padding: px(3) } }, `+${bonusMonths} ${lang.get("pricing.months_label")}`))
            : null;
    }
    renderPaymentIntervalControl(paymentInterval) {
        const paymentIntervalItems = [
            { name: lang.get("pricing.yearly_label"), value: 12 /* PaymentInterval.Yearly */ },
            { name: lang.get("pricing.monthly_label"), value: 1 /* PaymentInterval.Monthly */ },
        ];
        return paymentInterval
            ? m(SegmentControl, {
                selectedValue: paymentInterval(),
                items: paymentIntervalItems,
                onValueSelected: (v) => {
                    paymentInterval?.(v);
                    m.redraw();
                },
            })
            : null;
    }
    renderHeading(heading) {
        return m(".h4.text-center.dialog-header.flex.col.center-horizontally", {
            style: {
                // we need some margin for the discount banner for longer translations shown on the website
                "margin-right": px(30),
                "margin-left": px(30),
                "margin-bottom": px(6),
                "line-height": 1,
            },
        }, m("div", {
            style: {
                "font-size": heading.length > 20 ? "smaller" : undefined,
            },
        }, heading));
    }
}
/**
 * Loads the price information for the given feature type/amount and updates the price information on the BuyOptionBox.
 */
export async function updateBuyOptionBoxPriceInformation(bookingFacade, featureType, amount, attrs) {
    const newPrice = await bookingFacade.getPrice(featureType, amount, false);
    if (amount === getCountFromPriceData(newPrice.currentPriceNextPeriod, featureType)) {
        attrs.actionButton = getActiveSubscriptionActionButtonReplacement();
    }
    const futurePrice = newPrice.futurePriceNextPeriod;
    if (futurePrice) {
        const paymentInterval = asPaymentInterval(futurePrice.paymentInterval);
        const price = getPriceFromPriceData(futurePrice, featureType);
        attrs.price = formatMonthlyPrice(price, paymentInterval);
        attrs.helpLabel = paymentInterval === 12 /* PaymentInterval.Yearly */ ? "pricing.perMonthPaidYearly_label" : "pricing.perMonth_label";
        m.redraw();
    }
}
export class InfoIcon {
    constructor() {
        this.expanded = false;
        this.listener = (e) => {
            this.expanded = false;
            document.removeEventListener("click", this.listener, { capture: true });
            e.stopPropagation();
            m.redraw();
        };
    }
    view({ attrs }) {
        return m("div.flex.justify-center.no-grow-no-shrink.overflow-visible", {
            "aria-pressed": String(this.expanded),
            role: "button",
            style: {
                "margin-top": px(1),
            },
            // we can't really do the state with pure CSS on mobile
            onclick: () => this.expand(),
            onkeydown: (e) => {
                if (isKeyPressed(e.key, Keys.ESC)) {
                    this.listener(e);
                }
            },
            onfocusin: () => this.expand(),
            onfocusout: (e) => this.listener(e),
        }, m(".info-badge.tooltip", {
            expanded: String(this.expanded),
            tabindex: 0,
        }, "i", m("span.tooltiptext.break-word", {
            role: "tooltip",
            style: {
                width: px(120),
                marginLeft: px(-120),
            },
        }, attrs.text)));
    }
    expand() {
        this.expanded = true;
        document.addEventListener("click", this.listener, { capture: true });
    }
}
//# sourceMappingURL=BuyOptionBox.js.map