import m from "mithril";
import { lang } from "../misc/LanguageViewModel";
import { BOX_MARGIN, BuyOptionBox, BuyOptionDetails, getActiveSubscriptionActionButtonReplacement } from "./BuyOptionBox";
import { SegmentControl } from "../gui/base/SegmentControl";
import { formatMonthlyPrice } from "./PriceUtils";
import { getDisplayNameOfPlanType, } from "./FeatureListProvider";
import { ProgrammingError } from "../api/common/error/ProgrammingError";
import { Button } from "../gui/base/Button.js";
import { downcast } from "@tutao/tutanota-utils";
import { HighlightedPlans, LegacyPlans, NewBusinessPlans, NewPersonalPlans, PlanType } from "../api/common/TutanotaConstants.js";
import { px } from "../gui/size.js";
import { LoginButton } from "../gui/base/buttons/LoginButton.js";
const BusinessUseItems = [
    {
        name: lang.get("pricing.privateUse_label"),
        value: false,
    },
    {
        name: lang.get("pricing.businessUse_label"),
        value: true,
    },
];
export function getActionButtonBySubscription(actionButtons, subscription) {
    const ret = actionButtons[subscription];
    if (ret == null) {
        throw new ProgrammingError("Plan is not valid");
    }
    return () => m(LoginButton, ret());
}
export class SubscriptionSelector {
    constructor() {
        this.containerDOM = null;
        this.featuresExpanded = {
            [PlanType.Free]: false,
            [PlanType.Revolutionary]: false,
            [PlanType.Legend]: false,
            [PlanType.Essential]: false,
            [PlanType.Advanced]: false,
            [PlanType.Unlimited]: false,
            All: false,
        };
    }
    oninit(vnode) {
        const acceptedPlans = vnode.attrs.acceptedPlans;
        const onlyBusinessPlansAccepted = acceptedPlans.every((plan) => NewBusinessPlans.includes(plan));
        if (onlyBusinessPlansAccepted) {
            // if only business plans are accepted, we show them first even if the current plan is a personal plan
            vnode.attrs.options.businessUse(true);
        }
    }
    view(vnode) {
        // Add BuyOptionBox margin twice to the boxWidth received
        const columnWidth = vnode.attrs.boxWidth + BOX_MARGIN * 2;
        const inMobileView = (this.containerDOM && this.containerDOM.clientWidth < columnWidth * 2) == true;
        const featureExpander = this.renderFeatureExpanders(inMobileView, vnode.attrs.featureListProvider); // renders all feature expanders, both for every single subscription option but also for the whole list
        let additionalInfo;
        const acceptedPlans = vnode.attrs.acceptedPlans;
        let plans;
        const currentPlan = vnode.attrs.currentPlanType;
        const signup = currentPlan == null;
        const onlyBusinessPlansAccepted = acceptedPlans.every((plan) => NewBusinessPlans.includes(plan));
        const onlyPersonalPlansAccepted = acceptedPlans.every((plan) => NewPersonalPlans.includes(plan));
        // Show the business segmentControl for signup, if both personal & business plans are allowed
        const showBusinessSelector = !onlyBusinessPlansAccepted && !onlyPersonalPlansAccepted;
        let subscriptionPeriodInfoMsg = !signup && currentPlan !== PlanType.Free ? lang.get("switchSubscriptionInfo_msg") + " " : "";
        if (vnode.attrs.options.businessUse()) {
            plans = [PlanType.Essential, PlanType.Advanced, PlanType.Unlimited];
            subscriptionPeriodInfoMsg += lang.get("pricing.subscriptionPeriodInfoBusiness_msg");
            additionalInfo = m(".flex.flex-column.items-center", [
                featureExpander.All, // global feature expander
                m(".smaller.mb.center", subscriptionPeriodInfoMsg),
            ]);
        }
        else {
            if (inMobileView) {
                plans = [PlanType.Revolutionary, PlanType.Legend, PlanType.Free];
            }
            else {
                plans = [PlanType.Free, PlanType.Revolutionary, PlanType.Legend];
            }
            subscriptionPeriodInfoMsg += lang.get("pricing.subscriptionPeriodInfoPrivate_msg");
            additionalInfo = m(".flex.flex-column.items-center", [
                featureExpander.All, // global feature expander
                m(".smaller.mb.center", subscriptionPeriodInfoMsg),
            ]);
        }
        const buyBoxesViewPlacement = plans
            .filter((plan) => acceptedPlans.includes(plan) || vnode.attrs.currentPlanType === plan)
            .map((personalPlan, i) => {
            // only show category title for the leftmost item
            return [
                this.renderBuyOptionBox(vnode.attrs, inMobileView, personalPlan),
                this.renderBuyOptionDetails(vnode.attrs, i === 0, personalPlan, featureExpander),
            ];
        });
        const changePlanMessage = vnode.attrs.msg
            ? lang.getMaybeLazy(vnode.attrs.msg)
            : vnode.attrs.currentPlanType != null && LegacyPlans.includes(vnode.attrs.currentPlanType)
                ? lang.get("currentPlanDiscontinued_msg")
                : null;
        return m("", { lang: lang.code }, [
            showBusinessSelector
                ? m(SegmentControl, {
                    selectedValue: vnode.attrs.options.businessUse(),
                    onValueSelected: vnode.attrs.options.businessUse,
                    items: BusinessUseItems,
                })
                : null,
            m(".flex-center.items-center.mt", [
                vnode.attrs.priceInfoTextId && lang.exists(vnode.attrs.priceInfoTextId) ? m(".b.center", lang.get(vnode.attrs.priceInfoTextId)) : null,
            ]),
            changePlanMessage ? m(".b.center.mt", changePlanMessage) : null,
            m(".flex.center-horizontally.wrap", {
                oncreate: (vnode) => {
                    this.containerDOM = vnode.dom;
                    m.redraw();
                },
                style: {
                    "column-gap": px(BOX_MARGIN),
                },
            }, m(".plans-grid", buyBoxesViewPlacement.flat()), additionalInfo),
        ]);
    }
    renderBuyOptionBox(attrs, inMobileView, planType) {
        return m("", {
            style: {
                width: attrs.boxWidth ? px(attrs.boxWidth) : px(230),
            },
        }, m(BuyOptionBox, this.createBuyOptionBoxAttr(attrs, planType, inMobileView)));
    }
    renderBuyOptionDetails(attrs, renderCategoryTitle, planType, featureExpander) {
        return m("", {
            style: { width: attrs.boxWidth ? px(attrs.boxWidth) : px(230) },
        }, m(BuyOptionDetails, this.createBuyOptionBoxDetailsAttr(attrs, planType, renderCategoryTitle)), featureExpander[planType]);
    }
    createBuyOptionBoxAttr(selectorAttrs, targetSubscription, mobile) {
        const { priceAndConfigProvider } = selectorAttrs;
        // we highlight the center box if this is a signup or the current subscription type is Free
        const upgradingToPaidAccount = !selectorAttrs.currentPlanType || selectorAttrs.currentPlanType === PlanType.Free;
        const highlighted = upgradingToPaidAccount && HighlightedPlans.includes(targetSubscription);
        const subscriptionPrice = priceAndConfigProvider.getSubscriptionPrice(selectorAttrs.options.paymentInterval(), targetSubscription, "1" /* UpgradePriceType.PlanActualPrice */);
        const multiuser = NewBusinessPlans.includes(targetSubscription) || LegacyPlans.includes(targetSubscription) || selectorAttrs.multipleUsersAllowed;
        return {
            heading: getDisplayNameOfPlanType(targetSubscription),
            actionButton: selectorAttrs.currentPlanType === targetSubscription
                ? getActiveSubscriptionActionButtonReplacement()
                : getActionButtonBySubscription(selectorAttrs.actionButtons, targetSubscription),
            price: formatMonthlyPrice(subscriptionPrice, selectorAttrs.options.paymentInterval()),
            priceHint: getPriceHint(subscriptionPrice, selectorAttrs.options.paymentInterval(), multiuser),
            helpLabel: getHelpLabel(targetSubscription, selectorAttrs.options.businessUse()),
            width: selectorAttrs.boxWidth,
            height: selectorAttrs.boxHeight,
            paymentInterval: selectorAttrs.allowSwitchingPaymentInterval && targetSubscription !== PlanType.Free ? selectorAttrs.options.paymentInterval : null,
            highlighted: highlighted,
            showReferenceDiscount: selectorAttrs.allowSwitchingPaymentInterval,
            mobile,
            bonusMonths: targetSubscription !== PlanType.Free && selectorAttrs.options.paymentInterval() === 12 /* PaymentInterval.Yearly */
                ? Number(selectorAttrs.priceAndConfigProvider.getRawPricingData().bonusMonthsForYearlyPlan)
                : 0,
        };
    }
    createBuyOptionBoxDetailsAttr(selectorAttrs, targetSubscription, renderCategoryTitle) {
        const { featureListProvider } = selectorAttrs;
        const subscriptionFeatures = featureListProvider.getFeatureList(targetSubscription);
        const categoriesToShow = subscriptionFeatures.categories
            .map((fc) => {
            return localizeFeatureCategory(fc, targetSubscription, selectorAttrs);
        })
            .filter((fc) => fc != null);
        return {
            categories: categoriesToShow,
            featuresExpanded: this.featuresExpanded[targetSubscription] || this.featuresExpanded.All,
            renderCategoryTitle,
        };
    }
    /**
     * Renders the feature expanders depending on whether currently displaying the feature list in single-column layout or in multi-column layout.
     * If a specific expander is not needed and thus should not be renderer, null | undefined is returned
     */
    renderFeatureExpanders(inMobileView, featureListProvider) {
        if (!featureListProvider.featureLoadingDone()) {
            // the feature list is not available
            return {
                [PlanType.Free]: null,
                [PlanType.Revolutionary]: null,
                [PlanType.Legend]: null,
                [PlanType.Essential]: null,
                [PlanType.Advanced]: null,
                [PlanType.Unlimited]: null,
                All: null,
            };
        }
        if (inMobileView) {
            // In single-column layout every subscription type has its own feature expander.
            if (this.featuresExpanded.All) {
                for (const k in this.featuresExpanded) {
                    this.featuresExpanded[k] = true;
                }
            }
            return {
                [PlanType.Free]: this.renderExpander(PlanType.Free),
                [PlanType.Revolutionary]: this.renderExpander(PlanType.Revolutionary),
                [PlanType.Legend]: this.renderExpander(PlanType.Legend),
                [PlanType.Advanced]: this.renderExpander(PlanType.Advanced),
                [PlanType.Essential]: this.renderExpander(PlanType.Essential),
                [PlanType.Unlimited]: this.renderExpander(PlanType.Unlimited),
                All: null,
            };
        }
        else {
            for (const k in this.featuresExpanded) {
                this.featuresExpanded[k] = this.featuresExpanded.All; // in multi-column layout the specific feature expanders should follow the global one
            }
            return Object.assign({}, { All: this.renderExpander("All") });
        }
    }
    /**
     * Renders a single feature expander.
     * @param subType The current expander that should be rendered
     * @private
     */
    renderExpander(subType) {
        return this.featuresExpanded[subType]
            ? null
            : m(Button, {
                label: "pricing.showAllFeatures",
                type: "secondary" /* ButtonType.Secondary */,
                click: (event) => {
                    this.featuresExpanded[subType] = !this.featuresExpanded[subType];
                    event.stopPropagation();
                },
            });
    }
}
function localizeFeatureListItem(item, targetSubscription, attrs) {
    let text = tryGetTranslation(item.text, getReplacement(item.replacements, targetSubscription, attrs));
    if (text == null) {
        return null;
    }
    if (!item.toolTip) {
        return { text, key: item.text, antiFeature: item.antiFeature, omit: item.omit, heart: !!item.heart };
    }
    else {
        const toolTipText = tryGetTranslation(item.toolTip);
        if (toolTipText === null) {
            return null;
        }
        const toolTip = item.toolTip.endsWith("_markdown") ? m.trust(toolTipText) : toolTipText;
        return { text, toolTip, key: item.text, antiFeature: item.antiFeature, omit: item.omit, heart: !!item.heart };
    }
}
function localizeFeatureCategory(category, targetSubscription, attrs) {
    const title = tryGetTranslation(category.title);
    const features = downcast(category.features.map((f) => localizeFeatureListItem(f, targetSubscription, attrs)).filter((it) => it != null));
    return { title, key: category.title, features, featureCount: category.featureCount };
}
function tryGetTranslation(key, replacements) {
    try {
        return lang.get(key, replacements);
    }
    catch (e) {
        console.log("could not translate feature text for key", key, "hiding feature item");
        return null;
    }
}
/**
 * get a string to insert into a translation with a slot.
 * if no key is found, undefined is returned and nothing is replaced.
 */
export function getReplacement(key, subscription, attrs) {
    const { priceAndConfigProvider } = attrs;
    switch (key) {
        case "customDomains":
            return { "{amount}": priceAndConfigProvider.getPlanPricesForPlan(subscription).customDomains };
        case "mailAddressAliases":
            return { "{amount}": priceAndConfigProvider.getPlanPricesForPlan(subscription).includedAliases };
        case "storage":
            return { "{amount}": priceAndConfigProvider.getPlanPricesForPlan(subscription).includedStorage };
    }
}
function getHelpLabel(planType, businessUse) {
    if (planType === PlanType.Free)
        return "pricing.upgradeLater_msg";
    return businessUse ? "pricing.excludesTaxes_msg" : "pricing.includesTaxes_msg";
}
function getPriceHint(subscriptionPrice, paymentInterval, multiuser) {
    if (subscriptionPrice > 0) {
        if (multiuser) {
            return paymentInterval === 12 /* PaymentInterval.Yearly */ ? "pricing.perUserMonthPaidYearly_label" : "pricing.perUserMonth_label";
        }
        else {
            return paymentInterval === 12 /* PaymentInterval.Yearly */ ? "pricing.perMonthPaidYearly_label" : "pricing.perMonth_label";
        }
    }
    else {
        return "emptyString_msg";
    }
}
//# sourceMappingURL=SubscriptionSelector.js.map