import m from "mithril";
import { Dialog } from "../gui/base/Dialog";
import { lang } from "../misc/LanguageViewModel";
import { createSwitchAccountTypePostIn } from "../api/entities/sys/TypeRefs.js";
import { AccountType, Const, Keys, LegacyPlans, NewBusinessPlans, PlanType, PlanTypeToName, } from "../api/common/TutanotaConstants";
import { SubscriptionSelector } from "./SubscriptionSelector";
import stream from "mithril/stream";
import { showProgressDialog } from "../gui/dialogs/ProgressDialog";
import { SwitchSubscriptionDialogModel } from "./SwitchSubscriptionDialogModel";
import { locator } from "../api/main/MainLocator";
import { SwitchAccountTypeService } from "../api/entities/sys/Services.js";
import { BadRequestError, InvalidDataError, PreconditionFailedError } from "../api/common/error/RestError.js";
import { FeatureListProvider } from "./FeatureListProvider";
import { PriceAndConfigProvider } from "./PriceUtils";
import { defer, downcast } from "@tutao/tutanota-utils";
import { showSwitchToBusinessInvoiceDataDialog } from "./SwitchToBusinessInvoiceDataDialog.js";
import { getByAbbreviation } from "../api/common/CountryList.js";
import { formatNameAndAddress } from "../api/common/utils/CommonFormatter.js";
import { CancellationReasonInput } from "./CancellationReasonInput.js";
/**
 * Only shown if the user is already a Premium user. Allows cancelling the subscription (only private use) and switching the subscription to a different paid subscription.
 */
export async function showSwitchDialog(customer, customerInfo, accountingInfo, lastBooking, acceptedPlans, reason) {
    const deferred = defer();
    const [featureListProvider, priceAndConfigProvider] = await showProgressDialog("pleaseWait_msg", Promise.all([
        FeatureListProvider.getInitializedInstance(locator.domainConfigProvider().getCurrentDomainConfig()),
        PriceAndConfigProvider.getInitializedInstance(null, locator.serviceExecutor, null),
    ]));
    const model = new SwitchSubscriptionDialogModel(customer, accountingInfo, await locator.logins.getUserController().getPlanType(), lastBooking);
    const cancelAction = () => {
        dialog.close();
        deferred.resolve(customerInfo.plan);
    };
    const headerBarAttrs = {
        left: [
            {
                label: "cancel_action",
                click: cancelAction,
                type: "secondary" /* ButtonType.Secondary */,
            },
        ],
        right: [],
        middle: () => lang.get("subscription_label"),
    };
    const currentPlanInfo = model.currentPlanInfo;
    const businessUse = stream(currentPlanInfo.businessUse);
    const paymentInterval = stream(12 /* PaymentInterval.Yearly */); // always default to yearly
    const multipleUsersAllowed = model.multipleUsersStillSupportedLegacy();
    const dialog = Dialog.largeDialog(headerBarAttrs, {
        view: () => m("#upgrade-account-dialog.pt", m(SubscriptionSelector, {
            options: {
                businessUse,
                paymentInterval: paymentInterval,
            },
            priceInfoTextId: priceAndConfigProvider.getPriceInfoMessage(),
            msg: reason,
            boxWidth: 230,
            boxHeight: 270,
            acceptedPlans: acceptedPlans,
            currentPlanType: currentPlanInfo.planType,
            allowSwitchingPaymentInterval: currentPlanInfo.paymentInterval !== 12 /* PaymentInterval.Yearly */,
            actionButtons: subscriptionActionButtons,
            featureListProvider: featureListProvider,
            priceAndConfigProvider,
            multipleUsersAllowed,
        })),
    })
        .addShortcut({
        key: Keys.ESC,
        exec: cancelAction,
        help: "close_alt",
    })
        .setCloseHandler(cancelAction);
    const subscriptionActionButtons = {
        [PlanType.Free]: () => ({
            label: "pricing.select_action",
            onclick: () => cancelSubscription(dialog, currentPlanInfo, deferred, customer),
        }),
        [PlanType.Revolutionary]: createPlanButton(dialog, PlanType.Revolutionary, currentPlanInfo, deferred, paymentInterval, accountingInfo),
        [PlanType.Legend]: createPlanButton(dialog, PlanType.Legend, currentPlanInfo, deferred, paymentInterval, accountingInfo),
        [PlanType.Essential]: createPlanButton(dialog, PlanType.Essential, currentPlanInfo, deferred, paymentInterval, accountingInfo),
        [PlanType.Advanced]: createPlanButton(dialog, PlanType.Advanced, currentPlanInfo, deferred, paymentInterval, accountingInfo),
        [PlanType.Unlimited]: createPlanButton(dialog, PlanType.Unlimited, currentPlanInfo, deferred, paymentInterval, accountingInfo),
    };
    dialog.show();
    return deferred.promise;
}
async function doSwitchPlan(accountingInfo, newPaymentInterval, targetSubscription, dialog, currentPlanInfo, deferredPlan) {
    if (currentPlanInfo.paymentInterval !== newPaymentInterval) {
        await locator.customerFacade.changePaymentInterval(accountingInfo, newPaymentInterval);
    }
    await switchSubscription(targetSubscription, dialog, currentPlanInfo).then((newPlan) => deferredPlan.resolve(newPlan));
}
function createPlanButton(dialog, targetSubscription, currentPlanInfo, deferredPlan, newPaymentInterval, accountingInfo) {
    return () => ({
        label: "buy_action",
        onclick: async () => {
            // Show an extra dialog in the case that someone is upgrading from a legacy plan to a new plan because they can't revert.
            if (LegacyPlans.includes(currentPlanInfo.planType) &&
                !(await Dialog.confirm(() => lang.get("upgradePlan_msg", { "{plan}": PlanTypeToName[targetSubscription] })))) {
                return;
            }
            await showProgressDialog("pleaseWait_msg", doSwitchPlan(accountingInfo, newPaymentInterval(), targetSubscription, dialog, currentPlanInfo, deferredPlan));
        },
    });
}
function handleSwitchAccountPreconditionFailed(e) {
    const reason = e.data;
    if (reason == null) {
        return Dialog.message("unknownError_msg");
    }
    else {
        let detailMsg;
        switch (reason) {
            case "unsubscribe.too_many_users" /* UnsubscribeFailureReason.TOO_MANY_ENABLED_USERS */:
                detailMsg = lang.get("accountSwitchTooManyActiveUsers_msg");
                break;
            case "unsubscribe.custom_mail_address" /* UnsubscribeFailureReason.CUSTOM_MAIL_ADDRESS */:
                detailMsg = lang.get("accountSwitchCustomMailAddress_msg");
                break;
            case "unsubscribe.too_many_calendars" /* UnsubscribeFailureReason.TOO_MANY_CALENDARS */:
                detailMsg = lang.get("accountSwitchMultipleCalendars_msg");
                break;
            case "unsubscirbe.invalid_calendar_type" /* UnsubscribeFailureReason.CALENDAR_TYPE */:
                detailMsg = lang.get("accountSwitchSharedCalendar_msg");
                break;
            case "unsubscribe.too_many_aliases" /* UnsubscribeFailureReason.TOO_MANY_ALIASES */:
            case "bookingservice.too_many_aliases" /* BookingFailureReason.TOO_MANY_ALIASES */:
                detailMsg = lang.get("accountSwitchAliases_msg");
                break;
            case "unsubscribe.too_much_storage" /* UnsubscribeFailureReason.TOO_MUCH_STORAGE_USED */:
            case "bookingservice.too_much_storage_used" /* BookingFailureReason.TOO_MUCH_STORAGE_USED */:
                detailMsg = lang.get("storageCapacityTooManyUsedForBooking_msg");
                break;
            case "unsubscribe.too_many_domains" /* UnsubscribeFailureReason.TOO_MANY_DOMAINS */:
            case "bookingservice.too_many_domains" /* BookingFailureReason.TOO_MANY_DOMAINS */:
                detailMsg = lang.get("tooManyCustomDomains_msg");
                break;
            case "unsubscribe.has_template_group" /* UnsubscribeFailureReason.HAS_TEMPLATE_GROUP */:
            case "bookingservice.has_template_group" /* BookingFailureReason.HAS_TEMPLATE_GROUP */:
                detailMsg = lang.get("deleteTemplateGroups_msg");
                break;
            case "unsubscribe.whitelabel_domain_active" /* UnsubscribeFailureReason.WHITELABEL_DOMAIN_ACTIVE */:
            case "bookingservice.whitelabel_domain_active" /* BookingFailureReason.WHITELABEL_DOMAIN_ACTIVE */:
                detailMsg = lang.get("whitelabelDomainExisting_msg");
                break;
            case "unsubscribe.not_enough_credit" /* UnsubscribeFailureReason.NOT_ENOUGH_CREDIT */:
                return Dialog.message("insufficientBalanceError_msg");
            case "unsubscribe.invoice_not_paid" /* UnsubscribeFailureReason.INVOICE_NOT_PAID */:
                return Dialog.message("invoiceNotPaidSwitch_msg");
            default:
                throw e;
        }
        return Dialog.message(() => lang.get("accountSwitchNotPossible_msg", {
            "{detailMsg}": detailMsg,
        }));
    }
}
async function tryDowngradePremiumToFree(switchAccountTypeData, currentPlanInfo) {
    try {
        await locator.serviceExecutor.post(SwitchAccountTypeService, switchAccountTypeData);
        await locator.customerFacade.switchPremiumToFreeGroup();
        return PlanType.Free;
    }
    catch (e) {
        if (e instanceof PreconditionFailedError) {
            await handleSwitchAccountPreconditionFailed(e);
        }
        else if (e instanceof InvalidDataError) {
            await Dialog.message("accountSwitchTooManyActiveUsers_msg");
        }
        else if (e instanceof BadRequestError) {
            await Dialog.message("deactivatePremiumWithCustomDomainError_msg");
        }
        else {
            throw e;
        }
        return currentPlanInfo.planType;
    }
}
async function cancelSubscription(dialog, currentPlanInfo, planPromise, customer) {
    if (!(await Dialog.confirm("unsubscribeConfirm_msg"))) {
        return;
    }
    let reasonCategory = null;
    let reason = "";
    let result = await new Promise((resolve) => {
        let dialog = Dialog.showActionDialog({
            title: lang.get("survey_label"),
            child: {
                view: () => m(CancellationReasonInput, {
                    reason: reason,
                    reasonHandler: (enteredReason) => (reason = enteredReason),
                    category: reasonCategory,
                    categoryHandler: (category) => (reasonCategory = category),
                }),
            },
            okAction: () => {
                dialog.close();
                resolve(true);
            },
            cancelAction: () => {
                dialog.close();
                resolve(false);
            },
            cancelActionTextId: "skip_action",
        });
    });
    const switchAccountTypeData = createSwitchAccountTypePostIn({
        accountType: AccountType.FREE,
        date: Const.CURRENT_DATE,
        customer: customer._id,
        specialPriceUserSingle: null,
        referralCode: null,
        plan: PlanType.Free,
        reasonCategory: result ? reasonCategory : null,
        reason: result ? reason : null,
    });
    try {
        await showProgressDialog("pleaseWait_msg", tryDowngradePremiumToFree(switchAccountTypeData, currentPlanInfo).then((newPlan) => planPromise.resolve(newPlan)));
    }
    finally {
        dialog.close();
    }
}
async function switchSubscription(targetSubscription, dialog, currentPlanInfo) {
    if (targetSubscription === currentPlanInfo.planType) {
        return currentPlanInfo.planType;
    }
    const userController = locator.logins.getUserController();
    const customer = await userController.loadCustomer();
    if (!customer.businessUse && NewBusinessPlans.includes(downcast(targetSubscription))) {
        const accountingInfo = await userController.loadAccountingInfo();
        const invoiceData = {
            invoiceAddress: formatNameAndAddress(accountingInfo.invoiceName, accountingInfo.invoiceAddress),
            country: accountingInfo.invoiceCountry ? getByAbbreviation(accountingInfo.invoiceCountry) : null,
            vatNumber: accountingInfo.invoiceVatIdNo, // only for EU countries otherwise empty
        };
        const updatedInvoiceData = await showSwitchToBusinessInvoiceDataDialog(customer, invoiceData, accountingInfo);
        if (!updatedInvoiceData) {
            return currentPlanInfo.planType;
        }
    }
    try {
        const postIn = createSwitchAccountTypePostIn({
            accountType: AccountType.PAID,
            plan: targetSubscription,
            date: Const.CURRENT_DATE,
            referralCode: null,
            customer: customer._id,
            specialPriceUserSingle: null,
            reasonCategory: null,
            reason: null,
        });
        try {
            await showProgressDialog("pleaseWait_msg", locator.serviceExecutor.post(SwitchAccountTypeService, postIn));
            return targetSubscription;
        }
        catch (e) {
            if (e instanceof PreconditionFailedError) {
                await handleSwitchAccountPreconditionFailed(e);
                return currentPlanInfo.planType;
            }
            throw e;
        }
    }
    finally {
        dialog.close();
    }
}
//# sourceMappingURL=SwitchSubscriptionDialog.js.map