export type EccPrivateKey = Uint8Array;
export type EccPublicKey = Uint8Array;
/**
 * Contains a public key and its corresponding private key
 *
 * NOTE: Keys should be cleared from memory once they are no longer needed!
 */
export type EccKeyPair = {
    publicKey: EccPublicKey;
    privateKey: EccPrivateKey;
};
/**
 * Contains all information for deriving AES keys
 *
 * The shared secret MUST NEVER be used as a key directly as it is a biased (some bits are more likely to be set than others).
 * The involved public keys should also be included when deriving an AES key from these shared secrets.
 */
export type EccSharedSecrets = {
    ephemeralSharedSecret: Uint8Array;
    authSharedSecret: Uint8Array;
};
/**
 * @return randomly generated X25519 key pair
 */
export declare function generateEccKeyPair(): EccKeyPair;
/**
 * Derive a shared secret from the sender's private key and the recipient's public key to encrypt a message
 * @param senderIdentityPrivateKey	the sender's private identity key
 * @param ephemeralPrivateKey  the ephemeral private key generated by the sender for just one message (to one or more recipients)
 * @param recipientIdentityPublicKey the recipient's public identity key
 * @return the shared secrets
 */
export declare function eccEncapsulate(senderIdentityPrivateKey: EccPrivateKey, ephemeralPrivateKey: EccPrivateKey, recipientIdentityPublicKey: EccPublicKey): EccSharedSecrets;
/**
 * Derive a shared secret from the recipient's private key and the sender's public key to decrypt a message
 * @param senderIdentityPublicKey	the sender's public identity key
 * @param ephemeralPublicKey  the ephemeral public key generated by the sender for just one message (to one or more recipients)
 * @param recipientIdentityPrivateKey the recipient's private identity key
 * @return shared secret and the sender's public key
 */
export declare function eccDecapsulate(senderIdentityPublicKey: EccPublicKey, ephemeralPublicKey: EccPublicKey, recipientIdentityPrivateKey: EccPrivateKey): EccSharedSecrets;
