export interface ErrorInfo {
    readonly name: string | null;
    readonly message: string | null;
    readonly stack?: string | null;
}
export type lazy<T> = () => T;
export type lazyAsync<T> = () => Promise<T>;
export type Thunk = () => unknown;
/**
 * A group key and its version.
 */
export type Versioned<T> = {
    object: T;
    version: number;
};
/**
 * Create a versioned object with version 0
 */
export declare function freshVersioned<T>(object: T): Versioned<T>;
/** specifies a set of keys to be required, even if they're originally optional on a type.
 * requires nullable fields to be non-null, this may not be desired for all use cases.
 * Use "RequireNullable<K, T>" for cases where null is a meaningful value.
 *
 * `Require<"uid", Partial<CalendarEvent>>` */
export type Require<K extends keyof T, T> = T & {
    [P in K]-?: NonNullable<T[P]>;
};
export type DeferredObject<T> = {
    resolve: (arg0: T | PromiseLike<T>) => void;
    reject: (arg0: Error) => void;
    promise: Promise<T>;
};
export type DeferredObjectWithHandler<T, U> = {
    resolve: (arg0: T) => void;
    reject: (arg0: Error) => void;
    promise: Promise<U>;
};
export declare function defer<T>(): DeferredObject<T>;
export declare function deferWithHandler<T, U>(handler: (arg0: T) => U): DeferredObjectWithHandler<T, U>;
export declare function asyncFind<T>(array: ReadonlyArray<T>, finder: (item: T, index: number, arrayLength: number) => Promise<boolean>): Promise<T | null | undefined>;
export declare function asyncFindAndMap<T, R>(array: ReadonlyArray<T>, finder: (item: T, index: number, arrayLength: number) => Promise<R | null>): Promise<R | null | undefined>;
/**
 * Calls an executor function for slices of nbrOfElementsInGroup items of the given array until the executor function returns false.
 */
export declare function executeInGroups<T>(array: T[], nbrOfElementsInGroup: number, executor: (items: T[]) => Promise<boolean>): Promise<void>;
export declare function neverNull<T>(object: T): NonNullable<T>;
/**
 * returns its argument if it is not null, throws otherwise.
 * @param value the value to check
 * @param message optional error message
 */
export declare function assertNotNull<T>(value: T | null | undefined, message?: string): T;
/**
 * assertion function that only returns if the argument is non-null
 * (acts as a type guard)
 * @param value the value to check
 * @param message optional error message
 */
export declare function assertNonNull<T>(value: T | null | undefined, message?: string): asserts value is T;
export declare function isNotNull<T>(t: T | null | undefined): t is T;
export declare function assert(assertion: MaybeLazy<boolean>, message: string): void;
export declare function downcast<R = any>(object: any): R;
export declare function clone<T>(instance: T): T;
/**
 * Function which accepts another function. On first invocation
 * of this resulting function result will be remembered and returned
 * on consequent invocations.
 */
export declare function lazyMemoized<T>(source: () => T): lazy<T>;
export type Callback<T> = (arg: T) => void;
/**
 * accept a function taking exactly one argument and returning nothing and return a version of it
 * that will call the original function on the first call and ignore any further calls.
 * @param fn a function taking one argument and returning nothing
 */
export declare function makeSingleUse<T>(fn: Callback<T>): Callback<T>;
/**
 * Returns a cached version of {@param fn}.
 * Cached function checks that argument is the same (with ===) and if it is then it returns the cached result.
 * If the cached argument has changed then {@param fn} will be called with new argument and result will be cached again.
 * Only remembers the last argument.
 */
export declare function memoized<T, R>(fn: (arg0: T) => R): (arg0: T) => R;
/**
 * Like {@link memoized} but the argument is passed in via {@param argumentProvider}.
 * Useful for the cases where we want to keep only one field around e.g. for lazy getters
 */
export declare function memoizedWithHiddenArgument<T, R>(argumentProvider: () => T, computationFunction: (arg: T) => R): () => R;
/**
 * Function which returns what was passed into it
 */
export declare function identity<T>(t: T): T;
/**
 * Function which does nothing.
 */
export declare function noOp(): void;
/**
 * Return a function, which executed {@param toThrottle} only after it is not invoked for {@param timeout} ms.
 * Executes function with the last passed arguments
 * @return {Function}
 */
export declare function debounce<F extends (...args: any) => void>(timeout: number, toThrottle: F): F;
/**
 * Returns a debounced function. When invoked for the first time, will just invoke
 * {@param toThrottle}. On subsequent invocations it will either invoke it right away
 * (if {@param timeout} has passed) or will schedule it to be run after {@param timeout}.
 * So the first and the last invocations in a series of invocations always take place
 * but ones in the middle (which happen too often) are discarded.}
 */
export declare function debounceStart<F extends (...args: any) => void>(timeout: number, toThrottle: F): F;
export declare function randomIntFromInterval(min: number, max: number): number;
export declare function errorToString(error: ErrorInfo): string;
/**
 * Like {@link Object.entries} but preserves the type of the key and value
 */
export declare function objectEntries<A extends string | symbol, B>(object: Record<A, B>): Array<[A, B]>;
/**
 * modified deepEquals from ospec is only needed as long as we use custom classes (TypeRef) and Date is not properly handled
 */
export declare function deepEqual(a: any, b: any): boolean;
/**
 * returns an array of top-level properties that are in both objA and objB, but differ in value
 * does not handle functions or circular references
 * treats undefined and null as equal
 */
export declare function getChangedProps(objA: any, objB: any): Array<string>;
/**
 * Disallow set, delete and clear on Map.
 * Important: It is *not* a deep freeze.
 * @param myMap
 * @return {unknown}
 */
export declare function freezeMap<K, V>(myMap: ReadonlyMap<K, V>): ReadonlyMap<K, V>;
export declare function addressDomain(senderAddress: string): string;
/**
 * Ignores the fact that Object.keys returns also not owned properties.
 */
export declare function typedKeys<K extends string, V>(obj: Record<K, V>): Array<K>;
/**
 * Ignores the fact that Object.keys returns also not owned properties.
 */
export declare function typedEntries<K extends string, V>(obj: Record<K, V>): Array<[K, V]>;
/**
 * Ignores the fact that Object.keys returns also not owned properties.
 */
export declare function typedValues<K extends string, V>(obj: Record<K, V>): Array<V>;
export type MaybeLazy<T> = T | lazy<T>;
export declare function resolveMaybeLazy<T>(maybe: MaybeLazy<T>): T;
export declare function getAsLazy<T>(maybe: MaybeLazy<T>): lazy<T>;
export declare function mapLazily<T, U>(maybe: MaybeLazy<T>, mapping: (arg0: T) => U): lazy<U>;
/**
 * Stricter version of parseInt() from MDN. parseInt() allows some arbitrary characters at the end of the string.
 * Returns NaN in case there's anything non-number in the string.
 */
export declare function filterInt(value: string): number;
interface Positioned {
    x: number;
    y: number;
}
interface Sized {
    top: number;
    left: number;
    bottom: number;
    right: number;
}
export declare function insideRect(point: Positioned, rect: Sized): boolean;
/**
 * If val is non null, returns the result of val passed to action, else null
 */
export declare function mapNullable<T, U>(val: T | null | undefined, action: (arg0: T) => U | null | undefined): U | null;
/** Helper to take instead of `typeof setTimeout` which is hellish to reproduce */
export type TimeoutSetter = (fn: () => unknown, arg1: number) => ReturnType<typeof setTimeout>;
export declare function mapObject<K extends string | number | symbol, V, R>(mapper: (arg0: V) => R, obj: Record<K, V>): Record<K, R>;
/**
 * Run jobs with defined max parallelism.
 */
export declare class BoundedExecutor {
    private readonly maxParallelJobs;
    private runningJobsCount;
    private currentJob;
    constructor(maxParallelJobs: number);
    run<T>(job: () => Promise<T>): Promise<T>;
}
export {};
